/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * Copyright (c) 2002-2010 Neterion, Inc.
 * All right Reserved.
 *
 * FileName :   vxgehal-rnic.h
 *
 * Description:  API and data structures used by RNCI
 *
 * Created:       15 Sep 2006
 */

#ifndef	VXGE_HAL_RNIC_H
#define	VXGE_HAL_RNIC_H

__EXTERN_BEGIN_DECLS

#include "vxgehal-version.h"
#include "vxge-os-pal.h"
#include "vxgehal-status.h"
#include "vxgehal-types.h"
#include "vxge-list.h"
#include "vxge-defs.h"
#include "vxgehal-stats.h"

/*
 * This API is exported by HAL to support iWARP/RDMA module. The series of steps
 * involved in using this API are
 *	1. Get the information about Adapters and the virtual paths available on
 *	the Adapter for the host using the LL driver API.
 *	  2. For each virtual path that is going to be used by the iWARP/RDMA
 *	module, call vxge_hal_vpath_attach
 *	  -- This callback function passed to vxge_hal_vpath_attach() routine to
 *	   get replys to all asynchronous functions. The format of the reply
 *	   is a message along with the parameters that are common fro all
 *	   replys. The message handle passed to this callback is opaque for
 *	   the iWARP/RDMA module and the information from the message can be
 *	   got by calling appropriate get function depending on the message
 *	   type passed as one of the parameter to the callback. The message
 *	   types that are to be passed to the callback are the ones that ends
 *	   with UMQ in the message types enum.
 *	-- Each of the follwoing API is dnoted, if it is synchronous or
 *	   asynchronous.
 *	-- The vxge_hal_vpath_attach routine returns a handle, that is needed
 *	   to be passed a parameter to each of the API.
 *	-- Before exiting the module, vxge_hal_vpath_detach routine should be
 *	   called for each of the vpath handles got by calling
 *	   vxge_hal_vpath_attach routine.
 *	  3. vxge_hal_vpath_obj_count_get() routine can be used to find the
 *	current count of each of objects that are active on the adapter for
 *	the vpath (Virtual Path)
 *	  4. For SQ, SRQ, CQRQ and TCP/LRO/SPDM sessions, the reply callback for
 *	their respective create api calls provides an object id. This object
 *	id is opaque for the caller and should not be interpreted. The object
 *	id should be passed to all the API that need a referrence to one of
 *	these objects. For NCE objects, the object Id is index into NCE
 *	table. For STAG object this object Id is the object id passed in
 *	corresponding stag API.
 */

/*
 * Virtual Paths
 */

/*
 * vxge_hal_vpath_attach_f - Attach to a virtual path on a given adapter
 * @devh: handle to device object
 * @vp_id: Virtual path id
 * @cb_fn: Call back to be called to complete an asynchronous function call
 * @client_handle: Client's handle to be returned in the callback
 * @vpath_handle: Buffer to return a handle to the vpath
 *
 * This function is used to open access to virtual path of an
 * adapter for offload, LRO and SPDM operations. This function returns
 * synchronously.
 */
typedef vxge_hal_status_e (*vxge_hal_vpath_attach_f)(
			vxge_hal_device_h devh,
			u32 vp_id,
			vxge_hal_vpath_callback_f cb_fn,
			vxge_hal_client_h client_handle,
			vxge_hal_vpath_h *vpath_handle);

/*
 * vxge_hal_vpath_detach_f - Detach from a virtual path on a given adapter
 * @vpath_handle: Handle got from previous vpath open
 *
 * This function is used to close access to virtual path opened
 * earlier. This function returns synchronously.
 */
typedef vxge_hal_status_e (*vxge_hal_vpath_detach_f)(
			vxge_hal_vpath_h vpath_handle);

/*
 * struct vxge_hal_vpath_mgmt_if_t - Functions used to manage virtual paths
 * @vxge_hal_vpath_attach: Attach to a virtual path on a given adapter
 * @vxge_hal_vpath_detach: Detach from a virtual path on a given adapter
 *
 * This structure contains the functions for managing the virtual path
 */
typedef struct vxge_hal_vpath_mgmt_if_t {
	vxge_hal_vpath_attach_f		vxge_hal_vpath_attach;
	vxge_hal_vpath_detach_f		vxge_hal_vpath_detach;
} vxge_hal_vpath_mgmt_if_t;

/*
 * PB List
 */

/*
 * struct vxge_hal_pblist_t - PB list
 * @pble_list: List header.
 * @vpathh: This field is reserved for the HAL
 * @pble_size: size of PBLE.
 * @pbles_per_block: Number of pbles per block.
 * @num_pbles: Total number of PBLES in the list
 *
 * PBLE List data structure to keep track of the PBLE list
 * generated and passed to adapter in register and reregister memory region.
 *
 */
typedef struct vxge_hal_pblist_t {
	vxge_list_t			pble_list;
	vxge_hal_vpath_h		vpathh;
	u32				pble_size;
	u32				pbles_per_block;
	u32				num_pbles;
} vxge_hal_pblist_t;

/*
 * vxge_hal_pblist_allocate_f - Allocate PB list
 * @vpath_handle: Handle to vpath in which this pblist is used.
 * @pblist: pblist to be created.
 * @length: Length of the pblist.
 *
 * This function allocates PB list. This function returns synchronously
 */
typedef vxge_hal_status_e (*vxge_hal_pblist_allocate_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_pblist_t *pblist,
			u32 length);

/*
 * vxge_hal_pblist_set_f - Set buffer pointers in the pblist
 * @pblist: list of blocks for odlist.
 * @phy_addr_list: Array of DMA-able Physical addresses of buffers.
 * @num_addr: Number of addresses in phy_addr_list.
 *
 * This  function sets the buffers into the pblist appropriately. This
 * function returns synchronously.
 *
 */
typedef vxge_hal_status_e (*vxge_hal_pblist_set_f)(
			vxge_hal_pblist_t *pblist,
			u64 *phy_addr_list,
			u32 num_addr);

/*
 * vxge_hal_pblist_free_f - Free the PB List
 * @vpath_handle: Handle to vpath in which this pblist is used.
 * @pblist: pblist to be deallocated
 *
 * This function freeup the memory pool and removes the
 * pblist. This function returns synchronously.
 */
typedef void (*vxge_hal_pblist_free_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_pblist_t *pblist);

/*
 * struct vxge_hal_pblist_mgmt_if_t - Functions used to manage pblists
 * @vxge_hal_pblist_allocate: Allocate PB list
 * @vxge_hal_pblist_set: Set buffer pointers in the pblist
 * @vxge_hal_pblist_free: Free the PB List
 *
 * This structure contains the functions for managing the pb list
 */
typedef struct vxge_hal_pblist_mgmt_if_t {
	vxge_hal_pblist_allocate_f	vxge_hal_pblist_allocate;
	vxge_hal_pblist_set_f		vxge_hal_pblist_set;
	vxge_hal_pblist_free_f		vxge_hal_pblist_free;
} vxge_hal_pblist_mgmt_if_t;

/*
 * struct vxge_hal_odlist_t - OD list
 * @od_list: This field is reserved for the HAL
 * @vpathh: This field is reserved for the HAL
 * @num_ods: This field is reserved for tthe HAL
 * @od_size: This field is reserved for tthe HAL
 * @total_bytes: This field is reserved for tthe HAL
 *
 * Offload Descriptor(OD) List data structure to keep track of the OD list
 * generated and passed to adapter in towis.
 *
 */
typedef struct vxge_hal_odlist_t {
	void				*od_list;
	vxge_hal_vpath_h		vpathh;
	u32				num_ods;
#define	VXGE_HAL_ODLIST_NUM_ODS_MIN	2
#define	VXGE_HAL_ODLIST_NUM_ODS_MAX	256
	u32				od_size;
	u32				total_bytes;
} vxge_hal_odlist_t;

/*
 * vxge_hal_odlist_allocate_f - Allocate OD list
 * @vpath_handle: Handle to vpath in which this odlist is used.
 * @odlist: odlist to be created.
 * @length: Length of the odlist.
 *
 * This function allocates OD list. This function returns synchronously.
 */

typedef vxge_hal_status_e (*vxge_hal_odlist_allocate_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_odlist_t *odlist,
			u32 length);

/*
 * vxge_hal_odlist_set_f - Set buffer pointer in the odlist
 * @odlist: list of blocks for odlist.
 * @ptr_to: Physical address of the buffer or the tagged offset
 * @stag: stag
 * @length: Length of the buffer
 *
 * This  function sets the buffer into the odlist appropriately. This
 * function returns synchronously.
 *
 */
typedef vxge_hal_status_e (*vxge_hal_odlist_set_f)(
			vxge_hal_odlist_t *odlist,
			u64 ptr_to,
			u32 stag,
			u32 length);

/*
 * vxge_hal_odlist_free_f - Frees the OD List
 * @vpath_handle: Handle to vpath in which this odlist is used.
 * @odlist: odlist to be freed
 *
 * This function freeup the memory pool and removes the
 * odlist. This function returns synchronously.
 */

typedef void (*vxge_hal_odlist_free_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_odlist_t *odlist);

/*
 * struct vxge_hal_odlist_mgmt_if_t - Functions used to manage OD Lists
 * @vxge_hal_odlist_allocate: Allocate OD list
 * @vxge_hal_odlist_set: Set buffer pointers in the OD List
 * @vxge_hal_odlist_free: Free the OD List
 *
 * This structure contains the functions for managing the OD List
 */
typedef struct vxge_hal_odlist_mgmt_if_t {
	vxge_hal_odlist_allocate_f	vxge_hal_odlist_allocate;
	vxge_hal_odlist_set_f		vxge_hal_odlist_set;
	vxge_hal_odlist_free_f		vxge_hal_odlist_free;
} vxge_hal_odlist_mgmt_if_t;

/*
 * NCE
 */

/*
 * vxge_hal_nce_create_f - Create NCE
 *
 * @vpath_handle: Handle returned by virtual path open
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @nce_opaque_handle: NCE Opaque handle to be returned in vpath callbacks
 *		that are generated unsolicitedly upon any events related to this
 *		NCE. The memory allocated for this handle should not be freed
 *		until the NCE is deleted successfully and delete confirmation
 *		is received throuh the vpath callback for delete function.
 * @src_mac_addr: Source Mac Address
 * @rem_mac_addr: Remote Mac Address
 * @vlan_id: Valn Id
 * @host_rch_time: Host Reach Time
 *
 * Creates an NCE on the adapter. If this function returns VXGE_HAL_PENDING,then
 * the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_nce_create_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_opaque_handle_t *opaque_handle,
			vxge_hal_opaque_handle_t *nce_opaque_handle,
			u64 src_mac_addr,
			u64 rem_mac_addr,
			u32 vlan_id,
			u32 host_rch_time);

/*
 * vxge_hal_nce_update_mac_addr_f - Update NCE MAC Address
 *
 * @vpath_handle: Handle returned by virtual path open
 * @obj_id: Object Id of NCE
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @mac_addr: Mac Address
 *
 * Modifies the MAC address for an NCE on the adapter. If this function returns
 * VXGE_HAL_PENDING, then the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_nce_update_mac_addr_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t obj_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			u64 mac_addr);

/*
 * vxge_hal_nce_update_rch_time_f - Update NCE RCH Time
 *
 * @vpath_handle: Handle returned by virtual path open
 * @obj_id: Object Id of NCE
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @rch_time: Host reachability time
 *
 * Modifies the RCH Time for an NCE on the adapter. If this function returns
 * VXGE_HAL_PENDING, then the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_nce_update_rch_time_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t obj_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			u64 rch_time);

/*
 * vxge_hal_nce_query_f - Query NCE
 *
 * @vpath_handle: Handle returned by virtual path open
 * @obj_id: Object Id of NCE
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 *
 * PQueries a NCE on the adapter. If this function returns VXGE_HAL_PENDING,then
 * the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_nce_query_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t obj_id,
			vxge_hal_opaque_handle_t *opaque_handle);

/*
 * vxge_hal_nce_query_resp_get_f - Get values from NCE Query Response
 *
 * @umsgh: Handle to umsg.
 * @nce_opaque_handle: Buffer to return NCE opaque handle
 * @src_mac_addr: Buffer to return Source Mac Address
 * @rem_mac_addr: Buffer to return Remote Mac Address
 * @host_rch_time: Buffer to return Host Reachability Time
 * @adapter_rch_time: Buffer to return Adapter Reachability Time
 * @valid_flag: Buffer to return valid flag for NCE
 * @vlan_id: Buffer to return Valn Id
 *
 * Gets the values from NCE Query Response. This function returns synchronously.
 */
typedef void (*vxge_hal_nce_query_resp_get_f)(
			vxge_hal_up_msg_h umsgh,
			vxge_hal_opaque_handle_t **nce_opaque_handle,
			u64 *src_mac_addr,
			u64 *rem_mac_addr,
			u32 *host_rch_time,
			u32 *adapter_rch_time,
			u32 *valid_flag,
			u32 *vlan_id);

/*
 * vxge_hal_nce_delete_f - Delete NCE from the adapter
 *
 * @vpath_handle: Handle returned by virtual path open
 * @obj_id: Object Id of NCE
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 *
 * Delete an NCE on the adapter. If this function returns VXGE_HAL_PENDING, then
 * the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_nce_delete_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t obj_id,
			vxge_hal_opaque_handle_t *opaque_handle);

/*
 * struct vxge_hal_nce_mgmt_if_t - Functions used to manage NCE
 * @vxge_hal_nce_create: Create NCE
 * @vxge_hal_nce_update_mac_addr: Update NCE MAC Address
 * @vxge_hal_nce_update_rch_time: Update NCE RCH Time
 * @vxge_hal_nce_query: Query NCE
 * @vxge_hal_nce_query_resp_get: Get values from NCE Query Response
 * @vxge_hal_nce_delete: Delete NCE from the adapter
 *
 * This structure contains the functions for managing the NCE
 */
typedef struct vxge_hal_nce_mgmt_if_t {
	vxge_hal_nce_create_f			vxge_hal_nce_create;
	vxge_hal_nce_update_mac_addr_f		vxge_hal_nce_update_mac_addr;
	vxge_hal_nce_update_rch_time_f		vxge_hal_nce_update_rch_time;
	vxge_hal_nce_query_f			vxge_hal_nce_query;
	vxge_hal_nce_query_resp_get_f		vxge_hal_nce_query_resp_get;
	vxge_hal_nce_delete_f			vxge_hal_nce_delete;
} vxge_hal_nce_mgmt_if_t;

/*
 * STAGS
 */

/*
 * vxge_hal_stag_allocate_nsmr_f - Allocate Non-Shared Memory Region
 *
 * @vpath_handle: Handle returned by virtual path open
 * @obj_id: Object Id of STAG
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @pd_id: Protection domain to which the new NSMR belongs
 * @num_of_pbles: The number of PBLEs that the new NSMR must be able
 *		to accommodate.
 * @stag_index: The STAG Index of the NSMR to be created.
 * @remote_access_flag: Flag to Specify if this NSMR will be allowed to be used
 *		for any remote access operations (read,write,invalidate) when it
 *		is subsequently registered.
 *
 * Allocates Non-Shared Memory Region on the adapter. If this function returns
 * VXGE_HAL_PENDING, then the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_stag_allocate_nsmr_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t obj_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			u64 pd_id,
			u32 num_of_pbles,
			u32 stag_index,
			u32 remote_access_flag);

/*
 * vxge_hal_stag_allocate_nsmr_resp_get_f - Get the values from STAG Allocate
 *					NSMR Response
 *
 * @umsgh: Handle to umsg.
 * @stag_index: Buffer to return the STAG Index of the NSMR to be created.
 * @num_of_pbles: Buffer to return the number of PBLEs that the newly-allocated
 *		NSMR is capable of accommodating.
 *
 * Get the valueas from STAG Allocate NSMR Response used to communicate
 * the success or failure of the corresponding Allocate NSMR Request. This
 * function returns synchronously.
 */
typedef void (*vxge_hal_stag_allocate_nsmr_resp_get_f)(
			vxge_hal_up_msg_h umsgh,
			u32 *stag_index,
			u32 *num_of_pbles);

/*
 * vxge_hal_stag_allocate_mw_f - Allocate Memory Window
 *
 * @vpath_handle: Handle returned by virtual path open
 * @obj_id: Object Id of STAG
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @stag_index: The host assigned STag index of the MW being allocated on
 *		the adapter.
 * @pd_id: Protection domain to which the new NSMR belongs
 *
 * Allocate Memory Window on the adapter. If this function returns
 * VXGE_HAL_PENDING, then the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_stag_allocate_mw_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t obj_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			u32 stag_index,
			u64 pd_id);

/*
 * vxge_hal_stag_allocate_mw_resp_get_f - Get the values from STAG Allocate
 *				  MW Response
 *
 * @umsgh: Handle to umsg.
 * @stag_index: Buffer to return the host assigned STag index of the MW being
 *		allocated on the adapter.
 *
 * Get the values from STAG Allocate MW Response used to communicate the success
 * or failure of the corresponding STAG Allocate MW Request. This function
 * returns synchronously.
 */
typedef void (*vxge_hal_stag_allocate_mw_resp_get_f)(
			vxge_hal_up_msg_h umsgh,
			u32 *stag_index);

/*
 * struct vxge_hal_stag_register_nsmr_params_t - Register NSMR Request params
 *
 * @stag_index: The host assigned STag index to associate with the NSMR being
 *		registered on the adapter.
 * @stag_key: Instance identifier for this validation of this NSMR
 * @pble_size: Size in bytes of each physical buffer in the PBL.
 * @pd_id: Protection domain to which the new NSMR belongs
 * @pble_list: PB List prepared from Physicall addresses
 * @va_base_flag: Flag to Specify if the NSMR will use virtual addressing
 * @remote_access_flag: Flag to Specify if this NSMR will be allowed to be used
 *		for any remote access operations (read, write, invalidate) on
 *		this and subsequent registrations.
 * @local_write_flag: Flag to Specify if the adapter can write to this NSMR.
 * @local_read_flag: Flag to Specify if the adapter can read from this NSMR.
 * @remote_write_flag: Flag to Specify if remote peers are allowed to write into
 *		this NSMR
 * @remote_read_flag: Flag to Specify if remote peers are allowed to read from
 *		this NSMR.
 * @enable_mw_binding_flag: Flag to Specify if this NSMR may have MWs bound to
 *		it by subsequent PostSQ.BindMW work requests.
 * @first_byte_offset: Number of bytes into the first PBLE where the MR begins.
 * @mr_length: Size in bytes of the NSMR.
 * @base_to: Virtual address of the first byte of the NSMR.
 *
 * This structure is used to pass parameters to Register NSMR routine.
 */
typedef struct vxge_hal_stag_register_nsmr_params_t {
			u32 stag_index;
			u32 stag_key;
			u32 pble_size;
			u64 pd_id;
			u64 pble_list;
			u32 va_base_flag;
			u32 remote_access_flag;
			u32 local_write_flag;
			u32 local_read_flag;
			u32 remote_write_flag;
			u32 remote_read_flag;
			u32 enable_mw_binding_flag;
			u32 first_byte_offset;
			u64 mr_length;
			u64 base_to;
} vxge_hal_stag_register_nsmr_params_t;

/*
 * vxge_hal_stag_register_nsmr_f - Register Non-Shared Memory Region
 *
 * @vpath_handle: Handle returned by virtual path open
 * @obj_id: Object Id of STAG
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @params: Parameters
 *
 * Register Non-Shared Memory Region function. If this function returns
 * VXGE_HAL_PENDING, then the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_stag_register_nsmr_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t obj_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			vxge_hal_stag_register_nsmr_params_t *params);

/*
 * vxge_hal_stag_register_nsmr_resp_get_f - Get the values from STAG Register
 *					NSMR Response
 *
 * @umsgh: Handle to umsg.
 * @stag_index: Buffer to return the host assigned STag index of the NSMR that
 *		was registered via the original call.This is returned unmodified
 *		from the call for convenience.
 * @stag_key: Buffer to return the host assigned STag key of the NSMR that was
 *		registered via the original call.This is returned unmodified
 *		from the call for convenience.
 * @num_of_pbles: Buffer to return the number of PBLEs that the newly-allocated
 *		NSMR is capable of accommodating.
 *
 * Get the values from STAG Register NSMR Response used to communicate the
 * success or failure of the corresponding STAG Register NSMR. This function
 * returns synchronously.
 */
typedef void (*vxge_hal_stag_register_nsmr_resp_get_f)(
			vxge_hal_up_msg_h umsgh,
			u32 *stag_index,
			u32 *stag_key,
			u32 *num_of_pbles);

/*
 * vxge_hal_stag_re_register_nsmr_f - Re-register Non-Shared Memory Region
 *
 * @vpath_handle: Handle returned by virtual path open
 * @obj_id: Object Id of STAG
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @params: Parameters.
 *
 * Re-register Non-Shared Memory Region function. If this function returns
 * VXGE_HAL_PENDING, then the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_stag_re_register_nsmr_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t obj_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			vxge_hal_stag_register_nsmr_params_t *params);

/*
 * vxge_hal_stag_re_register_nsmr_resp_get_f - Get the values from STAG
 *					  Re_register NSMR Response
 *
 * @umsgh: Handle to umsg.
 * @stag_index: Buffer to return the STag index of the original NSMR that is
 *		being re-registered.
 * @stag_key: Buffer to return the Instance identifier for this validation of
 *		this NSMR
 * @num_of_pbles: Buffer to return the number of PBLEs that the newly-allocated
 *		NSMR is capable of accommodating.
 *
 * Get the values from STAG Re_register NSMR Response used to communicate the
 * success or failure of the corresponding STAG Re_register NSMR Request. This
 * function returns synchronously.
 */
typedef void (*vxge_hal_stag_re_register_nsmr_resp_get_f)(
			vxge_hal_up_msg_h umsgh,
			u32 *stag_index,
			u32 *stag_key,
			u32 *num_of_pbles);

/*
 * vxge_hal_stag_register_smr_f - Register Shared Memory Region
 *
 * @vpath_handle: Handle returned by virtual path open
 * @obj_id: Object Id of STAG
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @base_mr_stag_index: Unique identifier for the underlying MR on which this
 *		SMR will be created.
 * @va_base_flag: Flag to Specify if the SMR will use virtual addressing
 * @remote_access_flag: Flag to Specify if this SMR will be allowed to be used
 *		for any remote access operations (read, write, invalidate) on
 *		this and subsequent registrations.
 * @local_write_flag: Flag to Specify if the adapter can write to this SMR.
 * @local_read_flag: Flag to Specify if the adapter can read from this SMR.
 * @remote_write_flag: Flag to Specify if remote peers are allowed to write into
 *		this SMR
 * @remote_read_flag: Flag to Specify if remote peers are allowed to read from
 *		this SMR.
 * @enable_mw_binding_flag: Flag to Specify if this SMR may have MWs bound to
 *		it by subsequent PostSQ.BindMW work requests.
 * @stag_index: STag Index of the new STag record that is to be created.
 * @stag_key: Instance identifier for this validation of this SMR
 * @pd_id: Protection domain to which the new SMR belongs
 * @base_to: Virtual address of the first byte of the SMR.
 *
 * Register Shared Memory Region routine. If this function returns
 * VXGE_HAL_PENDING, then the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_stag_register_smr_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t obj_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			u32 base_mr_stag_index,
			u32 va_base_flag,
			u32 remote_access_flag,
			u32 local_write_flag,
			u32 local_read_flag,
			u32 remote_write_flag,
			u32 remote_read_flag,
			u32 enable_mw_binding_flag,
			u32 stag_index,
			u32 stag_key,
			u64 pd_id,
			u64 base_to);

/*
 * vxge_hal_stag_register_smr_resp_get_f - Get the values from STAG Register
 *					 SMR Response
 *
 * @umsgh: Handle to umsg.
 * @stag_index: Buffer to return the host assigned STag index of the NSMR that
 *		was registered via the original call.
 * @stag_key: Buffer to return the host assigned STag key of the NSMR that was
 *		registered via the original call.
 *
 * Get the values from STAG Register SMR Response used to communicate the
 * success or failure of the corresponding STAG Register SMR Request. This
 * function returns synchronously.
 */
typedef void (*vxge_hal_stag_register_smr_resp_get_f)(
			vxge_hal_up_msg_h umsgh,
			u32 *stag_index,
			u32 *stag_key);

/*
 * vxge_hal_stag_query_f - Query stag on the adapter
 *
 * @vpath_handle: Handle returned by virtual path open
 * @obj_id: Object Id of STAG
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @stag_index: The STag index of the existing on-adapter STag that is being
 *		queried
 *
 * Queries an STAG information on the adapter. If this function returns
 * VXGE_HAL_PENDING, then the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_stag_query_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t obj_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			u32 stag_index);

/*
 * struct vxge_hal_stag_info_t - STAG Query Response Information
 *
 * @stag_index: Field uniquely identifies the STag record.
 * @stag_key: Instance identifier for the STag.
 * @is_allocated: Record in use flag
 * @is_window: STag a MR or an MW flag
 * @is_zero_based: STag zero-based flag
 * @is_valid: STag in the valid state flag
 * @is_shared: MR an SMR flag
 * @allow_bind: MWs be bound to this STag flag
 * @allow_remote_access: STag be used for remote access (read,write,invalidate)
 *		flag
 * @allow_remote_read: STag be used for remote read access flag
 * @allow_remote_write: STag be used for remote write access flag
 * @allow_local_read: STag be used for local read access flag
 * @allow_local_write: STag be used for local write access flag
 * @pd_id:  Protection Domain ID
 * @sid: Session ID
 * @sin:  Session Instance Number
 * @length: length (in bytes) of the MR or MW described by this STag.
 * @mr_base_to: TO of first byte in the MR
 * @mw_base_to: TO of the first byte in the MW relative to the underlying MR.
 * @fbo: The offset in bytes into the first physical buffer where the MR begins.
 * @pble_size: Size in bytes of each underlying physical buffer
 * @root_pbl_record_id: Identify of the Underlying PBL
 * @pbl_tree_depth: Depth of the PBL tree
 * @parent_stag_record_id: Identity of the parent STag
 * @parent_stag_valid: Parent record id flag
 * @left_subtree_valid: Left subtree for this STag record flag
 * @right_subtree_valid: Right subtree for this STag record flag
 * @left_subtree_record_id: Identity of the STag record at the top of the
 *		left subtree
 * @right_subtree_record_id: Identity of the STag record at the top of the
 *		right subtree
 * @num_mws_bound: The number of MWs bound to this MR
 * @subtree_imbalance: Sub tree depths different flag
 * @max_pbl_length: Maximum PBLE capacity of the current PBL tree.
 * @actual_pbl_length: Actual number of PBLEs in the current PBL tree.
 * @underlying_mr_record_id: Identity of the underlying MR
 *
 * The values from STAG Query Response message used by the adapter to
 * return the configuration of an STAG on the adapter
 */
typedef struct vxge_hal_stag_info_t {
	u32 stag_index;
	u32 stag_key;
	u32 is_allocated;
	u32 is_window;
	u32 is_zero_based;
	u32 is_valid;
	u32 is_shared;
	u32 allow_bind;
	u32 allow_remote_access;
	u32 allow_remote_read;
	u32 allow_remote_write;
	u32 allow_local_read;
	u32 allow_local_write;
	u64 pd_id;
	u32 sid;
	u32 sin;
	u64 length;
	u64 mr_base_to;
	u64 mw_base_to;
	u32 fbo;
	u32 pble_size;
	u32 root_pbl_record_id;
	u32 pbl_tree_depth;
	u32 parent_stag_record_id;
	u32 parent_stag_valid;
	u32 left_subtree_valid;
	u32 right_subtree_valid;
	u32 left_subtree_record_id;
	u32 right_subtree_record_id;
	u32 num_mws_bound;
	u32 subtree_imbalance;
	u32 max_pbl_length;
	u32 actual_pbl_length;
	u32 underlying_mr_record_id;
} vxge_hal_stag_info_t;

/*
 * vxge_hal_stag_query_resp_get_f - Get the values from STAG Query Response
 *
 * @umsgh: Handle to umsg.
 * @stag_info: Poniter to structure to return the query reposne values
 *
 * Get the values from STAG Query Response message used by the adapter to
 * return the configuration of an STAG on the adapter. This function returns
 * synchronously.
 */
typedef void (*vxge_hal_stag_query_resp_get_f)(
			vxge_hal_up_msg_h umsgh,
			vxge_hal_stag_info_t *stag_info);

/*
 * vxge_hal_stag_validate_f - Validate a stag
 *
 * @vpath_handle: Handle returned by virtual path open
 * @obj_id: Object Id of STAG
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @stag_index: The index of the existing on-adapter STag that is being
 *		validated.
 * @stag_key: The key that to assign to the STag when it is validated.
 *
 * Validate stag request. If this function returns VXGE_HAL_PENDING, then
 * the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_stag_validate_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t obj_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			u32 stag_index,
			u32 stag_key);

/*
 * vxge_hal_stag_validate_resp_get_f - Get the values from STAG Validate
 *				 Response
 *
 * @umsgh: Handle to umsg.
 * @stag_index: Buffer to return the host assigned STag index of the STag
 *		that was validated via the original call.
 * @stag_key: Buffer to return the host assigned STag key of the STag that
 *		was validated via the original call.
 *
 * Get the values from the STAG Validate Response used to communicate the
 * success or failure of the corresponding STAG Validate Request. This
 * function returns synchronously.
 */
typedef void (*vxge_hal_stag_validate_resp_get_f)(
			vxge_hal_up_msg_h umsgh,
			u32 *stag_index,
			u32 *stag_key);

/*
 * vxge_hal_stag_invalidate_f - Invalidate stag Request
 *
 * @vpath_handle: Handle returned by virtual path open
 * @obj_id: Object Id of STAG
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @stag_index: The index of the existing on-adapter STag that is being
 *		invalidated.
 * @stag_key: The key that is currently assigned to the STag.
 * @pd_id: The PD ID of the session or QP.
 *
 * Invalidate STAG Request. If this function returns VXGE_HAL_PENDING, then
 * the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_stag_invalidate_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t obj_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			u32 stag_index,
			u32 stag_key,
			u64 pd_id);

/*
 * vxge_hal_stag_invalidate_resp_get_f - Get the values from STAG InValidate
 *				  Response
 *
 * @umsgh: Handle to umsg.
 * @stag_index: Buffer to return he host assigned STag index of the STag that
 *		was invalidated.
 * @stag_key: Buffer to return the host assigned STag key of the STag that was
 *		invalidated.
 *
 * Get the values from the STAG InValidate Response used to communicate the
 * success or failure of the corresponding STAG InValidate Request. This
 * function returns synchronously.
 */
typedef void (*vxge_hal_stag_invalidate_resp_get_f)(
			vxge_hal_up_msg_h umsgh,
			u32 *stag_index,
			u32 *stag_key);

/*
 * enum vxge_hal_mw_ctrl_flags_e - Memory Window control flags
 *
 * @VXGE_HAL_MW_CTRL_VADDR_TYPE: Virtual Addressing Type
 * @VXGE_HAL_MW_CTRL_RW_ENABLE: Remote Write Enabled
 * @VXGE_HAL_MW_CTRL_RR_ENABLE:  Remote Read Enabled
 *
 * The flags that are valid for Memory window control
 */
typedef enum vxge_hal_mw_ctrl_flags_e {
	VXGE_HAL_MW_CTRL_VADDR_TYPE		= 0x1,
	VXGE_HAL_MW_CTRL_RW_ENABLE		= 0x2,
	VXGE_HAL_MW_CTRL_RR_ENABLE		= 0x4
} vxge_hal_mw_ctrl_flags_e;

/*
 * vxge_hal_stag_bind_mw_f - Bind Memory Window
 *
 * @vpath_handle: Handle returned by virtual path open
 * @obj_id: Object Id of STAG
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @mr_stag_index: The STag Index of the underlying MR to bind to.
 * @mr_stag_key: The STag Key of the underlying MR to bind to.
 * @mw_stag_index: The existing Memory Window's STag Index.
 * @mw_stag_key: The STag Key to use for the MW.
 * @mr_stag: The existing Memory Region's STAG index.
 *		Least significant 8 bits is the key.
 * @mw_ctrl: Control flags for Memory Window.
 *		3'bxx1-Virtual Addressing Type,
 *		3'bx1x Remote Write Enabled
 *		3'b1xx Remote Read Enabled
 * @mw_length: Memory window length in bytes.
 * @tagged_offset: Memory window tagged offset in bytes.
 *
 * Binds a memory window to a memory region. If this function returns
 * VXGE_HAL_PENDING, then the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_stag_bind_mw_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t obj_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			u32 mr_stag_index,
			u32 mr_stag_key,
			u32 mw_stag_index,
			u32 mw_stag_key,
			u32 mr_stag,
			u32 mw_ctrl,
			u64 mw_length,
			u64 tagged_offset);

/*
 * vxge_hal_stag_bind_mw_resp_Get_f - Get the values from STAG Bind MW
 *				Response
 *
 * @umsgh: Handle to umsg.
 * @stag_index: Buffer to return the host assigned STag index MW that was
 *		bound.
 * @stag_key: Buffer to return the host assigned STag key of the MW that
 *		was bound.
 * @result: Result code
 *
 * Get the values from STAG Bind MW Response used to communicate to the host
 * the success or failure of the corresponding STAG Bind MW Request. This
 * function returns synchronously.
 */
typedef void (*vxge_hal_stag_bind_mw_resp_get_f)(
			vxge_hal_up_msg_h umsgh,
			u32 *stag_index,
			u32 *stag_key);

/*
 * vxge_hal_stag_fast_register_nsmr_f - Fast Register NSMR Request
 *
 * @vpath_handle: Handle returned by virtual path open
 * @obj_id: Object Id of STAG
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @mr_control_size: Length in bytes of the Memory Region control structure.
 * @mr_control_ptr: Pointer to Memory Region control structure in host memory.
 *
 * Fast Register NSMR request. If this function returns VXGE_HAL_PENDING, then
 * the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_stag_fast_register_nsmr_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t obj_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			u32 mr_control_size,
			u64 mr_control_ptr);

/*
 * vxge_hal_stag_fast_register_nsmr_resp_get_f - Get values from the STAG Fast
 *						Register NSMR Response
 *
 * @umsgh: Handle to umsg.
 * @stag_index: Buffer to return the host assigned STag index of the NSMR that
 *		was fast-registered.
 * @stag_key: Buffer to return the host assigned STag key of the NSMR that was
 *		fast-registerd.
 *
 * Get the values from STAG Fast Register Response used to communicate the
 * success or failure of the corresponding STAG Fast Register Request. This
 * function returns synchronously.
 */
typedef void (*vxge_hal_stag_fast_register_nsmr_resp_get_f)(
			vxge_hal_up_msg_h umsgh,
			u32 *stag_index,
			u32 *stag_key);

/*
 * vxge_hal_stag_deallocate_f - Deallocate a STAG
 *
 * @vpath_handle: Handle returned by virtual path open
 * @obj_id: Object Id of STAG
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @stag_index: The index of the STag being de-allocated on the adapter.
 * @dealloc_pbl: Flag to Specify if the PBL that underlies this STag is
 *		also to be deallocated with the STag.
 *
 * Deallocates the specified STag, freeing up any on-adapter resources. This
 * function returns VXGE_HAL_PENDING, then the vpath callback will be called
 * upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_stag_deallocate_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t obj_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			u32 stag_index,
			u32 dealloc_pbl);

/*
 * vxge_hal_stag_deallocate_resp_get_f - Get the values from STAG deallocate
 *				  Response
 *
 * @umsgh: Handle to umsg.
 * @stag_index: Buffer to return The host assigned STag index of the
 *		 de-allocated STag
 *
 * Get the values from STAG Adeallocate Response used to communicate the success
 * or failure of the corresponding STAG deallocate Request.This function returns
 * synchronously.
 */
typedef void (*vxge_hal_stag_deallocate_resp_get_f)(
			vxge_hal_up_msg_h umsgh,
			u32 *stag_index);

/*
 * struct vxge_hal_stag_mgmt_if_t - Functions used to manage Stags
 * @vxge_hal_stag_allocate_nsmr: Function to allocate non shared memory region
 * @vxge_hal_stag_allocate_nsmr_resp_get: Function to get values from
 *		allocate non shared memory region response
 * @vxge_hal_stag_allocate_mw: Function to allocate memory window
 * @vxge_hal_stag_allocate_mw_resp_get: Function to get values from allocate
 *		memory window response
 * @vxge_hal_stag_register_nsmr: Function to register non shared memory region
 * @vxge_hal_stag_register_nsmr_resp_get: Function to get values from
 *		register non shared memory region response
 * @vxge_hal_stag_re_register_nsmr: Function to re-register non shared memory
 *		region
 * @vxge_hal_stag_re_register_nsmr_resp_get: Function to get values
 *		from re-register non shared memory region response
 * @vxge_hal_stag_register_smr: Function to register shared memory region
 * @vxge_hal_stag_register_smr_resp_get: Function to get values from
 *		a register shared memory region response
 * @vxge_hal_stag_query: Function to query stag
 * @vxge_hal_stag_query_resp_get: Function to get values from query stag
 *		response
 * @vxge_hal_stag_validate: Function to post validate stag
 * @vxge_hal_stag_validate_resp_get: Function used to get values from
 *		validate stag response
 * @vxge_hal_stag_invalidate: Function used to invalidate stag request
 * @vxge_hal_stag_invalidate_resp_get: Function used to get values from
 *		invalidate stag response
 * @vxge_hal_stag_bind_mw: Function to post bind memory window
 * @vxge_hal_stag_bind_mw_resp_get: Function to get values from bind memory
 *		window response
 * @vxge_hal_stag_fast_register_nsmr: Function to post fast register non shared
 *		memory region
 * @vxge_hal_stag_fast_register_nsmr_resp_get: Function to get values from
 *		fast register non shared memory region response
 * @vxge_hal_stag_deallocate: Function to deallocate stag
 * @vxge_hal_stag_deallocate_resp_get: Function to get values from deallocate
 *		stag reponse
 *
 * This structure contains the functions for managing the STAGs
 */
typedef struct vxge_hal_stag_mgmt_if_t {
    vxge_hal_stag_allocate_nsmr_f	   vxge_hal_stag_allocate_nsmr;
    vxge_hal_stag_allocate_nsmr_resp_get_f vxge_hal_stag_allocate_nsmr_resp_get;
    vxge_hal_stag_allocate_mw_f		   vxge_hal_stag_allocate_mw;
    vxge_hal_stag_allocate_mw_resp_get_f   vxge_hal_stag_allocate_mw_resp_get;
    vxge_hal_stag_register_nsmr_f	   vxge_hal_stag_register_nsmr;
    vxge_hal_stag_register_nsmr_resp_get_f vxge_hal_stag_register_nsmr_resp_get;
    vxge_hal_stag_re_register_nsmr_f	   vxge_hal_stag_re_register_nsmr;
    vxge_hal_stag_re_register_nsmr_resp_get_f
					vxge_hal_stag_re_register_nsmr_resp_get;
    vxge_hal_stag_register_smr_f	   vxge_hal_stag_register_smr;
    vxge_hal_stag_register_smr_resp_get_f  vxge_hal_stag_register_smr_resp_get;
    vxge_hal_stag_query_f		   vxge_hal_stag_query;
    vxge_hal_stag_query_resp_get_f	   vxge_hal_stag_query_resp_get;
    vxge_hal_stag_validate_f		   vxge_hal_stag_validate;
    vxge_hal_stag_validate_resp_get_f	   vxge_hal_stag_validate_resp_get;
    vxge_hal_stag_invalidate_f		   vxge_hal_stag_invalidate;
    vxge_hal_stag_invalidate_resp_get_f	   vxge_hal_stag_invalidate_resp_get;
    vxge_hal_stag_bind_mw_f		   vxge_hal_stag_bind_mw;
    vxge_hal_stag_bind_mw_resp_get_f	   vxge_hal_stag_bind_mw_resp_get;
    vxge_hal_stag_fast_register_nsmr_f	   vxge_hal_stag_fast_register_nsmr;
    vxge_hal_stag_fast_register_nsmr_resp_get_f
				    vxge_hal_stag_fast_register_nsmr_resp_get;
    vxge_hal_stag_deallocate_f		   vxge_hal_stag_deallocate;
    vxge_hal_stag_deallocate_resp_get_f	   vxge_hal_stag_deallocate_resp_get;
} vxge_hal_stag_mgmt_if_t;

/*
 * CQRQ
 */

/*
 * struct vxge_hal_cqrq_hw_cqe_Info_t - CQE contents
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @push: Return push flah
 * @send_type: Return the send type from CQE
 * @ino: Return ino flag
 * @immed_data_len: Return Immediate data length
 * @immed_data: Buffer to return immediate data
 * @stag: Return stag
 * @invalidate: Return invalidate flag
 * @msn: Return MSN
 * @bytes_placed: Return the number of bytes placed in the WQE
 *
 * This is used to return the contents of CQE
 */
typedef struct vxge_hal_cqrq_hw_cqe_Info_t {
	u64 opaque_handle;
	u32 push;
	u32 send_type;
	u32 ino;
	u32 immed_data_len;
	u8 *immed_data;
	u32 stag;
	u32 invalidate;
	u32 msn;
	u32 bytes_placed;
} vxge_hal_cqrq_hw_cqe_Info_t;

/*
 * function vxge_hal_cqrq_hw_cqe_callback_f - Callback routine for CQRQ.
 * @vpath_handle: Handle returned by virtual path open
 * @cqrq_id: Object Id of CQRQ that has CQE ready.
 * @cqe_info: Buffer to return CQE contents
 * @userdata: Pointer to user data passed when queue is created
 *
 * Callback function is called when there is CQE ready to be processed in a
 * CQRQ. The client can call vxge_hal_cqrq_next_ready_hw_cqe_get_f() function to
 * retrieve the next ready CQEs one by one in a loop.
 */
typedef vxge_hal_status_e (*vxge_hal_cqrq_hw_cqe_callback_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t cqrq_id,
			vxge_hal_cqrq_hw_cqe_Info_t *cqe_info,
			void *userdata);

/*
 * vxge_hal_cqrq_create_f - Create CQRQ queue
 *
 * @vpath_handle: Handle returned by virtual path open
 * @attr: CQRQ attributes in which the callback field set to
 *		vxge_hal_cqrq_hw_cqe_callback_f.
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @cq_opaque_handle: CQ Opaque handle to be returned in vpath callbacks
 *		that are generated unsolicitedly upon any events related to this
 *		CQRQ. The memory allocated for this handle should not be freed
 *		until the CQRQ is deleted successfully and delete confirmation
 *		is received throuh the vpath callback for delete function.
 *
 * Creates a CQRQ on the adapter. If this function returns VXGE_HAL_PENDING,
 * then the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_cqrq_create_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_cqrq_attr_t *attr,
			vxge_hal_opaque_handle_t *opaque_handle,
			vxge_hal_opaque_handle_t *cq_opaque_handle);

/*
 * vxge_hal_cqrq_modify_f - Modify a CQRQ with new configuration
 *
 * @vpath_handle: Handle returned by virtual path open
 * @cqrq_id: Object Id of CQRQ that has CQE ready.
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 *
 * Modify a CQRQ on the adapter with new configuration parameters in the HAL.
 * If this function returns VXGE_HAL_PENDING, then the vpath callback will be
 * called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_cqrq_modify_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t cqrq_id,
			vxge_hal_opaque_handle_t *opaque_handle);

/*
 * vxge_hal_cqrq_query_f - Query the information of a CQRQ
 *
 * @vpath_handle: Handle returned by virtual path open
 * @cqrq_id: Object Id of CQRQ that has CQE ready.
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 *
 * Query the information of a CQRQ on the adapter. If this function
 * returns VXGE_HAL_PENDING, then the vpath callback will be called
 * upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_cqrq_query_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t cqrq_id,
			vxge_hal_opaque_handle_t *opaque_handle);

/*
 * vxge_hal_cqrq_query_resp_get_f - Get the values from CQRQ Query Response
 *
 * @umsgh: Handle to umsg.
 * @cqrq_info: Pointer to structure to return values from CQRQ query response
 *
 * Gets the values from CQRQ Query Response. If this function returns
 * VXGE_HAL_PENDING, then the vpath callback will be called upon completion.
 */
typedef void (*vxge_hal_cqrq_query_resp_get_f)(
			vxge_hal_up_msg_h umsgh,
			vxge_hal_cqrq_info_t *cqrq_info);

/*
 * vxge_hal_cqrq_arm_f - Arm a CQRQ
 *
 * @vpath_handle: Handle returned by virtual path open
 * @cqrq_id: Object Id of CQRQ that has CQE ready.
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @arm_type: The armed state of the CQRQ.
 *
 * Changes the armed state of CQRQ on the adapter. If this function returns
 * VXGE_HAL_PENDING, then the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_cqrq_arm_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t cqrq_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			u32 arm_type);

/*
 * vxge_hal_cqrq_event_notification_get_f - Get values from CQRQ Event
 *						Notification
 *
 * @umsgh: Handle to umsg.
 * @event_type : Buffer to return Event Type
 *
 * Gets the values from CQRQ Event Notification. This function returns
 * synchronously.
 */
typedef void (*vxge_hal_cqrq_event_notification_get_f)(
			vxge_hal_up_msg_h umsgh,
			u32 *event_type);

/*
 * vxge_hal_cqrq_ready_hw_cqe_check_f - Check the CQRQ for any CQEs
 *		 currently indicated and available for processing
 * @vpath_handle: Handle returned by virtual path open
 * @cqrq_id: Object Id of CQRQ that has CQE ready.
 *
 * This function return (VXGE_HAL_OK) if there is a CQE available for
 * processing by the host. This function returns synchronously.
 */
typedef vxge_hal_status_e (*vxge_hal_cqrq_ready_hw_cqe_check_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t cqrq_id);

/*
 * vxge_hal_cqrq_next_ready_hw_cqe_get_f - Return the next ready CQE
 * @vpath_handle: Handle returned by virtual path open
 * @cqrq_id: Object Id of CQRQ that has CQE ready.
 * @cqe_info: Buffer to return CQE contents
 *
 * This function returns the next ready CQE in the CQRQ. This function
 * returns synchronously.
 */
typedef vxge_hal_status_e (*vxge_hal_cqrq_next_ready_hw_cqe_get_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t cqrq_id,
			vxge_hal_cqrq_hw_cqe_Info_t *cqe_info);

/*
 * vxge_hal_cqrq_delete_f - Delete CQRQ queue
 *
 * @vpath_handle: Handle returned by virtual path open
 * @cqrq_id: Object Id of CQRQ that has CQE ready.
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 *
 * Deletes an CQRQ on the adapter. If this function returns VXGE_HAL_PENDING,
 * then the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_cqrq_delete_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t cqrq_id,
			vxge_hal_opaque_handle_t *opaque_handle);

/*
 * struct vxge_hal_cqrq_mgmt_if_t - Functions used to manage CQRQ
 * @vxge_hal_cqrq_create: Function to create CQRQ
 * @vxge_hal_cqrq_modify: Function to modify a CQRQ
 * @vxge_hal_cqrq_query: Function to query a CQRQ
 * @vxge_hal_cqrq_query_resp_get:Function to get values from CQRQ query response
 * @vxge_hal_cqrq_arm: Function to arm CQRQ
 * @vxge_hal_cqrq_event_notification_get: Function to get values from CQRQ event
 *		notification
 * @vxge_hal_cqrq_ready_hw_cqe_check: Function to check if HW-CQE is ready
 * @vxge_hal_cqrq_next_ready_hw_cqe_get: Function to get next ready HW-CQE
 * @vxge_hal_cqrq_delete: Function to delete a CQRQ
 *
 * This structure contains the functions for managing the CQRQs
 */
typedef struct vxge_hal_cqrq_mgmt_if_t {
	vxge_hal_cqrq_create_f		   vxge_hal_cqrq_create;
	vxge_hal_cqrq_modify_f		   vxge_hal_cqrq_modify;
	vxge_hal_cqrq_query_f		   vxge_hal_cqrq_query;
	vxge_hal_cqrq_query_resp_get_f	   vxge_hal_cqrq_query_resp_get;
	vxge_hal_cqrq_arm_f		   vxge_hal_cqrq_arm;
	vxge_hal_cqrq_event_notification_get_f
			    vxge_hal_cqrq_event_notification_get;
	vxge_hal_cqrq_ready_hw_cqe_check_f vxge_hal_cqrq_ready_hw_cqe_check;
	vxge_hal_cqrq_next_ready_hw_cqe_get_f
			    vxge_hal_cqrq_next_ready_hw_cqe_get;
	vxge_hal_cqrq_delete_f		   vxge_hal_cqrq_delete;
} vxge_hal_cqrq_mgmt_if_t;

/*
 * SRQ
 */
/*
 * vxge_hal_srq_create_f - Create SRQ
 *
 * @vpath_handle: Handle returned by virtual path open
 * @attr: SRQ configuration parameters structure
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @srq_opaque_handle: SRQ Opaque handle to be returned in vpath callbacks
 *		that are generated unsolicitedly upon any events related to this
 *		SRQ. The memory allocated for this handle should not be freed
 *		until the SRQ is deleted successfully and delete confirmation
 *		is received throuh the vpath callback for delete function.
 *
 * Creates a SRQ on the adapter. If this function returns VXGE_HAL_PENDING, then
 * the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_srq_create_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_srq_attr_t *attr,
			vxge_hal_opaque_handle_t *opaque_handle,
			vxge_hal_opaque_handle_t *srq_opaque_handle);

/*
 * vxge_hal_srq_modify_f - Modify an SRQ with new configuration
 *
 * @vpath_handle: Handle returned by virtual path open
 * @srq_id: Object Id of SRQ
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 *
 * Modifies a SRQ on the adapter with new configuration.If this function returns
 * VXGE_HAL_PENDING, then the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_srq_modify_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t srq_id,
			vxge_hal_opaque_handle_t *opaque_handle);

/*
 * vxge_hal_srq_query_f - Query the information of SRQ
 *
 * @vpath_handle: Handle returned by virtual path open
 * @srq_id: Object Id of SRQ
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 *
 * Queries a SRQ on the adapter. If this function returns VXGE_HAL_PENDING, then
 * the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_srq_query_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t srq_id,
			vxge_hal_opaque_handle_t *opaque_handle);

/*
 * vxge_hal_srq_query_resp_get_f - Get the values from SRQ Query Response
 *
 * @umsgh: Handle to umsg.
 * @srq_info: Pointer to structure to return information in SRQ Query response
 *
 * Queries the configuration of an SRQ on the adapter. This function returns
 * synchronously.
 */
typedef void (*vxge_hal_srq_query_resp_get_f)(
			vxge_hal_up_msg_h umsgh,
			vxge_hal_srq_info_t *srq_info);

/*
 * vxge_hal_srq_arm_f - Arm a SRQ
 *
 * @vpath_handle: Handle returned by virtual path open
 * @srq_id: Object Id of SRQ
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 *
 * Changes the armed state of SRQ on the adapter. If this function returns
 * VXGE_HAL_PENDING, then the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_srq_arm_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t srq_id,
			vxge_hal_opaque_handle_t *opaque_handle);

/*
 * vxge_hal_srq_event_notification_get_f - Get values from
 *                       SRQ Event Notification
 *
 * @umsgh: Handle to umsg.
 * @event_type : Buffer to return Event Type
 *
 * Gets the values from SRQ Event Notification. This function returns
 * synchronously
 */
typedef void (*vxge_hal_srq_event_notification_get_f)(
			vxge_hal_up_msg_h umsgh,
			u32 *event_type);

/*
 * vxge_hal_srq_wqebadd_f - Add WQE Blocks to SRQ
 *
 * @vpath_handle: Handle returned by virtual path open
 * @srq_id: Object Id of SRQ
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @nblocks: Number of additional blocks
 *
 * This function is used to add additional WQE blocks to an SRQ.
 * If this function returns VXGE_HAL_PENDING, then the vpath callback
 * will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_srq_wqebadd_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t srq_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			u32 nblocks);

/*
 * vxge_hal_srq_hw_wqe_post_f - Post a WQE
 * @vpath_handle: Handle returned by virtual path open
 * @srq_id: Object Id of SRQ
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @sg_list: List of physical buffer to be set.
 * @num_sgle: Number of elements in sgle.
 *
 * This function posts a WQE for work request. This function
 * returns synchronously.
 */
typedef vxge_hal_status_e (*vxge_hal_srq_hw_wqe_post_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t srq_id,
			u64 opaque_handle,
			vxge_hal_sg_list_t *sg_list,
			u32 num_sgle);

/*
 * vxge_hal_srq_return_unused_wqes_f - Return unused WQEs Request
 *
 * @vpath_handle: Handle returned by virtual path open
 * @srq_id: Object Id of SRQ
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @number_of_wqes: Number of WQEs requested
 *
 * Requests the adapter to return unused WQEs. If this function returns
 * VXGE_HAL_PENDING, then the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_srq_return_unused_wqes_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t srq_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			u32 number_of_wqes);

/*
 * vxge_hal_srq_return_wqes_umsg_get_f - Get the values from SRQ Return WQEs
 *
 * @umsgh: Handle to umsg.
 * @srq_opaque_handle: Buffer to return SRQ opaque handle echoed in response
 * @number_of_wqes_in_msg: Buffer to return The number of WQE opaque handles
 *		in this message
 * @number_of_wqes_remaining: Buffer to return the number of WQE opaque handles
 *		left to send
 * @wqe_opaque_handles: Array to return the WQE Opaque Handles
 *
 * Get the values from SRQ Return WQEs used when deleting a TCP session or
 * tearing down an SRQ to return any WQEs that might still be in play for
 * the SRQ. This function returns synchronously.
 */
typedef void (*vxge_hal_srq_return_unused_wqes_resp_get_f)(
			vxge_hal_up_msg_h umsgh,
			vxge_hal_opaque_handle_t **srq_opaque_handle,
			u32 *number_of_wqes_in_msg,
			u32 *number_of_wqes_remaining,
			u64 *wqe_opaque_handles);

/*
 * vxge_hal_srq_delete_f - Delete SRQ
 *
 * @vpath_handle: Handle returned by virtual path open
 * @srq_id: Object Id of SRQ
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 *
 * Deletes an SRQ on the adapter. If this function returns VXGE_HAL_PENDING,then
 * the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_srq_delete_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t srq_id,
			vxge_hal_opaque_handle_t *opaque_handle);

/*
 * struct vxge_hal_srq_mgmt_if_t - Functions used to manage SRQ
 * @vxge_hal_srq_create: Function to create a SRQ
 * @vxge_hal_srq_modify: Function to modify a SRQ
 * @vxge_hal_srq_query: Function to query information of an SRQ
 * @vxge_hal_srq_query_resp_get: Function to get values from a SRQ query
 *		response
 * @vxge_hal_srq_arm: Function to arm a SRQ
 * @vxge_hal_srq_event_notification_get: Function to get values from
 *		SRQ event notification
 * @vxge_hal_srq_wqebadd: Function to add WQE block
 * @vxge_hal_srq_hw_wqe_post: Function to post a HW-WQE into CQRQ
 * @vxge_hal_srq_return_unused_wqes: Function to request the return of
 *		unused WQES
 * @vxge_hal_srq_return_unused_wqes_resp_get: Function to get values
 *		from return unused WQEs response
 * @vxge_hal_srq_delete: Function to delete a SRQ
 *
 * This structure contains the functions for managing the SRQs
 */
typedef struct vxge_hal_srq_mgmt_if_t {
	vxge_hal_srq_create_f			vxge_hal_srq_create;
	vxge_hal_srq_modify_f			vxge_hal_srq_modify;
	vxge_hal_srq_query_f			vxge_hal_srq_query;
	vxge_hal_srq_query_resp_get_f		vxge_hal_srq_query_resp_get;
	vxge_hal_srq_arm_f			vxge_hal_srq_arm;
	vxge_hal_srq_event_notification_get_f
				    vxge_hal_srq_event_notification_get;
	vxge_hal_srq_wqebadd_f			vxge_hal_srq_wqebadd;
	vxge_hal_srq_hw_wqe_post_f		vxge_hal_srq_hw_wqe_post;
	vxge_hal_srq_return_unused_wqes_f	vxge_hal_srq_return_unused_wqes;
	vxge_hal_srq_return_unused_wqes_resp_get_f
				    vxge_hal_srq_return_unused_wqes_resp_get;
	vxge_hal_srq_delete_f			vxge_hal_srq_delete;
} vxge_hal_srq_mgmt_if_t;

/*
 * SQ
 */
/*
 * function vxge_hal_sq_callback_f - Callback routine for SQ.
 * @vpath_handle: Handle returned by virtual path open
 * @sq_id: Object Id of SQ that has TOWI ready.
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @userdata: user data passed when the SQ is created.
 *
 * Callback function is called when there is TOWI is completed in a SQ. The
 * client can call vxge_hal_sq_next_completed_towi_get_f() function to retrieve
 * the next ready towis one by one in a loop.
 */
typedef vxge_hal_status_e (*vxge_hal_towi_callback_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t sq_id,
			u64 opaque_handle,
			void *userdata);

/*
 * struct vxge_hal_sq_attr_t - Configuration of TOWI List/SQ
 * @sq_length: The length of the SQ/TOWI List.
 * @towi_completion: The indication type when a TOWI is completed.
 * @rdma_read_completion:The indication type when RDMA READ complete.
 * @callback: Callback function to be called when a TOWI is completed.
 * @userdata: userdata to be passed back in callback
 *
 * Configuration of SQ. This structure represents the characteristics
 * of SQ. This structure needs to be filled in and passed to
 * vxge_hal_sq_create in attr to create send queue.
 *
 */
typedef struct vxge_hal_sq_attr_t {
	u32				sq_length;
#define	VXGE_HAL_SQ_MIN_LENGTH		0
#define	VXGE_HAL_SQ_MAX_LENGTH		8192

	u32				towi_completion;
	u32				rdma_read_completion;
#define	VXGE_HAL_SQ_GENERATE_INTERRUPT	0x10
#define	VXGE_HAL_SQ_UPDATE_HW_CQ_SQ	0x01
#define	VXGE_HAL_SQ_GENERATE_UPDATE	0x11

	vxge_hal_towi_callback_f		callback;

	void				*userdata;
} vxge_hal_sq_attr_t;

/*
 * vxge_hal_sq_create_f - Create a SQ
 * @vpath_handle: Handle returned by virtual path open
 * @attr: SQ configuration parameters structure
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 *
 * This function creates SQ. If this function returns VXGE_HAL_PENDING, then
 * the vpath callback will be called upon completion.
 *
 */
typedef vxge_hal_status_e (*vxge_hal_sq_create_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_sq_attr_t *attr,
			vxge_hal_opaque_handle_t *opaque_handle);

/*
 * vxge_hal_send_towi_post_f - Post Send TOWIs from stag or buf ptr
 * @vpath_handle: Handle returned by virtual path open
 * @sq_id: Object Id of SQ
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @push: Flag to include push in TOWI.
 * @msn: Message Sequence Number
 * @last: Is it last message in sequence
 * @start_msg_offset: Starting message offset.
 * @read_fencing: Flag to indicate read Fencing
 * @remote_invalidate: Flag to Invalidate Remote Stag
 * @solicited_event: Flag for solicited event
 * @local_stag: Local stag where payload is located.
 * @ptr_to: Physical address of a buffer or tagged offset
 * @length: length of data
 * @remote_stag: Remote stag to write to.
 *
 * This function reserves, sets and posts write TOWI to SQ. This function
 * returns synchronously.
 *
 */
typedef vxge_hal_status_e (*vxge_hal_send_towi_post_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t sq_id,
			u64 opaque_handle,
			u32 push,
			u32 msn,
			u32 last,
			u32 start_msg_offset,
			u32 read_fencing,
			u32 remote_invalidate,
			u32 solicited_event,
			u32 local_stag,
			u32 ptr_to,
			u32 length,
			u32 remote_stag);

/*
 * vxge_hal_send_towi_odlist_post_f - Post Send TOWIs from od list
 * @vpath_handle: Handle returned by virtual path open
 * @sq_id: Object Id of SQ
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @push: Flag to include push in TOWI.
 * @msn: Message Sequence Number
 * @last: Is it last message in sequence
 * @start_msg_offset: Starting message offset.
 * @read_fencing: Flag to indicate read Fencing
 * @remote_invalidate: Flag to Invalidate Remote Stag
 * @solicited_event: Flag for solicited event
 * @remote_stag: Remote stag to write to.
 * @odlist: OD List prepared
 *
 * This function reserves, sets and posts write TOWI to SQ. This function
 * returns synchronously.
 *
 */
typedef vxge_hal_status_e (*vxge_hal_send_towi_odlist_post_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t sq_id,
			u64 opaque_handle,
			u32 push,
			u32 msn,
			u32 last,
			u32 start_msg_offset,
			u32 read_fencing,
			u32 remote_invalidate,
			u32 solicited_event,
			u32 remote_stag,
			vxge_hal_odlist_t *odlist);

/*
 * vxge_hal_write_towi_post_f - Post Write TOWIs with local stag or buffer ptr
 * @vpath_handle: Handle returned by virtual path open
 * @sq_id: Object Id of SQ
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @push: Flag to include push in TOWI.
 * @last: Is it last message in sequence
 * @read_fencing: Flag to indicate read Fencing
 * @local_stag: Local stag where payload is located.
 * @ptr_to: Physical address of a buffer or tagged offset
 * @length: length of data
 * @remote_stag: Remote stag to write to.
 * @remote_tagged_offset: Tagged Offset in remote buffer.
 *
 * This function reserves, sets and posts write TOWI to SQ.  This function
 * returns synchronously.
 *
 */
typedef vxge_hal_status_e (*vxge_hal_write_towi_post_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t sq_id,
			u64 opaque_handle,
			u32 push,
			u32 last,
			u32 read_fencing,
			u32 local_stag,
			u64 ptr_to,
			u32 length,
			u32 remote_stag,
			u64 remote_tagged_offset);

/*
 * vxge_hal_write_towi_odlist_post_f - Post Write TOWIs using odlist
 * @vpath_handle: Handle returned by virtual path open
 * @sq_id: Object Id of SQ
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @push: Flag to include push in TOWI.
 * @last: Is it last message in sequence
 * @read_fencing: Flag to indicate read Fencing
 * @remote_stag: Remote stag to write to.
 * @remote_tagged_offset: Tagged Offset in remote buffer.
 * @odlist: OD List prepared early
 *
 * This function reserves, sets and posts write TOWI to SQ.  This function
 * returns synchronously.
 *
 */
typedef vxge_hal_status_e (*vxge_hal_write_towi_odlist_post_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t sq_id,
			u64 opaque_handle,
			u32 push,
			u32 last,
			u32 read_fencing,
			u32 remote_stag,
			u64 remote_tagged_offset,
			vxge_hal_odlist_t *odlist);

/*
 * vxge_hal_read_towi_post_f - Post Read TOWIs
 * @vpath_handle: Handle returned by virtual path open
 * @sq_id: Object Id of SQ
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @push: Flag to include push in TOWI.
 * @msn: Message sequence number of message.
 * @invalidate_local_stag: Flag to invalidate Local Stag
 * @local_fencing: Flag to invalidate Local Fencing
 * @local_stag: Local stag where RDMA Read Request is located.
 * @length: Length of Physical buffer containing RDMA Read Request
 * @ptr_to: Physical buffer containing RDMA Read Request(if
 *		local_stag=0) or tagged offset(if local_stag > 0)
 *
 * This function reserves, sets and posts Read TOWI to SQ.  This function
 * returns synchronously.
 *
 */
typedef vxge_hal_status_e (*vxge_hal_read_towi_post_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t sq_id,
			u64 opaque_handle,
			u32 push,
			u32 msn,
			u32 invalidate_local_stag,
			u32 local_fencing,
			u32 local_stag,
			u32 length,
			u64 ptr_to);

/*
 * vxge_hal_bind_mw_towi_post_f - Post Bind Memory Window TOWI
 * @vpath_handle: Handle returned by virtual path open
 * @sq_id: Object Id of SQ
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @mw_stag_index: The existing Memory Window's STag Index.
 * @mw_stag_key: The STag Key to use for the MW.
 * @mr_stag: The existing Memory Region's STAG index.
 *		Least significant 8 bits is the key.
 * @mw_ctrl: Control flags for Memory Window.
 * @mw_length: Memory window length in bytes.
 * @tagged_offset: Memory window tagged offset in bytes.
 *
 * This function reserves, sets and posts Bind Memory Window TOWI to SQ.  This
 * function returns synchronously.
 *
 */
typedef vxge_hal_status_e (*vxge_hal_bind_mw_towi_post_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t sq_id,
			u64 opaque_handle,
			u32 mw_stag_index,
			u32 mw_stag_key,
			u32 mr_stag,
			u32 mw_ctrl,
			u64 mw_length,
			u64 tagged_offset);

/*
 * vxge_hal_fmr_towi_post_f - Post Fast Memory Registration TOWI
 * @vpath_handle: Handle returned by virtual path open
 * @sq_id: Object Id of SQ
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @stag_index: Stag Index
 * @stag_key: Stag Key
 * @pb_size: Size of each physical buffer
 * @pble_list: pble list allocated previously
 * @local_write_flag: Flag to enable local write
 * @remote_write_flag: Flag to enable remote write
 * @local_read_flag: Flag to enable local read
 * @remote_read_flag: Flag to enable remote read
 * @enable_mw_binding_flag: Flag to enable MW binding
 * @first_byte_offset: Byte offset of first byte in the first physical buffer
 * @mr_length: Length of memory region
 * @base_to: Base To of virtual Address
 *
 * This function reserves, sets and posts Fast Memory Registration TOWI to SQ.
 * This function returns synchronously.
 */
typedef vxge_hal_status_e (*vxge_hal_fmr_towi_post_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t sq_id,
			u64 opaque_handle,
			u32 stag_index,
			u32 stag_key,
			u32 pb_size,
			u64 pble_list,
			u32 local_write_flag,
			u32 remote_write_flag,
			u32 local_read_flag,
			u32 remote_read_flag,
			u32 enable_mw_binding_flag,
			u32 first_byte_offset,
			u64 mr_length,
			u64 base_to);

/*
 * vxge_hal_local_invalidate_towi_post_f - Post Local Invalidate MR or MW TOWI
 * @vpath_handle: Handle returned by virtual path open
 * @sq_id: Object Id of SQ
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @stag: Stag to be invalidated.
 * @local_fencing: Falg to enable local fencing.
 *
 * This function reserves, sets and posts Local Invalidate MR or MW TOWI to SQ.
 * This function returns synchronously.
 */
typedef vxge_hal_status_e (*vxge_hal_local_invalidate_towi_post_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t sq_id,
			u64 opaque_handle,
			u32 stag,
			u32 local_fencing);

/*
 * vxge_hal_terminate_towi_post_f - Post Terminate TOWIs
 * @vpath_handle: Handle returned by virtual path open
 * @sq_id: Object Id of SQ
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @push: Flag to include push in TOWI.
 * @msn: Message sequence number of message.
 * @local_stag: Local stag where TERMINATE Request is located.
 * @length: Length of Physical buffer containing TERMINATE Request
 * @ptr_to: Physical buffer containing TERMINATE Request(if
 *		local_stag=0) or tagged offset(if local_stag > 0)
 *
 * This function reserves, sets and posts Terminate TOWI to SQ.
 * This function returns synchronously.
 */
typedef vxge_hal_status_e (*vxge_hal_terminate_towi_post_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t sq_id,
			u64 opaque_handle,
			u32 push,
			u32 msn,
			u32 local_stag,
			u32 length,
			u64 ptr_to);

/*
 * vxge_hal_bytestream_towi_post_f - Post Bytestream TOWI from buffer pointer
 * @vpath_handle: Handle returned by virtual path open
 * @sq_id: Object Id of SQ
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @push: Flag to include push in TOWI.
 * @ptr: Physical address of buffer
 * @length: Length of buffer
 *
 * This function reserves, sets and posts bytestream TOWI to SQ.
 * This function returns synchronously.
 */
typedef vxge_hal_status_e (*vxge_hal_bytestream_towi_post_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t sq_id,
			u64 opaque_handle,
			u32 push,
			u64 ptr,
			u32 length);

/*
 * vxge_hal_bytestream_towi_odlist_post_f - Post Bytestream TOWI from od list
 * @vpath_handle: Handle returned by virtual path open
 * @sq_id: Object Id of SQ
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @push: Flag to include push in TOWI.
 * @odlist: OD List prepared earlier
 *
 * This function reserves, sets and posts bytestream TOWI to SQ.
 * This function returns synchronously.
 */
typedef vxge_hal_status_e (*vxge_hal_bytestream_towi_odlist_post_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t sq_id,
			u64 opaque_handle,
			u32 push,
			vxge_hal_odlist_t *odlist);

/*
 * vxge_hal_tcp_fin_towi_post_f - Post TCP Fin TOWI
 * @vpath_handle: Handle returned by virtual path open
 * @sq_id: Object Id of SQ
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @push: Flag to include push in TOWI.
 *
 * This function reserves, sets and posts TCP Fin TOWI to SQ.
 * This function returns synchronously.
 */
typedef vxge_hal_status_e (*vxge_hal_tcp_fin_towi_post_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t sq_id,
			u64 opaque_handle,
			u32 push);

/*
 * vxge_hal_sq_completed_towi_check_f - Check the HW_CQ_SQ_IR for
 *		 completed TOWIs
 * @vpath_handle: Handle returned by virtual path open
 * @sq_id: Object Id of SQ
 *
 * This function return (VXGE_HAL_OK) if a new TOWI completed.
 * This function returns synchronously.
 */
typedef vxge_hal_status_e (*vxge_hal_sq_completed_towi_check_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t sq_id);

/*
 * vxge_hal_sq_next_completed_towi_get_f - Return the next completed TOWI
 * @vpath_handle: Handle returned by virtual path open
 * @sq_id: Object Id of SQ
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 *
 * This function returns the next completed TOWI in the SQ. This function
 * returns synchronously.
 */
typedef vxge_hal_status_e (*vxge_hal_sq_next_completed_towi_get_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t sq_id,
			u64 *opaque_handle);

/*
 * vxge_hal_sq_delete_f - Delete a SQ
 * @vpath_handle: Handle returned by virtual path open
 * @sq_id: Object Id of SQ
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 *
 * This function deletes SQ. If this function returns VXGE_HAL_PENDING, then
 * the vpath callback will be called upon completion.
 *
 */
typedef vxge_hal_status_e (*vxge_hal_sq_delete_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t sq_id,
			vxge_hal_opaque_handle_t *opaque_handle);

/*
 * struct vxge_hal_sq_mgmt_if_t - Functions used to manage SQ
 * @vxge_hal_sq_create: Function to create a sq
 * @vxge_hal_send_towi_post: Function used to post a send towi
 * @vxge_hal_send_towi_odlist_post: Function used to post a send towi with od
 *		list
 * @vxge_hal_write_towi_post: Function used to post a write towi
 * @vxge_hal_write_towi_odlist_post: Function used to post a write towi with od
 *		list
 * @vxge_hal_read_towi_post: Function used to post a read towi
 * @vxge_hal_bind_mw_towi_post: Function used to post a bind memory window towi
 * @vxge_hal_fmr_towi_post: Function used to post a fast memory register towi
 * @vxge_hal_local_invalidate_towi_post: Function used to post local invalidate
 *		towi
 * @vxge_hal_terminate_towi_post: Function used to post terminate towi
 * @vxge_hal_bytestream_towi_post: Function used to post bytestream towi
 * @vxge_hal_bytestream_towi_odlist_post: Function used to post a bytestream
 *		towi with od list
 * @vxge_hal_tcp_fin_towi_post: Function used to post a fin towi
 * @vxge_hal_sq_completed_towi_check: Function used to check if any twois are
 *		completed
 * @vxge_hal_sq_next_completed_towi_get: Function used to get next completed
 *		towi
 * @vxge_hal_sq_delete: Function to delete a SQ
 *
 * This structure contains the functions for managing the SQs
 */
typedef struct vxge_hal_sq_mgmt_if_t {
	vxge_hal_sq_create_f		   vxge_hal_sq_create;
	vxge_hal_send_towi_post_f	   vxge_hal_send_towi_post;
	vxge_hal_send_towi_odlist_post_f   vxge_hal_send_towi_odlist_post;
	vxge_hal_write_towi_post_f	   vxge_hal_write_towi_post;
	vxge_hal_write_towi_odlist_post_f  vxge_hal_write_towi_odlist_post;
	vxge_hal_read_towi_post_f	   vxge_hal_read_towi_post;
	vxge_hal_bind_mw_towi_post_f	   vxge_hal_bind_mw_towi_post;
	vxge_hal_fmr_towi_post_f	   vxge_hal_fmr_towi_post;
	vxge_hal_local_invalidate_towi_post_f
				    vxge_hal_local_invalidate_towi_post;
	vxge_hal_terminate_towi_post_f	   vxge_hal_terminate_towi_post;
	vxge_hal_bytestream_towi_post_f	   vxge_hal_bytestream_towi_post;
	vxge_hal_bytestream_towi_odlist_post_f
				    vxge_hal_bytestream_towi_odlist_post;
	vxge_hal_tcp_fin_towi_post_f	   vxge_hal_tcp_fin_towi_post;
	vxge_hal_sq_completed_towi_check_f vxge_hal_sq_completed_towi_check;
	vxge_hal_sq_next_completed_towi_get_f
				    vxge_hal_sq_next_completed_towi_get;
	vxge_hal_sq_delete_f		   vxge_hal_sq_delete;
} vxge_hal_sq_mgmt_if_t;

/*
 * struct vxge_hal_tcp_open_params_t - Parameters for Post TCP Open Request Down
 *				Message routine
 *
 * @sq_id: Object Id of SQ
 * @srq_id: Object Id of BS SRQ
 * @cqrq_id: Object Id of BS CQRQQ
 * @active_opener: Flags to specify if active opener
 * @tcp_src_port: TCP source port
 * @tcp_dest_port: TCP destination port
 * @use_ipv6: Flag to specify if this session uses IPv6
 * @use_vlan: Indicates whether or not this session supports VLAN.
 * @vlan_id: The VLAN ID used by this session. Only valid if use_vlan == 1.
 * @ip_sa_low: Lower 64 bits of IPv4 or IPv6 source address
 * @ip_sa_high: Higher 64 bits of IPv4 or IPv6 source address
 * @ip_da_low: Lower 64 bits of IPv4 or IPv6 destination address
 * @ip_da_high: Higher 64 bits of IPv4 or IPv6 destination address
 * @break_merge_on_psh: If == 1, a merge will be broken if a frame with
 *		PSH == 1 is received.
 * @data_merging_enabled: Flag to enable merging of TCP payloads.
 * @send_vlan_up: Flag to force VLAN tag insertion on transmit frames
 *		for this session.
 * @wscale_ok: Flag to enable for window scaling.
 * @sack_ok: Flag to enable for SACKs.
 * @ecn_ok: Flag to enable for Explicit Congestion Notification
 * @tstamp_ok: Flag to enable for timestamps.
 * @kpalive_ok: Flag to enable for keepalive probes for this connection.
 * @nce_id: NCE # for this session.
 * @vlan_user_priority: Value to insert in the user_priority field when VLAN
 *		tags are generated on tx frames.
 * @dup_ack_thres: Duplicate ACK threshold for fast retransmit.  Default is 3.
 * @ip_tos: Value to place in the IPv4 TOS field or the Ipv6 Traffic Class field
 * @ip_flags: Value to place in the IPv4 Flags field.
 * @ip_flowlabel: Value to place in the IPv6 Flow Label field.
 * @kpa_idle_time: Keepalive idle timer
 * @kpa_retry_time: Keepalive retry time interval
 * @retx_time_max: maximum time to spent retransmitting one segment.
 * @snd_wscale: Sender window shift scale value
 * @kpa_probes_max: Keepalive maximum probe count
 * @snd_nxt: Next sequence number to send.
 * @snd_mss: Max segment size for transmit.
 * @snd_mss_limit: Value of MSS option received from far end on the SYN or
 *		SYN/ACK segment.
 * @rcv_nxt: Next byte expected to be received.
 * @rcv_wscale: Receiver window shift scale value
 * @rcv_mss: Max segment size for receive.
 * @rcv_buf_size: Maximum receive buffer size that the connection is allowed to
 *		consume.
 * @rcv_max_wnd: Maximum value for receive window to be advertised.
 * @rcv_buf: Lowest TCP sequence number buffered in the connection's receive
 *		buffer.
 * @rttmin: Minimum RTT value for this session.
 * @ts_recent: Timestamps Enabled
 *
 * Parameters structure to be passed to Open TCP routine to open a TCP
 * connection on the adapter.
 */
typedef struct vxge_hal_tcp_open_params_t {
	vxge_hal_obj_id_t sq_id;
	vxge_hal_obj_id_t srq_id;
	vxge_hal_obj_id_t cqrq_id;
	u32 active_opener;
	u32 tcp_src_port;
	u32 tcp_dest_port;
	u32 use_ipv6;
	u32 use_vlan;
	u32 vlan_id;
	u64 ip_sa_high;
	u64 ip_sa_low;
	u64 ip_da_high;
	u64 ip_da_low;
	u32 break_merge_on_psh;
	u32 data_merging_enabled;
	u32 send_vlan_up;
	u32 wscale_ok;
	u32 sack_ok;
	u32 ecn_ok;
	u32 tstamp_ok;
	u32 kpalive_ok;
	vxge_hal_obj_id_t nce_id;
	u32 vlan_user_priority;
	u32 dup_ack_thres;
	u32 ip_tos;
	u32 ip_flags;
	u32 ip_flowlabel;
	u32 kpa_idle_time;
	u32 kpa_retry_time;
	u32 retx_time_max;
	u32 snd_wscale;
	u32 kpa_probes_max;
	u32 snd_nxt;
	u32 snd_mss;
	u32 snd_mss_limit;
	u32 rcv_nxt;
	u32 rcv_wscale;
	u32 rcv_mss;
	u32 rcv_buf_size;
	u32 rcv_max_wnd;
	u32 rcv_buf;
	u32 rttmin;
	u32 ts_recent;
} vxge_hal_tcp_open_params_t;

/*
 * vxge_hal_tcp_open_f - Open a TCP session
 *
 * @vpath_handle: Handle returned by virtual path open
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @session_opaque_handle:Session Opaque handle to be returned in vpath callback
 *		that are generated unsolicitedly upon any events related to this
 *		Session.The memory allocated for this handle should not be freed
 *		until the Session is deleted successfully and delete
 *		confirmation is received throuh the vpath callback for delete
 *		function.
 * @params: Parameters for the TCP session
 *
 * Opens a TCP connection on the adapter. If this function returns
 * VXGE_HAL_PENDING, then the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_tcp_open_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_opaque_handle_t *opaque_handle,
			vxge_hal_opaque_handle_t *session_opaque_handle,
			vxge_hal_tcp_open_params_t *params);

/*
 * struct vxge_hal_tcp_promote_to_iwarp_params_t - Parameters for Post TCP
 *				Promote to iWARP Request Down Message
 *
 * @srq_id: Object Id of IWARP SRQ
 * @cqrq_id: Object Id of IWARP CQRQQ
 * @mpa_crc_valid: Flag to specify MPA CRCs are expected to be valid on
 *		received frames.
 * @tx_marker_status: Flag to specify markers exist in the transmit direction.
 * @rx_marker_status: Flag to specify markers exist in the receive direction.
 * @iwarp_local_stag_zero_allowed: Flag to specify the session is allowed to
 *		use STag=0 in Rx WQEs.
 * @iwarp_wire_stag_zero_allowed: Flag to specify session to accept STag=0 in
 *		receive frames.
 * @inbound_write_readresp_enable: Flag to specify the session has remote write
 *		permission, incoming RDMA WRITE or RDMA READ RESPONSE messages
 *		can be placed
 * @inbound_readreq_enable: Flag to specify when a SEND* WITH INVALIDATE message
 *		is received, the STag can only be invalidated if the session has
 *		remote write and/or remote read access.  Also used to ensure
 *		that RIRRs are allowed on this session
 * @srq_pd_id: Protection domain to which the associated iWARP SRQ belongs.
 *		This is needed to do access checking.
 * @iwarp_qp_pd_id: Protection domain to which this session belongs.
 * @cqrq_ll_send_frame_data_offset: Ooffset from the CQE base address.
 * @iwarp_qp_limit_value: Ifn-progress WQEs limit value.
 * @rq_max_length: maximum number of WQEs that this session is allowed to have
 *		in progress on this SRQ.
 * @iwarp_qp_limit_armed: Flag to specify the Queue Pair limit is armed at the
 *		iwarp_qp_limit_value
 * @rdmap_version:The RDMA version number 01 - IETF RNICs  00 - RDMAC RNICs
 * @ddp_version: Direct Data Placement Protocol Version
 *		01 - for protocol described in
 *		http://ietf.org/internet-drafts/draft-ietf-rddp-ddp-06.txt
 * @mpa_reply_frame_size: This field represents the size of TCP payload of MPA
 *		reply frame (in bytes) on the Responder. This value is relevant
 *		only for the MPA Responder. It allows the Tx PE to determine the
 *		TCP SN at which iWarp will begin (ie the TxIMSN).
 * @is_mpa_initiator: This allows the PE to take distinquish whether it is the
 *		MPA Initiator or the MPA Responder.
 * @block_permissions: Used to restrict session's ability to use special TOWI
 *		atoms. It is used to enforce checks such as preventing the QP
 *		from performing a Fast Register NSMR operation.
 * @ird_limit: Host supplied IRD limit for this session.
 *
 * Parameters for Promote TCP to iWARP  routine called in order to migrate
 * an existing bytestream session to iWARP mode.
 */
typedef struct  vxge_hal_tcp_promote_to_iwarp_params_t {
	vxge_hal_obj_id_t srq_id;
	vxge_hal_obj_id_t cqrq_id;
	u32 mpa_crc_valid;
	u32 tx_marker_status;
	u32 rx_marker_status;
	u32 iwarp_local_stag_zero_allowed;
	u32 iwarp_wire_stag_zero_allowed;
	u32 inbound_write_readresp_enable;
	u32 inbound_readreq_enable;
	u64 srq_pd_id;
	u64 iwarp_qp_pd_id;
	u32 cqrq_ll_send_frame_data_offset;
	u32 iwarp_qp_limit_value;
	u32 rq_max_length;
	u32 iwarp_qp_limit_armed;
	u32 rdmap_version;
	u32 ddp_version;
	u32 mpa_reply_frame_size;
	u32 is_mpa_initiator;
	u32 block_permissions;
	u32 ird_limit;
} vxge_hal_tcp_promote_to_iwarp_params_t;

/*
 * vxge_hal_tcp_promote_to_iwarp_req_dmsg_post_f - Post TCP Promote to iWARP
 *					 Request Down Message
 *
 * @vpath_handle: Handle returned by virtual path open
 * @session_id: Object Id of Session
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @params: Parameters for promote
 *
 * This routine is used to migrate an existing bytestream session to iWARP mode.
 * If this function returns VXGE_HAL_PENDING, then the vpath callback will be
 * called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_tcp_promote_to_iwarp_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t session_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			vxge_hal_tcp_promote_to_iwarp_params_t *params);

/*
 * vxge_hal_tcp_promote_to_iwarp_resp_get_f - Get the values from TCP Promote to
 *					 iWARP Response
 *
 * @umsgh: Handle to umsg.
 * @rx_imsn: Buffer to return 32b per-session constant that identifies TCP SN
 *
 * Get the valueas from TCP Promote to iWARP Response used to communicate
 * the success or failure of the corresponding TCP Promote to iWARP Request.
 * This function returns asynchronously.
 */
typedef void (*vxge_hal_tcp_promote_to_iwarp_resp_get_f)(
			vxge_hal_up_msg_h umsgh,
			u32 *rx_imsn);

/*
 * vxge_hal_tcp_modify_f - Modifies the TCP session
 *
 * @vpath_handle: Handle returned by virtual path open
 * @session_id: Object Id of Session
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @mpa_crc_valid: If set then MPA CRCs are expected to be valid on received
 *		frames. The adapter always calculates MPA CRCs in the transmit
 *		direction. This field is ignored for a bytestream session.
 * @tx_marker_mode: Whether MPA markers exist in the transmit direction or not.
 *		-- 1 - enable MPA markers in transmit frames
 *		-- 0 - disable MPA markers in transmit frames
 *		This field is ignored for a bytestream session.
 * @rx_marker_mode: Whether markers exist in the receive direction or not.
 *		-- 1 - enable MPA markers in receive frames
 *		-- 0 - disable MPA markers in receive frames
 *		If this is set to zero for an RDMA session, the RxPE drops all
 *		out of order receive segments. This field is ignored for a
 *		bytestream session.
 * @rdmap_version: The RDMA version number
 *		  -- 01 - for IETF RNICs
 *		  -- 00 - for RDMAC RNICs
 * @ddp_version: Direct Data Placement Protocol Version
 *		  -- 01 - for protocol described in
 *		http://ietf.org/internet-drafts/draft-ietf-rddp-ddp-06.txt
 * @iwarp_qp_limit_armed: If set, the Queue Pair limit is armed at the
 *		iwarp_qp_limit_value. This field is ignored for a bytestream
 *		session.
 * @iwarp_qp_limit_value: If the number of in-progress WQEs reaches this
 *		value and the Queue Pair limit is armed, the session produces
 *		an asynchronous event. This field is ignored for a bytestream
 *		session.
 * @rq_max_length: The maximum number of WQEs that this session is
 *		allowed to have in progress on this SRQ.
 * @block_permissions: Used to restrict session's ability to use special
 *		TOWI atoms. It is used to enforce checks such as preventing the
 *		QP from performing a Fast Register NSMR operation.
 * @ird_limit: Host supplied IRD limit for this session.
 *
 * Modifies the attributes associated with a bytestream or iWARP session.
 * If this function returns VXGE_HAL_PENDING, then the vpath callback will be
 * called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_tcp_modify_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t session_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			u32 mpa_crc_valid,
			u32 tx_marker_mode,
			u32 rx_marker_mode,
			u32 rdmap_version,
			u32 ddp_version,
			u32 iwarp_qp_limit_armed,
			u32 iwarp_qp_limit_value,
			u32 rq_max_length,
			u32 block_permissions,
			u32 ird_limit);

/*
 * vxge_hal_tcp_abort_f - Abort a TCP Session
 *
 * @vpath_handle: Handle returned by virtual path open
 * @session_id: Object Id of Session
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 *
 * Aborts a TCP session on the adapter. If this function returns
 * VXGE_HAL_PENDING, then the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_tcp_abort_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t session_id,
			vxge_hal_opaque_handle_t *opaque_handle);

/*
 * vxge_hal_tcp_update_rxwin_f - Update TCP Rxwin
 *
 * @vpath_handle: Handle returned by virtual path open
 * @session_id: Object Id of Session
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @rcv_buf: For bytestream sessions: lowest TCP sequence number buffered in
 *		the connection's receive buffer.  This is updated by the host
 *		as data is consumed by the ULP/application.  Must increase
 *		monotonically to avoid window shrinks (either check in PE when
 *		host updates it, or just make it a requirement of the host).
 *		For iWARP sessions: the DDP mainland RHS TCP sequence number.
 *		Represents the last aligned DDP mainland placement.  The DDP
 *		mainland edge differs from the TCP mainland edge (rcv_nxt)
 *		if MPA misaligned frames are arriving (ie session is ""DDP OOO""
 *		and iwarp_align_state != aligned).  This field is only updated
 *		when misaligned.  When aligned, including in the normal fast
 *		path DDP in order case, this is not updated to save fast path
 *		instructions; instead rcv_nxt defines both the TCP and DDP
 *		mainland edge.
 *		Supplied by host for passive open.  Initialized by adapter for
 *		active open on receipt of SYN/ACK or SYN segment.
 *		In both cases the proper initial value is IRS + 1
 *		(same as rcv_nxt). This is set equal to rcv_nxt on promote to
 *	        iWARP"
 * @rcv_buf_size: Maximum receive buffer size that the connection is allowed
 *		to consume.  Units are TCP payload bytes (ie sequence number
 *		bytes).
 * @rcv_max_wnd: Maximum value for receive window to be advertised.  This can
 *		be less than rcv_buf_size, so that in cases when the transmit
 *		rate is not limited by the receive application, the advertised
 *		window will remain constant. This is allows less frequent buffer
 *		usage updates from host without causing bursty behaviour from
 *		the sender.
 *
 * Updates TCP Rxwin.If the update fails, then the vpath callback will be called
 */
typedef vxge_hal_status_e (*vxge_hal_tcp_update_rxwin_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t session_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			u32 rcv_buf,
			u32 rcv_buf_size,
			u32 rcv_max_wnd);

/*
 * vxge_hal_tcp_update_mss_f - Updates MSS Request
 *
 * @vpath_handle: Handle returned by virtual path open
 * @session_id: Object Id of Session
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @snd_mss: Max segment size for transmit.This is defined as the maximum
 *		permitted TCP payload size,accounting for the maximum TCP option
 *		size (40B if both SACK and timestamps are enabled)
 *		  - snd_mss = MSS - max_tcp_options max_tcp_options is 40B
 *		assuming both SACK and timestamps are enabled (note that the
 *		firmware might choose to use a fixed value of 40 even if SACK or
 *		timestamps are not enabled - this is TBD). The definition is
 *		slightly different from most TCP stacks which store the true MSS
 *		value (max payload assuming no IP or TCP options). It allows the
 *		same amount of payload whether SACK is sent or not - important
 *		for sending aligned DDP segments. Most stacks typically pack the
 *		max possible data into a segment, and resegment on retransmit
 *		if the options size has grown.When sending the MSS option on SYN
 *		or SYN/ACK segments, the PE firmware calculates the true MSS
 *		value based on this field (ie MSS = snd_mss + max_tcp_options).
 *		Supplied by host for both active and passive open.In active open
 *		case, adapter may adjust value to the minimum of the host
 *		supplied value and the value from the received MSS option.
 * @snd_mss_limit: Value of MSS option received from far end on the
 *		SYN or SYN/ACK segment.The PE applies this as a limit on snd_mss
 *		when the host signals path MTU changes as follows -
 *		snd_mss = min(new_mss_from_host,snd_mss_limit) - max_tcp_options
 *		Supplied by host for passive open.
 *		Set by adapter for active open based on the MSS option value on
 *		the received SYN or SYN/ACK segment.
 *
 * Updates TCP MSS.  If the update fails, then the vpath callback will be called
 */
typedef vxge_hal_status_e (*vxge_hal_tcp_update_mss_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t session_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			u32 snd_mss,
			u32 snd_mss_limit);

/*
 * vxge_hal_tcp_update_ip_header_f - Update TCP IP Header
 *
 * @vpath_handle: Handle returned by virtual path open
 * @session_id: Object Id of Session
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @ip_tos: Value to place in the IPv4 TOS field or the Ipv6 Traffic Class
 *		field.  The 2 lower bits are overwritten if ECN is enabled.
 * @ip_flowlabel: Value to place in the IPv6 Flow Label field.  This might
 *		be moved to the H2L context.
 *
 * Updates TCP IP Header.  If the update fails, then the vpath callback will
 * be called.
 */
typedef vxge_hal_status_e (*vxge_hal_tcp_update_ip_header_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t session_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			u32 ip_tos,
			u32 ip_flowlabel);

/*
 * vxge_hal_tcp_update_keepalive_f - Update TCP  Keepalive
 *
 * @vpath_handle: Handle returned by virtual path open
 * @session_id: Object Id of Session
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @kpa_ok: Enable for keepalive probes for this connection.
 * @probes_max: Keepalive maximum probe count: maximum number of unanswered
 *		keepalive probes before giving up on session.
 * @idle_time: Keepalive idle timer: the idle time before sending the first
 *		keepalive probe (default is 2 hours).
 * @retry_time: Keepalive retry time interval: idle time before resending an
 *		unanswered keepalive probe.
 *
 * Update TCP Keepalive. If the update fails, then the vpath callback will be
 * called.
 */
typedef vxge_hal_status_e (*vxge_hal_tcp_update_keepalive_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t session_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			u32 kpa_ok,
			u32 probes_max,
			u32 idle_time,
			u32 retry_time);

/*
 * vxge_hal_tcp_relink_to_nce_f - Relinks TCP session to NCE
 *
 * @vpath_handle: Handle returned by virtual path open
 * @session_id: Object Id of Session
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 * @nce_id: NCE Id.
 *
 * Relinks a TCP session to NCE.  If this function returns VXGE_HAL_PENDING,
 * then the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_tcp_relink_to_nce_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t session_id,
			vxge_hal_opaque_handle_t *opaque_handle,
			u32 nce_id);

/*
 * vxge_hal_tcp_delete_f - Post TCP Delete Request Down Message
 *
 * @vpath_handle: Handle returned by virtual path open
 * @session_id: Object Id of Session
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 *
 * Deletes a TCP session on the adapter. If this function returns
 * VXGE_HAL_PENDING, then the vpath callback will be called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_tcp_delete_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t session_id,
			vxge_hal_opaque_handle_t *opaque_handle);

/*
 * struct vxge_hal_tcp_mgmt_if_t - Functions to manage TCP Session
 * @vxge_hal_tcp_open: Function to open a TCP session
 * @vxge_hal_tcp_promote_to_iwarp: Function to promote a TCP session to iWarp
 *		session
 * @vxge_hal_tcp_promote_to_iwarp_resp_get: Function used to get values from
 *		promote to iWarp response
 * @vxge_hal_tcp_modify: Function to modify TCP session
 * @vxge_hal_tcp_abort: Function to abort a TCP session
 * @vxge_hal_tcp_update_rxwin: Function to update TCP RX window
 * @vxge_hal_tcp_update_mss: Function to update TCP MSS
 * @vxge_hal_tcp_update_ip_header: Function to update TCP IP header request
 * @vxge_hal_tcp_update_keepalive: Function to keep TCP Session alive
 * @vxge_hal_tcp_relink_to_nce: Function to relink a TCP session to NCE
 * @vxge_hal_tcp_delete: Function to delete TCP session
 *
 * This structure contains the functions for managing the TCP Session
 */
typedef struct vxge_hal_tcp_mgmt_if_t {
	vxge_hal_tcp_open_f		    vxge_hal_tcp_open;
	vxge_hal_tcp_promote_to_iwarp_f	    vxge_hal_tcp_promote_to_iwarp;
	vxge_hal_tcp_promote_to_iwarp_resp_get_f
				    vxge_hal_tcp_promote_to_iwarp_resp_get;
	vxge_hal_tcp_modify_f		    vxge_hal_tcp_modify;
	vxge_hal_tcp_abort_f		    vxge_hal_tcp_abort;
	vxge_hal_tcp_update_rxwin_f	    vxge_hal_tcp_update_rxwin;
	vxge_hal_tcp_update_mss_f	    vxge_hal_tcp_update_mss;
	vxge_hal_tcp_update_ip_header_f	    vxge_hal_tcp_update_ip_header;
	vxge_hal_tcp_update_keepalive_f	    vxge_hal_tcp_update_keepalive;
	vxge_hal_tcp_relink_to_nce_f	    vxge_hal_tcp_relink_to_nce;
	vxge_hal_tcp_delete_f		    vxge_hal_tcp_delete;
} vxge_hal_tcp_mgmt_if_t;

/*
 * Session
 */

/*
 * vxge_hal_session_query_f - Query a session parameters
 *
 * @vpath_handle: Handle returned by virtual path open
 * @session_id: Object Id of session
 * @opaque_handle: Opaque handle returned in vpath callback upon completion of
 *		this function. The memory allocated for this handle can be freed
 *		up on receiving the vpath callback for this function. (Note that
 *		vpath callback will be invoked only if this function returns
 *		VXGE_HAL_PENDING)
 *
 * Queries the information about a session on the adapter.
 * If this function returns VXGE_HAL_PENDING, then the vpath callback will be
 * called upon completion.
 */
typedef vxge_hal_status_e (*vxge_hal_session_query_f)(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_obj_id_t session_id,
			vxge_hal_opaque_handle_t *opaque_handle);

/*
 * vxge_hal_session_query_resp_get_f - Get the values from a Session query
 *					  response
 *
 * @umsgh: Handle to umsg.
 * @sess_info: Pointer to structure to return the session attributes
 *
 * Get the values from the Session Query Response to return the attributes
 * associated with the specified session. This function returns synchronously.
 */
typedef void (*vxge_hal_session_query_resp_get_f)(
			vxge_hal_up_msg_h umsgh,
			vxge_hal_session_info_t *sess_info);
/*
 * vxge_hal_session_return_in_progress_wqes_get_f - Get the values from
 *			 Session Return in progress WQEs Message
 *
 * @umsgh: Handle to umsg.
 * @number_of_wqes_in_msg: Buffer to return The number of WQE opaque handles
 *		in this message
 * @number_of_wqes_remaining: Buffer to return the number of WQE opaque handles
 *		left to send
 * @wqe_opaque_handles: Array to return the WQE Opaque Handles
 *
 * Get the values from Session Return in progress WQEs used when deleting a
 * session to return any WQEs that might still be in play for the SRQ. This
 * function returns synchronously.
 */
typedef void (*vxge_hal_session_return_in_progress_wqes_get_f)(
			vxge_hal_up_msg_h umsgh,
			u32 *number_of_wqes_in_msg,
			u32 *number_of_wqes_remaining,
			u64 *wqe_opaque_handles);


/*
 * vxge_hal_session_frame_write_get_f - Get the values from Session Frame Write
 *		 Message
 *
 * @umsgh: Handle to umsg.
 * @umq_opaque_data: Buffer to return a code provided to the RPE by the RxPE
 *		firmware.
 * @num_markers: Buffer to return the number of markers contained in the frame,
 *		or portion of the frame, that will be passed up the UMQ.
 * @first_marker_index: Buffer to return the first byte of frame data passed
 *		to the host is considered to be Byte 0.
 * @fdata: Buffer to return the pointer to variable length data of the frame
 *		pulled from the frame buffer by the ROCRC.
 * @fdata_length: Length of fdata
 *
 * Get the values from Session Frame Write used by adapter to send
 * certain frames to the host. This function returns synchronously.
 */
typedef void (*vxge_hal_session_frame_write_get_f)(
			vxge_hal_up_msg_h umsgh,
			u32 *umq_opaque_data,
			u32 *num_markers,
			u32 *first_marker_index,
			u64 *fdata,
			u32 *fdata_length);

/*
 * struct vxge_hal_session_mgmt_if_t - Functions to manage Session
 * @vxge_hal_session_query: Function to Query the session information
 * @vxge_hal_session_query_resp_get: Function to get values from
 *	   Query session response
 * @vxge_hal_session_return_in_progress_wqes_get: Function to get
 *	   values from sessiion return in progress WQEs message
 * @vxge_hal_session_frame_write_get: Function to get values from
 *	   session frame write up message.
 *
 * This structure contains the functions for managing the Session
 */
typedef struct vxge_hal_session_mgmt_if_t {
    vxge_hal_session_query_f vxge_hal_session_query;
    vxge_hal_session_query_resp_get_f vxge_hal_session_query_resp_get;
    vxge_hal_session_return_in_progress_wqes_get_f
			vxge_hal_session_return_in_progress_wqes_get;
    vxge_hal_session_frame_write_get_f vxge_hal_session_frame_write_get;
} vxge_hal_session_mgmt_if_t;

/*
 * struct vxge_hal_rnic_if_t - Function pointers for use by RNICPI
 * @vxge_hal_version: Hal version
 * @vxge_hal_vpath_mgmt_if: Functions used to manage virtual paths
 * @vxge_hal_pblist_mgmt_if: Functions used to manage PB List
 * @vxge_hal_odlist_mgmt_if: Functions used to manage OD List
 * @vxge_hal_nce_mgmt_if: Functions used to manage NCEs
 * @vxge_hal_stag_mgmt_if: Functions used to manage stags
 * @vxge_hal_cqrq_mgmt_if: Functions used to manage CQRQs
 * @vxge_hal_srq_mgmt_if: Functions used to manage SRQs
 * @vxge_hal_sq_mgmt_if: Functions used to manage SQs
 * @vxge_hal_tcp_mgmt_if: Functions used to manage TCP sessions
 * @vxge_hal_lro_mgmt_if: Functions used to manage LRO sessions
 * @vxge_hal_spdm_mgmt_if: Functions used to manage SPDM entries
 * @vxge_hal_session_mgmt_if: Functions used to manage features of sessions
 *
 * This structure contains the structures that contain functions to manage
 * various types of objects in the RNIC
 */
typedef struct vxge_hal_rnic_if_t {
	vxge_hal_version_t		vxge_hal_version;
	vxge_hal_vpath_mgmt_if_t	vxge_hal_vpath_mgmt_if;
	vxge_hal_pblist_mgmt_if_t	vxge_hal_pblist_mgmt_if;
	vxge_hal_odlist_mgmt_if_t	vxge_hal_odlist_mgmt_if;
	vxge_hal_nce_mgmt_if_t		vxge_hal_nce_mgmt_if;
	vxge_hal_stag_mgmt_if_t		vxge_hal_stag_mgmt_if;
	vxge_hal_cqrq_mgmt_if_t		vxge_hal_cqrq_mgmt_if;
	vxge_hal_srq_mgmt_if_t		vxge_hal_srq_mgmt_if;
	vxge_hal_sq_mgmt_if_t		vxge_hal_sq_mgmt_if;
	vxge_hal_tcp_mgmt_if_t		vxge_hal_tcp_mgmt_if;
	vxge_hal_session_mgmt_if_t	vxge_hal_session_mgmt_if;
} vxge_hal_rnic_if_t;

__EXTERN_END_DECLS

#endif /* VXGE_HAL_RNIC_H */
