/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * Copyright (c) 2002-2010 Neterion, Inc.
 * All right Reserved.
 *
 * FileName :   vxgehal-cqrq.c
 *
 * Description:  HAL offload receive Completion Queue functionality
 *
 * Created:       14 December 2005
 */
#include "vxgehal.h"

/*
 * __hal_cqe_mempool_item_alloc - Allocate List blocks for CQE list
 * callback
 * @mempoolh: Handle to memory pool
 * @memblock: Address of this memory block
 * @memblock_index: Index of this memory block
 * @dma_object: dma object for this block
 * @item: Pointer to this item
 * @index: Index of this item in memory block
 * @is_last: If this is last item in the block
 * @userdata: Specific data of user
 *
 * This function is callback passed to __hal_mempool_create to create memory
 * pool for CQE list
 */
static vxge_hal_status_e
__hal_cqe_mempool_item_alloc(
	vxge_hal_mempool_h mempoolh,
	void *memblock,
	u32 memblock_index,
	vxge_hal_mempool_dma_t *dma_object,
	void *item,
	u32 index,
	u32 is_last,
	void *userdata)
{
	u32 i;
	void *block_priv;
	u32 memblock_item_idx;
	__hal_cqrq_t *cqrqdata = (__hal_cqrq_t *)userdata;
	__hal_list_block_t *prev_block, *next_block;
	__hal_list_block_wrapper_t *prev_wrapper, *wrapper;
	u32 cqes_block_size = cqrqdata->cqe_size * cqrqdata->cqes_per_block;
	__hal_device_t *hldev = (__hal_device_t *)cqrqdata->channel.devh;

	vxge_assert(cqrqdata != NULL);

	vxge_hal_trace_log_pool(hldev,
	    ((__hal_vpath_handle_t *)cqrqdata->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_pool(hldev,
	    ((__hal_vpath_handle_t *)cqrqdata->channel.vph)->vpath->vp_id,
	    "mempoolh = 0x"VXGE_OS_STXFMT", memblock = 0x"VXGE_OS_STXFMT", "
	    "memblock_index = %d, dma_object = 0x"VXGE_OS_STXFMT", "
	    "item = 0x"VXGE_OS_STXFMT", index = %d, is_last = %d, "
	    "userdata = 0x"VXGE_OS_STXFMT, (ptr_t)mempoolh, (ptr_t)memblock,
	    memblock_index, (ptr_t)dma_object, (ptr_t)item, index, is_last,
	    (ptr_t)userdata);

	block_priv = __hal_mempool_item_priv(mempoolh, memblock_index, item,
	    &memblock_item_idx);

	vxge_assert(block_priv != NULL);

	for (i = 0; i < cqrqdata->cqes_per_block; i++) {
		__hal_cqe_priv_t *cqe_priv;
		void *cqe;

		int reserve_index = cqrqdata->channel.reserve_ptr - 1 -
		    (index * cqrqdata->cqes_per_block + i);

		cqe = (void *)((char *)item + i * cqrqdata->cqe_size);

		cqrqdata->channel.reserve_arr[reserve_index] = cqe;

		cqe_priv = (__hal_cqe_priv_t *)(void *)
		    ((char *)block_priv + cqrqdata->cqe_priv_size * i);

		cqe_priv->dma_offset = (char *)cqe - (char *)memblock;
		cqe_priv->dma_addr = dma_object->addr +  cqe_priv->dma_offset;
		cqe_priv->dma_handle = dma_object->handle;
		cqe_priv->memblock = memblock;
		cqe_priv->wrapper = (__hal_list_block_wrapper_t *)
		    (((char *)memblock) + cqes_block_size);
		cqe_priv->cqrqh = (vxge_hal_cqrq_h)userdata;
	}

	__hal_list_block_memblock_idx_set(memblock, memblock_index);

	wrapper = (__hal_list_block_wrapper_t *)
	    (((char *)memblock) + cqes_block_size);

	if (index > 0) {
		prev_block = __hal_mempool_item(mempoolh, index-1);
		next_block = __hal_mempool_item(mempoolh, 0);
		prev_wrapper = (__hal_list_block_wrapper_t *)
		    (((char *)prev_block) + cqes_block_size);
		__hal_wrapper_link(prev_wrapper, wrapper,
		    dma_object->addr, cqes_block_size);
	} else {
		prev_block = memblock;
		next_block = memblock;
		__hal_wrapper_link(wrapper, wrapper,
		    dma_object->addr, cqes_block_size);
	}

	if (cqrqdata->first_block == NULL)
		cqrqdata->first_block = memblock;

	if (is_last) {
		wrapper->control_0 = 0;
		cqrqdata->last_block = memblock;
	}

	__hal_list_block_link(memblock, prev_block, next_block);

	vxge_hal_trace_log_pool(hldev,
	    ((__hal_vpath_handle_t *)cqrqdata->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * __hal_cqe_mempool_item_free - Free List blocks for hw_cqe list callback
 * @mempoolh: Handle to memory pool
 * @memblock: Address of this memory block
 * @memblock_index: Index of this memory block
 * @dma_object: dma object for this block
 * @item: Pointer to this item
 * @index: Index of this item in memory block
 * @is_last: If this is last item in the block
 * @userdata: Specific data of user
 *
 * This function is callback passed to __hal_mempool_free to destroy memory
 * pool for hw_cqe list
 */
static vxge_hal_status_e
__hal_cqe_mempool_item_free(
	vxge_hal_mempool_h mempoolh,
	void *memblock,
	u32 memblock_index,
	vxge_hal_mempool_dma_t *dma_object,
	void *item,
	u32 index,
	u32 is_last,
	void *userdata)
{
	__hal_cqrq_t *cqrqdata = (__hal_cqrq_t *)userdata;
	__hal_list_block_t *prev_block;
	__hal_list_block_wrapper_t *prev_wrapper, *wrapper;
	u32 cqes_block_size = cqrqdata->cqe_size * cqrqdata->cqes_per_block;
	__hal_device_t *hldev = (__hal_device_t *)cqrqdata->channel.devh;

	vxge_assert(cqrqdata != NULL);

	vxge_hal_trace_log_pool(hldev,
	    ((__hal_vpath_handle_t *)cqrqdata->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_pool(hldev,
	    ((__hal_vpath_handle_t *)cqrqdata->channel.vph)->vpath->vp_id,
	    "mempoolh = 0x"VXGE_OS_STXFMT", memblock = 0x"VXGE_OS_STXFMT", "
	    "memblock_index = %d, dma_object = 0x"VXGE_OS_STXFMT", "
	    "item = 0x"VXGE_OS_STXFMT", index = %d, is_last = %d, "
	    "userdata = 0x"VXGE_OS_STXFMT, (ptr_t)mempoolh, (ptr_t)memblock,
	    memblock_index, (ptr_t)dma_object, (ptr_t)item, index, is_last,
	    (ptr_t)userdata);

	prev_block = __hal_list_block_prev_pointer(memblock);

	vxge_assert(prev_block != NULL);

	prev_wrapper = (__hal_list_block_wrapper_t *)
	    (((char *)prev_block) + cqes_block_size);

	wrapper = (__hal_list_block_wrapper_t *)
	    (((char *)memblock) + cqes_block_size);

	__hal_wrapper_unlink(prev_wrapper, wrapper);

	__hal_list_block_unlink(memblock);

	vxge_hal_trace_log_pool(hldev,
	    ((__hal_vpath_handle_t *)cqrqdata->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * __hal_cqrq_create - Create Completion Queue
 * @vpath_handle: Virtual path handle to which this queue belongs
 * @cqrqr: CQRQ
 *
 * This function creates completion queue with the attributes passed
 */

vxge_hal_status_e
__hal_cqrq_create(
	vxge_hal_vpath_h vpath_handle,
	__hal_cqrq_t *cqrq)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_dmq_t *dmq;
	__hal_device_t *hldev;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;
	vxge_hal_mempool_dma_t *dma_object;
	vxge_hal_cqe_h cqeh;
	u32 event_notification_mask = 0;

	vxge_assert((vpath_handle != NULL) && (cqrq != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", cqrq = 0x"VXGE_OS_STXFMT,
	    (ptr_t)vpath_handle, (ptr_t)cqrq);

	if ((vpath_handle == NULL) || (cqrq == NULL)) {
		vxge_hal_info_log_cqrq(hldev, vp->vpath->vp_id,
		    "null pointer passed == > %s : %d", __func__, __LINE__);
		vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: 1", __FILE__, __func__, __LINE__);
		return (VXGE_HAL_FAIL);
	}

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	cqrq->mempool = __hal_mempool_create(
	    vp->vpath->hldev,
	    VXGE_OS_HOST_PAGE_SIZE,
	    VXGE_OS_HOST_PAGE_SIZE,
	    cqrq->cqeblock_priv_size,
	    cqrq->cqrq_length/cqrq->cqes_per_block,
	    cqrq->cqrq_length/cqrq->cqes_per_block,
	    __hal_cqe_mempool_item_alloc,
	    __hal_cqe_mempool_item_free,
	    cqrq);

	if (cqrq->mempool == NULL) {
		vxge_hal_info_log_cqrq(hldev, vp->vpath->vp_id,
		    "Memory allocation for cqrq failed == > %s : %d",
		    __func__, __LINE__);
		vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
		    __LINE__, VXGE_HAL_ERR_OUT_OF_MEMORY);
		return (VXGE_HAL_ERR_OUT_OF_MEMORY);
	}

	status = __hal_channel_initialize(&cqrq->channel);

	if (status != VXGE_HAL_OK) {
		vxge_hal_err_log_cqrq(hldev, vp->vpath->vp_id,
		    "cqrq channel initialization failed == > %s : %d",
		    __func__, __LINE__);
		goto cqrq_create_error;
	}

	while (cqrq->channel.reserve_ptr > cqrq->channel.reserve_top) {
		cqeh = cqrq->channel.reserve_arr[--cqrq->channel.reserve_ptr];

		vxge_hal_info_log_cqrq(hldev, vp->vpath->vp_id,
		    "cqrq CQE %d  addr 0x"VXGE_OS_STXFMT", length %d",
		    cqrq->channel.post_index, (ptr_t)cqeh, cqrq->cqe_size);

		__hal_channel_dtr_post(&cqrq->channel, cqeh);
	}

	cqrq->cq_rq_ir->control_0 = 0;

	dma_object = __hal_mempool_memblock_dma(cqrq->mempool, 0);

	vxge_assert(dma_object != NULL);

	cqrq->first_block_address = dma_object->addr;

	cqrq->first_block_length = VXGE_OS_HOST_PAGE_SIZE;

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(cqrq->opaque_handle,
	    vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(cqrq->opaque_handle,
	    (ptr_t)cqrq);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(cqrq->cqrq_opaque_handle,
	    vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(cqrq->cqrq_opaque_handle,
	    (ptr_t)cqrq);

	__hal_cqrq_create_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    cqrq->cqrq_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    cqrq->opaque_handle,
	    cqrq->cqrq_opaque_handle,
	    vp->vpath->bmap_intr_num,
	    VXGE_HAL_RNIC_OID_DB_SID_GET(cqrq->cqrq_id),
	    vp->vpath->hldev->header.config.max_cqe_groups,
	    vp->vpath->hldev->header.config.refill_threshold_high,
	    vp->vpath->hldev->header.config.refill_threshold_low,
	    cqrq->low_threshold_notify,
	    event_notification_mask,
	    cqrq->cqe_size,
	    cqrq->cq_rq_ir_reg_entry->dma_addr);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK) {
		vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__,
		    status);
		return (status);
	}

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "cqrq channel dmq post failed == > %s : %d", __func__, __LINE__);

cqrq_create_error:

	__hal_cqrq_delete(vpath_handle, cqrq);

	vp->cb_fn(vp->client_handle,
	    NULL,
	    VXGE_HAL_MSG_TYPE_CQRQ_CREATE_RESP,
	    0,
	    status,
	    cqrq->opaque_handle);

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);

	return (status);
}

/*
 * __hal_cqrq_delete - Removes the completion Queue
 * @vpath_handle: Virtual path handle to which this queue belongs
 * @cqrq: CQRQ to be deleted
 *
 * This function freeup the memory pool and removes the
 * completion queue.
 */
vxge_hal_status_e
__hal_cqrq_delete(
	vxge_hal_vpath_h vpath_handle,
	__hal_cqrq_t *cqrq)
{
	__hal_device_t *hldev;
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;

	vxge_assert((vpath_handle != NULL) && (cqrq != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", cqrq = 0x"VXGE_OS_STXFMT,
	    (ptr_t)vpath_handle, (ptr_t)cqrq);

	if (cqrq->mempool) {
		__hal_mempool_destroy(cqrq->mempool);
	}

	if (cqrq->cq_rq_ir_reg_entry) {
		__hal_regpool_free(&vp->vpath->hldev->reg_pool,
		    cqrq->cq_rq_ir_reg_entry);
	}

	__hal_vp_oid_free(vp->vpath, &vp->vpath->cqrq_oid_db, cqrq->cqrq_id);

	__hal_channel_terminate(&cqrq->channel);

#if defined(VXGE_HAL_CQRQ_POST)
	vxge_os_spin_lock_destroy(&cqrq->channel.post_lock,
	    vp->vpath->hldev->header.pdev);
#elif defined(VXGE_HAL_CQRQ_POST_IRQ)
	vxge_os_spin_lock_destroy_irq(&cqrq->channel.post_lock,
	    vp->vpath->hldev->header.pdev);
#endif

#if defined(VXGE_HAL_CQRQ_FREE)
	vxge_os_spin_lock_destroy(&cqrq->channel.free_lock,
	    vp->vpath->hldev->header.pdev);
#elif defined(VXGE_HAL_CQRQ_FREE_IRQ)
	vxge_os_spin_lock_destroy_irq(&cqrq->channel.free_lock,
	    vp->vpath->hldev->header.pdev);
#endif

#if defined(VXGE_HAL_CQRQ_RESERVE)
	vxge_os_spin_lock_destroy(&cqrq->channel.reserve_lock,
	    vp->vpath->hldev->header.pdev);
#elif defined(VXGE_HAL_CQRQ_RESERVE_IRQ)
	vxge_os_spin_lock_destroy_irq(&cqrq->channel.reserve_lock,
	    vp->vpath->hldev->header.pdev);
#endif

	__hal_channel_free(&cqrq->channel);

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * __hal_cqrq_replenish - Replenish the CQRQ with CQES
 * @cqrq: CQRQ
 *
 * This function checks the reserve_arr for a full block
 * of CQEs. If a full block of CQEs are available, it will
 * post the block to CQE by setting the wrapper length.
 */

vxge_hal_status_e
__hal_cqrq_replenish(
	__hal_cqrq_t *cqrq)
{
	u32 i;
	vxge_hal_cqe_h cqeh;
	void *first, *last;
	__hal_device_t *hldev;
	__hal_list_block_wrapper_t *wrapper;

	vxge_assert(cqrq != NULL);

	hldev = (__hal_device_t *)cqrq->channel.devh;

	vxge_hal_trace_log_cqrq(hldev,
	    ((__hal_vpath_handle_t *)cqrq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_cqrq(hldev,
	    ((__hal_vpath_handle_t *)cqrq->channel.vph)->vpath->vp_id,
	    "cqrq = 0x"VXGE_OS_STXFMT, (ptr_t)cqrq);

	if ((u32)((cqrq->channel.reserve_ptr - cqrq->channel.reserve_top) +
	    (cqrq->channel.length - cqrq->channel.free_ptr)) <
	    cqrq->cqes_per_block) {
		vxge_hal_trace_log_cqrq(hldev,
		    ((__hal_vpath_handle_t *)cqrq->channel.vph)->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
		    __LINE__, VXGE_HAL_INF_NOT_ENOUGH_HW_CQES);
		return (VXGE_HAL_INF_NOT_ENOUGH_HW_CQES);
	}

	first = cqrq->first_block;
	last = ((char *)cqrq->first_block) +
	    cqrq->cqes_per_block * cqrq->cqe_size;

	for (i = 0; i < cqrq->cqes_per_block; i++) {

		if (__hal_channel_dtr_alloc(&cqrq->channel, &cqeh)
		    == VXGE_HAL_OK) {

			vxge_assert((first <= cqeh)&& (cqeh <= last));

			__hal_channel_dtr_post(&cqrq->channel, cqeh);
		}
	}

	wrapper = (__hal_list_block_wrapper_t *)
	    (((char *)cqrq->first_block) +
	    cqrq->cqes_per_block *  cqrq->cqe_size);

	wrapper->control_0 = 0;

	wrapper = (__hal_list_block_wrapper_t *)
	    (((char *)cqrq->last_block) +
	    cqrq->cqes_per_block * cqrq->cqe_size);

	wrapper->control_0 =
	    VXGE_HAL_LBW_NEXT_BLOCK_LENGTH(cqrq->cqes_per_block *
	    cqrq->cqe_size);

	cqrq->last_block = cqrq->first_block;
	cqrq->first_block =
	    __hal_list_block_next_pointer(cqrq->first_block);

	vxge_hal_trace_log_cqrq(hldev,
	    ((__hal_vpath_handle_t *)cqrq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * vxge_hal_cqrq_create - Create CQRQ queue
 *
 * @vpath_handle: Virtual path handle to which this queue belongs
 * @attr: CQRQ attributes
 * @opaque_handle: HAL Opaque handle
 * @cq_opaque_handle: HAL CQ Opaque Handle
 *
 * Creates a CQRQ on the adapter.
 */
vxge_hal_status_e
vxge_hal_cqrq_create(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_cqrq_attr_t *attr,
	vxge_hal_opaque_handle_t *opaque_handle,
	vxge_hal_opaque_handle_t *cq_opaque_handle)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_cqrq_t *cqrq;
	vxge_hal_status_e status;
	u32 immed_data_length;
	u32 cqe_size;
	u32 cqes_per_block;
	u32 cqrq_length;

	vxge_assert((vpath_handle != NULL) && (attr != NULL) &&
	    (opaque_handle != NULL) && (cq_opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", attr = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", "
	    "cq_opaque_handle = 0x"VXGE_OS_STXFMT, (ptr_t)vpath_handle,
	    (ptr_t)attr, (ptr_t)opaque_handle, (ptr_t)cq_opaque_handle);

	if (((vxge_hal_cqrq_attr_t *)attr)->immed_data_length > 0) {
		immed_data_length =
		    ((((vxge_hal_cqrq_attr_t *)attr)->immed_data_length +
		    __vxge_os_cacheline_size - 1)/__vxge_os_cacheline_size) *
		    __vxge_os_cacheline_size;
	} else {
		immed_data_length = 0;
	}

	if (((vxge_hal_cqrq_attr_t *)attr)->cqrq_type ==
	    VXGE_HAL_CQRQ_TYPE_BYTE_STREAM_IWARP) {
		cqe_size = sizeof (__hal_hw_cqe_t) + immed_data_length;
	} else {
		cqe_size = sizeof (__hal_lro_cqe_t);
	}


	cqes_per_block = VXGE_HAL_LIST_BLOCK_ITEMS_SIZE / cqe_size;

	cqrq_length = ((attr->cqrq_length + cqes_per_block - 1) /
	    cqes_per_block) * cqes_per_block;

	cqrq = (__hal_cqrq_t *)__hal_channel_allocate(
	    (vxge_hal_device_h)vp->vpath->hldev,
	    vpath_handle,
	    VXGE_HAL_CHANNEL_TYPE_COMPLETION_QUEUE,
	    cqrq_length,
	    TRUE,
	    TRUE,
	    TRUE,
	    0,
	    attr->userdata);

	if (cqrq == NULL) {
		vxge_hal_err_log_cqrq(hldev, vp->vpath->vp_id,
		    "Memory allocation failed %s : %d", __func__, __LINE__);
		vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
		    __LINE__, VXGE_HAL_ERR_OUT_OF_MEMORY);
		return (VXGE_HAL_ERR_OUT_OF_MEMORY);
	}

	cqrq->cqrq_length = cqrq_length;
	cqrq->immed_data_length = immed_data_length;
	cqrq->callback = attr->callback;
	cqrq->cqrq_type = ((vxge_hal_cqrq_attr_t *)attr)->cqrq_type;
	cqrq->is_shared = ((vxge_hal_cqrq_attr_t *)attr)->is_shared;
	cqrq->low_threshold_notify =
	    ((vxge_hal_cqrq_attr_t *)attr)->low_threshold_notify;

#if defined(VXGE_HAL_CQRQ_POST)
	vxge_os_spin_lock_init(&cqrq->channel.post_lock,
	    hldev->header.pdev);
#elif defined(VXGE_HAL_CQRQ_POST_IRQ)
	vxge_os_spin_lock_init_irq(&cqrq->channel.post_lock,
	    hldev->header.irqh);
#endif

#if defined(VXGE_HAL_CQRQ_FREE)
	vxge_os_spin_lock_init(&cqrq->channel.free_lock,
	    hldev->header.pdev);
#elif defined(VXGE_HAL_CQRQ_FREE_IRQ)
	vxge_os_spin_lock_init_irq(&cqrq->channel.free_lock,
	    hldev->header.irqh);
#endif

#if defined(VXGE_HAL_CQRQ_RESERVE)
	vxge_os_spin_lock_init(&cqrq->channel.reserve_lock,
	    hldev->header.pdev);
#elif defined(VXGE_HAL_CQRQ_RESERVE_IRQ)
	vxge_os_spin_lock_init_irq(&cqrq->channel.reserve_lock,
	    hldev->header.irqh);
#endif

	cqrq->first_block = NULL;
	cqrq->last_block = NULL;

	cqrq->high_cqe_number = 0;
	cqrq->cqe_size = cqe_size;
	cqrq->cqes_per_block = cqes_per_block;
	cqrq->cqe_priv_size = sizeof (__hal_cqe_priv_t);
	cqrq->cqeblock_priv_size = cqes_per_block * cqrq->cqe_priv_size;

	status = __hal_vp_oid_allocate(
	    vp->vpath,
	    &vp->vpath->cqrq_oid_db,
	    &cqrq->cqrq_id);

	if (status != VXGE_HAL_OK) {
		vxge_hal_err_log_cqrq(hldev, vp->vpath->vp_id,
		    "Memory allocation for cqrq oid db failed %s : %d",
		    __func__, __LINE__);
		__hal_cqrq_delete(vpath_handle, cqrq);
		vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
		    __LINE__, VXGE_HAL_ERR_TOO_MANY);
		return (VXGE_HAL_ERR_TOO_MANY);
	}

	status = __hal_regpool_allocate(&vp->vpath->hldev->reg_pool,
	    &cqrq->cq_rq_ir_reg_entry);

	if (status != VXGE_HAL_OK) {
		vxge_hal_err_log_cqrq(hldev, vp->vpath->vp_id,
		    "Memory allocation for cq_rq__ir failed == > %s : %d",
		    __func__, __LINE__);
		__hal_cqrq_delete(vpath_handle, cqrq);
		vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d",
		    __FILE__, __func__, __LINE__, status);
		return (status);
	}

	cqrq->cq_rq_ir = cqrq->cq_rq_ir_reg_entry->reg_addr;

	cqrq->cqrq_opaque_handle = cq_opaque_handle;

	cqrq->opaque_handle = opaque_handle;

	__hal_channel_insert_pending_list(&cqrq->channel);

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);

	return (status);
}

/*
 * vxge_hal_cqrq_delete - Removes the completion Queue
 * @vpath_handle: Virtual path handle to which this queue belongs
 * @cqrq_id: CQRQ to be deleted
 * @opaque_handle: HAL Opaque handle
 *
 * This function freeup the memory pool and removes the
 * completion queue.
 */
vxge_hal_status_e
vxge_hal_cqrq_delete(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t cqrq_id,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	__hal_cqrq_t *cqrq;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (cqrq_id != 0) &&
	    (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", cqrq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT, (ptr_t)vpath_handle,
	    (ptr_t)cqrq_id, (ptr_t)opaque_handle);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	cqrq = (__hal_cqrq_t *)(ptr_t)cqrq_id;

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, cqrq_id);

	__hal_cqrq_delete_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    cqrq->cqrq_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_cqrq_modify - Modify the attributes of a cqrq
 *
 * @vpath_handle: Virtual path handle to which this queue belongs
 * @cqrq_id: CQRQ to be modified
 * @opaque_handle: HAL Opaque handle
 *
 * Modifies the attributes of a CQRQ on the adapter.
 */
vxge_hal_status_e
vxge_hal_cqrq_modify(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t cqrq_id,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	u32 event_notification_mask = 0;
	__hal_cqrq_t *cqrq;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (cqrq_id != 0) &&
	    (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", cqrq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT, (ptr_t)vpath_handle,
	    (ptr_t)cqrq_id, (ptr_t)opaque_handle);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	cqrq = (__hal_cqrq_t *)(ptr_t)cqrq_id;

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, cqrq_id);

	__hal_cqrq_modify_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    cqrq->cqrq_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    vp->vpath->hldev->header.config.max_cqe_groups,
	    vp->vpath->hldev->header.config.refill_threshold_high,
	    vp->vpath->hldev->header.config.refill_threshold_low,
	    cqrq->low_threshold_notify,
	    event_notification_mask);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_cqrq_query - Queries the information of a CQRQ
 *
 * @vpath_handle: Virtual path handle to which this queue belongs
 * @cqrq_id: CQRQ to be queries
 * @opaque_handle: HAL Opaque handle
 *
 * Queries the information of a CQRQ.
 */
vxge_hal_status_e
vxge_hal_cqrq_query(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t cqrq_id,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	__hal_cqrq_t *cqrq;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (cqrq_id != 0) &&
	    (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", cqrq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT, (ptr_t)vpath_handle,
	    (ptr_t)cqrq_id, (ptr_t)opaque_handle);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	cqrq = (__hal_cqrq_t *)(ptr_t)cqrq_id;

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, cqrq_id);

	__hal_cqrq_query_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    cqrq->cqrq_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_cqrq_arm - Arm a CQRQ
 *
 * @vpath_handle: Virtual path handle to which this queue belongs
 * @cqrq_id: CQRQ to be armed
 * @opaque_handle: HAL Opaque handle
 * @arm_type: The armed state of the CQRQ.
 *
 * Arms a CQRQ on the adapter
 */
vxge_hal_status_e
vxge_hal_cqrq_arm(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t cqrq_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 arm_type)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	__hal_cqrq_t *cqrq;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (cqrq_id != 0) &&
	    (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT",  cqrq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT, (ptr_t)vpath_handle,
	    (ptr_t)cqrq_id, (ptr_t)opaque_handle);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	cqrq = (__hal_cqrq_t *)(ptr_t)cqrq_id;

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, cqrq_id);

	__hal_cqrq_arm_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    cqrq->cqrq_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    arm_type);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * ___hal_cqrq_fcqebw - Post First CQE Block to CQRQ
 *
 * @vpath_handle: Virtual path handle to which this queue belongs
 * @cqrq_id: CQRQ to to post to
 * @opaque_handle: HAL Opaque handle
 *
 * Posts first CQE block to CQRQ
 */
vxge_hal_status_e
__hal_cqrq_fcqebw(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t cqrq_id,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	__hal_cqrq_t *cqrq;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (cqrq_id != 0) &&
	    (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT",  cqrq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT, (ptr_t)vpath_handle,
	    (ptr_t)cqrq_id, (ptr_t)opaque_handle);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	cqrq = (__hal_cqrq_t *)(ptr_t)cqrq_id;

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, cqrq_id);

	__hal_cqrq_fcqebw_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    cqrq->cqrq_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    cqrq->first_block_address,
	    cqrq->first_block_length);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_cqrq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * __hal_cqrq_ready_cqe_check - Check the CQRQ for any CQEs
 *		 currently indicated and available for processing
 * @cqrq: cqrq.
 *
 * This function return (VXGE_HAL_OK) if there is a CQE available for
 * processing by the host.
 */
vxge_hal_status_e
__hal_cqrq_ready_cqe_check(
	vxge_hal_obj_id_t cqrq_id)
{
	__hal_cqrq_t *cqrq = (__hal_cqrq_t *)(ptr_t)cqrq_id;
	__hal_device_t *hldev;
	vxge_hal_status_e status;

	vxge_assert(cqrq != NULL);

	hldev = (__hal_device_t *)cqrq->channel.devh;

	vxge_hal_trace_log_cqrq(hldev,
	    ((__hal_vpath_handle_t *)cqrq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_cqrq(hldev,
	    ((__hal_vpath_handle_t *)cqrq->channel.vph)->vpath->vp_id,
	    "cqrq = 0x"VXGE_OS_STXFMT, (ptr_t)cqrq);

#if defined(VXGE_HAL_CQRQ_POST_IRQ)
	vxge_os_spin_lock_irq(&cqrq->channel.post_lock, flags);
#elif defined(VXGE_HAL_CQRQ_POST)
	vxge_os_spin_lock(&cqrq->channel.post_lock);
#endif

	cqrq->high_cqe_number =
	    VXGE_HAL_CQ_RQ_IR_GET_NUM_CQE(cqrq->cq_rq_ir->control_0);

	vxge_hal_info_log_cqrq(hldev,
	    ((__hal_vpath_handle_t *)cqrq->channel.vph)->vpath->vp_id,
	    "high_cqe_number = 0x"VXGE_OS_LLXFMT, cqrq->high_cqe_number);

	if ((cqrq->high_cqe_number -
	    ((cqrq->high_cqe_number / cqrq->cqrq_length) * cqrq->cqrq_length))
	    != cqrq->channel.compl_index)
		status = VXGE_HAL_OK;
	else
		status = VXGE_HAL_INF_NO_MORE_COMPLETED_DESCRIPTORS;

#if defined(VXGE_HAL_CQRQ_POST_IRQ)
	vxge_os_spin_unlock_irq(&cqrq->channel.post_lock, flags);
#elif defined(VXGE_HAL_CQRQ_POST)
	vxge_os_spin_unlock(&cqrq->channel.post_lock);
#endif

	vxge_hal_trace_log_cqrq(hldev,
	    ((__hal_vpath_handle_t *)cqrq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);

	return (status);
}

/*
 * vxge_hal_cqrq_ready_hw_cqe_check - Check the CQRQ for any CQEs
 *		 currently indicated and available for processing
 * @vpath_handle: Handle returned by virtual path open
 * @cqrq_id: Object Id of CQRQ that has CQE ready.
 *
 * This function return (VXGE_HAL_OK) if there is a CQE available for
 * processing by the host. This function returns synchronously.
 */
vxge_hal_status_e
vxge_hal_cqrq_ready_hw_cqe_check(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t cqrq_id)
{
	return (__hal_cqrq_ready_cqe_check(cqrq_id));
}

/*
 * vxge_hal_cqrq_ready_lro_cqe_check - Check the CQRQ for any CQEs
 *		 currently indicated and available for processing
 * @vpath_handle: Handle returned by virtual path open
 * @cqrq_id: Object Id of CQRQ that has CQE ready.
 *
 * This function return (VXGE_HAL_OK) if there is a CQE available for
 * processing by the host. This function returns synchronously.
 */
vxge_hal_status_e
vxge_hal_cqrq_ready_lro_cqe_check(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t cqrq_id)
{
	return (__hal_cqrq_ready_cqe_check(cqrq_id));
}

/*
 * vxge_hal_cqrq_next_ready_hw_cqe_get - Return the next ready CQE info
 * @vpath_handle: virtual path to which this queue belongs
 * @cqrq_id: Object Id of CQRQ that has CQE ready.
 * @cqe_info: Buffer to return HW-CQE contents
 *
 * This function returns the next ready CQE in the CQRQ.
 */
vxge_hal_status_e
vxge_hal_cqrq_next_ready_hw_cqe_get(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t cqrq_id,
	vxge_hal_cqrq_hw_cqe_Info_t *cqe_info)
{
	__hal_cqrq_t *cqrq = (__hal_cqrq_t *)(ptr_t)cqrq_id;
	__hal_hw_cqe_t *cqe;
	__hal_device_t *hldev;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (cqrq_id != 0) &&
	    (cqe_info != NULL));

	hldev = (__hal_device_t *)cqrq->channel.devh;

	vxge_hal_trace_log_cqrq(hldev,
	    ((__hal_vpath_handle_t *)cqrq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_cqrq(hldev,
	    ((__hal_vpath_handle_t *)cqrq->channel.vph)->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", cqrq_id = 0x"VXGE_OS_STXFMT", "
	    "cqe_info = 0x"VXGE_OS_STXFMT, (ptr_t)vpath_handle,
	    (ptr_t)cqrq_id, (ptr_t)cqe_info);

#if defined(VXGE_HAL_CQRQ_POST_IRQ)
	vxge_os_spin_lock_irq(&cqrq->channel.post_lock, flags);
#elif defined(VXGE_HAL_CQRQ_POST)
	vxge_os_spin_lock(&cqrq->channel.post_lock);
#endif

	if ((cqrq->high_cqe_number -
	    ((cqrq->high_cqe_number / cqrq->cqrq_length) * cqrq->cqrq_length))
	    == cqrq->channel.compl_index) {

		status = VXGE_HAL_INF_NO_MORE_COMPLETED_DESCRIPTORS;

	} else {
		__hal_channel_dtr_try_complete(&cqrq->channel,
		    (__hal_dtr_h *)&cqe);

		__hal_channel_dtr_complete(&cqrq->channel);

		cqe_info->opaque_handle = cqe->opaque_handle_1;
		cqe_info->push = (u32)VXGE_HAL_HW_CQE_GET_PUSH(cqe->control_0);
		cqe_info->send_type =
		    (u32)VXGE_HAL_HW_CQE_GET_SEND_TYPE(cqe->control_0);
		cqe_info->ino = (u32)VXGE_HAL_HW_CQE_GET_INO(cqe->control_0);
		cqe_info->immed_data_len =
		    (u32)VXGE_HAL_HW_CQE_GET_IMMEDIATE_DATA_LEN(cqe->control_0);
		vxge_os_memcpy(cqe_info->immed_data,
		    (((char *)cqe)+sizeof (__hal_hw_cqe_t)),
		    cqe_info->immed_data_len);
		cqe_info->stag = (u32)VXGE_HAL_HW_CQE_GET_STAG(cqe->control_0);
		cqe_info->invalidate =
		    (u32)VXGE_HAL_HW_CQE_GET_INVALIDATE(cqe->control_0);
		cqe_info->msn = (u32)VXGE_HAL_HW_CQE_GET_MSN(cqe->control_1);
		cqe_info->bytes_placed =
		    (u32)VXGE_HAL_HW_CQE_GET_BYTES_PLACED(cqe->control_1);

		__hal_channel_dtr_free(&cqrq->channel, cqe);

		__hal_cqrq_replenish(cqrq);

		status = VXGE_HAL_OK;
	}

#if defined(VXGE_HAL_CQRQ_POST_IRQ)
	vxge_os_spin_unlock_irq(&cqrq->channel.post_lock, flags);
#elif defined(VXGE_HAL_CQRQ_POST)
	vxge_os_spin_unlock(&cqrq->channel.post_lock);
#endif

	vxge_hal_trace_log_cqrq(hldev,
	    ((__hal_vpath_handle_t *)cqrq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);

	return (status);
}

/*
 * vxge_hal_cqrq_next_ready_lro_cqe_get - Return the next ready CQE info
 * @vpath_handle: virtual path to which this queue belongs
 * @cqrq_id: Object Id of CQRQ that has CQE ready.
 * @cqe_info: Buffer to return HW-CQE contents
 *
 * This function returns the next ready CQE in the CQRQ.
 */
vxge_hal_status_e
vxge_hal_cqrq_next_ready_lro_cqe_get(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t cqrq_id,
	vxge_hal_cqrq_lro_cqe_Info_t *cqe_info)
{
	__hal_cqrq_t *cqrq = (__hal_cqrq_t *)(ptr_t)cqrq_id;
	__hal_lro_cqe_t *cqe;
	__hal_device_t *hldev;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (cqrq_id != 0) &&
	    (cqe_info != NULL));

	hldev = (__hal_device_t *)cqrq->channel.devh;

	vxge_hal_trace_log_cqrq(hldev,
	    ((__hal_vpath_handle_t *)cqrq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_cqrq(hldev,
	    ((__hal_vpath_handle_t *)cqrq->channel.vph)->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", cqrq_id = 0x"VXGE_OS_STXFMT", "
	    "cqe_info = 0x"VXGE_OS_STXFMT, (ptr_t)vpath_handle,
	    (ptr_t)cqrq_id, (ptr_t)cqe_info);

#if defined(VXGE_HAL_CQRQ_POST_IRQ)
	vxge_os_spin_lock_irq(&cqrq->channel.post_lock, flags);
#elif defined(VXGE_HAL_CQRQ_POST)
	vxge_os_spin_lock(&cqrq->channel.post_lock);
#endif

	if ((cqrq->high_cqe_number -
	    ((cqrq->high_cqe_number / cqrq->cqrq_length) * cqrq->cqrq_length))
	    == cqrq->channel.compl_index) {

		status = VXGE_HAL_INF_NO_MORE_COMPLETED_DESCRIPTORS;

	} else {
		__hal_channel_dtr_try_complete(&cqrq->channel,
		    (__hal_dtr_h *)&cqe);

		__hal_channel_dtr_complete(&cqrq->channel);

		cqe_info->opaque_handle =
		    __hal_srq_lro_wqe_opaque_handle_get(cqe->wqe_opaque_handle);
		cqe_info->ipv4_header_cksum = (u32)
		    VXGE_HAL_LRO_CQE_GET_IPV4_HEADER_CHECKSUM(cqe->control_0);
		cqe_info->tcp_cksum =
		    (u32)VXGE_HAL_LRO_CQE_GET_TCP_CHECKSUM(cqe->control_0);
		cqe_info->vlan_tag =
		    (u32)VXGE_HAL_LRO_CQE_GET_VLAN_TAG(cqe->control_0);
		cqe_info->buf0_bytes_placed =
		    (u32)VXGE_HAL_LRO_CQE_GET_BUF0_BYTES_PLACED(cqe->control_0);
		cqe_info->buf1_bytes_placed =
		    (u32)VXGE_HAL_LRO_CQE_GET_BUF1_BYTES_PLACED(cqe->control_1);
		cqe_info->buf2_bytes_placed =
		    (u32)VXGE_HAL_LRO_CQE_GET_BUF2_BYTES_PLACED(cqe->control_1);
		cqe_info->buf3_bytes_placed =
		    (u32)VXGE_HAL_LRO_CQE_GET_BUF3_BYTES_PLACED(cqe->control_1);
		cqe_info->buf4_bytes_placed =
		    (u32)VXGE_HAL_LRO_CQE_GET_BUF4_BYTES_PLACED(cqe->control_1);
		cqe_info->is_ipv6 =
		    (u32)VXGE_HAL_LRO_CQE_GET_IPV6(cqe->control_2);
		cqe_info->num_frames =
		    (u32)VXGE_HAL_LRO_CQE_GET_NUMBER_OF_FRAMES(cqe->control_2);
		cqe_info->eth_encapsulation =
		    (u32)VXGE_HAL_LRO_CQE_GET_ETH_ENCAPSULATION(cqe->control_2);
		cqe_info->reason =
		    (u32)VXGE_HAL_LRO_CQE_GET_REASON(cqe->control_2);
		cqe_info->vlan_flag =
		    (u32)VXGE_HAL_LRO_CQE_GET_VLAN_FLAG(cqe->control_2);

		__hal_srq_lro_wqe_owner_host_set(cqe->wqe_opaque_handle);

		__hal_channel_dtr_free(&cqrq->channel, cqe);

		__hal_cqrq_replenish(cqrq);

		status = VXGE_HAL_OK;
	}

#if defined(VXGE_HAL_CQRQ_POST_IRQ)
	vxge_os_spin_unlock_irq(&cqrq->channel.post_lock, flags);
#elif defined(VXGE_HAL_CQRQ_POST)
	vxge_os_spin_unlock(&cqrq->channel.post_lock);
#endif

	vxge_hal_trace_log_cqrq(hldev,
	    ((__hal_vpath_handle_t *)cqrq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);

	return (status);
}
