/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */
/*
 * Copyright (c) 2002-2010 Neterion, Inc.
 * All right Reserved.
 *
 * FileName :   vxgehal-lbwrapper.h
 *
 * Description:  HAL HW Full offload structures
 *
 * Created:       05 Dec 2005
 */

#ifndef	VXGE_HAL_LB_WRAPPER_H
#define	VXGE_HAL_LB_WRAPPER_H

__EXTERN_BEGIN_DECLS

/*
 * struct __hal_list_block_wrapper_t - List Block Wrapper
 * @next_block_ptr: The location in host memory where the next block starts
 * @control_0: Bits 0 to 31 - Reserved.
 *	       Bits 32 to 63 - The length in bytes of the next block
 *			  - not including that block's list block wrapper.
 *
 * @pad: Padding to make it cache-aligned.
 *
 * The adapter DMA reads a list block wrapper to traverse from one block
 * structure to the next. In particular, list block wrappers are used to
 * tie TOWI, OD, SRQ, CQRQ, UMQ, DMQ lists together. The SRQ, DMQ, and UMQ
 * are the only lists in which an individual item is allowed to span blocks.
 * The adapter never writes to a list block wrapper.
 */
typedef __vxge_os_attr_cacheline_aligned  struct __hal_list_block_wrapper_t {
	u64		control_0;
#define	VXGE_HAL_LBW_GET_NEXT_BLOCK_LENGTH(ctrl0)	bVAL32(ctrl0, 32)
#define	VXGE_HAL_LBW_NEXT_BLOCK_LENGTH(val)		vBIT(val, 32, 32)

	u64		next_block_ptr;

	u8		pad[(((15+__vxge_os_cacheline_size)/
			__vxge_os_cacheline_size)*__vxge_os_cacheline_size)-16];
} __hal_list_block_wrapper_t;

#define	VXGE_HAL_LIST_BLOCK_ITEMS_SIZE		(VXGE_OS_HOST_PAGE_SIZE - \
				    sizeof (__hal_list_block_wrapper_t) - \
				    (((31+__vxge_os_cacheline_size)/ \
			__vxge_os_cacheline_size)*__vxge_os_cacheline_size))

typedef u8 __hal_list_block_t[VXGE_OS_HOST_PAGE_SIZE];

#define	VXGE_HAL_LIST_BLOCK_NEXT_POINTER_OFFSET	(VXGE_OS_HOST_PAGE_SIZE-8)
#define	VXGE_HAL_LIST_BLOCK_PREV_POINTER_OFFSET	(VXGE_OS_HOST_PAGE_SIZE-16)
#define	VXGE_HAL_LIST_BLOCK_IDX_OFFSET		(VXGE_OS_HOST_PAGE_SIZE-24)
#define	VXGE_HAL_LIST_BLOCK_PRIV_DATA_OFFSET	(VXGE_OS_HOST_PAGE_SIZE-32)

/*
 * __hal_list_block_priv_data - Get the private data of this list block
 * @block: Pointer to list block.
 *
 * This function is used to get the priv data of list block.
 */
static inline u64
__hal_list_block_priv_data(__hal_list_block_t *block)
{
	return ((u64)*((u64 *)(void *)((char *)block +
	    VXGE_HAL_LIST_BLOCK_PRIV_DATA_OFFSET)));
}

/*
 * __hal_list_block_memblock_priv_data_set - Set the private data
 *					   of this list block
 * @block: Pointer to list block.
 * @priv_data: Private Data.
 *
 * This function is used to set the private data of list block.
 */
static inline void
__hal_list_block_memblock_priv_data_set(__hal_list_block_t *block,
					u64 priv_data)
{
	*((u64 *)(void *)((char *)block +
	    VXGE_HAL_LIST_BLOCK_PRIV_DATA_OFFSET)) = priv_data;
}

/*
 * __hal_list_block_memblock_idx - Get the index of this list block
 * @block: Pointer to list block.
 *
 * This function is used to get the index of list block.
 */
static inline int
__hal_list_block_memblock_idx(__hal_list_block_t *block)
{
	return ((int)*((u64 *)(void *)((char *)block +
	    VXGE_HAL_LIST_BLOCK_IDX_OFFSET)));
}

/*
 * __hal_list_block_memblock_idx_set - Set the index of this list block
 * @block: Pointer to list block.
 * @memblock_idx: List block index.
 *
 * This function is used to set the index of list block.
 */
static inline void
__hal_list_block_memblock_idx_set(__hal_list_block_t *block, int memblock_idx)
{
	*((u64 *)(void *)((char *)block +
	    VXGE_HAL_LIST_BLOCK_IDX_OFFSET)) = memblock_idx;
}

/*
 * __hal_list_block_next_pointer_set - Get the next list block
 * @block: Pointer to list block.
 *
 * This function is used to get the address of next list block.
 */
static inline void*
__hal_list_block_next_pointer(__hal_list_block_t *block)
{
	return ((void *)(*((ptr_t *)(void *)((char *)block +
	    VXGE_HAL_LIST_BLOCK_NEXT_POINTER_OFFSET))));
}

/*
 * __hal_list_block_next_pointer_set - Set the next list block
 * @block: Pointer to list block.
 * @next_block: Next list block address.
 *
 * This function is used to set the address of next list block.
 */
static inline void
__hal_list_block_next_pointer_set(__hal_list_block_t *block,
			__hal_list_block_t *next_block)
{
	*((u64 *)(void *)((char *)block +
	    VXGE_HAL_LIST_BLOCK_NEXT_POINTER_OFFSET)) = (u64)(ptr_t)next_block;
}

/*
 * __hal_list_block_prev_pointer_set - Get the prev list block
 * @block: Pointer to list block.
 *
 * This function is used to get the address of prev list block.
 */
static inline void*
__hal_list_block_prev_pointer(__hal_list_block_t *block)
{
	return ((void *)(*((ptr_t *)(void *)((char *)block +
	    VXGE_HAL_LIST_BLOCK_PREV_POINTER_OFFSET))));
}

/*
 * __hal_list_block_prev_pointer_set - Set the prev list block
 * @block: Pointer to list block.
 * @prev_block: prev list block address.
 *
 * This function is used to set the prev list block address.
 */
static inline void
__hal_list_block_prev_pointer_set(__hal_list_block_t *block,
			__hal_list_block_t *prev_block)
{
	*((u64 *)(void *)((char *)block +
	    VXGE_HAL_LIST_BLOCK_PREV_POINTER_OFFSET)) = (u64)(ptr_t)prev_block;
}

/*
 * __hal_list_block_link - Links the list blocks
 * @block: Pointer to list block.
 * @prev_block: prev list block.
 * @next_block: next list block.
 *
 * This function is used to link the list blocks.
 */
static inline void
__hal_list_block_link(__hal_list_block_t *block,
		__hal_list_block_t *prev_block,
		__hal_list_block_t *next_block)
{
	__hal_list_block_next_pointer_set(prev_block, block);

	__hal_list_block_next_pointer_set(block, next_block);

	__hal_list_block_prev_pointer_set(next_block, block);

	__hal_list_block_prev_pointer_set(block, prev_block);
}

/*
 * __hal_list_block_unlink - Unlinks the list blocks
 * @block: Pointer to list block.
 *
 * This function is used to link the list blocks.
 */
static inline void
__hal_list_block_unlink(__hal_list_block_t *block)
{
	__hal_list_block_t *next_item, *prev_item;

	/* get "prev" RxD block */
	prev_item = (__hal_list_block_t *)__hal_list_block_prev_pointer(block);

	/* get "next" RxD block */
	next_item = (__hal_list_block_t *)__hal_list_block_next_pointer(block);

	__hal_list_block_next_pointer_set(prev_item, next_item);

	__hal_list_block_prev_pointer_set(next_item, prev_item);
}


/*
 * __hal_wrapper_link - link current block wrapper with
 *			 previous block wrapper
 * @prev_wrapper: Previous block wrapper.
 * @wrapper: Current block wrapper.
 * @dma_addr: dma address of curent block.
 * @size: List size(don't include wrapper size).
 *
 * This function is used to link the list block wrappers.
 */
static inline void
__hal_wrapper_link(__hal_list_block_wrapper_t *prev_wrapper,
			__hal_list_block_wrapper_t *wrapper,
			u64 dma_addr, u32 size)
{
	wrapper->next_block_ptr = prev_wrapper->next_block_ptr;
	wrapper->control_0 = prev_wrapper->control_0;
	prev_wrapper->next_block_ptr = dma_addr;
	prev_wrapper->control_0 = VXGE_HAL_LBW_NEXT_BLOCK_LENGTH(size);
}

/*
 * __hal_wrapper_unlink - unlink current block wrapper from
 *			 previous block wrapper
 * @prev_wrapper: Previous block wrapper.
 * @wrapper: Current block wrapper.
 *
 * This function is used to unlink the list block wrappers.
 */
static inline void
__hal_wrapper_unlink(__hal_list_block_wrapper_t *prev_wrapper,
			__hal_list_block_wrapper_t *wrapper)
{
	prev_wrapper->next_block_ptr = wrapper->next_block_ptr;
	prev_wrapper->control_0 = wrapper->control_0;
}

/*
 * __hal_wrapper_htonll - Swap the bytes in the wrapper from
 *	   host to network byte order
 * @wrapper: Current block wrapper.
 *
 * This function is used to Swap the bytes in the wrapper.
 */
static inline void
__hal_wrapper_htonll(__hal_list_block_wrapper_t *wrapper)
{
	wrapper->next_block_ptr = vxge_os_htonll(wrapper->next_block_ptr);
	wrapper->control_0 = vxge_os_htonll(wrapper->control_0);
}

/*
 * __hal_wrapper_ntohll - Swap the bytes in the wrapper from
 *	   network to host byte order
 * @wrapper: Current block wrapper.
 *
 * This function is used to Swap the bytes in the wrapper.
 */
static inline void
__hal_wrapper_ntohll(__hal_list_block_wrapper_t *wrapper)
{
	wrapper->next_block_ptr = vxge_os_ntohll(wrapper->next_block_ptr);
	wrapper->control_0 = vxge_os_ntohll(wrapper->control_0);
}

__EXTERN_END_DECLS

#endif /* VXGE_HAL_LB_WRAPPER_H */
