/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */
/*
 * Copyright (c) 2002-2010 Neterion, Inc.
 * All right Reserved.
 *
 * FileName :   vxgehal-lro.c
 *
 * Description:  LRO offload related routines
 *
 * Created:       11 January 2006
 */
#include "vxgehal.h"

#ifdef	VXGE_HAL_INTERNAL_COMPILER_ERROR

#pragma optimize("", off)

#endif

#ifdef	VXGE_HAL_TITAN_2

/*
 * __hal_lro_session_pool_create - Create lro session pool
 * @devh: HAL Device object.
 * @vpath: Virtual path.
 * @pool_size: Number of LRO sessions in the pool.
 *
 * This function creates LRO session pool
 */
vxge_hal_status_e
__hal_lro_session_pool_create(
	vxge_hal_device_h devh,
	__hal_virtualpath_t *vpath,
	u32 pool_size)
{
	__hal_lro_session_t *lro_session;
	__hal_device_t *hldev = (__hal_device_t *)devh;
	u32 i;

	vxge_assert((hldev != NULL) && (vpath != NULL) && (pool_size > 0));

	vxge_hal_trace_log_lro(hldev, vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_lro(hldev, vpath->vp_id,
	    "devh = 0x"VXGE_OS_STXFMT", vpath = "
	    "0x"VXGE_OS_STXFMT", pool_size = %d",
	    (ptr_t)devh, (ptr_t)vpath, pool_size);

	if (vpath == NULL) {
		vxge_hal_err_log_lro(hldev, NULL_VPID,
		    "%s:%d null pointer passed vpath = null",
		    __FILE__, __LINE__);
		vxge_hal_trace_log_lro(hldev, NULL_VPID,
		    "<==  %s:%s:%d  Result: %d",
		    __FILE__, __func__, __LINE__,
		    VXGE_HAL_ERR_INVALID_HANDLE);
		return (VXGE_HAL_ERR_INVALID_HANDLE);
	}

	for (i = 0; i < pool_size; i++) {
		lro_session = (__hal_lro_session_t *)vxge_os_malloc(
		    hldev->header.pdev,
		    sizeof (__hal_lro_session_t));
		if (lro_session == NULL) {
			vxge_hal_err_log_lro(hldev, vpath->vp_id,
			    "%s:%d Memory allocation Failed",
			    __FILE__, __LINE__);
			vxge_hal_trace_log_lro(hldev, vpath->vp_id,
			    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
			    __LINE__, VXGE_HAL_ERR_OUT_OF_MEMORY);
			__hal_lro_session_pool_destroy(devh, vpath);
			return (VXGE_HAL_ERR_OUT_OF_MEMORY);
		}

		vxge_os_memzero(lro_session, sizeof (__hal_lro_session_t));

		VXGE_HAL_LRO_SESSION_STATE_SET(lro_session,
		    VXGEHAL_LRO_SESS_STATE_INITED);

		vxge_list_insert(&lro_session->item, &vpath->free_lros);

		vpath->free_lro_count++;
	}

	vxge_hal_trace_log_lro(hldev, vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * __hal_lro_session_pool_reset - Resets lro session pool
 * @devh: HAL Device object.
 * @vpath: Virtual path.
 *
 * This function resets LRO session pool
 */

vxge_hal_status_e
__hal_lro_session_pool_reset(
	vxge_hal_device_h devh,
	__hal_virtualpath_t *vpath)
{
	vxge_list_t *n, *p;
	__hal_lro_session_t *lro_session = NULL;
	__hal_device_t *hldev = (__hal_device_t *)devh;

	vxge_assert((hldev != NULL) && (vpath != NULL));

	vxge_hal_trace_log_lro(hldev, vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_lro(hldev, vpath->vp_id,
	    "hldev = 0x"VXGE_OS_STXFMT", vpath = 0x"VXGE_OS_STXFMT,
	    (ptr_t)hldev, (ptr_t)vpath);

	if (vpath == NULL) {
		vxge_hal_err_log_lro(hldev, NULL_VPID,
		    "%s:%d null pointer passed vpath = null",
		    __FILE__, __LINE__);
		vxge_hal_trace_log_lro(hldev, NULL_VPID,
		    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
		    __LINE__, VXGE_HAL_ERR_INVALID_HANDLE);
		return (VXGE_HAL_ERR_INVALID_HANDLE);
	}

#if defined(VXGE_HAL_VP_CHANNELS)
	vxge_os_spin_lock(&vpath->lro_lock);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
	vxge_os_spin_lock_irq(&vpath->lro_lock, flags);
#endif

	if (vpath->active_lro_count > 0) {

		vxge_list_for_each_safe(p, n, &vpath->active_lros) {

			vxge_list_remove(p);

			vpath->active_lro_count--;

			__hal_vp_oid_free(vpath,
			    &vpath->session_oid_db,
			    ((__hal_lro_session_t *)p)->session_id);

			vxge_os_memzero(p, sizeof (__hal_lro_session_t));

			VXGE_HAL_LRO_SESSION_STATE_SET(
			    ((__hal_lro_session_t *)p),
			    VXGEHAL_LRO_SESS_STATE_INITED);

			vxge_list_insert(p, &vpath->free_lros);

			vpath->free_lro_count++;

		}

	}

#if defined(VXGE_HAL_VP_CHANNELS)
	vxge_os_spin_unlock(&vpath->lro_lock);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
	vxge_os_spin_unlock_irq(&vpath->lro_lock, flags);
#endif

	vxge_hal_trace_log_lro(hldev, vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);
	return (VXGE_HAL_OK);
}

/*
 * __hal_lro_session_pool_destroy - Destroys lro session pool
 * @devh: HAL Device object.
 * @vpath: Virtual path.
 *
 * This function destroys LRO session pool
 */

vxge_hal_status_e
__hal_lro_session_pool_destroy(
	vxge_hal_device_h devh,
	__hal_virtualpath_t *vpath)
{
	vxge_list_t *n, *p;
	__hal_device_t *hldev = (__hal_device_t *)devh;

	vxge_assert((hldev != NULL) && (vpath != NULL));

	vxge_hal_trace_log_lro(hldev, vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_lro(hldev, vpath->vp_id,
	    "hldev = 0x"VXGE_OS_STXFMT", vpath = 0x"VXGE_OS_STXFMT,
	    (ptr_t)hldev, (ptr_t)vpath);

	if (vpath == NULL) {
		vxge_hal_err_log_lro(hldev, NULL_VPID,
		    "%s:%d null pointer passed vpath = null",
		    __FILE__, __LINE__);
		vxge_hal_trace_log_lro(hldev, NULL_VPID,
		    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
		    __LINE__, VXGE_HAL_ERR_INVALID_HANDLE);
		return (VXGE_HAL_ERR_INVALID_HANDLE);
	}

	vxge_assert(vpath->active_lro_count == 0);

	if (vpath->active_lro_count > 0) {

		vxge_list_for_each_safe(p, n, &vpath->active_lros) {

			vxge_list_remove(p);

			vxge_os_free(hldev->header.pdev, p,
			    sizeof (__hal_lro_session_t));

			vpath->active_lro_count--;

		}

	}

	if (vpath->free_lro_count > 0) {

		vxge_list_for_each_safe(p, n, &vpath->free_lros) {

			vxge_list_remove(p);

			vxge_os_free(hldev->header.pdev, p,
			    sizeof (__hal_lro_session_t));

			vpath->free_lro_count--;

		}

	}

	vxge_hal_trace_log_lro(hldev, vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * __hal_lro_session_allocate - Allocate a LRO session from the pool
 * @devh: HAL Device object.
 * @vpath: Virtual path.
 * @session_opaque_handle: HAL Opaque hanlde
 * @src_port: Source TCP Port
 * @dest_port: Destination TCP Port
 * @ip_sa_high: Higer 64 bits of source ip address
 * @ip_sa_low: Lower 64 bits of source ip address
 * @ip_da_high: Higher 64 bits of destination ip address
 * @ip_da_low: Lower 64 bits of destination ip address
 *
 * This function allocates LRO session from pool
 */
__hal_lro_session_t *
__hal_lro_session_allocate(
	vxge_hal_device_h devh,
	__hal_virtualpath_t *vpath,
	vxge_hal_opaque_handle_t *session_opaque_handle,
	u32 src_port,
	u32 dest_port,
	u64 ip_sa_high,
	u64 ip_sa_low,
	u64 ip_da_high,
	u64 ip_da_low)
{
	vxge_hal_status_e status = VXGE_HAL_OK;
	__hal_device_t *hldev = (__hal_device_t *)devh;
	__hal_lro_session_t *lro_session = NULL;

	vxge_assert((devh != NULL) && (vpath != NULL));

	vxge_hal_trace_log_lro(hldev, vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_lro(hldev, vpath->vp_id,
	    "devh = 0x"VXGE_OS_STXFMT", vpath = 0x"VXGE_OS_STXFMT", "
	    "src_port = %d, dest_port = %d, ip_sa_high = 0x"VXGE_OS_LLXFMT
	    ", ip_sa_low = 0x"VXGE_OS_LLXFMT", ip_da_high = 0x"VXGE_OS_LLXFMT
	    ", ip_da_low = 0x"VXGE_OS_LLXFMT, (ptr_t)devh, (ptr_t)vpath,
	    src_port, dest_port, ip_sa_high, ip_sa_low, ip_da_high, ip_da_low);

#if defined(VXGE_HAL_VP_CHANNELS)
	vxge_os_spin_lock(&vpath->lro_lock);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
	vxge_os_spin_lock_irq(&vpath->lro_lock, flags);
#endif
	if (vpath->free_lro_count <= 0) {
#if defined(VXGE_HAL_VP_CHANNELS)
		vxge_os_spin_unlock(&vpath->lro_lock);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
		vxge_os_spin_unlock_irq(&vpath->lro_lock, flags);
#endif

		status = __hal_lro_sessions_timeout(devh, vpath);

#if defined(VXGE_HAL_VP_CHANNELS)
		vxge_os_spin_lock(&vpath->lro_lock);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
		vxge_os_spin_lock_irq(&vpath->lro_lock, flags);
#endif
	}

	if (status == VXGE_HAL_OK) {
		if (vpath->free_lro_count > 0) {

			lro_session = (__hal_lro_session_t *)
			    vxge_list_first_get(&vpath->free_lros);

			if (lro_session == NULL) {
				status = VXGE_HAL_FAIL;
			} else {
				status = __hal_vp_oid_allocate(
				    vpath,
				    &vpath->session_oid_db,
				    &lro_session->session_id);
			}

			if (status == VXGE_HAL_OK) {

				vxge_list_remove(&lro_session->item);
				vpath->free_lro_count--;
				vxge_list_insert(&lro_session->item,
				    &vpath->active_lros);
				vpath->active_lro_count++;

				lro_session->session_state =
				    VXGEHAL_LRO_SESS_STATE_INITED;
				lro_session->session_opaque_handle =
				    session_opaque_handle;
				lro_session->ip_sa.high = ip_sa_high;
				lro_session->ip_sa.low = ip_sa_low;
				lro_session->src_port = src_port;
				lro_session->ip_da.high = ip_da_high;
				lro_session->ip_da.low = ip_da_low;
				lro_session->dest_port = dest_port;
				lro_session->time_last_accessed =
				    vxge_os_curr_time();
				lro_session->total_bytes_recvd = 0;
				lro_session->total_cqes_recvd = 0;
			} else {
				lro_session = NULL;
			}
		} else {
			status = VXGE_HAL_ERR_TOO_MANY;
		}
	}

#if defined(VXGE_HAL_VP_CHANNELS)
	vxge_os_spin_unlock(&vpath->lro_lock);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
	vxge_os_spin_unlock_irq(&vpath->lro_lock, flags);
#endif

	vxge_hal_trace_log_lro(hldev, vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);

	return (lro_session);
}

/*
 * __hal_lro_session_timestamp - Time stamp a LRO session
 * @devh: HAL Device object.
 * @vpath: Virtual path.
 * @session_id: Object Id of session
 *
 * This function allocates LRO session from pool
 */
vxge_hal_status_e
__hal_lro_session_timestamp(
	vxge_hal_device_h devh,
	__hal_virtualpath_t *vpath,
	vxge_hal_obj_id_t session_id)
{
	vxge_hal_status_e status = VXGE_HAL_OK;
	__hal_device_t *hldev = (__hal_device_t *)devh;
	__hal_lro_session_t *lro_session =
	    (__hal_lro_session_t *)(ptr_t)session_id;

	vxge_assert((devh != NULL) && (vpath != NULL) && (session_id != 0));

	vxge_hal_trace_log_lro(hldev, vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_lro(hldev, vpath->vp_id,
	    "devh = 0x"VXGE_OS_STXFMT", vpath = 0x"VXGE_OS_STXFMT", "
	    "session_id = 0x"VXGE_OS_STXFMT, (ptr_t)devh, (ptr_t)vpath,
	    (ptr_t)session_id);

#if defined(VXGE_HAL_VP_CHANNELS)
	vxge_os_spin_lock(&vpath->lro_lock);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
	vxge_os_spin_lock_irq(&vpath->lro_lock, flags);
#endif
	if (lro_session->session_state != VXGEHAL_LRO_SESS_STATE_OPENED) {
		status = VXGE_HAL_ERR_INVALID_STATE;
	} else {
		lro_session->time_last_accessed = vxge_os_curr_time();
		vxge_list_remove(&lro_session->item);
		vxge_list_insert(&lro_session->item, &vpath->active_lros);
	}

#if defined(VXGE_HAL_VP_CHANNELS)
	vxge_os_spin_unlock(&vpath->lro_lock);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
	vxge_os_spin_unlock_irq(&vpath->lro_lock, flags);
#endif

	vxge_hal_trace_log_lro(hldev, vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);

	return (status);
}

/*
 * __hal_lro_sessions_timeout - Release the timed out sessions
 * @devh: HAL Device object.
 * @vpath: Virtual path.
 *
 * This function checks the active sessions and starts the
 * declassify and delete.
 */
vxge_hal_status_e
__hal_lro_sessions_timeout(
	vxge_hal_device_h devh,
	__hal_virtualpath_t *vpath)
{
	u64 cur_time = vxge_os_curr_time();
	vxge_list_t *p, *n;
	__hal_device_t *hldev = (__hal_device_t *)devh;
	vxge_hal_status_e  status = VXGE_HAL_OK;

	vxge_assert((devh != NULL) && (vpath != NULL));

	vxge_hal_trace_log_lro(hldev, vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_lro(hldev, vpath->vp_id,
	    "devh = 0x"VXGE_OS_STXFMT", vpath = 0x"VXGE_OS_STXFMT,
	    (ptr_t)devh, (ptr_t)vpath);

#if defined(VXGE_HAL_VP_CHANNELS)
	vxge_os_spin_lock(&vpath->lro_lock);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
	vxge_os_spin_lock_irq(&vpath->lro_lock, flags);
#endif
	vxge_list_for_each_prev_safe(p, n, &vpath->active_lros) {

		__hal_lro_session_t *lro_session = (__hal_lro_session_t *)p;

		if ((cur_time - lro_session->time_last_accessed) <
		    vpath->vp_config->lro.sessions_timeout)
			break;

		status =  vxge_hal_lro_end_classification(
		    VXGE_HAL_OPAQUE_HANDLE_GET_VPATH_HANDLE(
		    lro_session->session_opaque_handle),
		    lro_session->session_id,
		    lro_session->session_opaque_handle);

		if (status != VXGE_HAL_PENDING) {
			break;
		}

	}

#if defined(VXGE_HAL_VP_CHANNELS)
	vxge_os_spin_unlock(&vpath->lro_lock);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
	vxge_os_spin_unlock_irq(&vpath->lro_lock, flags);
#endif

	vxge_hal_trace_log_lro(hldev, vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);

	return (status);
}

/*
 * __hal_lro_session_free - Free a LRO session to the pool
 * @devh: HAL Device object.
 * @vpath: Virtual path.
 * @session_id: Object Id of session
 *
 * This function allocates LRO session from pool
 */
vxge_hal_status_e
__hal_lro_session_free(
	vxge_hal_device_h devh,
	__hal_virtualpath_t *vpath,
	vxge_hal_obj_id_t session_id)
{
	vxge_hal_status_e status = VXGE_HAL_OK;
	__hal_device_t *hldev = (__hal_device_t *)devh;
	__hal_lro_session_t *lro_session =
	    (__hal_lro_session_t *)(ptr_t)session_id;

	vxge_assert((devh != NULL) && (vpath != NULL) && (session_id != 0));

	vxge_hal_trace_log_lro(hldev, vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_lro(hldev, vpath->vp_id,
	    "devh = 0x"VXGE_OS_STXFMT", vpath = 0x"VXGE_OS_STXFMT
	    " , session_id = 0x"VXGE_OS_STXFMT, (ptr_t)devh,
	    (ptr_t)vpath, (ptr_t)session_id);

#if defined(VXGE_HAL_VP_CHANNELS)
	vxge_os_spin_lock(&vpath->lro_lock);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
	vxge_os_spin_lock_irq(&vpath->lro_lock, flags);
#endif

	if (lro_session->session_state != VXGEHAL_LRO_SESS_STATE_CLOSED) {

		status = VXGE_HAL_ERR_INVALID_STATE;

	} else {
		vxge_list_remove(&lro_session->item);

		vpath->active_lro_count--;

		__hal_vp_oid_free(vpath,
		    &vpath->session_oid_db,
		    lro_session->session_id);

		vxge_os_memzero(lro_session, sizeof (__hal_lro_session_t));

		VXGE_HAL_LRO_SESSION_STATE_SET(lro_session,
		    VXGEHAL_LRO_SESS_STATE_INITED);

		vxge_list_insert(&lro_session->item, &vpath->free_lros);

		vpath->free_lro_count++;
	}

#if defined(VXGE_HAL_VP_CHANNELS)
	vxge_os_spin_unlock(&vpath->lro_lock);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
	vxge_os_spin_unlock_irq(&vpath->lro_lock, flags);
#endif

	vxge_hal_trace_log_lro(hldev, vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);

	return (status);
}

/*
 * vxge_hal_lro_open - Post LRO Open Request Down Message
 *
 * @vpath_handle: Handle returned by message queue register
 * @session_opaque_handle: Opaque handle used by the host to reference data
 *		structures in host memory related to this session
 * @params: Parameters for LRO
 *
 * Posts the LRO Open Request DMQ message sent by the host to open
 * a LRO connection on the adapter.
 */
vxge_hal_status_e
vxge_hal_lro_open(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_opaque_handle_t *session_opaque_handle,
	vxge_hal_lro_open_params_t *params)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_srq_t *srq;
	__hal_cqrq_t *cqrq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;
	__hal_lro_session_t *lro_session;

	vxge_assert((vpath_handle != NULL) && (session_opaque_handle != NULL) &&
	    (params != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_lro(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_lro(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "session_opaque_handle = 0x"VXGE_OS_STXFMT", "
	    "params = 0x"VXGE_OS_STXFMT, (ptr_t)vpath_handle,
	    (ptr_t)session_opaque_handle, (ptr_t)params);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	srq = (__hal_srq_t *)(ptr_t)params->srq_id;

	cqrq = (__hal_cqrq_t *)(ptr_t)params->cqrq_id;

	vxge_assert((dmq != NULL) && (srq != NULL) && (cqrq != NULL));

	lro_session = __hal_lro_session_allocate(
	    (vxge_hal_device_h)vp->vpath->hldev,
	    vp->vpath,
	    session_opaque_handle,
	    0,
	    0,
	    0,
	    0,
	    params->ip_da_high,
	    params->ip_da_low);

	if (lro_session == NULL) {
		vxge_hal_trace_log_lro(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d",
		    __FILE__, __func__, __LINE__, VXGE_HAL_ERR_TOO_MANY);
		return (VXGE_HAL_ERR_TOO_MANY);
	}

	__hal_vpath_adaptive_lro_filter_set(
	    vpath_handle,
	    params->use_ipv6,
	    params->ip_da_low,
	    params->ip_da_high,
	    params->use_vlan,
	    params->vlan_id);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(session_opaque_handle,
	    vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(session_opaque_handle,
	    (ptr_t)lro_session);

	__hal_down_msg_header_set((vxge_hal_down_msg_h)&dmsg,
	    lro_session->session_id,
	    VXGE_HAL_MSG_TYPE_LRO_OPEN_REQ,
	    dmq->intr_ctrl, dmq->gen_compl,
	    session_opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_lro_open_req_dmq_t));

	dmsg.m.lro_open.session_opaque_handle = vxge_os_htonll(
	    (u64)(ptr_t)session_opaque_handle);

	dmsg.m.lro_open.control_0 = vxge_os_htonll(
	    VXGE_HAL_LRO_OPEN_REQ_LRO_MODE(VXGE_HAL_LRO_MODE_ADAPTIVE) |
	    VXGE_HAL_LRO_OPEN_REQ_NUM_LRO_SESSIONS(
	    vp->vpath->vp_config->lro.sessions_max) |
	    VXGE_HAL_LRO_OPEN_REQ_TCP_SRC_PORT(0) |
	    VXGE_HAL_LRO_OPEN_REQ_TCP_DEST_PORT(0));


	dmsg.m.lro_open.ip_sa.high = vxge_os_htonll(0);

	dmsg.m.lro_open.ip_sa.low = vxge_os_htonll(0);

	dmsg.m.lro_open.ip_da.high = vxge_os_htonll(params->ip_da_high);

	dmsg.m.lro_open.ip_da.low = vxge_os_htonll(params->ip_da_low);


	dmsg.m.lro_open.control_1 =
	    VXGE_HAL_LRO_OPEN_REQ_VLAN_ID(params->vlan_id) |
	    VXGE_HAL_LRO_OPEN_REQ_NO_WQE_THRESHOLD(
	    vp->vpath->vp_config->lro.no_wqe_threshold);

	if (params->use_ipv6)
		dmsg.m.lro_open.control_1 |= VXGE_HAL_LRO_OPEN_REQ_USE_IPV6;

	if (params->use_vlan)
		dmsg.m.lro_open.control_1 |= VXGE_HAL_LRO_OPEN_REQ_USE_VLAN;

	dmsg.m.lro_open.control_1 =
	    vxge_os_htonll(dmsg.m.lro_open.control_1);

	dmsg.m.lro_open.control_2 = vxge_os_htonll(
	    VXGE_HAL_LRO_OPEN_REQ_LRO_CQRQ_INST_NUM(
	    VXGE_HAL_RNIC_OID_DB_SIN_GET(cqrq->cqrq_id)) |
	    VXGE_HAL_LRO_OPEN_REQ_LRO_CQRQ_ID(
	    VXGE_HAL_RNIC_OID_DB_SID_GET(cqrq->cqrq_id)) |
	    VXGE_HAL_LRO_OPEN_REQ_LRO_SRQ_INST_NUM(
	    VXGE_HAL_RNIC_OID_DB_SIN_GET(srq->srq_id)) |
	    VXGE_HAL_LRO_OPEN_REQ_LRO_SRQ_ID(
	    VXGE_HAL_RNIC_OID_DB_SID_GET(srq->srq_id)));

	dmsg.m.lro_open.control_3 =
	    VXGE_HAL_LRO_OPEN_REQ_LLC_HDR_MODE(
	    vp->vpath->vp_config->lro.llc_hdr_mode) |
	    VXGE_HAL_LRO_OPEN_REQ_SNAP_HDR_MODE(
	    vp->vpath->vp_config->lro.snap_hdr_mode) |
	    VXGE_HAL_LRO_OPEN_REQ_SCATTER_MODE(
	    vp->vpath->vp_config->lro.scatter_mode) |
	    VXGE_HAL_LRO_OPEN_REQ_BUFFER_MODE(
	    vp->vpath->vp_config->lro.buffer_mode);

	if (vp->vpath->vp_config->lro.dupack_detection_enabled)
		dmsg.m.lro_open.control_3 |=
		    VXGE_HAL_LRO_OPEN_REQ_DUP_ACK_DET_EN;

	if (vp->vpath->vp_config->lro.data_merging_enabled)
		dmsg.m.lro_open.control_3 |=
		    VXGE_HAL_LRO_OPEN_REQ_DATA_MERGE_EN;

	if (vp->vpath->vp_config->lro.ack_merging_enabled)
		dmsg.m.lro_open.control_3 |= VXGE_HAL_LRO_OPEN_REQ_ACK_MERGE_EN;

	if (vp->vpath->vp_config->lro.session_ecn_enabled)
		dmsg.m.lro_open.control_3 |= VXGE_HAL_LRO_OPEN_REQ_SESS_ECN_EN;

	if (vp->vpath->vp_config->lro.session_ecn_enabled_nonce)
		dmsg.m.lro_open.control_3 |=
		    VXGE_HAL_LRO_OPEN_REQ_S_ECN_EN_NONCE;

	dmsg.m.lro_open.control_3 = vxge_os_htonll(dmsg.m.lro_open.control_3);

	dmsg.m.lro_open.control_4 = vxge_os_htonll(
	    VXGE_HAL_LRO_OPEN_REQ_DATA_LM_MASK(
	    vp->vpath->vp_config->lro.data_lm_mask) |
	    VXGE_HAL_LRO_OPEN_REQ_DUPACK_MASK(
	    vp->vpath->vp_config->lro.dupack_mask) |
	    VXGE_HAL_LRO_OPEN_REQ_LRO_NM_MASK(
	    vp->vpath->vp_config->lro.lro_nm_mask) |
	    VXGE_HAL_LRO_OPEN_REQ_ACK_LM_MASK(
	    vp->vpath->vp_config->lro.ack_lm_mask));

	dmsg.m.lro_open.control_5 = vxge_os_htonll(
	    VXGE_HAL_LRO_OPEN_REQ_MAX_IP_DG_SIZE(
	    vp->vpath->vp_config->lro.max_ip_datagram_size) |
	    VXGE_HAL_LRO_OPEN_REQ_LRO_RM_MASK(
	    vp->vpath->vp_config->lro.lro_rm_mask)|
	    VXGE_HAL_LRO_OPEN_REQ_FRAME_TH(
	    vp->vpath->vp_config->lro.frame_threshold) |
	    VXGE_HAL_LRO_OPEN_REQ_PSH_TH(
	    vp->vpath->vp_config->lro.psh_threshold));

	dmsg.m.lro_open.control_6 = vxge_os_htonll(
	    VXGE_HAL_LRO_OPEN_REQ_MTU_TH(
	    vp->vpath->vp_config->lro.mtu_threshold) |
	    VXGE_HAL_LRO_OPEN_REQ_MSS_TH(
	    vp->vpath->vp_config->lro.mss_threshold) |
	    VXGE_HAL_LRO_OPEN_REQ_MAX_TCP_TSV_DEL(
	    vp->vpath->vp_config->lro.max_tcp_tsval_delta));

	dmsg.m.lro_open.control_7 = vxge_os_htonll(
	    VXGE_HAL_LRO_OPEN_REQ_MAX_ACK_NBRV_DEL(
	    vp->vpath->vp_config->lro.max_ack_nbr_delta) |
	    VXGE_HAL_LRO_OPEN_REQ_BUF0_LENGTH(params->buffer0_length) |
	    VXGE_HAL_LRO_OPEN_REQ_BUF1_LENGTH(params->buffer1_length));

	dmsg.m.lro_open.control_8 = vxge_os_htonll(
	    VXGE_HAL_LRO_OPEN_REQ_BUF2_LENGTH(params->buffer2_length) |
	    VXGE_HAL_LRO_OPEN_REQ_BUF3_LENGTH(params->buffer3_length) |
	    VXGE_HAL_LRO_OPEN_REQ_BUF4_LENGTH(params->buffer4_length) |
	    VXGE_HAL_LRO_OPEN_REQ_MIN_SPARE_WQE_CAPACITY(
	    vp->vpath->vp_config->lro.min_spare_wqe_capacity));

	VXGE_HAL_LRO_SESSION_STATE_SET(lro_session,
	    VXGEHAL_LRO_SESS_STATE_OPEN_PENDING);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK) {
		status = VXGE_HAL_PENDING;
	} else {
		__hal_lro_session_free((vxge_hal_device_h)vp->vpath->hldev,
		    vp->vpath,
		    lro_session->session_id);
	}

	vxge_hal_trace_log_lro(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_lro_end_classification - End LRO Classification
 *
 * @vpath_handle: Handle returned by message queue register
 * @session_id: Object Id of LRO session
 * @session_opaque_handle: HAL Opaque handle
 *
 * Ends the classification of LRO session on the adapter
 */
vxge_hal_status_e
vxge_hal_lro_end_classification(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t session_id,
	vxge_hal_opaque_handle_t *session_opaque_handle)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;
	__hal_lro_session_t *lro_session =
	    (__hal_lro_session_t *)(ptr_t)session_id;

	vxge_assert((vpath_handle != NULL) && (session_id != 0) &&
	    (session_opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_lro(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_lro(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", session_id = 0x"VXGE_OS_STXFMT
	    ", session_opaque_handle = 0x"VXGE_OS_STXFMT,
	    (ptr_t)vpath_handle, (ptr_t)session_id,
	    (ptr_t)session_opaque_handle);

	if (VXGE_HAL_LRO_SESSION_STATE_GET(lro_session) !=
	    VXGEHAL_LRO_SESS_STATE_OPENED) {
		vxge_hal_trace_log_lro(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
		    __LINE__, VXGE_HAL_INF_QUEUE_IS_NOT_READY);
		return (VXGE_HAL_INF_QUEUE_IS_NOT_READY);
	}

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(session_opaque_handle,
	    vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(session_opaque_handle,
	    session_id);

	__hal_lro_end_classification_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    lro_session->session_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    session_opaque_handle);

	VXGE_HAL_LRO_SESSION_STATE_SET(lro_session,
	    VXGEHAL_LRO_SESS_STATE_DECLASS_PENDING);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK) {
		status = VXGE_HAL_PENDING;
	} else {
		VXGE_HAL_LRO_SESSION_STATE_SET(lro_session,
		    VXGEHAL_LRO_SESS_STATE_OPENED);
	}

	vxge_hal_trace_log_lro(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_lro_delete - Delete a LRO session
 *
 * @vpath_handle: Handle returned by message queue register
 * @session_id: Object Id of LRO session
 * @session_opaque_handle: HAL Opaque handle
 *
 * Deletes a LRO session on the adapter.
 */
vxge_hal_status_e
vxge_hal_lro_delete(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t session_id,
	vxge_hal_opaque_handle_t *session_opaque_handle)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_dmq_t *dmq;
	__hal_device_t *hldev;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;
	__hal_lro_session_t *lro_session =
	    (__hal_lro_session_t *)(ptr_t)session_id;

	vxge_assert((vpath_handle != NULL) && (session_id != 0) &&
	    (session_opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_lro(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_lro(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", session_id = 0x"VXGE_OS_STXFMT
	    ", session_opaque_handle = 0x"VXGE_OS_STXFMT,
	    (ptr_t)vpath_handle, (ptr_t)session_id,
	    (ptr_t)session_opaque_handle);

	if (VXGE_HAL_LRO_SESSION_STATE_GET(lro_session) !=
	    VXGEHAL_LRO_SESS_STATE_DECLASSED) {
		vxge_hal_trace_log_lro(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
		    __LINE__, VXGE_HAL_INF_QUEUE_IS_NOT_READY);
		return (VXGE_HAL_INF_QUEUE_IS_NOT_READY);
	}

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(session_opaque_handle,
	    vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(session_opaque_handle,
	    session_id);

	__hal_lro_delete_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    lro_session->session_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    session_opaque_handle);

	VXGE_HAL_LRO_SESSION_STATE_SET(lro_session,
	    VXGEHAL_LRO_SESS_STATE_CLOSE_PENDING);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK) {
		status = VXGE_HAL_PENDING;
	} else {
		VXGE_HAL_LRO_SESSION_STATE_SET(lro_session,
		    VXGEHAL_LRO_SESS_STATE_DECLASSED);
	}

	vxge_hal_trace_log_lro(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);
	return (status);
}

#endif  /* VXGE_HAL_TITAN_2 */

#if defined(VXGE_HAL_USE_SW_LRO)

/*
 * __hal_l4_pyld_length_get : Find the tcp seg len.
 * @tcp: tcp header.
 * @ip: ip header.
 *
 * Finds the tcp seg len.
 */
static inline u32
__hal_l4_pyld_length_get(vxge_hal_tcp_hdr_t *tcp, vxge_hal_ip_hdr_t *ip)
{
	u32	 ret;
	ret = (vxge_os_ntohs(ip->tot_len) - ((ip->version_ihl & 0x0F) <<2) -
	    (((tcp->doff_res & 0xF0) >> 4) << 2));
	return (ret);
}

vxge_hal_status_e
__hal_lro_capable(__hal_virtualpath_t *vpath, u8 *buffer,
			vxge_hal_ip_hdr_t **ip,  vxge_hal_tcp_hdr_t **tcp,
			vxge_hal_ring_rxd_info_t *ext_info)
{
	u8 ip_off, ip_length;
	u32 daddr;
	u16 vlan_id = 0;

	/* Check whether it is a TCP Packet Non TCP packets can not be LROed */
	if (!(ext_info->proto & VXGE_HAL_FRAME_PROTO_TCP)) {
		return (VXGE_HAL_FAIL);
	}

	if (!*ip) {

		if ((ext_info->frame == VXGE_HAL_FRAME_TYPE_DIX)||
		    (ext_info->is_vlan ==  1)) {

			ip_off = VXGE_HAL_HEADER_ETHERNET_II_802_3_SIZE;
			/* Get the Vlan ID of the frame */
			vlan_id =
			    (u16)(VXGE_HAL_VLAN_VID_MASK & ext_info->vlan);
			/*
			 * If vlan stripping is disabled and the frame is VLAN
			 * tagged, shift the offset by the VLAN header size
			 * bytes.
			 */
			if ((vpath->vp_config->rpa_strip_vlan_tag ==
			    VXGE_HAL_VPATH_RPA_STRIP_VLAN_TAG_DISABLE) &&
			    vlan_id)
				ip_off += VXGE_HAL_HEADER_VLAN_SIZE;

		} else {
			/* LLC, SNAP etc are considered non-mergeable */
			return (VXGE_HAL_FAIL);
		}


		/* Grab ip headers */
		*ip = (vxge_hal_ip_hdr_t *)((char *)buffer + ip_off);
	} /* !*ip */

	ip_length = (u8)((*ip)->version_ihl & 0x0F);
	ip_length = ip_length << 2;

	/* Grab the tcp header */
	*tcp = (vxge_hal_tcp_hdr_t *)((char *)*ip + ip_length);

	if (vpath->hldev->header.config.lro_enable ==
	    VXGE_HAL_LRO_DONT_AGGR_FWD_PKTS) {
		daddr = (*ip)->daddr;
		/* Check if it is a broadcast or multicast ip */
		if (!vxge_os_in_multicast(daddr) &&
		    (VXGE_OS_INADDR_BROADCAST != daddr)) {

			/*
			 * Does this packets destined for this interface?
			 */
			if (!vxge_os_is_my_packet(
			    vpath->hldev->header.pdev, daddr))
				return (VXGE_HAL_OK);

			/* Check if it is a vlan packet */
			/* FIXME Need to be ported */

			return (VXGE_HAL_FAIL);

		}

	}
	return (VXGE_HAL_OK);

}

static int __hal_update_tcp_timestamp_slow(vxge_hal_tcp_hdr_t *th,
				vxge_hal_sw_lro_t *lro, int save)
{
	unsigned char *ptr;
	int opt_cnt = 0;
	int length = (((th->doff_res & 0xf0) >> 4) << 2) -
	    sizeof (vxge_hal_tcp_hdr_t);

	ptr = (unsigned char *)(th + 1);

	while (length > 0) {
		int opcode = *ptr++;
		int opsize;

		switch (opcode) {
		case VXGE_HAL_TCPOPT_EOL:
			return (1);
		case VXGE_HAL_TCPOPT_NOP:
			length--;
			continue;
		default:
			/* Not sure about this check, but not taking a chance */
			if ((opcode == VXGE_HAL_TCPOPT_SACK_PERM) ||
			    (opcode == VXGE_HAL_TCPOPT_SACK))
				return (1);
			opsize = *ptr++;
			if (opsize < 2)
				return (1);
			/* don't parse partial options */
			if (opsize > length)
				return (1);
			if (++opt_cnt > 3)
				return (1);
			if (opcode == VXGE_HAL_TCPOPT_TIMESTAMP) {
				if (opsize == VXGE_HAL_TCPOLEN_TIMESTAMP) {

				if (save == VXGE_HAL_TS_SAVE) {
					lro->cur_tsval =
					    vxge_os_ntohl(*(vxge_os_be32 *)ptr);
					lro->cur_tsecr =
					    *(vxge_os_be32 *)(ptr + 4);
				} else if (save == VXGE_HAL_TS_VERIFY) {
					/*
					 * Ensure timestamp value increases
					 * monotonically
					 */
					if (lro->cur_tsval > vxge_os_ntohl(
					    *((vxge_os_be32 *)ptr)))
						return (-1);
					/*
					 * timestamp echo reply should be
					 * non-zero
					 */
					if (*((vxge_os_be32 *)(ptr + 4)) == 0)
						return (-1);
				} else {
					vxge_os_be32 *tmp_ptr =
					    (vxge_os_be32 *)(ptr + 4);
					*tmp_ptr = lro->cur_tsecr;
				}
				return (0);
			}
		}
		ptr += opsize-2;
		length -= opsize;
		}
	}
	return (1);
}

int __hal_update_tcp_timestamp(vxge_hal_tcp_hdr_t *tcph,
				vxge_hal_sw_lro_t *lro, int save) {

	if (((tcph->doff_res & 0xf0) >> 4) ==
	    sizeof (struct vxge_hal_tcp_hdr_t) >> 2) {
		return (VXGE_HAL_FAIL);

	} else if (((tcph->doff_res & 0xf0) >> 4) ==
	    (sizeof (struct vxge_hal_tcp_hdr_t) >> 2)
	    + (VXGE_HAL_TCPOLEN_TSTAMP_ALIGNED >> 2)) {

		vxge_os_be32 *ptr = (vxge_os_be32 *)(tcph + 1);

		if (*ptr == vxge_os_htonl((VXGE_HAL_TCPOPT_NOP << 24) |
		    (VXGE_HAL_TCPOPT_NOP << 16) |
		    (VXGE_HAL_TCPOPT_TIMESTAMP << 8) |
		    VXGE_HAL_TCPOLEN_TIMESTAMP)) {

			++ptr;
			if (save == VXGE_HAL_TS_SAVE) {
				lro->cur_tsval =
				    vxge_os_ntohl(*(vxge_os_be32 *)ptr);
				lro->cur_tsecr = *(vxge_os_be32 *)(ptr + 1);
			} else if (save == VXGE_HAL_TS_VERIFY) {
			/* Ensure timestamp value increases monotonically */
				if (lro->cur_tsval >
					vxge_os_ntohl(*((vxge_os_be32 *)ptr)))
					return (VXGE_HAL_FAIL);

				/* timestamp echo reply should be non-zero */
				if (*((vxge_os_be32 *)(ptr + 1)) == 0)
					return (VXGE_HAL_FAIL);
			} else
				*(ptr + 1) = lro->cur_tsecr;

		return (VXGE_HAL_OK);
		}
	}
	return (__hal_update_tcp_timestamp_slow(tcph, lro, save));
}

vxge_hal_status_e
__hal_lro_l3_l4_lro_capable(vxge_hal_ip_hdr_t *ip,
				vxge_hal_ring_rxd_info_t *rxd_info,
				vxge_hal_sw_lro_t *lro,
				vxge_hal_tcp_hdr_t *tcp, u32 tcp_pyld_len,
				u8 aggr_ack) {
	u8 ip_length;
	if (!aggr_ack && !tcp_pyld_len) {
		/* Pure ACK */
		return (VXGE_HAL_FAIL);
	}

	/* Ensure there are no IP options */
	ip_length = (u8)(ip->version_ihl & 0x0F);
	ip_length = ip_length << 2;
	if (ip_length != sizeof (*ip))
		return (VXGE_HAL_FAIL);

	/* IP packet is not fragmented */
	if (rxd_info->proto & VXGE_HAL_FRAME_PROTO_IP_FRAG)
		return (VXGE_HAL_FAIL);

	/* If we see CE codepoint in IP header, packet is not mergeable */
	if ((ip->tos & VXGE_HAL_INET_ECN_MASK) == VXGE_HAL_INET_ECN_CE)
		return (VXGE_HAL_FAIL);

	/*
	 * If we see ECE or CWR flags or CTRL flags in TCP header,
	 * packet is not mergeable
	 */
	if ((tcp->ctrl & VXGE_HAL_TCP_CTRL_URG_MASK) ||
	    (tcp->ctrl & VXGE_HAL_TCP_CTRL_PSH_MASK) ||
	    (tcp->ctrl & VXGE_HAL_TCP_CTRL_RST_MASK) ||
	    (tcp->ctrl & VXGE_HAL_TCP_CTRL_SYN_MASK) ||
	    (tcp->ctrl & VXGE_HAL_TCP_CTRL_FIN_MASK) ||
	    (tcp->ctrl & VXGE_HAL_TCP_ECN_ECE_MASK) ||
	    (tcp->ctrl & VXGE_HAL_TCP_ECN_CWR_MASK) ||
	    (!(tcp->ctrl &VXGE_HAL_TCP_CTRL_ACK_MASK))) {
		/*
		 * Currently recognize only the ack control word and
		 * any other control field being set would result in
		 * flushing the LRO session
		 */
		return (VXGE_HAL_FAIL);
	}
	/* FIXME : Update Time stamp details */
	if (lro) {
		if (__hal_update_tcp_timestamp(tcp, lro, VXGE_HAL_TS_VERIFY)
		    == -1)
			return (VXGE_HAL_FAIL);
	}
	return (VXGE_HAL_OK);
}
/*
 * __hal_append_lro: Appends new frame to existing LRO session.
 * @lro: lro pointer
 * @tcp: IN tcp	header,	OUT tcp	payload.
 * @ip:	ip header.
 * @seg_len: tcp payload length.
 *
 * Appends new frame to existing LRO session.
 */
static inline vxge_hal_status_e
__hal_append_lro(vxge_hal_sw_lro_t *lro,
		vxge_hal_tcp_hdr_t **tcp,
		vxge_hal_ip_hdr_t *ip,
		u32 * tcp_seg_len)
{
	*tcp_seg_len = __hal_l4_pyld_length_get(*tcp, ip);
	lro->total_length += *tcp_seg_len;
	lro->frags_len += *tcp_seg_len;
	lro->tcp_next_seq_num += *tcp_seg_len;
	lro->window = (*tcp)->window;
	if (lro->saw_ts)
		__hal_update_tcp_timestamp((*tcp), lro, VXGE_HAL_TS_SAVE);
	/*
	 * Update mbuf chain will be done in ll	driver.
	 * xge_hal_accumulate_large_rx on success of appending new frame to
	 * lro will return to ll driver	tcpdata	pointer, and tcp payload length.
	 * along with return code lro frame appended.
	 */
	lro->tcp_seq_num = (*tcp)->seq;
	lro->tcp_ack_num = (*tcp)->ack_seq;

	lro->sg_num++;
	*tcp = (vxge_hal_tcp_hdr_t *)((char *)*tcp +
	    ((((*tcp)->doff_res & 0xF0) >> 4) << 2));

	return (VXGE_HAL_OK);
}

void vxge_hal_update_L3L4_header(vxge_hal_vpath_h vpath_handle,
				vxge_hal_sw_lro_t *lro)
{
	vxge_hal_tcp_hdr_t *tcp = lro->tcp_hdr;
	vxge_hal_ip_hdr_t *ip = lro->ip_hdr;
	u16 nchk;
	__hal_ring_t *ring;
	__hal_virtualpath_t *vpath;

	vxge_assert(vpath_handle != NULL);

	vpath = (__hal_virtualpath_t *)
	    ((__hal_vpath_handle_t *)vpath_handle)->vpath;

	ring = (__hal_ring_t *)vpath->ringh;

	/* Update L3 Header */
	ip->tot_len = vxge_os_htons(lro->total_length);
	ip->check = 0;
	nchk = VXGE_LL_IP_FAST_CSUM((u8 *)ip,  ip->version_ihl & 0x0F);
	ip->check = nchk;

	/* Update L4 Header */
	tcp->ack_seq = lro->tcp_ack_num;
	tcp->window = lro->window;

	/* Update tsecr field if this session has timestamps enabled */
	if (lro->saw_ts)
		__hal_update_tcp_timestamp(tcp, lro, VXGE_HAL_TS_UPDATE);

	ring->stats->lro_num_aggregations++;
	ring->stats->lro_sum_avg_pkts_aggregated += lro->sg_num;
}

/*
 * __hal_lro_check_for_session_match: Check if frame belongs to given lro.
 * @lro: LRO session
 * @tcp: tcp header.
 * @ip: ip header.
 *
 * Check if frame belongs to given lro.
 */
static inline vxge_hal_status_e
__hal_lro_check_for_session_match(vxge_hal_sw_lro_t *lro,
				vxge_hal_tcp_hdr_t *tcp, vxge_hal_ip_hdr_t *ip)
{
	if ((lro->ip_hdr->saddr != ip->saddr) ||
	    (lro->ip_hdr->daddr != ip->daddr) ||
	    (lro->tcp_hdr->source != tcp->source) ||
	    (lro->tcp_hdr->dest != tcp->dest))
		return (VXGE_HAL_FAIL);

	return (VXGE_HAL_OK);
}

/*
 * __hal_lro_under_optimal_thresh: Finds whether combined session is optimal.
 * @ring: Ring
 * @lro: lro pointer
 * @tcp: tcp header.
 * @ip:	ip header.
 *
 * Finds whether combined session is optimal.
 */
static inline vxge_hal_status_e
__hal_lro_under_optimal_thresh(__hal_ring_t *ring, vxge_hal_sw_lro_t *lro,
				vxge_hal_tcp_hdr_t *tcp, vxge_hal_ip_hdr_t *ip)
{
	if (!lro)
		return (VXGE_HAL_FAIL);

	if ((lro->total_length + (int)__hal_l4_pyld_length_get(tcp, ip)) >
	    ring->config->sw_lro_frm_len) {
		return (VXGE_HAL_FAIL);
	}

	if (lro->sg_num	 == ring->config->sw_lro_sg_size) {
		return (VXGE_HAL_FAIL);
	}

	return (VXGE_HAL_OK);
}

/*
 * __hal_open_lro_session: Open a new LRO session.
 * @ring: ring
 * @buffer: Ethernet frame.
 * @tcp: tcp header.
 * @ip: ip header.
 * @lro: lro pointer
 * @tcp_seg_len: Length of tcp segment.
 *
 * Opens a new LRO session.
 */
static inline  vxge_hal_status_e
__hal_open_lro_session(__hal_ring_t *ring, u8 *buffer,
			vxge_hal_tcp_hdr_t *tcp, vxge_hal_ip_hdr_t *ip,
			vxge_hal_sw_lro_t **lro, u32 tcp_seg_len)
{
	vxge_list_t *p;

#if defined(VXGE_HAL_VP_CHANNELS)
	vxge_os_spin_lock(&ring->sw_lro_lock);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
	vxge_os_spin_lock_irq(&ring->sw_lro_lock, flags);
#endif

	p = vxge_list_first_get(&ring->free_sw_lros);

	if (p != NULL) {
		vxge_list_remove(p);
		vxge_list_insert(p, &ring->active_sw_lros);
		*lro = VXGE_HAL_SW_LRO_SESSION(p);
	} else {
		*lro = NULL;
		return (VXGE_HAL_INF_SW_LRO_UNCAPABLE);
	}

#if defined(VXGE_HAL_VP_CHANNELS)
	vxge_os_spin_unlock(&ring->sw_lro_lock);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
	vxge_os_spin_unlock_irq(&ring->sw_lro_lock, flags);
#endif

	(*lro)->ip_hdr			= ip;
	(*lro)->tcp_hdr			= tcp;
	(*lro)->window			= tcp->window;
	(*lro)->tcp_next_seq_num	= tcp_seg_len + vxge_os_ntohl(tcp->seq);
	(*lro)->tcp_seq_num		= tcp->seq;
	(*lro)->tcp_ack_num		= tcp->ack_seq;
	(*lro)->sg_num			= 1;
	(*lro)->total_length		= vxge_os_ntohs(ip->tot_len);
	(*lro)->frags_len		= 0;

	/*
	 * check if we saw TCP timestamp. Other consistency checks have
	 * already been done.
	 */
	if (!__hal_update_tcp_timestamp(tcp, *lro, VXGE_HAL_TS_SAVE))
		(*lro)->saw_ts = 1;

return (VXGE_HAL_OK);

}

/*
 * __hal_get_lro_session: Gets matching LRO session or creates one.
 * @ring: Ring
 * @rxd_info: Descriptor info.
 * @eth_hdr:	Ethernet header.
 * @tcp: tcp header.
 * @ip: ip header.
 * @lro: lro pointer
 *
 * Gets matching LRO session or creates one.
 */
static inline  vxge_hal_status_e
__hal_get_lro_session(__hal_ring_t *ring, vxge_hal_ring_rxd_info_t *rxd_info,
			u8 *eth_hdr, vxge_hal_tcp_hdr_t *tcp,
			vxge_hal_ip_hdr_t *ip, vxge_hal_sw_lro_t **lro)
{
	vxge_list_t *p, *n;
	u32	 tcp_seg_len;
	/* FIXME Need to add this aggr_ack as configuration param */
	u8 aggr_ack = 0;

	*lro = NULL;

	/*
	 * Search in the pool of LROs for the session that matches
	 * the incoming frame.
	 */
	vxge_list_for_each_safe(p, n, &ring->active_sw_lros) {
		if (__hal_lro_check_for_session_match(
		    VXGE_HAL_SW_LRO_SESSION(p), tcp, ip) == VXGE_HAL_OK) {
			*lro = VXGE_HAL_SW_LRO_SESSION(p);
			break;
		}
	}

	tcp_seg_len = __hal_l4_pyld_length_get(tcp, ip);
	if (*lro) {
		/*
		 * Matching LRO Session found
		 */
		if ((*lro)->tcp_next_seq_num != vxge_os_ntohl(tcp->seq)) {
			/* Out of Order Packets */
			ring->stats->lro_outof_sequence_pkts++;
			return (VXGE_HAL_INF_SW_LRO_FLUSH_BOTH);
		}
		if (__hal_lro_l3_l4_lro_capable(ip, rxd_info, *lro, tcp,
		    tcp_seg_len, aggr_ack))
			return (VXGE_HAL_INF_SW_LRO_FLUSH_BOTH);

		/*
		 * The frame is good, in-sequence, can be LRO-ed;
		 * take its (latest) ACK - unless it is a dupack.
		 * Note: to be exact need to check window size as well..
		 */
		if ((*lro)->tcp_ack_num == tcp->ack_seq &&
		    (*lro)->tcp_seq_num == tcp->seq) {
			return (VXGE_HAL_INF_SW_LRO_FLUSH_BOTH);
		}

		return (VXGE_HAL_INF_SW_LRO_CONT);
	}

	/* ********** New Session ************** */

	if (__hal_lro_l3_l4_lro_capable(ip, rxd_info, *lro, tcp,
	    tcp_seg_len, aggr_ack))
		return (VXGE_HAL_INF_SW_LRO_UNCAPABLE);


	if (__hal_open_lro_session(ring, eth_hdr, tcp, ip, lro, tcp_seg_len) !=
	    VXGE_HAL_OK)
		return (VXGE_HAL_INF_SW_LRO_UNCAPABLE);

	return (VXGE_HAL_INF_SW_LRO_BEGIN);
}

/*
 * vxge_hal_sw_lro_session_close: Close LRO session
 * @vpath_handle: Virtual Path handle.
 * @sw_lro: LRO Session. Please see vxge_hal_sw_lro_t {}
 *
 * Closes the LRO session
 */
void
vxge_hal_sw_lro_session_close(vxge_hal_vpath_h vpath_handle,
				vxge_hal_sw_lro_t *sw_lro)
{
	__hal_ring_t *ring;
	__hal_virtualpath_t *vpath;
	__hal_device_t *hldev;

	vxge_assert(vpath_handle != NULL);

	vpath = (__hal_virtualpath_t *)((__hal_vpath_handle_t *)
	    vpath_handle)->vpath;

	hldev = (__hal_device_t *)vpath->hldev;

	vxge_hal_trace_log_ring(hldev, vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	ring = (__hal_ring_t *)vpath->ringh;

	vxge_hal_trace_log_ring(hldev, vpath->vp_id,
	    "ring = 0x"VXGE_OS_STXFMT, (ptr_t)ring);

#if defined(VXGE_HAL_VP_CHANNELS)
	vxge_os_spin_lock(&ring->sw_lro_lock);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
	vxge_os_spin_lock_irq(&ring->sw_lro_lock, flags);
#endif

	vxge_list_remove((vxge_list_t *)VXGE_HAL_SW_LRO_SESSION_ITEM(sw_lro));

	ring->active_sw_lro_count--;

	vxge_os_memzero(sw_lro, VXGE_HAL_SW_LRO_SESSION_SIZE);

	vxge_list_insert((vxge_list_t *)VXGE_HAL_SW_LRO_SESSION_ITEM(sw_lro),
	    &ring->free_sw_lros);

	ring->free_sw_lro_count++;

#if defined(VXGE_HAL_VP_CHANNELS)
	vxge_os_spin_unlock(&ring->sw_lro_lock);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
	vxge_os_spin_unlock_irq(&ring->sw_lro_lock, flags);
#endif

	vxge_hal_trace_log_ring(hldev, vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__,
	    __LINE__);

}


/*
 * xge_hal_sw_lro_next_session_get: get the next LRO session
 * @vpath_handle: Virtual Path handle.
 * @sw_lro: Current LRO Session. If zero, the function will return the
 *		first LRO session
 *
 * Returns the next LRO session
 */
vxge_hal_sw_lro_t *
vxge_hal_sw_lro_next_session_get(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_sw_lro_t *sw_lro)
{
	__hal_ring_t *ring;
	__hal_virtualpath_t *vpath;
	vxge_hal_sw_lro_t *next_lro = NULL;
	__hal_device_t *hldev;

	vxge_assert(vpath_handle != NULL);

	vpath = (__hal_virtualpath_t *)
	    ((__hal_vpath_handle_t *)vpath_handle)->vpath;
	hldev = (__hal_device_t *)vpath->hldev;

	vxge_hal_trace_log_ring(hldev, vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	ring = (__hal_ring_t *)vpath->ringh;
	vxge_hal_trace_log_ring(hldev, vpath->vp_id,
	    "ring = 0x"VXGE_OS_STXFMT, (ptr_t)ring);

#if defined(VXGE_HAL_VP_CHANNELS)
	vxge_os_spin_lock(&ring->sw_lro_lock);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
	vxge_os_spin_lock_irq(&ring->sw_lro_lock, flags);
#endif

	if (sw_lro == NULL)
		next_lro = VXGE_HAL_SW_LRO_SESSION(vxge_list_first_get(
		    &ring->active_sw_lros));
	else
		next_lro = VXGE_HAL_SW_LRO_SESSION(vxge_list_first_get(
		    (vxge_list_t *)
		    VXGE_HAL_SW_LRO_SESSION_ITEM(sw_lro)));

#if defined(VXGE_HAL_VP_CHANNELS)
	vxge_os_spin_unlock(&ring->sw_lro_lock);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
	vxge_os_spin_unlock_irq(&ring->sw_lro_lock, flags);
#endif

	if (next_lro != NULL) {
		next_lro->ip_hdr->tot_len = vxge_os_htons(
		    next_lro->total_length);
		next_lro->ip_hdr->check = vxge_os_htons(0);
		next_lro->ip_hdr->check = VXGE_LL_IP_FAST_CSUM(
		    ((u8 *)(next_lro->ip_hdr)),
		    (next_lro->ip_hdr->version_ihl & 0x0F));
	}

	vxge_hal_trace_log_ring(hldev, vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (next_lro);
}

/*
 * __hal_sw_lro_terminate - Terminate lro resources.
 * @ring: Hal ring structure.
 *
 * Terminate lro resources.
 */
vxge_hal_status_e
__hal_sw_lro_terminate(__hal_ring_t *ring)
{
	vxge_list_t *n, *p;
	__hal_virtualpath_t *vpath;
	__hal_device_t *hldev;
	vxge_assert(ring != NULL);

	vpath = (__hal_virtualpath_t *)
	    ((__hal_vpath_handle_t *)ring->channel.vph)->vpath;

	hldev = (__hal_device_t *)vpath->hldev;

	vxge_hal_trace_log_ring(hldev, vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_ring(hldev, vpath->vp_id,
	    "ring = 0x"VXGE_OS_STXFMT, (ptr_t)ring);

#if defined(VXGE_HAL_VP_CHANNELS)
	vxge_os_spin_lock(&ring->sw_lro_lock);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
	vxge_os_spin_lock_irq(&ring->sw_lro_lock, flags);
#endif

	if (ring->active_sw_lro_count > 0) {

		vxge_list_for_each_safe(p, n, &ring->active_sw_lros) {

			vxge_list_remove(p);

			vxge_os_free(vpath->hldev->header.pdev,
			    VXGE_HAL_SW_LRO_SESSION(p),
			    VXGE_HAL_SW_LRO_SESSION_SIZE);

			ring->active_sw_lro_count--;

		}

	}

	if (ring->free_sw_lro_count > 0) {

		vxge_list_for_each_safe(p, n, &ring->free_sw_lros) {

			vxge_list_remove(p);

			vxge_os_free(vpath->hldev->header.pdev,
			    VXGE_HAL_SW_LRO_SESSION(p),
			    VXGE_HAL_SW_LRO_SESSION_SIZE);

			ring->free_sw_lro_count--;

		}

	}

#if defined(VXGE_HAL_VP_CHANNELS)
	vxge_os_spin_unlock(&ring->sw_lro_lock);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
	vxge_os_spin_unlock_irq(&ring->sw_lro_lock, flags);
#endif

#if defined(VXGE_HAL_VP_CHANNELS)
	vxge_os_spin_lock_destroy(&ring->sw_lro_lock,
	    vpath->hldev->header.pdev);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
	vxge_os_spin_lock_destroy_irq(&ring->sw_lro_lock,
	    vpath->hldev->header.pdev);
#endif

	vxge_hal_trace_log_ring(vpath->hldev, vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * __hal_sw_lro_init - Initiate lro resources.
 * @ring: Hal ring structure.
 *
 * Initiate lro resources.
 */

vxge_hal_status_e
__hal_sw_lro_init(__hal_ring_t *ring)
{
	int i;
	vxge_hal_sw_lro_t *lro_session;
	__hal_virtualpath_t *vpath;
	__hal_device_t *hldev;
	vxge_assert(ring != NULL);

	vpath = (__hal_virtualpath_t *)
	    ((__hal_vpath_handle_t *)ring->channel.vph)->vpath;
	hldev = (__hal_device_t *)vpath->hldev;

	vxge_hal_trace_log_ring(hldev, vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_ring(hldev, vpath->vp_id,
	    "ring = 0x"VXGE_OS_STXFMT, (ptr_t)ring);

	vxge_list_init(&ring->active_sw_lros);

	ring->active_sw_lro_count = 0;

	vxge_list_init(&ring->free_sw_lros);

	ring->free_sw_lro_count = 0;

#if defined(VXGE_HAL_VP_CHANNELS)
	vxge_os_spin_lock_init(&ring->sw_lro_lock, vpath->hldev->pdev);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
	vxge_os_spin_lock_init_irq(&ring->sw_lro_lock, vpath->hldev->irqh);
#endif

	for (i = 0; i < ring->config->sw_lro_sessions; i++) {
		lro_session = (vxge_hal_sw_lro_t *)vxge_os_malloc(
		    vpath->hldev->header.pdev,
		    VXGE_HAL_SW_LRO_SESSION_SIZE);
		if (lro_session == NULL) {
			vxge_hal_trace_log_ring(hldev, vpath->vp_id,
			    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
			    __LINE__, VXGE_HAL_ERR_OUT_OF_MEMORY);
			__hal_sw_lro_terminate(ring);
			return (VXGE_HAL_ERR_OUT_OF_MEMORY);
		}

		vxge_os_memzero(lro_session, VXGE_HAL_SW_LRO_SESSION_SIZE);

		vxge_list_insert((vxge_list_t *)
		    VXGE_HAL_SW_LRO_SESSION_ITEM(lro_session),
		    &ring->free_sw_lros);

		ring->free_sw_lro_count++;
	}

	vxge_hal_trace_log_ring(hldev, vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * __hal_sw_lro_reset - Reset lro resources.
 * @ring: Hal ring structure.
 *
 * Reset lro resources.
 */
vxge_hal_status_e
__hal_sw_lro_reset(__hal_ring_t *ring)
{
	vxge_list_t *n, *p;
	__hal_virtualpath_t *vpath;
	vxge_hal_sw_lro_t *lro;
	__hal_device_t *hldev;
	vxge_assert(ring != NULL);

	vpath = (__hal_virtualpath_t *)
	    ((__hal_vpath_handle_t *)ring->channel.vph)->vpath;

	hldev = (__hal_device_t *)vpath->hldev;
	vxge_hal_trace_log_ring(hldev, vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_ring(hldev, vpath->vp_id,
	    "ring = 0x"VXGE_OS_STXFMT, (ptr_t)ring);

#if defined(VXGE_HAL_VP_CHANNELS)
	vxge_os_spin_lock(&ring->sw_lro_lock);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
	vxge_os_spin_lock_irq(&ring->sw_lro_lock, flags);
#endif

	if (ring->active_sw_lro_count > 0) {

		vxge_list_for_each_safe(p, n, &ring->active_sw_lros) {

			vxge_list_remove(p);

			ring->active_sw_lro_count--;
			lro = VXGE_HAL_SW_LRO_SESSION(p);
			if (lro)
				vxge_os_memzero(lro,
				    VXGE_HAL_SW_LRO_SESSION_SIZE);

			vxge_list_insert(p, &ring->free_sw_lros);

			ring->free_sw_lro_count++;

		}

	}

#if defined(VXGE_HAL_VP_CHANNELS)
	vxge_os_spin_unlock(&ring->sw_lro_lock);
#elif defined(VXGE_HAL_VP_CHANNELS_IRQ)
	vxge_os_spin_unlock_irq(&ring->sw_lro_lock, flags);
#endif

	vxge_hal_trace_log_ring(hldev, vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__,
	    __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * vxge_hal_sw_lro_rx_process: Process Rx Buffer
 * @vpath_handle: Virtual Path handle.
 * @rxd_info: Please see vxge_hal_ring_rxd_info_t {}
 * @eth_hdr: Ethernet header (start of frame buffer)
 * @ip: Ip Header. Please see vxge_hal_ip_hdr_t {}
 * @tcp: Buffer to return pointer to vxge_hal_tcp_hdr_t {}
 * @sw_lro: Buffer to return pointer to vxge_hal_sw_lro_t {}
 *
 * Processes the rx buffer and either creates new lro session or
 * appends to the existing lro session
 */
vxge_hal_status_e
vxge_hal_sw_lro_rx_process(
			vxge_hal_vpath_h vpath_handle,
			vxge_hal_ring_rxd_info_t *rxd_info,
			u8 *eth_hdr,
			u32 *seglen,
			vxge_hal_sw_lro_t **sw_lro)
{
	__hal_ring_t *ring;
	__hal_virtualpath_t *vpath;
	vxge_hal_status_e status;
	vxge_hal_ip_hdr_t *ip = NULL;
	vxge_hal_tcp_hdr_t *tcp = NULL;
	__hal_device_t *hldev;

	vxge_assert(vpath_handle != NULL);

	vpath = (__hal_virtualpath_t *)
	    ((__hal_vpath_handle_t *)vpath_handle)->vpath;
	hldev = (__hal_device_t *)vpath->hldev;

	vxge_hal_trace_log_ring(hldev, vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	ring = (__hal_ring_t *)vpath->ringh;
	vxge_hal_trace_log_ring(hldev, vpath->vp_id,
	    "ring = 0x"VXGE_OS_STXFMT, (ptr_t)ring);

	if (__hal_lro_capable(vpath, eth_hdr, &ip, &tcp, rxd_info) !=
	    VXGE_HAL_OK) {
		vxge_hal_trace_log_ring(hldev, vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
		    __LINE__, VXGE_HAL_INF_SW_LRO_UNCAPABLE);
		return (VXGE_HAL_INF_SW_LRO_UNCAPABLE);
	}

	/*
	 * This	function shall get matching LRO or else
	 * create one and return it
	 */
	status = __hal_get_lro_session(ring, rxd_info, eth_hdr,
	    tcp, ip, sw_lro);

	if (status == VXGE_HAL_INF_SW_LRO_CONT) {
		if (__hal_lro_under_optimal_thresh(ring, *sw_lro, tcp, ip) ==
		    VXGE_HAL_OK) {
			(void) __hal_append_lro(*sw_lro, &tcp, ip, seglen);
			ring->stats->lro_clubbed_frms_cnt++;

			if ((*sw_lro)->sg_num >= ring->config->sw_lro_sg_size) {
				status = VXGE_HAL_INF_SW_LRO_FLUSH_SESSION;
			}

		} else {
			status = VXGE_HAL_INF_SW_LRO_FLUSH_BOTH;
		}
	}

	/*
	 * Since its time to flush,
	 * update ip header	so that	it can be sent up
	 */
	if (status == VXGE_HAL_INF_SW_LRO_FLUSH_SESSION) {
		vxge_hal_update_L3L4_header(vpath_handle, *sw_lro);
		ring->stats->lro_flush_max_pkts++;
	} else if (status == VXGE_HAL_INF_SW_LRO_FLUSH_BOTH) {
		vxge_hal_update_L3L4_header(vpath_handle, *sw_lro);
		ring->stats->lro_sending_both++;
	}


	vxge_hal_trace_log_ring(hldev, vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
	    __LINE__, status);
	if ((*sw_lro) && ((*sw_lro)->sg_num > ring->stats->lro_max_pkts_aggr))
		ring->stats->lro_max_pkts_aggr = (*sw_lro)->sg_num;
	return (status);
}

#endif
