/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */
/*
 * Copyright (c) 2002-2010 Neterion, Inc.
 * All right Reserved.
 *
 * FileName :   vxgehal-nce.c
 *
 * Description:  NCE related routines
 *
 * Created:       11 January 2006
 */
#include "vxgehal.h"


/*
 * vxge_hal_nce_create - Create NCE on adapter
 *
 * @vpath_handle: Virtual path handle
 * @opaque_handle: HAL Opaque handle
 * @src_mac_addr: Source Mac Address
 * @rem_mac_addr: Remote Mac Address
 * @vlan_id: Valn Id
 * @host_rch_time: Host Reach Time
 *
 * Creates an NCE on adapter
 */
vxge_hal_status_e
vxge_hal_nce_create(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_opaque_handle_t *opaque_handle,
	vxge_hal_opaque_handle_t *nce_opaque_handle,
	u64 src_mac_addr,
	u64 rem_mac_addr,
	u32 vlanId,
	u32 host_rch_time)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_dmq_t *dmq;
	__hal_device_t *hldev;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;
	u32 msg_id;

	vxge_assert((vpath_handle != NULL) && (opaque_handle != NULL) &&
	    (nce_opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_nce(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_nce(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", "
	    "nce_opaque_handle = 0x"VXGE_OS_STXFMT", "
	    "src_mac_addr = 0x"VXGE_OS_STXFMT", "
	    "rem_mac_addr = 0x"VXGE_OS_STXFMT", vlanId = %d, "
	    "host_rch_time = %d", (ptr_t)vpath_handle,
	    (ptr_t)opaque_handle, (ptr_t)nce_opaque_handle,
	    (ptr_t)src_mac_addr, (ptr_t)rem_mac_addr,
	    vlanId, host_rch_time);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	status = __hal_vp_oid_allocate(
	    vp->vpath,
	    &vp->vpath->nce_oid_db,
	    &msg_id);

	if (status != VXGE_HAL_OK) {
		vxge_assert(FALSE);
		vxge_hal_trace_log_nce(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
		    __LINE__, VXGE_HAL_ERR_TOO_MANY);
		return (VXGE_HAL_ERR_TOO_MANY);
	}

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, msg_id);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(nce_opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(nce_opaque_handle, msg_id);

	__hal_nce_create_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    msg_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    nce_opaque_handle,
	    src_mac_addr,
	    rem_mac_addr,
	    vlanId,
	    host_rch_time);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_nce(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);

	return (status);
}

/*
 * vxge_hal_nce_delete - Delete an NCE entry
 *
 * @vpath_handle: Virtual path handle
 * @obj_id: Object Id of NCE
 * @opaque_handle: HAL Opaque handle
 *
 * Deletes the given NCE entry from adapter
 */
vxge_hal_status_e
vxge_hal_nce_delete(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t obj_id,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_nce(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_nce(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "obj_id = 0x"VXGE_OS_STXFMT", opaque_handle = 0x"VXGE_OS_STXFMT,
	    (ptr_t)vpath_handle, (ptr_t)obj_id, (ptr_t)opaque_handle);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, obj_id);

	__hal_nce_delete_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    (u32)obj_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_nce(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_nce_update_mac_addr - Post NCE Update MAC Request DMQ
 *
 * @vpath_handle: Virtual path handle
 * @obj_id: Object Id of NCE
 * @opaque_handle: HAL Opaque handle
 * @mac_addr: Mac Address
 *
 * Updates the MAC address of an NCE entry on the adapter
 */
vxge_hal_status_e
vxge_hal_nce_update_mac_addr(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t obj_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	u64 mac_addr)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_nce(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_nce(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "obj_id = 0x"VXGE_OS_STXFMT", opaque_handle = 0x"VXGE_OS_STXFMT
	    ", mac_addr = 0x"VXGE_OS_STXFMT, (ptr_t)vpath_handle,
	    (ptr_t)obj_id, (ptr_t)opaque_handle, (ptr_t)mac_addr);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, obj_id);

	__hal_nce_update_mac_addr_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    (u32)obj_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    mac_addr);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_nce(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_nce_update_rch_time - Update RCH time
 *
 * @vpath_handle: Virtual path handle
 * @obj_id: Object Id of NCE
 * @opaque_handle: HAL Opaque handle
 * @rch_time: Host reachability time
 *
 * Updates and rch time of an NCE entry on adapter
 */
vxge_hal_status_e
vxge_hal_nce_update_rch_time(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t obj_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	u64 rch_time)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_nce(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_nce(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "obj_id = 0x"VXGE_OS_STXFMT", opaque_handle = 0x"VXGE_OS_STXFMT
	    ", rch_time = 0x"VXGE_OS_LLXFMT, (ptr_t)vpath_handle,
	    (ptr_t)obj_id, (ptr_t)opaque_handle, rch_time);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, obj_id);

	__hal_nce_update_rch_time_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    (u32)obj_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    rch_time);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_nce(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_nce_query - Dueries the info of an NCE entry on adapter
 *
 * @vpath_handle: Virtual path handle
 * @obj_id: Object Id of NCE
 * @opaque_handle: HAL Opaque handle
 *
 * Post the NCE Query Request DMQ message used by the host to query a NCE
 * on the adapter.
 */
vxge_hal_status_e
vxge_hal_nce_query(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t obj_id,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_nce(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_nce(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "obj_id = 0x"VXGE_OS_STXFMT", opaque_handle = 0x"VXGE_OS_STXFMT,
	    (ptr_t)vpath_handle, (ptr_t)obj_id, (ptr_t)opaque_handle);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, obj_id);

	__hal_nce_query_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    (u32)obj_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_nce(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}
