/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */
/*
 * Copyright (c) 2002-2010 Neterion, Inc.
 * All right Reserved.
 *
 * FileName :   vxgehal-odlist.c
 *
 * Description:  HAL Object descriptor functionality
 *
 * Created:       14 December 2005
 */
#include "vxgehal.h"

/*
 * vxge_hal_odlist_allocate - Allocate OD list
 * @vpathh: Virtual Path in which this odlist is used.
 * @odlist: odlist to be created.
 * @length: Length of the odlist.
 *
 * This function allocates OD list
 */
vxge_hal_status_e
vxge_hal_odlist_allocate(vxge_hal_vpath_h vpathh,
			vxge_hal_odlist_t *odlist,
			u32 length)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpathh;
	__hal_device_t *hldev;

	vxge_assert((vpathh != NULL) && (odlist != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
	    "vpathh = 0x"VXGE_OS_STXFMT", odlist = 0x"VXGE_OS_STXFMT", "
	    "length = %d", (ptr_t)vpathh, (ptr_t)odlist, length);

#if 0
	if ((length < VXGE_HAL_ODLIST_NUM_ODS_MIN) ||
	    (length > VXGE_HAL_ODLIST_NUM_ODS_MAX)) {
		vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: 1",
		    __FILE__, __func__, __LINE__);
		return (VXGE_HAL_FAIL);
	}
#endif

	odlist->num_ods = 0;
	odlist->od_size = sizeof (__hal_od_t);
	odlist->total_bytes = 0;

	odlist->vpathh = vpathh;

	odlist->od_list = (void *)__hal_blockpool_block_allocate(
	    vp->vpath->hldev,
	    VXGE_OS_HOST_PAGE_SIZE);

	if (odlist->od_list == NULL) {

		vxge_hal_err_log_pool(hldev, vp->vpath->vp_id,
		    "%s:%d Memory allocation for odlist failed",
		    __FILE__, __LINE__);
		vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d",
		    __FILE__, __func__, __LINE__,
		    VXGE_HAL_ERR_OUT_OF_MEMORY);
		return (VXGE_HAL_ERR_OUT_OF_MEMORY);

	}

	vxge_os_memzero(((__hal_blockpool_entry_t *)odlist->od_list)->memblock,
	    VXGE_OS_HOST_PAGE_SIZE);

	vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * vxge_hal_odlist_set - Set buffer pointers in the odlist
 * @odlist: list of blocks for odlist.
 * @ptr_to: Physical address of the buffer or the tagged offset
 * @stag: stag
 * @length: Length of the buffer
 *
 * This  function sets the buffer into the odlist appropriately.
 *
 */
vxge_hal_status_e
vxge_hal_odlist_set(vxge_hal_odlist_t *odlist,
		    u64 ptr_to,
		    u32 stag,
		    u32 length)
{
	__hal_od_t *od;
	__hal_vpath_handle_t *vp;
	__hal_device_t *hldev;
	__hal_blockpool_entry_t *entry;

	vxge_assert(odlist != NULL);

	vp = (__hal_vpath_handle_t *)odlist->vpathh;

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
	    "odlist = 0x"VXGE_OS_STXFMT", ptr_to = 0x"VXGE_OS_STXFMT", "
	    "stag = %d, length = %d", (ptr_t)odlist, (ptr_t)ptr_to,
	    stag, length);

	if (odlist == NULL) {
		vxge_hal_err_log_pool(hldev, vp->vpath->vp_id,
		    "null pointer passed odlist = 0x"VXGE_OS_STXFMT,
		    (ptr_t)odlist);
		vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: 1",
		    __FILE__, __func__, __LINE__);
		return (VXGE_HAL_FAIL);
	}

	entry = (__hal_blockpool_entry_t *)odlist->od_list;

	od =  ((__hal_od_t *)entry->memblock) + odlist->num_ods;

	od->ptr_to = ptr_to;

	od->control_0 = VXGE_HAL_OD_STAG(stag);
	od->control_0 |= VXGE_HAL_OD_LENGTH(length);
	odlist->total_bytes += length;

	odlist->num_ods++;

	vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);
	return (VXGE_HAL_OK);
}

/*
 * vxge_hal_odlist_free - Frees the OD List
 * @vpathh: Virtual Path in which this odlist is used.
 * @odlist: odlist to be freed
 *
 * This function freeup the memory pool and removes the
 * odlist.
 */

void
vxge_hal_odlist_free(vxge_hal_vpath_h vpathh,
		    vxge_hal_odlist_t *odlist)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpathh;
	__hal_device_t *hldev;

	vxge_assert(vpathh != NULL);

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
	    "vpathh = 0x"VXGE_OS_STXFMT", odlist = 0x"VXGE_OS_STXFMT,
	    (ptr_t)vpathh, (ptr_t)odlist);

	if (odlist == NULL) {
		vxge_hal_err_log_pool(hldev, vp->vpath->vp_id,
		    "%s:%d null pointer passed odlist = null",
		    __FILE__, __LINE__);
		vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: 1",
		    __FILE__, __func__, __LINE__);
		return;
	}

	__hal_blockpool_block_free(
	    ((__hal_vpath_handle_t *)vpathh)->vpath->hldev,
	    (__hal_blockpool_entry_t *)odlist->od_list);

	vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);
}
