/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */
/*
 * Copyright (c) 2002-2010 Neterion, Inc.
 * All right Reserved.
 *
 * FileName :   vxgehal-pblist.c
 *
 * Description:  HAL PB List functionality
 *
 * Created:       28 June 2006
 */
#include "vxgehal.h"

/*
 * __hal_pblist_dma_addr_get - Get the dma address of first block in the list
 * @pblist: pblist.
 *
 * This function is returns the dma address of the first block.
 */
u64
__hal_pblist_dma_addr_get(
	vxge_hal_pblist_t *pblist)
{
	__hal_blockpool_entry_t *entry;

	entry =
	    (__hal_blockpool_entry_t *)vxge_list_first_get(&pblist->pble_list);

	if (entry != NULL)
		return (entry->dma_addr);

	return (0);
}

/*
 * __hal_pblist_length_get - Get the size of the PBLE List block
 *			  to be placed in the LB Wrapper
 * @pblist: pblist.
 * @num: Number of PBLEs.
 *
 * This function is returns the dma address of the first block.
 */
u32
__hal_pblist_length_get(
	vxge_hal_pblist_t *pblist,
	u32 num)
{
	return ((num <= pblist->pbles_per_block) ?
	    num * sizeof (__hal_pble_t):
	    pblist->pbles_per_block *sizeof (__hal_pble_t));
}

/*
 * vxge_hal_pblist_allocate - Allocate PB list
 * @vpath_handle: Handle to vpath in which this pblist is used.
 * @pblist: pblist to be created.
 * @length: Length of the pblist.
 *
 * This function allocates PB list
 */

vxge_hal_status_e
vxge_hal_pblist_allocate(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_pblist_t *pblist,
	u32	length)
{
	int nblocks;
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;

	vxge_assert((vpath_handle != NULL) && (pblist != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", pblist = 0x"VXGE_OS_STXFMT", "
	    "length = %d", (ptr_t)vpath_handle, (ptr_t)pblist, length);

	if (pblist == NULL) {
		vxge_hal_err_log_pool(hldev, vp->vpath->vp_id,
		    "%s:%d null pointer passed pblist = null",
		    __FILE__, __LINE__);
		vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: 1",
		    __FILE__, __func__, __LINE__);
		return (VXGE_HAL_FAIL);
	}

	pblist->pble_size = sizeof (__hal_pble_t);
	pblist->pbles_per_block = (VXGE_OS_HOST_PAGE_SIZE -
	    sizeof (__hal_list_block_wrapper_t)) / pblist->pble_size;
	pblist->num_pbles = length;

	nblocks =
	    (length + pblist->pbles_per_block - 1) / pblist->pbles_per_block;

	vxge_list_init(&pblist->pble_list);

	if (__hal_blockpool_list_allocate(
	    ((__hal_vpath_handle_t *)vpath_handle)->vpath,
	    &pblist->pble_list, nblocks) != VXGE_HAL_OK) {

		vxge_hal_err_log_pool(hldev, vp->vpath->vp_id,
		    "%s:%d Memory allocation for pblist failed",
		    __FILE__, __LINE__);
		vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
		    __LINE__, VXGE_HAL_ERR_OUT_OF_MEMORY);
		return (VXGE_HAL_ERR_OUT_OF_MEMORY);

	}

	vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * vxge_hal_pblist_set - Set buffer pointers in the pblist
 * @pblist: list of blocks for odlist.
 * @phy_addr_list: Array of DMA-able Physical addresses of buffers.
 * @num_addr: Number of addresses in phy_addr_list.
 *
 * This  function sets the buffers into the pblist appropriately.
 *
 */
vxge_hal_status_e
vxge_hal_pblist_set(
	vxge_hal_pblist_t *pblist,
	u64	*phy_addr_list,
	u32 num_addr)
{

	vxge_list_t *item, *next_item;
	__hal_blockpool_entry_t *block_entry, *block_entry_next;
	__hal_list_block_wrapper_t *wrapper;
	__hal_pble_t *pble;
	__hal_vpath_handle_t *vp;
	__hal_device_t *hldev;
	int i, j, next_block_length;

	vxge_assert((pblist != NULL) && (phy_addr_list != NULL));

	vp = (__hal_vpath_handle_t *)pblist->vpathh;

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
	    "pblist = 0x"VXGE_OS_STXFMT", phy_addr_list = 0x"VXGE_OS_STXFMT", "
	    "num_addr = %d", (ptr_t)pblist, (ptr_t)phy_addr_list, num_addr);

	if ((pblist == NULL) || (phy_addr_list == NULL)) {
		vxge_hal_err_log_pool(hldev, vp->vpath->vp_id,
		    "null pointer passed pblist = 0x"VXGE_OS_STXFMT", "
		    "phy_addr_list = 0x"VXGE_OS_STXFMT, (ptr_t)pblist,
		    (ptr_t)phy_addr_list);
		vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: 1",
		    __FILE__, __func__, __LINE__);
		return (VXGE_HAL_FAIL);
	}

	vxge_list_for_each_safe(item, next_item, &pblist->pble_list) {

		block_entry = (__hal_blockpool_entry_t *)item;

		block_entry_next = (__hal_blockpool_entry_t *)next_item;

		pble =  (__hal_pble_t *)block_entry->memblock;

		if (num_addr <= pblist->pbles_per_block) {
			i = num_addr;
		} else {
			i = pblist->pbles_per_block;
		}

		for (j = 0; j < i; j++) {
			*pble++	= (__hal_pble_t)*phy_addr_list++;
		}

		num_addr -= i;

		if (num_addr > 0) {
			wrapper = (__hal_list_block_wrapper_t *)
			    (((char *)block_entry->memblock) +
			    pblist->pble_size *
			    pblist->pbles_per_block);
			next_block_length =
			    ((num_addr > pblist->pbles_per_block) ?
			    (pblist->pble_size * pblist->pbles_per_block) :
			    (num_addr * pblist->pbles_per_block));
			wrapper->next_block_ptr = block_entry_next->dma_addr;
			wrapper->control_0 =
			    VXGE_HAL_LBW_NEXT_BLOCK_LENGTH(next_block_length);
		} else {
			wrapper = (__hal_list_block_wrapper_t *)
			    (((char *)block_entry->memblock) +
			    pblist->pble_size * j);
			wrapper->next_block_ptr = 0;
			wrapper->control_0 =
			    VXGE_HAL_LBW_NEXT_BLOCK_LENGTH(0);
			break;
		}
	}

	vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);
	return (VXGE_HAL_OK);

}
/*
 * vxge_hal_pblist_free - Deallocates the PB List
 * @vpath_handle: Handle to vpath in which this pblist is used.
 * @pblist: pblist to be deallocated
 *
 * This function freeup the memory pool and removes the
 * pblist.
 */

void vxge_hal_pblist_free(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_pblist_t *pblist)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;

	vxge_assert((vpath_handle != NULL) && (pblist != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", pblist = 0x"VXGE_OS_STXFMT,
	    (ptr_t)vpath_handle, (ptr_t)pblist);

	if (pblist == NULL) {
		vxge_hal_err_log_pool(hldev, vp->vpath->vp_id,
		    "%s:%d null pointer passed pblist = null",
		    __FILE__, __LINE__);
		vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: 1", __FILE__, __func__, __LINE__);
		return;
	}

	__hal_blockpool_list_free(
	    ((__hal_vpath_handle_t *)vpath_handle)->vpath->hldev,
	    &pblist->pble_list);

	vxge_hal_trace_log_pool(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);
}
