/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */
/*
 * Copyright (c) 2002-2010 Neterion, Inc.
 * All right Reserved.
 *
 * FileName :   vxgehal-regpool.c
 *
 * Description:  HAL Register pool for IR Register and NSMR registers
 *
 * Created:       15 September 2006
 */
#include "vxgehal.h"

/*
 * __hal_rp_regpool_item_alloc - Allocate List blocks for Reg Pool
 * @mempoolh: Handle to memory pool
 * @memblock: Address of this memory block
 * @memblock_index: Index of this memory block
 * @dma_object: dma object for this block
 * @item: Pointer to this item
 * @index: Index of this item in memory block
 * @is_last: If this is last item in the block
 * @userdata: Specific data of user
 *
 * This function is callback passed to __hal_mempool_create to create memory
 * pool for reg pool
 */
static vxge_hal_status_e
__hal_rp_mempool_item_alloc(
	vxge_hal_mempool_h mempoolh,
	void *memblock,
	u32 memblock_index,
	vxge_hal_mempool_dma_t *dma_object,
	void *item,
	u32 index,
	u32 is_last,
	void *userdata)
{
	u32 i;
	void *block_priv;
	u32 memblock_item_idx;
	vxge_hal_mempool_t *mempool = (vxge_hal_mempool_t *)mempoolh;
	__hal_regpool_t *pooldata = (__hal_regpool_t *)userdata;
	__hal_device_t *hldev;

	vxge_assert((pooldata != NULL) && (mempool != NULL));

	hldev = (__hal_device_t *)mempool->devh;

	vxge_hal_trace_log_pool(hldev, NULL_VPID,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_pool(hldev, NULL_VPID,
	    "mempoolh = 0x"VXGE_OS_STXFMT", memblock = 0x"VXGE_OS_STXFMT", "
	    "memblock_index = %d, dma_object = 0x"VXGE_OS_STXFMT", "
	    "item = 0x"VXGE_OS_STXFMT", index = %d, is_last = %d, "
	    "userdata = 0x"VXGE_OS_STXFMT, (ptr_t)mempoolh, (ptr_t)memblock,
	    memblock_index, (ptr_t)dma_object, (ptr_t)item, index, is_last,
	    (ptr_t)userdata);

	block_priv = __hal_mempool_item_priv(mempoolh, memblock_index, item,
	    &memblock_item_idx);

	vxge_assert(block_priv != NULL);

	for (i = 0; i < pooldata->regs_per_block; i++) {
		__hal_reg_entry_t *reg_entry;

		reg_entry = (__hal_reg_entry_t *)(void *)
		    ((char *)block_priv + pooldata->reg_entry_size * i);

		reg_entry->reg_addr =
		    (void *)((char *)item + i * pooldata->reg_size);
		reg_entry->dma_offset =
		    (char *)reg_entry->reg_addr - (char *)memblock;
		reg_entry->dma_addr = dma_object->addr +  reg_entry->dma_offset;
		reg_entry->dma_handle = dma_object->handle;
		reg_entry->memblock = memblock;
		reg_entry->memblock_index = memblock_index;

		vxge_list_insert(&reg_entry->item, &pooldata->free_reg_list);

	}

	vxge_hal_trace_log_pool(hldev, NULL_VPID,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * __hal_rp_mempool_item_free - Free blocks from the reg pool
 * @mempoolh: Handle to memory pool
 * @memblock: Address of this memory block
 * @memblock_index: Index of this memory block
 * @dma_object: dma object for this block
 * @item: Pointer to this item
 * @index: Index of this item in memory block
 * @is_last: If this is last item in the block
 * @userdata: Specific data of user
 *
 * This function is callback passed to __hal_mempool_free to destroy memory
 * pool for block pool
 */
static vxge_hal_status_e
__hal_rp_mempool_item_free(
	vxge_hal_mempool_h mempoolh,
	void *memblock,
	u32 memblock_index,
	vxge_hal_mempool_dma_t *dma_object,
	void *item,
	u32 index,
	u32 is_last,
	void *userdata)
{
	__hal_device_t *hldev;
	vxge_hal_mempool_t *mempool = (vxge_hal_mempool_t *)mempoolh;

	vxge_assert(mempool != NULL);

	hldev = (__hal_device_t *)mempool->devh;

	vxge_hal_trace_log_pool(hldev, NULL_VPID,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_pool(hldev, NULL_VPID,
	    "mempoolh = 0x"VXGE_OS_STXFMT", memblock = 0x"VXGE_OS_STXFMT", "
	    "memblock_index = %d, dma_object = 0x"VXGE_OS_STXFMT", "
	    "item = 0x"VXGE_OS_STXFMT", index = %d, is_last = %d, "
	    "userdata = 0x"VXGE_OS_STXFMT, (ptr_t)mempoolh, (ptr_t)memblock,
	    memblock_index, (ptr_t)dma_object, (ptr_t)item, index, is_last,
	    (ptr_t)userdata);

	vxge_hal_trace_log_pool(hldev, NULL_VPID,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * __hal_regpool_create - Create reg pool
 * @devh: Pointer to HAL Device object.
 * @regpool: Reg pool to be created.
 * @pool_size: Number of registers in the pool.
 *
 * This function creates reg pool
 */

vxge_hal_status_e
__hal_regpool_create(
	vxge_hal_device_h devh,
	__hal_regpool_t *regpool,
	u32 pool_size)
{
	__hal_device_t *hldev = (__hal_device_t *)devh;

	vxge_assert((hldev != NULL) && (regpool != NULL));

	vxge_hal_trace_log_pool(devh, NULL_VPID,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_pool(devh, NULL_VPID,
	    "devh = 0x"VXGE_OS_STXFMT", regpool = 0x"VXGE_OS_STXFMT", "
	    "pool_size = %d", (ptr_t)devh, (ptr_t)regpool, pool_size);

	if (regpool == NULL) {
		vxge_hal_info_log_pool(devh, NULL_VPID,
		    "%s:%d null pointer passed regpool = null",
		    __FILE__, __LINE__);
		vxge_hal_trace_log_pool(devh, NULL_VPID,
		    "<==  %s:%s:%d  Result: 1",
		    __FILE__, __func__, __LINE__);
		return (VXGE_HAL_FAIL);
	}

	regpool->pool_size = pool_size;

#if defined(VXGE_HAL_RP_POST)
	vxge_os_spin_lock_init(&regpool->pool_lock, hldev->pdev);
#elif defined(VXGE_HAL_RP_POST_IRQ)
	vxge_os_spin_lock_init_irq(&regpool->pool_lock, hldev->irqh);
#endif

	regpool->reg_size = VXGE_HAL_REGPOOL_REG_SIZE;
	regpool->regs_per_block = VXGE_OS_HOST_PAGE_SIZE/regpool->reg_size;
	regpool->reg_entry_size = sizeof (__hal_reg_entry_t);
	regpool->regblock_priv_size = regpool->regs_per_block *
	    regpool->reg_entry_size;

	vxge_list_init(&regpool->free_reg_list);

	regpool->mempool = __hal_mempool_create(
	    hldev,
	    VXGE_OS_HOST_PAGE_SIZE,
	    VXGE_OS_HOST_PAGE_SIZE,
	    regpool->regblock_priv_size,
	    regpool->pool_size/regpool->regs_per_block,
	    regpool->pool_size/regpool->regs_per_block,
	    __hal_rp_mempool_item_alloc,
	    __hal_rp_mempool_item_free,
	    regpool);

	if (regpool->mempool == NULL) {
		vxge_hal_info_log_pool(devh, NULL_VPID,
		    "%s:%d Memory allocation for regpool failed",
		    __FILE__, __LINE__);
		vxge_hal_trace_log_pool(devh, NULL_VPID,
		    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
		    __LINE__, VXGE_HAL_ERR_OUT_OF_MEMORY);
		return (VXGE_HAL_ERR_OUT_OF_MEMORY);
	}

	vxge_hal_info_log_pool(devh, NULL_VPID,
	    "Regpool  reg size:%d reg entry size:%d reg pool size:%d",
	    regpool->reg_size, regpool->reg_entry_size,
	    regpool->pool_size);

	vxge_hal_trace_log_pool(devh, NULL_VPID,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * __hal_regpool_destroy - Deallocates the reg pool
 * @devh: Pointer to HAL Device object.
 * @regpool: regpool to be deallocated
 *
 * This function freeup the memory pool and removes the
 * block pool.
 */

void
__hal_regpool_destroy(
	vxge_hal_device_h devh,
	__hal_regpool_t *regpool)
{
	__hal_device_t *hldev = (__hal_device_t *)devh;

	vxge_assert((devh != NULL) && (regpool != NULL));

	vxge_hal_trace_log_pool(devh, NULL_VPID,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_pool(devh, NULL_VPID,
	    "devh = 0x"VXGE_OS_STXFMT", regpool = 0x"VXGE_OS_STXFMT,
	    (ptr_t)devh, (ptr_t)regpool);

	if (regpool == NULL) {
		vxge_hal_info_log_pool(devh, NULL_VPID,
		    "%s:%d null pointer passed regpool = null",
		    __FILE__, __LINE__);
		vxge_hal_trace_log_pool(devh, NULL_VPID,
		    "<==  %s:%s:%d  Result: %d",
		    __FILE__, __func__, __LINE__,
		    VXGE_HAL_ERR_INVALID_HANDLE);
		return;
	}

	if (regpool->mempool) {
		__hal_mempool_destroy(regpool->mempool);
	}

#if defined(VXGE_HAL_RP_POST)
	vxge_os_spin_lock_destroy(&regpool->pool_lock,
	    ((__hal_device_t *)devh)->header.pdev);
#elif defined(VXGE_HAL_RP_POST_IRQ)
	vxge_os_spin_lock_destroy_irq(&regpool->pool_lock,
	    ((__hal_device_t *)devh)->header.pdev);
#endif

	vxge_hal_trace_log_pool(devh, NULL_VPID,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);
}


/*
 * __hal_regpool_allocate - Allocate a register from reg pool
 * @regpool: regpool from which register to be allocated
 * @reg_entry: Pointer to a buffer to return registry entry
 *
 * This function allocates a register from the register pool
 */

vxge_hal_status_e
__hal_regpool_allocate(
	__hal_regpool_t *regpool,
	__hal_reg_entry_t **reg_entry)
{
	vxge_hal_status_e status;
	__hal_device_t *hldev;

	vxge_assert((regpool != NULL) && (reg_entry != NULL) &&
	    (regpool->mempool != NULL));

	hldev = (__hal_device_t *)regpool->mempool->devh;

	vxge_hal_trace_log_pool(hldev, NULL_VPID,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_pool(hldev, NULL_VPID,
	    "regpool = 0x"VXGE_OS_STXFMT", reg_entry = 0x"VXGE_OS_STXFMT,
	    (ptr_t)regpool, (ptr_t)reg_entry);

	if ((regpool == NULL) || (reg_entry == NULL)) {
		vxge_hal_info_log_pool(hldev, NULL_VPID,
		    "null pointer passed regpool = 0x"VXGE_OS_STXFMT", "
		    "reg_entry = 0x"VXGE_OS_STXFMT, (ptr_t)regpool,
		    (ptr_t)reg_entry);
		vxge_hal_trace_log_pool(hldev, NULL_VPID,
		    "<==  %s:%s:%d  Result: 1", __FILE__, __func__, __LINE__);
		return (VXGE_HAL_FAIL);
	}

#if defined(VXGE_HAL_RP_POST)
	vxge_os_spin_lock(&regpool->pool_lock);
#elif defined(VXGE_HAL_RP_POST_IRQ)
	vxge_os_spin_lock_irq(&regpool->pool_lock, flags);
#endif

	*reg_entry = (__hal_reg_entry_t *)
	    vxge_list_first_get(&regpool->free_reg_list);

	if (*reg_entry == NULL) {
		vxge_hal_trace_log_pool(hldev, NULL_VPID,
		    "%s:%d Regpool out of registers", __FILE__, __LINE__);

		vxge_assert(FALSE);

		status = VXGE_HAL_ERR_OUT_OF_MEMORY;
	} else {
		vxge_list_remove(&(*reg_entry)->item);

		status = VXGE_HAL_OK;
	}

#if defined(VXGE_HAL_RP_POST)
	vxge_os_spin_unlock(&regpool->pool_lock);
#elif defined(VXGE_HAL_RP_POST_IRQ)
	vxge_os_spin_unlock_irq(&regpool->pool_lock, flags);
#endif

	vxge_hal_trace_log_pool(hldev, NULL_VPID,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * __hal_regpool_free - Free a register from reg pool
 * @regpool: regpool to which register to be freed
 * @reg_entry: register entry to be freed
 *
 * This function frees a register to the register pool
 */

vxge_hal_status_e
__hal_regpool_free(
	__hal_regpool_t *regpool,
	__hal_reg_entry_t *reg_entry)
{
	__hal_device_t *hldev;

	vxge_assert((regpool != NULL) && (reg_entry != NULL) &&
	    (regpool->mempool != NULL));

	hldev = (__hal_device_t *)regpool->mempool->devh;

	vxge_hal_trace_log_pool(hldev, NULL_VPID,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_pool(hldev, NULL_VPID,
	    "regpool = 0x"VXGE_OS_STXFMT", reg_entry = 0x"VXGE_OS_STXFMT,
	    (ptr_t)regpool, (ptr_t)reg_entry);

	if ((regpool == NULL) || (reg_entry == NULL)) {
		vxge_hal_info_log_pool(hldev, NULL_VPID,
		    "null pointer passed regpool = 0x"VXGE_OS_STXFMT", "
		    "reg_entry = 0x"VXGE_OS_STXFMT, (ptr_t)regpool,
		    (ptr_t)reg_entry);
		vxge_hal_trace_log_pool(hldev, NULL_VPID,
		    "<==  %s:%s:%d  Result: 1", __FILE__, __func__, __LINE__);
		return (VXGE_HAL_FAIL);
	}

#if defined(VXGE_HAL_RP_POST)
	vxge_os_spin_lock(&regpool->pool_lock);
#elif defined(VXGE_HAL_RP_POST_IRQ)
	vxge_os_spin_lock_irq(&regpool->pool_lock, flags);
#endif

	vxge_os_memzero(reg_entry->reg_addr, regpool->reg_size);

	vxge_list_insert(&reg_entry->item, &regpool->free_reg_list);

#if defined(VXGE_HAL_RP_POST)
	vxge_os_spin_unlock(&regpool->pool_lock);
#elif defined(VXGE_HAL_RP_POST_IRQ)
	vxge_os_spin_unlock_irq(&regpool->pool_lock, flags);
#endif

	vxge_hal_trace_log_pool(hldev, NULL_VPID,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);

}
