/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */
/*
 * Copyright (c) 2002-2010 Neterion, Inc.
 * All right Reserved.
 *
 * FileName :   ivxgehal-spdm.c
 *
 * Description:  SPDM related routines
 *
 * Created:       11 January 2006
 */
#include "vxgehal.h"

/*
 * vxge_hal_spdm_open - Open a SPDM session
 *
 * @vpath_handle: Handle returned by message queue register
 * @opaque_handle: HAL Opaque handle
 * @session_opaque_handle: HAL Opaque handle used by the host to reference data
 *		structures in host memory related to this session
 * @l4sp: Layer 4 source port
 * @l4dp: Layer 4 destination port
 * @use_ipv6: Flag to specify if IPv6 is used
 * @use_vlan: Flag to specify if vlan is used
 * @vlan_id: vlan id
 * @ip_sa_low: Lower 64 bits of IPv4 or IPv6 source address
 * @ip_sa_high: Higher 64 bits of IPv4 or IPv6 source address
 * @ip_da_low: Lower 64 bits of IPv4 or IPv6 destination address
 * @ip_da_high: Higher 64 bits of IPv4 or IPv6 destination address
 *
 * Open a SPDM session on the adapter.
 */
vxge_hal_status_e
vxge_hal_spdm_open(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_opaque_handle_t *opaque_handle,
	vxge_hal_opaque_handle_t *session_opaque_handle,
	u32 l4sp,
	u32 l4dp,
	u32 use_ipv6,
	u32 use_vlan,
	u32 vlan_id,
	u64 ip_sa_high,
	u64 ip_sa_low,
	u64 ip_da_high,
	u64 ip_da_low)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;
	u32 session_id;

	vxge_assert((vpath_handle != NULL) && (opaque_handle != NULL) &&
	    (session_opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_spdm(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_spdm(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", "
	    "session_opaque_handle = 0x"VXGE_OS_STXFMT", l4sp = %d, "
	    "l4dp = %x, use_ipv6 = %d, use_vlan = %d, vlan_id = %d, "
	    " ip_sa_high = 0x"VXGE_OS_LLXFMT", "
	    "ip_sa_low = 0x"VXGE_OS_LLXFMT", ip_da_high = "
	    "0x"VXGE_OS_LLXFMT", ip_da_low = 0x"VXGE_OS_LLXFMT,
	    (ptr_t)vpath_handle, (ptr_t)opaque_handle,
	    (ptr_t)session_opaque_handle, l4sp, l4dp,
	    use_ipv6, use_vlan, vlan_id, ip_sa_high, ip_sa_low,
	    ip_da_high, ip_da_low);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	status = __hal_vp_oid_allocate(
	    vp->vpath,
	    &vp->vpath->session_oid_db,
	    &session_id);

	if (status != VXGE_HAL_OK) {
		vxge_hal_trace_log_spdm(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d",
		    __FILE__, __func__, __LINE__, VXGE_HAL_ERR_TOO_MANY);
		return (VXGE_HAL_ERR_TOO_MANY);
	}

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle,
	    vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle,
	    session_id);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(session_opaque_handle,
	    vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(session_opaque_handle,
	    session_id);

	__hal_spdm_open_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    session_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    session_opaque_handle,
	    l4sp,
	    l4dp,
	    use_ipv6,
	    use_vlan,
	    vlan_id,
	    ip_sa_high,
	    ip_sa_low,
	    ip_da_high,
	    ip_da_low);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_spdm(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_spdm_delete - Delete a SPDM session
 *
 * @vpath_handle: Handle returned by message queue register
 * @session_id: Object Id of SPDM session
 * @opaque_handle: HAL Opaque handle
 *
 * Deletes a SPDM session on the adapter
 */
vxge_hal_status_e
vxge_hal_spdm_delete(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t session_id,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (session_id != 0) &&
	    (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_spdm(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_spdm(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "session_id = 0x"VXGE_OS_LLDFMT", opaque_handle = 0x"VXGE_OS_STXFMT,
	    (ptr_t)vpath_handle, session_id, (ptr_t)opaque_handle);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, session_id);

	__hal_spdm_delete_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    (u32)session_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_spdm(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);
	return (status);
}
