/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */
/*
 * Copyright (c) 2002-2010 Neterion, Inc.
 * All right Reserved.
 *
 * FileName :   vxgehal-sq.c
 *
 * Description:  HAL offload send queue functionality
 *
 * Created:       14 December 2005
 */
#include "vxgehal.h"

/*
 * __hal_towi_mempool_item_alloc - Allocate List blocks for towi list callback
 * @mempoolh: Handle to memory pool
 * @memblock: Address of this memory block
 * @memblock_index: Index of this memory block
 * @dma_object: dma object for this block
 * @item: Pointer to this item
 * @index: Index of this item in memory block
 * @is_last: If this is last item in the block
 * @userdata: Specific data of user
 *
 * This function is callback passed to __hal_mempool_create to create memory
 * pool for towi list
 */
static vxge_hal_status_e
__hal_towi_mempool_item_alloc(
	vxge_hal_mempool_h mempoolh,
	void *memblock,
	u32 memblock_index,
	vxge_hal_mempool_dma_t *dma_object,
	void *item,
	u32 index,
	u32 is_last,
	void *userdata)
{
	u32 i;
	void *block_priv;
	u32 memblock_item_idx;
	__hal_sq_t *sqdata = (__hal_sq_t *)userdata;
	__hal_list_block_t *prev_block, *next_block;
	__hal_list_block_wrapper_t *prev_wrapper, *wrapper;
	u32 towis_block_size = sqdata->towi_size * sqdata->towis_per_block;
	__hal_device_t *hldev = (__hal_device_t *)sqdata->channel.devh;

	vxge_assert(sqdata != NULL);

	vxge_hal_trace_log_pool(hldev,
	    ((__hal_vpath_handle_t *)sqdata->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_pool(hldev,
	    ((__hal_vpath_handle_t *)sqdata->channel.vph)->vpath->vp_id,
	    "mempoolh = 0x"VXGE_OS_STXFMT", memblock = 0x"VXGE_OS_STXFMT", "
	    "memblock_index = %d, dma_object = 0x"VXGE_OS_STXFMT", "
	    "item = 0x"VXGE_OS_STXFMT", index = %d, is_last = %d, "
	    "userdata = 0x"VXGE_OS_STXFMT, (ptr_t)mempoolh,
	    (ptr_t)memblock, memblock_index, (ptr_t)dma_object,
	    (ptr_t)item, index, is_last, (ptr_t)userdata);

	block_priv = __hal_mempool_item_priv(mempoolh, memblock_index, item,
	    &memblock_item_idx);

	vxge_assert(block_priv != NULL);

	for (i = 0; i < sqdata->towis_per_block; i++) {
		__hal_towi_priv_t *towi_priv;
		__hal_towi_t *towi;
		int reserve_index = sqdata->channel.reserve_ptr - 1 -
		    (index * sqdata->towis_per_block + i);

		towi = (__hal_towi_t *)((char *)item + i *sqdata->towi_size);

		sqdata->channel.reserve_arr[reserve_index] = towi;

		towi_priv = (__hal_towi_priv_t *)(void *)
		    ((char *)block_priv + sqdata->towi_priv_size * i);

		towi_priv->dma_offset = (char *)towi - (char *)memblock;
		towi_priv->dma_addr = dma_object->addr +  towi_priv->dma_offset;
		towi_priv->dma_handle = dma_object->handle;
		towi_priv->memblock = memblock;
		towi_priv->wrapper = (__hal_list_block_wrapper_t *)
		    (((char *)memblock) + towis_block_size);
		towi_priv->sqh = (vxge_hal_sq_h)userdata;

		towi->host_control_0 = (u64)(ptr_t)towi_priv;
	}

	__hal_list_block_memblock_idx_set(memblock, memblock_index);

	wrapper = (__hal_list_block_wrapper_t *)
	    (((char *)memblock) + towis_block_size);

	if (index > 0) {
		prev_block = __hal_mempool_item(mempoolh, index-1);
		next_block = __hal_mempool_item(mempoolh, 0);
		prev_wrapper = (__hal_list_block_wrapper_t *)
		    (((char *)prev_block) + towis_block_size);
		__hal_wrapper_link(prev_wrapper, wrapper,
		    dma_object->addr, towis_block_size);

	} else {
		prev_block = memblock;
		next_block = memblock;
		__hal_wrapper_link(wrapper, wrapper,
		    dma_object->addr, towis_block_size);
	}

	__hal_list_block_link(memblock, prev_block, next_block);

	vxge_hal_trace_log_pool(hldev,
	    ((__hal_vpath_handle_t *)sqdata->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * __hal_towi_mempool_item_free - Free List blocks for towi list callback
 * @mempoolh: Handle to memory pool
 * @memblock: Address of this memory block
 * @memblock_index: Index of this memory block
 * @dma_object: dma object for this block
 * @item: Pointer to this item
 * @index: Index of this item in memory block
 * @is_last: If this is last item in the block
 * @userdata: Specific data of user
 *
 * This function is callback passed to __hal_mempool_free to destroy memory
 * pool for towi list
 */
static vxge_hal_status_e
__hal_towi_mempool_item_free(
	vxge_hal_mempool_h mempoolh,
	void *memblock,
	u32 memblock_index,
	vxge_hal_mempool_dma_t *dma_object,
	void *item,
	u32 index,
	u32 is_last,
	void *userdata)
{
	__hal_sq_t *sqdata = (__hal_sq_t *)userdata;
	__hal_list_block_t *prev_block;
	__hal_list_block_wrapper_t *prev_wrapper, *wrapper;
	u32 towis_block_size = sqdata->towi_size * sqdata->towis_per_block;
	__hal_device_t *hldev = (__hal_device_t *)sqdata->channel.devh;

	vxge_assert(sqdata != NULL);

	vxge_hal_trace_log_pool(hldev,
	    ((__hal_vpath_handle_t *)sqdata->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_pool(hldev,
	    ((__hal_vpath_handle_t *)sqdata->channel.vph)->vpath->vp_id,
	    "mempoolh = 0x"VXGE_OS_STXFMT", memblock = 0x"VXGE_OS_STXFMT", "
	    "memblock_index = %d, dma_object = 0x"VXGE_OS_STXFMT", "
	    "item = 0x"VXGE_OS_STXFMT", index = %d, is_last = %d, "
	    "userdata = 0x"VXGE_OS_STXFMT, (ptr_t)mempoolh, (ptr_t)memblock,
	    memblock_index, (ptr_t)dma_object, (ptr_t)item, index, is_last,
	    (ptr_t)userdata);

	prev_block = __hal_list_block_prev_pointer(memblock);

	prev_wrapper = (__hal_list_block_wrapper_t *)
	    (((char *)prev_block) + towis_block_size);
	wrapper = (__hal_list_block_wrapper_t *)
	    (((char *)memblock) + towis_block_size);

	__hal_wrapper_unlink(prev_wrapper, wrapper);

	__hal_list_block_unlink(memblock);

	vxge_hal_trace_log_pool(hldev,
	    ((__hal_vpath_handle_t *)sqdata->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * __hal_sq_create - Create a SQ
 * @vpath_handle: Handle returned by virtual path open
 * @sq: SQ
 *
 * This function allocates DMA memory pool for SQ and completes
 * SQ Creation.
 *
 */
vxge_hal_status_e
__hal_sq_create(
	vxge_hal_vpath_h vpath_handle,
	__hal_sq_t *sq)
{
	vxge_hal_status_e status;
	vxge_hal_mempool_dma_t *dma_object;
	__hal_device_t *hldev;
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;

	vxge_assert((vpath_handle != NULL) && (sq != NULL));

	hldev = vp->vpath->hldev;

	vxge_hal_trace_log_sq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_sq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", sq = 0x"VXGE_OS_STXFMT,
	    (ptr_t)vpath_handle, (ptr_t)sq);

	sq->mempool = __hal_mempool_create(
	    sq->channel.devh,
	    VXGE_OS_HOST_PAGE_SIZE,
	    VXGE_OS_HOST_PAGE_SIZE,
	    sq->towiblock_priv_size,
	    sq->sq_length/sq->towis_per_block,
	    sq->sq_length/sq->towis_per_block,
	    __hal_towi_mempool_item_alloc,
	    __hal_towi_mempool_item_free,
	    sq);

	if (sq->mempool == NULL) {
		vxge_hal_info_log_sq(hldev, vp->vpath->vp_id,
		    "Memory allocation for sq failed %s : %d",
		    __func__, __LINE__);
		vxge_hal_trace_log_sq(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
		    __LINE__, VXGE_HAL_ERR_OUT_OF_MEMORY);
		return (VXGE_HAL_ERR_OUT_OF_MEMORY);
	}

	status = __hal_channel_initialize(&sq->channel);

	if (status != VXGE_HAL_OK) {
		vxge_hal_info_log_sq(hldev, vp->vpath->vp_id,
		    "sq channel initialization failed == > %s : %d",
		    __func__, __LINE__);
		__hal_sq_delete(vpath_handle, sq);
		vxge_hal_trace_log_sq(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d",
		    __FILE__, __func__, __LINE__, status);
		sq = NULL;
	} else {
		dma_object = __hal_mempool_memblock_dma(sq->mempool, 0);

		vxge_assert(dma_object != NULL);

		sq->first_block_address = dma_object->addr;

		sq->first_block_length = VXGE_OS_HOST_PAGE_SIZE;

		VXGE_HAL_SQ_STATE_SET(sq, VXGEHAL_SQ_STATE_INITIALIZED);
	}

	vp->cb_fn(vp->client_handle,
	    NULL,
	    VXGE_HAL_MSG_TYPE_SQ_CREATE_RESP,
	    (vxge_hal_obj_id_t)(ptr_t)sq,
	    status,
	    sq->opaque_handle);

	vxge_hal_trace_log_sq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * vxge_hal_sq_create - Create a SQ
 * @vpath_handle: Virtual path handle to which this queue belongs
 * @attr: SQ configuration parameters structure
 * @opaque_handle: HAL Opaque handle
 *
 * This function creates SQ.
 *
 */
vxge_hal_status_e
vxge_hal_sq_create(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_sq_attr_t *attr,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	vxge_hal_status_e status;
	__hal_sq_t *sq;
	__hal_device_t *hldev;
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	u32 towi_size;
	u32 towis_per_block;
	u32 sq_length;

	vxge_assert((vpath_handle != NULL) && (attr != NULL) &&
	    (opaque_handle != NULL));

	hldev = vp->vpath->hldev;

	vxge_hal_trace_log_sq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_sq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", attr = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT, (ptr_t)vpath_handle,
	    (ptr_t)attr, (ptr_t)opaque_handle);

	if ((vpath_handle == NULL) ||
	    (attr == NULL) || (opaque_handle == NULL)) {
		vxge_hal_info_log_sq(hldev, vp->vpath->vp_id,
		    "null pointer passed %s : %d", __func__, __LINE__);
		vxge_hal_trace_log_sq(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
		    __LINE__, VXGE_HAL_ERR_INVALID_HANDLE);
		return (VXGE_HAL_ERR_INVALID_HANDLE);
	}

	towi_size = sizeof (__hal_towi_t);

	towis_per_block = VXGE_HAL_LIST_BLOCK_ITEMS_SIZE / towi_size;

	sq_length = ((attr->sq_length + towis_per_block - 1) /
	    towis_per_block) * towis_per_block;

	sq = (__hal_sq_t *)__hal_channel_allocate(
	    (vxge_hal_device_h)vp->vpath->hldev,
	    vpath_handle,
	    VXGE_HAL_CHANNEL_TYPE_SEND_QUEUE,
	    sq_length,
	    TRUE,
	    TRUE,
	    TRUE,
	    0,
	    attr->userdata);

	if (sq == NULL) {
		vxge_hal_info_log_sq(hldev, vp->vpath->vp_id,
		    "Memory allocation failed == > %s : %d",
		    __func__, __LINE__);
		vxge_hal_trace_log_sq(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d",
		    __FILE__, __func__,
		    __LINE__, VXGE_HAL_ERR_OUT_OF_MEMORY);
		return (VXGE_HAL_ERR_OUT_OF_MEMORY);
	}

	VXGE_HAL_SQ_STATE_SET(sq, VXGEHAL_SQ_STATE_UNKNOWN);

	sq->sq_length = sq_length;
	sq->callback = attr->callback;
	sq->towi_completion =
	    ((vxge_hal_sq_attr_t *)attr)->towi_completion;
	sq->rdma_read_completion =
	    ((vxge_hal_sq_attr_t *)attr)->rdma_read_completion;

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_lock_init(&sq->channel.post_lock, hldev->pdev);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_lock_init_irq(&sq->channel.post_lock, hldev->irqh);
#endif

#if defined(VXGE_HAL_SQ_FREE)
	vxge_os_spin_lock_init(&sq->channel.free_lock, hldev->pdev);
#elif defined(VXGE_HAL_SQ_FREE_IRQ)
	vxge_os_spin_lock_init_irq(&sq->channel.free_lock, hldev->irqh);
#endif

#if defined(VXGE_HAL_SQ_RESERVE)
	vxge_os_spin_lock_init(&sq->channel.reserve_lock, hldev->pdev);
#elif defined(VXGE_HAL_SQ_RESERVE_IRQ)
	vxge_os_spin_lock_init_irq(&sq->channel.reserve_lock, hldev->irqh);
#endif

	sq->high_towi_number_posted = 0;
	sq->high_towi_number_completed = 0;
	sq->towi_size = towi_size;
	sq->towis_per_block = towis_per_block;
	sq->towi_priv_size = sizeof (__hal_towi_priv_t);
	sq->towiblock_priv_size = sq->towis_per_block * sq->towi_priv_size;

	status = __hal_vp_oid_allocate(
	    vp->vpath,
	    &vp->vpath->session_oid_db,
	    &sq->session_id);
	if (status != VXGE_HAL_OK) {
		vxge_hal_info_log_sq(hldev, vp->vpath->vp_id,
		    "Memory allocation for session_oid_db failed %s : %d",
		    __func__, __LINE__);
		__hal_sq_delete(vpath_handle, sq);
		vxge_hal_trace_log_sq(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d",
		    __FILE__, __func__, __LINE__,
		    VXGE_HAL_ERR_TOO_MANY);
		return (VXGE_HAL_ERR_TOO_MANY);
	}

	sq->doorbell = __hal_offload_db_ptr_get(sq, sq->session_id);

	status = __hal_regpool_allocate(&vp->vpath->hldev->reg_pool,
	    &sq->cq_sq_ir_reg_entry);


	if (status != VXGE_HAL_OK) {
		vxge_hal_info_log_sq(hldev, vp->vpath->vp_id,
		    "Memory allocation for cq_sq_ir failed %s : %d",
		    __func__, __LINE__);
		__hal_sq_delete(vpath_handle, sq);
		vxge_hal_trace_log_sq(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d",
		    __FILE__, __func__, __LINE__, status);
		return (status);
	}

	sq->cq_sq_ir = sq->cq_sq_ir_reg_entry->reg_addr;

	status = __hal_regpool_allocate(&vp->vpath->hldev->reg_pool,
	    &sq->cq_rr_ir_reg_entry);

	if (status != VXGE_HAL_OK) {
		vxge_hal_info_log_sq(hldev, vp->vpath->vp_id,
		    "Memory allocation for cq_rr_ir failed == > %s : %d",
		    __func__, __LINE__);
		__hal_sq_delete(vpath_handle, sq);
		vxge_hal_trace_log_sq(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d",
		    __FILE__, __func__, __LINE__, status);
		return (status);
	}

	sq->cq_rr_ir = sq->cq_rr_ir_reg_entry->reg_addr;

	sq->opaque_handle = opaque_handle;

	__hal_channel_insert_pending_list(&sq->channel);

	vxge_hal_trace_log_sq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_PENDING);
}

/*
 * __hal_sq_delete - Removes the HW Send Queue
 * @vpath_handle: Virtual path handle to which this queue belongs
 * @sq: send queue to be deleted
 *
 * This function freeup the memory pool and removes the
 * hw send queue.
 */

vxge_hal_status_e
__hal_sq_delete(
	vxge_hal_vpath_h vpath_handle,
	__hal_sq_t *sq)
{
	__hal_device_t *hldev;
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;

	vxge_assert((vpath_handle != NULL) && (sq != NULL));

	hldev = vp->vpath->hldev;

	vxge_hal_trace_log_sq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_sq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", sq = 0x"VXGE_OS_STXFMT,
	    (ptr_t)vpath_handle, (ptr_t)sq);

	if (sq->mempool) {
		__hal_mempool_destroy(sq->mempool);
	}

	if (sq->cq_sq_ir_reg_entry) {
		__hal_regpool_free(&vp->vpath->hldev->reg_pool,
		    sq->cq_sq_ir_reg_entry);
	}

	if (sq->cq_rr_ir_reg_entry) {
		__hal_regpool_free(&vp->vpath->hldev->reg_pool,
		    sq->cq_rr_ir_reg_entry);
	}

	__hal_vp_oid_free(vp->vpath,
	    &vp->vpath->session_oid_db,
	    sq->session_id);

	__hal_channel_terminate(&sq->channel);

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_lock_destroy(&sq->channel.post_lock,
	    vp->vpath->hldev->header.pdev);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_lock_destroy_irq(&sq->channel.post_lock,
	    vp->vpath->hldev->header.pdev);
#endif

#if defined(VXGE_HAL_SQ_FREE)
	vxge_os_spin_lock_destroy(&sq->channel.free_lock,
	    vp->vpath->hldev->header.pdev);
#elif defined(VXGE_HAL_SQ_FREE_IRQ)
	vxge_os_spin_lock_destroy_irq(&sq->channel.free_lock,
	    vp->vpath->hldev->header.pdev);
#endif

#if defined(VXGE_HAL_SQ_RESERVE)
	vxge_os_spin_lock_destroy(&sq->channel.reserve_lock,
	    vp->vpath->hldev->header.pdev);
#elif defined(VXGE_HAL_SQ_RESERVE_IRQ)
	vxge_os_spin_lock_destroy_irq(&sq->channel.reserve_lock,
	    vp->vpath->hldev->header.pdev);
#endif

	__hal_channel_free(&sq->channel);

	vxge_hal_trace_log_sq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * vxge_hal_sq_delete - Deletes a SQ
 * @vpath_handle: Virtual path handle to which this queue belongs
 * @sq_id: SQ to be deleted
 * @opaque_handle: HAL Opaque handle
 *
 * This function deletes SQ.
 *
 */
vxge_hal_status_e
vxge_hal_sq_delete(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	vxge_hal_status_e status;
	__hal_device_t *hldev;
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_sq_t *sq = (__hal_sq_t *)(ptr_t)sq_id;

	vxge_assert((vpath_handle != NULL) && (sq_id != 0) &&
	    (opaque_handle != NULL));

	hldev = vp->vpath->hldev;

	vxge_hal_trace_log_sq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_sq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", sq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT, (ptr_t)vpath_handle,
	    (ptr_t)sq_id, (ptr_t)opaque_handle);

	status = __hal_sq_delete(vpath_handle, sq);

	vp->cb_fn(vp->client_handle,
	    NULL,
	    VXGE_HAL_MSG_TYPE_SQ_DELETE_RESP,
	    sq_id,
	    status,
	    opaque_handle);

	vxge_hal_trace_log_sq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_PENDING);
}


/*
 * __hal_towi_nsmr_ctrl_block_set - Set nsmr contro block in the TOWI.
 * @sq: Send Queue
 * @towi: TOWI
 * @stag_index: Stag Index
 * @stag_key: Stag Key
 * @pb_size: Size of each physical buffer
 * @pble_list: pble list allocated previously
 * @local_write_flag: Flag to enable local write
 * @remote_write_flag: Flag to enable remote write
 * @local_read_flag: Flag to enable local read
 * @remote_read_flag: Flag to enable remote read
 * @enable_mw_binding_flag: Flag to enable MW binding
 * @first_byte_offset: Byte offset of first byte in the first physical buffer
 * @mr_length: Length of memory region
 * @base_to: Base To of virtual Address
 *
 * This function allocates the nsmr control block and pble list and sets
 * into TOWI.
 *
 * See also: vxge_hal_towi_buffer_set_aligned(), vxge_hal_towi_reserve(),
 *	  vxge_hal_towi_post(),
 */
vxge_hal_status_e
__hal_towi_nsmr_ctrl_block_set(
	__hal_sq_t *sq,
	__hal_towi_t *towi,
	u32 stag_index,
	u32 stag_key,
	u32 pb_size,
	u64 pble_list,
	u32 local_write_flag,
	u32 remote_write_flag,
	u32 local_read_flag,
	u32 remote_read_flag,
	u32 enable_mw_binding_flag,
	u32 first_byte_offset,
	u64 mr_length,
	u64 base_to)
{
	vxge_hal_pblist_t *pblist = (vxge_hal_pblist_t *)(ptr_t)pble_list;
	__hal_towi_priv_t *towi_priv;
	__hal_device_t *hldev;
	vxge_hal_status_e status;

	vxge_assert((sq != NULL) && (towi != NULL) && (pblist != NULL));

	hldev = (__hal_device_t *)sq->channel.devh;

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "sq = 0x"VXGE_OS_STXFMT", towi = 0x"VXGE_OS_STXFMT
	    ", stag_index = %d, stag_key = %d, pb_size = %d, "
	    "pble_list = 0x"VXGE_OS_LLXFMT", local_write_flag = %d, "
	    "remote_write_flag = %d, local_read_flag = %d, "
	    "remote_read_flag = %d, enable_mw_binding_flag = %d, "
	    "first_byte_offset = %d, mr_length = 0x"VXGE_OS_LLDFMT", "
	    "base_to = 0x"VXGE_OS_LLXFMT, (ptr_t)sq, (ptr_t)towi,
	    stag_index, stag_key, pb_size, pble_list, local_write_flag,
	    remote_write_flag, local_read_flag, remote_read_flag,
	    enable_mw_binding_flag, first_byte_offset, mr_length, base_to);

	towi_priv = (__hal_towi_priv_t *)(ptr_t)towi->host_control_0;

	vxge_assert(towi_priv != NULL);

	towi->payload_locator_0 = VXGE_HAL_TOWI_DDP_FR_NMSR_CTRL_LENGTH(
	    sizeof (__hal_nsmr_ctrl_block_t));

	status = __hal_regpool_allocate(&hldev->reg_pool,
	    &towi_priv->nsmr_ctrl_block_reg_entry);

	if (status != VXGE_HAL_OK) {
		vxge_hal_info_log_sq(hldev,
		    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
		    "sq channel nsmr cb allocation failed %s : %d",
		    __func__, __LINE__);
		vxge_hal_trace_log_sq(hldev,
		    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d",
		    __FILE__, __func__, __LINE__, status);
		return (status);
	}

	towi_priv->nsmr_ctrl_block =
	    towi_priv->nsmr_ctrl_block_reg_entry->reg_addr;

	towi_priv->nsmr_ctrl_block->control_0 =
	    VXGE_HAL_NSMR_STAG_INDEX(stag_index) |
	    VXGE_HAL_NSMR_STAG_KEY(stag_key) |
	    VXGE_HAL_NSMR_STAG_FBO(first_byte_offset);

	towi_priv->nsmr_ctrl_block->control_1 = mr_length;

	towi_priv->nsmr_ctrl_block->control_2 = base_to;

	towi_priv->nsmr_ctrl_block->control_3 =
	    VXGE_HAL_NSMR_ADDR_TYPE(VXGE_HAL_NSMR_ADDR_TYPE_VA_BASED);

	if (local_write_flag)
		towi_priv->nsmr_ctrl_block->control_3 |=
		    VXGE_HAL_NSMR_LOCAL_WRITE(VXGE_HAL_NSMR_LOCAL_WRITE_ENABLE);

	if (remote_write_flag)
		towi_priv->nsmr_ctrl_block->control_3 |=
		    VXGE_HAL_NSMR_REMOTE_WRITE(
		    VXGE_HAL_NSMR_REMOTE_WRITE_ENABLE);

	if (local_read_flag)
		towi_priv->nsmr_ctrl_block->control_3 |=
		    VXGE_HAL_NSMR_LOCAL_READ(VXGE_HAL_NSMR_LOCAL_READ_ENABLE);

	if (remote_read_flag)
		towi_priv->nsmr_ctrl_block->control_3 |=
		    VXGE_HAL_NSMR_REMOTE_READ(VXGE_HAL_NSMR_REMOTE_READ_ENABLE);

	if (enable_mw_binding_flag)
		towi_priv->nsmr_ctrl_block->control_3 |=
		    VXGE_HAL_NSMR_MW_BINDING(VXGE_HAL_NSMR_MW_BINDING_ENABLE);

	towi_priv->nsmr_ctrl_block->control_4 =
	    VXGE_HAL_NSMR_ADDR_LIST_LENGTH(pblist->num_pbles) |
	    VXGE_HAL_NSMR_PBLE_SIZE(pb_size);

	towi_priv->nsmr_ctrl_block->pbl_block_wrapper0 =
	    __hal_pblist_dma_addr_get(pblist);

	towi_priv->nsmr_ctrl_block->pbl_block_wrapper1 =
	    VXGE_HAL_NSMR_PBL_BLOCK_LENGTH(__hal_pblist_length_get(
	    pblist, pblist->num_pbles));

	towi->payload_locator_1 =
	    towi_priv->nsmr_ctrl_block_reg_entry->dma_addr;

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * vxge_hal_send_towi_post - Post Send TOWIs from stag or buf pointer
 * @vpath_handle: Handle of Virtual path to which this queue belongs.
 * @sq_id: Handle of sq
 * @opaque_handle: RNICPI private handle.
 * @push: Flag to include push in TOWI.
 * @msn: Message Sequence Number
 * @last: Is it last message in sequence
 * @start_msg_offset: Starting message offset.
 * @read_fencing: Flag to indicate read Fencing
 * @remote_invalidate: Flag to Invalidate Remote Stag
 * @solicted_event: Flag for solicited event
 * @local_stag: Local stag where payload is located.
 * @ptr_to: Physical address of buffer or tagged offset
 * @length: length of data
 * @remote_stag: Remote stag to write to.
 *
 * This function reserves, sets and posts write TOWI to SQ.
 *
 */
vxge_hal_status_e
vxge_hal_send_towi_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 push,
	u32 msn,
	u32 last,
	u32 start_msg_offset,
	u32 read_fencing,
	u32 remote_invalidate,
	u32 solicited_event,
	u32 local_stag,
	u32 ptr_to,
	u32 length,
	u32 remote_stag)
{
	__hal_device_t *hldev;
	__hal_sq_t *sq = (__hal_sq_t *)(ptr_t)sq_id;
	__hal_towi_t *towi1 = NULL, *towi2 = NULL;
	vxge_hal_towi_h towih1 = NULL, towih2 = NULL;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (sq_id != 0));

	hldev = (__hal_device_t *)sq->channel.devh;

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", sq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", push = %d, msn = %d, "
	    "last = %d, start_msg_offset = %x, read_fencing = %d, "
	    "remote_invalidate = %d, solicited_event = %d, local_stag = %d, "
	    "ptr_to = %d, length = %d, remote_stag = %d", (ptr_t)vpath_handle,
	    (ptr_t)sq_id, (ptr_t)opaque_handle, push, msn, last,
	    start_msg_offset, read_fencing, remote_invalidate,
	    solicited_event, local_stag, ptr_to, length, remote_stag);

	vxge_assert(sq != NULL);

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_lock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_lock_irq(&sq->channel.post_lock, flags);
#endif

	if (read_fencing == 1) {

		status = __hal_channel_dtr_alloc(&sq->channel, &towih2);

		if (status != VXGE_HAL_OK) {
			vxge_hal_err_log_sq(hldev, ((__hal_vpath_handle_t *)
			    sq->channel.vph)->vpath->vp_id,
			    "sq channel dtr allocation failed %s : %d",
			    __func__, __LINE__);
			goto done;
		}

		towi2 = (__hal_towi_t *)towih2;


		towi2->control_0 = VXGE_HAL_TOWI_MSG_TYPE(
		    VXGE_HAL_TOWI_MSG_TYPE_BLOCK_FOR_LIRR) |
		    VXGE_HAL_TOWI_COMPLETION(
		    VXGE_HAL_TOWI_COMPLETION_UPDATE_HW_CQ_SQ);

		towi2->control_1 = 0;

		towi2->ddp_info_0 = 0;

		towi2->ddp_info_1 = 0;

		towi2->payload_locator_0 = 0;

		towi2->payload_locator_1 = 0;

		towi2->host_control_1 = opaque_handle;

	}

	status = __hal_channel_dtr_alloc(&sq->channel, &towih1);

	if (status != VXGE_HAL_OK) {

		vxge_hal_err_log_sq(hldev,
		    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
		    "sq channel dtr allocation failed %s : %d",
		    __func__, __LINE__);
		if (towih2 != NULL)
			__hal_channel_dtr_restore(&sq->channel, towih2);

		goto done;
	}

	towi1 = (__hal_towi_t *)towih1;

	towi1->control_0 =
	    VXGE_HAL_TOWI_TYPE(VXGE_HAL_TOWI_TYPE_OD) |
	    VXGE_HAL_TOWI_TCP_PUSH(((push)? VXGE_HAL_TOWI_PUSH_INCLUDED :
	    VXGE_HAL_TOWI_TCP_PUSH_NOT_INCLUDED)) |
	    VXGE_HAL_TOWI_MSG_TYPE(VXGE_HAL_TOWI_MSG_TYPE_DDP_UNTAGGED) |
	    VXGE_HAL_TOWI_COMPLETION(VXGE_HAL_TOWI_COMPLETION_UPDATE_HW_CQ_SQ);

	towi1->control_1 = VXGE_HAL_TOWI_PAYLOAD_BYTES(length);

	if (remote_invalidate == 0) {

		if (solicited_event == 0) {
			towi1->ddp_info_0 =
			    VXGE_HAL_TOWI_DDP_RDMA_OPCODE(
			    VXGE_HAL_TOWI_DDP_RDMA_OPCODE_SEND) |
			    VXGE_HAL_TOWI_DDP_LAST_SEGMENT_BIT(
			    VXGE_HAL_TOWI_DDP_LAST_SEGMENT_SET);
		} else {
			towi1->ddp_info_0 =
			    VXGE_HAL_TOWI_DDP_RDMA_OPCODE(
			    VXGE_HAL_TOWI_DDP_RDMA_OPCODE_SEND_SE) |
			    VXGE_HAL_TOWI_DDP_LAST_SEGMENT_BIT(
			    VXGE_HAL_TOWI_DDP_LAST_SEGMENT_SET);
		}

	} else {

		if (solicited_event == 0) {
			towi1->ddp_info_0 =
			    VXGE_HAL_TOWI_DDP_RDMA_OPCODE(
			    VXGE_HAL_TOWI_DDP_RDMA_OPCODE_SEND_INVALIDATE) |
			    VXGE_HAL_TOWI_DDP_LAST_SEGMENT_BIT(
			    VXGE_HAL_TOWI_DDP_LAST_SEGMENT_SET) |
			    VXGE_HAL_TOWI_DDP_REMOTE_STAG(remote_stag);
		} else {
			towi1->ddp_info_0 =
			    VXGE_HAL_TOWI_DDP_RDMA_OPCODE(
			    VXGE_HAL_TOWI_DDP_RDMA_OPCODE_SEND_SE_INVALIDATE) |
			    VXGE_HAL_TOWI_DDP_LAST_SEGMENT_BIT(
			    VXGE_HAL_TOWI_DDP_LAST_SEGMENT_SET) |
			    VXGE_HAL_TOWI_DDP_REMOTE_STAG(remote_stag);
		}

	}


	towi1->ddp_info_1 =
	    VXGE_HAL_TOWI_DDP_START_MSG_OFFSET(start_msg_offset) |
	    VXGE_HAL_TOWI_DDP_MSG_SEQ_NO(msn);

	towi1->payload_locator_0 =
	    VXGE_HAL_TOWI_LOCAL_STAG(local_stag) |
	    VXGE_HAL_TOWI_LENGTH(length);

	towi1->payload_locator_1 = ptr_to;

	towi1->host_control_1 = opaque_handle;

	if (towih2 != NULL) {
		__hal_channel_dtr_post(&sq->channel, towih2);
		INCREMENT_HIGH_TOWI_NUMBER_POSTED(sq);
	}

	__hal_channel_dtr_post(&sq->channel, towih1);
	INCREMENT_HIGH_TOWI_NUMBER_POSTED(sq);

	__hal_offload_db_post(sq,
	    VXGE_HAL_RNIC_OID_DB_SID_GET(sq->session_id),
	    VXGE_HAL_RNIC_OID_DB_SIN_GET(sq->session_id),
	    sq->high_towi_number_posted);

	status = VXGE_HAL_OK;

done:

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_unlock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_unlock_irq(&sq->channel.post_lock, flags);
#endif

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);

	return (status);
}

/*
 * vxge_hal_send_towi_odlist_post - Post Send TOWIs from od list
 * @vpath_handle: Handle of Virtual path to which this queue belongs.
 * @sq_id: Handle of sq
 * @opaque_handle: RNICPI private handle.
 * @push: Flag to include push in TOWI.
 * @msn: Message Sequence Number
 * @last: Is it last message in sequence
 * @start_msg_offset: Starting message offset.
 * @read_fencing: Flag to indicate read Fencing
 * @remote_invalidate: Flag to Invalidate Remote Stag
 * @solicted_event: Flag for solicited event
 * @remote_stag: Remote stag to write to.
 * @od_list: OD List prepared earlier
 *
 * This function reserves, sets and posts write TOWI to SQ.
 *
 */
vxge_hal_status_e
vxge_hal_send_towi_odlist_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 push,
	u32 msn,
	u32 last,
	u32 start_msg_offset,
	u32 read_fencing,
	u32 remote_invalidate,
	u32 solicited_event,
	u32 remote_stag,
	vxge_hal_odlist_t *odlist)
{
	__hal_device_t *hldev;
	__hal_sq_t *sq = (__hal_sq_t *)(ptr_t)sq_id;
	__hal_towi_t *towi1 = NULL, *towi2 = NULL;
	vxge_hal_towi_h towih1 = NULL, towih2 = NULL;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (sq_id != 0));

	hldev = (__hal_device_t *)sq->channel.devh;

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", sq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", push = %d, msn = %d, "
	    "last = %d, start_msg_offset = %x, read_fencing = %d, "
	    "remote_invalidate = %d, solicited_event = %d, "
	    "remote_stag = %d, odlist = 0x"VXGE_OS_STXFMT, (ptr_t)vpath_handle,
	    (ptr_t)sq_id, (ptr_t)opaque_handle, push, msn, last,
	    start_msg_offset, read_fencing, remote_invalidate,
	    solicited_event, remote_stag, (ptr_t)odlist);

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_lock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_lock_irq(&sq->channel.post_lock, flags);
#endif

	if (read_fencing == 1) {

		status = __hal_channel_dtr_alloc(&sq->channel, &towih2);

		if (status != VXGE_HAL_OK) {
			vxge_hal_err_log_sq(hldev, ((__hal_vpath_handle_t *)
			    sq->channel.vph)->vpath->vp_id,
			    "sq channel dtr allocation failed %s : %d",
			    __func__, __LINE__);
			goto done;
		}

		towi2 = (__hal_towi_t *)towih2;


		towi2->control_0 = VXGE_HAL_TOWI_MSG_TYPE(
		    VXGE_HAL_TOWI_MSG_TYPE_BLOCK_FOR_LIRR) |
		    VXGE_HAL_TOWI_COMPLETION(
		    VXGE_HAL_TOWI_COMPLETION_UPDATE_HW_CQ_SQ);

		towi2->control_1 = 0;

		towi2->ddp_info_0 = 0;

		towi2->ddp_info_1 = 0;

		towi2->payload_locator_0 = 0;

		towi2->payload_locator_1 = 0;

		towi2->host_control_1 = opaque_handle;

	}

	status = __hal_channel_dtr_alloc(&sq->channel, &towih1);

	if (status != VXGE_HAL_OK) {

		vxge_hal_err_log_sq(hldev,
		    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
		    "sq channel dtr allocation failed %s : %d",
		    __func__, __LINE__);
		if (towih2 != NULL)
			__hal_channel_dtr_restore(&sq->channel, towih2);

		goto done;
	}

	towi1 = (__hal_towi_t *)towih1;

	towi1->control_0 =
	    VXGE_HAL_TOWI_TYPE(VXGE_HAL_TOWI_TYPE_OD_LB_WRAPPER) |
	    VXGE_HAL_TOWI_TCP_PUSH(((push)? VXGE_HAL_TOWI_PUSH_INCLUDED :
	    VXGE_HAL_TOWI_TCP_PUSH_NOT_INCLUDED)) |
	    VXGE_HAL_TOWI_MSG_TYPE(VXGE_HAL_TOWI_MSG_TYPE_DDP_UNTAGGED) |
	    VXGE_HAL_TOWI_COMPLETION(VXGE_HAL_TOWI_COMPLETION_UPDATE_HW_CQ_SQ);

	towi1->control_1 = VXGE_HAL_TOWI_PAYLOAD_BYTES(odlist->total_bytes);

	if (remote_invalidate == 0) {

		if (solicited_event == 0) {
			towi1->ddp_info_0 =
			    VXGE_HAL_TOWI_DDP_RDMA_OPCODE(
			    VXGE_HAL_TOWI_DDP_RDMA_OPCODE_SEND) |
			    VXGE_HAL_TOWI_DDP_LAST_SEGMENT_BIT(
			    VXGE_HAL_TOWI_DDP_LAST_SEGMENT_SET);
		} else {
			towi1->ddp_info_0 =
			    VXGE_HAL_TOWI_DDP_RDMA_OPCODE(
			    VXGE_HAL_TOWI_DDP_RDMA_OPCODE_SEND_SE) |
			    VXGE_HAL_TOWI_DDP_LAST_SEGMENT_BIT(
			    VXGE_HAL_TOWI_DDP_LAST_SEGMENT_SET);
		}

	} else {

		if (solicited_event == 0) {
			towi1->ddp_info_0 =
			    VXGE_HAL_TOWI_DDP_RDMA_OPCODE(
			    VXGE_HAL_TOWI_DDP_RDMA_OPCODE_SEND_INVALIDATE) |
			    VXGE_HAL_TOWI_DDP_LAST_SEGMENT_BIT(
			    VXGE_HAL_TOWI_DDP_LAST_SEGMENT_SET) |
			    VXGE_HAL_TOWI_DDP_REMOTE_STAG(remote_stag);
		} else {
			towi1->ddp_info_0 =
			    VXGE_HAL_TOWI_DDP_RDMA_OPCODE(
			    VXGE_HAL_TOWI_DDP_RDMA_OPCODE_SEND_SE_INVALIDATE) |
			    VXGE_HAL_TOWI_DDP_LAST_SEGMENT_BIT(
			    VXGE_HAL_TOWI_DDP_LAST_SEGMENT_SET) |
			    VXGE_HAL_TOWI_DDP_REMOTE_STAG(remote_stag);
		}

	}


	towi1->ddp_info_1 =
	    VXGE_HAL_TOWI_DDP_START_MSG_OFFSET(start_msg_offset) |
	    VXGE_HAL_TOWI_DDP_MSG_SEQ_NO(msn);

	towi1->payload_locator_0 =
	    VXGE_HAL_TOWI_LENGTH(__hal_odlist_length_get(odlist));

	towi1->payload_locator_1 = __hal_odlist_dma_addr_get(odlist);

	towi1->host_control_1 = opaque_handle;

	if (towih2 != NULL) {
		__hal_channel_dtr_post(&sq->channel, towih2);
		INCREMENT_HIGH_TOWI_NUMBER_POSTED(sq);
	}

	__hal_channel_dtr_post(&sq->channel, towih1);
	INCREMENT_HIGH_TOWI_NUMBER_POSTED(sq);

	__hal_offload_db_post(sq,
	    VXGE_HAL_RNIC_OID_DB_SID_GET(sq->session_id),
	    VXGE_HAL_RNIC_OID_DB_SIN_GET(sq->session_id),
	    sq->high_towi_number_posted);

	status = VXGE_HAL_OK;

done:

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_unlock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_unlock_irq(&sq->channel.post_lock, flags);
#endif
	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);

	return (status);
}

/*
 * vxge_hal_write_towi_post - Post Write TOWIs with local stag or buffer pointer
 * @vpath_handle: Handle of Virtual path to which this queue belongs.
 * @sq_id: Handle of sq
 * @opaque_handle: RNICPI private handle.
 * @push: Flag to include push in TOWI.
 * @last: Is it last message in sequence
 * @read_fencing: Flag to indicate read Fencing
 * @local_stag: Local stag where payload is located.
 * @ptr_to: Tagged offset
 * @length: length of data
 * @remote_stag: Remote stag to write to.
 * @remote_tagged_offset: Tagged Offset in remote buffer.
 *
 * This function reserves, sets and posts write TOWI to SQ.
 *
 */
vxge_hal_status_e
vxge_hal_write_towi_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 push,
	u32 last,
	u32 read_fencing,
	u32 local_stag,
	u64 ptr_to,
	u32 length,
	u32 remote_stag,
	u64 remote_tagged_offset)
{
	__hal_device_t *hldev;
	__hal_sq_t *sq = (__hal_sq_t *)(ptr_t)sq_id;
	__hal_towi_t *towi1 = NULL, *towi2 = NULL;
	vxge_hal_towi_h towih1 = NULL, towih2 = NULL;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (sq_id != 0));

	hldev = (__hal_device_t *)sq->channel.devh;

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", sq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", push = %d, last = %d, "
	    "read_fencing = %d, local_stag = %d, ptr_to = 0x"VXGE_OS_LLXFMT", "
	    "length = %d, remote_stag = %d, "
	    "remote_tagged_offset = 0x"VXGE_OS_LLXFMT, (ptr_t)vpath_handle,
	    (ptr_t)sq_id, (ptr_t)opaque_handle, push, last, read_fencing,
	    local_stag, ptr_to, length, remote_stag, remote_tagged_offset);

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_lock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_lock_irq(&sq->channel.post_lock, flags);
#endif

	if (read_fencing == 1) {

		status = __hal_channel_dtr_alloc(&sq->channel, &towih2);

		if (status != VXGE_HAL_OK) {
			vxge_hal_err_log_sq(hldev, ((__hal_vpath_handle_t *)
			    sq->channel.vph)->vpath->vp_id,
			    "sq channel dtr allocation failed %s : %d",
			    __func__, __LINE__);
			goto done;
		}

		towi2 = (__hal_towi_t *)towih2;


		towi2->control_0 = VXGE_HAL_TOWI_MSG_TYPE(
		    VXGE_HAL_TOWI_MSG_TYPE_BLOCK_FOR_LIRR) |
		    VXGE_HAL_TOWI_COMPLETION(
		    VXGE_HAL_TOWI_COMPLETION_UPDATE_HW_CQ_SQ);

		towi2->control_1 = 0;

		towi2->ddp_info_0 = 0;

		towi2->ddp_info_1 = 0;

		towi2->payload_locator_0 = 0;

		towi2->payload_locator_1 = 0;

		towi2->host_control_1 = opaque_handle;

	}

	status = __hal_channel_dtr_alloc(&sq->channel, &towih1);

	if (status != VXGE_HAL_OK) {

		vxge_hal_err_log_sq(hldev,
		    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
		    "sq channel dtr allocation failed %s : %d",
		    __func__, __LINE__);
		if (towih2 != NULL)
			__hal_channel_dtr_restore(&sq->channel, towih2);

		goto done;
	}

	towi1 = (__hal_towi_t *)towih1;

	towi1->control_0 =
	    VXGE_HAL_TOWI_TYPE(VXGE_HAL_TOWI_TYPE_OD) |
	    VXGE_HAL_TOWI_TCP_PUSH(((push)? VXGE_HAL_TOWI_PUSH_INCLUDED :
	    VXGE_HAL_TOWI_TCP_PUSH_NOT_INCLUDED)) |
	    VXGE_HAL_TOWI_MSG_TYPE(VXGE_HAL_TOWI_MSG_TYPE_DDP_TAGGED) |
	    VXGE_HAL_TOWI_COMPLETION(VXGE_HAL_TOWI_COMPLETION_UPDATE_HW_CQ_SQ);

	towi1->control_1 = VXGE_HAL_TOWI_PAYLOAD_BYTES(length);

	towi1->ddp_info_0 =
	    VXGE_HAL_TOWI_DDP_RDMA_OPCODE(VXGE_HAL_TOWI_DDP_RDMA_OPCODE_WRITE) |
	    VXGE_HAL_TOWI_DDP_LAST_SEGMENT_BIT(last) |
	    VXGE_HAL_TOWI_DDP_REMOTE_STAG(remote_stag);

	towi1->ddp_info_1 = remote_tagged_offset;

	towi1->payload_locator_0 =
	    VXGE_HAL_TOWI_LOCAL_STAG(local_stag) |
	    VXGE_HAL_TOWI_LENGTH(length);

	towi1->payload_locator_1 = ptr_to;

	towi1->host_control_1 = opaque_handle;

	if (towih2 != NULL) {
		__hal_channel_dtr_post(&sq->channel, towih2);
		INCREMENT_HIGH_TOWI_NUMBER_POSTED(sq);
	}

	__hal_channel_dtr_post(&sq->channel, towih1);
	INCREMENT_HIGH_TOWI_NUMBER_POSTED(sq);

	__hal_offload_db_post(sq,
	    VXGE_HAL_RNIC_OID_DB_SID_GET(sq->session_id),
	    VXGE_HAL_RNIC_OID_DB_SIN_GET(sq->session_id),
	    sq->high_towi_number_posted);

	status = VXGE_HAL_OK;

done:

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_unlock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_unlock_irq(&sq->channel.post_lock, flags);
#endif
	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);

	return (status);
}

/*
 * vxge_hal_write_towi_odlist_post - Post Write TOWIs using odlist
 * @vpath_handle: Handle of Virtual path to which this queue belongs.
 * @sq_id: Handle of sq
 * @opaque_handle: RNICPI private handle.
 * @push: Flag to include push in TOWI.
 * @last: Is it last message in sequence
 * @read_fencing: Flag to indicate read Fencing
 * @remote_stag: Remote stag to write to.
 * @remote_tagged_offset: Tagged Offset in remote buffer.
 * @odlist: OD List prepared earlier
 *
 * This function reserves, sets and posts write TOWI to SQ.
 *
 */
vxge_hal_status_e
vxge_hal_write_towi_odlist_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 push,
	u32 last,
	u32 read_fencing,
	u32 remote_stag,
	u64 remote_tagged_offset,
	vxge_hal_odlist_t *odlist)
{
	__hal_device_t *hldev;
	__hal_sq_t *sq = (__hal_sq_t *)(ptr_t)sq_id;
	__hal_towi_t *towi1 = NULL, *towi2 = NULL;
	vxge_hal_towi_h towih1 = NULL, towih2 = NULL;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (sq_id != 0));

	hldev = (__hal_device_t *)sq->channel.devh;

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", sq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", push = %d, last = %d, "
	    "read_fencing = %d, remote_stag = %d, "
	    "remote_tagged_offset = 0x"VXGE_OS_LLXFMT", "
	    "odlist = 0x"VXGE_OS_STXFMT, (ptr_t)vpath_handle, (ptr_t)sq_id,
	    (ptr_t)opaque_handle, push, last, read_fencing, remote_stag,
	    remote_tagged_offset, (ptr_t)odlist);

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_lock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_lock_irq(&sq->channel.post_lock, flags);
#endif

	if (read_fencing == 1) {

		status = __hal_channel_dtr_alloc(&sq->channel, &towih2);

		if (status != VXGE_HAL_OK) {
			vxge_hal_err_log_sq(hldev, ((__hal_vpath_handle_t *)
			    sq->channel.vph)->vpath->vp_id,
			    "sq channel dtr allocation failed %s : %d",
			    __func__, __LINE__);
			goto done;
		}

		towi2 = (__hal_towi_t *)towih2;


		towi2->control_0 = VXGE_HAL_TOWI_MSG_TYPE(
		    VXGE_HAL_TOWI_MSG_TYPE_BLOCK_FOR_LIRR) |
		    VXGE_HAL_TOWI_COMPLETION(
		    VXGE_HAL_TOWI_COMPLETION_UPDATE_HW_CQ_SQ);

		towi2->control_1 = 0;

		towi2->ddp_info_0 = 0;

		towi2->ddp_info_1 = 0;

		towi2->payload_locator_0 = 0;

		towi2->payload_locator_1 = 0;

		towi2->host_control_1 = opaque_handle;

	}

	status = __hal_channel_dtr_alloc(&sq->channel, &towih1);

	if (status != VXGE_HAL_OK) {

		vxge_hal_err_log_sq(hldev,
		    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
		    "sq channel dtr allocation failed %s : %d",
		    __func__, __LINE__);
		if (towih2 != NULL)
			__hal_channel_dtr_restore(&sq->channel, towih2);

		goto done;
	}

	towi1 = (__hal_towi_t *)towih1;

	towi1->control_0 =
	    VXGE_HAL_TOWI_TYPE(VXGE_HAL_TOWI_TYPE_OD_LB_WRAPPER) |
	    VXGE_HAL_TOWI_TCP_PUSH(((push)? VXGE_HAL_TOWI_PUSH_INCLUDED :
	    VXGE_HAL_TOWI_TCP_PUSH_NOT_INCLUDED)) |
	    VXGE_HAL_TOWI_MSG_TYPE(VXGE_HAL_TOWI_MSG_TYPE_DDP_TAGGED) |
	    VXGE_HAL_TOWI_COMPLETION(VXGE_HAL_TOWI_COMPLETION_UPDATE_HW_CQ_SQ);

	towi1->control_1 = VXGE_HAL_TOWI_PAYLOAD_BYTES(odlist->total_bytes);

	towi1->ddp_info_0 =
	    VXGE_HAL_TOWI_DDP_RDMA_OPCODE(VXGE_HAL_TOWI_DDP_RDMA_OPCODE_WRITE) |
	    VXGE_HAL_TOWI_DDP_LAST_SEGMENT_BIT(last) |
	    VXGE_HAL_TOWI_DDP_REMOTE_STAG(remote_stag);

	towi1->ddp_info_1 = remote_tagged_offset;

	towi1->payload_locator_0 =
	    VXGE_HAL_TOWI_LENGTH(__hal_odlist_length_get(odlist));

	towi1->payload_locator_1 = __hal_odlist_dma_addr_get(odlist);

	towi1->host_control_1 = opaque_handle;

	if (towih2 != NULL) {
		__hal_channel_dtr_post(&sq->channel, towih2);
		INCREMENT_HIGH_TOWI_NUMBER_POSTED(sq);
	}

	__hal_channel_dtr_post(&sq->channel, towih1);

	INCREMENT_HIGH_TOWI_NUMBER_POSTED(sq);

	__hal_offload_db_post(sq,
	    VXGE_HAL_RNIC_OID_DB_SID_GET(sq->session_id),
	    VXGE_HAL_RNIC_OID_DB_SIN_GET(sq->session_id),
	    sq->high_towi_number_posted);

	status = VXGE_HAL_OK;

done:

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_unlock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_unlock_irq(&sq->channel.post_lock, flags);
#endif
	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);

	return (status);
}

/*
 * vxge_hal_read_towi_post - Post Read TOWIs
 * @vpath_handle: Handle of Virtual path to which this queue belongs.
 * @sq_id: Handle of sq
 * @opaque_handle: RNICPI private handle.
 * @push: Flag to include push in TOWI.
 * @msn: Message sequence number of message.
 * @invalidate_local_stag: Flag to invalidate Local Stag
 * @local_fencing: Flag to invalidate Local Fencing
 * @local_stag: Local stag where RDMA Read Request is located.
 * @length: Length of Physical buffer containing RDMA Read Request
 * @ptr_to: Physical buffer containing RDMA Read Request(if
 *		local_stag = 0) or tagged offset(if local_stag > 0)
 *
 * This function reserves, sets and posts Read TOWI to SQ.
 *
 */
vxge_hal_status_e
vxge_hal_read_towi_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 push,
	u32 msn,
	u32 invalidate_local_stag,
	u32 local_fencing,
	u32 local_stag,
	u32 length,
	u64 ptr_to)
{
	__hal_device_t *hldev;
	__hal_sq_t *sq = (__hal_sq_t *)(ptr_t)sq_id;
	__hal_towi_t *towi1, *towi2 = NULL;
	vxge_hal_towi_h towih1, towih2 = NULL;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (sq_id != 0));

	hldev = (__hal_device_t *)sq->channel.devh;

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", sq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", push = %d, msn = %d, "
	    "invalidate_local_stag = %d, local_fencing = %d, "
	    "local_stag = %d, length = %d, ptr_to = 0x"VXGE_OS_LLDFMT,
	    (ptr_t)vpath_handle, (ptr_t)sq_id, (ptr_t)opaque_handle,
	    push, msn, invalidate_local_stag,
	    local_fencing, local_stag, length, ptr_to);

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_lock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_lock_irq(&sq->channel.post_lock, flags);
#endif

	status = __hal_channel_dtr_alloc(&sq->channel, &towih1);

	if (status != VXGE_HAL_OK) {
		vxge_hal_err_log_sq(hldev,
		    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
		    "sq channel dtr allocation failed %s : %d",
		    __func__, __LINE__);
		goto done;
	}

	towi1 = (__hal_towi_t *)towih1;

	if ((invalidate_local_stag == 1) && (local_fencing == 0)) {

		towi1->control_0 = VXGE_HAL_TOWI_MSG_TYPE(
		    VXGE_HAL_TOWI_MSG_TYPE_LIRR_INVALID);

		towi1->control_1 =
		    VXGE_HAL_TOWI_PAYLOAD_BYTES(length) |
		    VXGE_HAL_TOWI_LOCAL_STAG_TO_INVALID(local_stag);

	} else {

		towi1->control_0 = VXGE_HAL_TOWI_MSG_TYPE(
		    VXGE_HAL_TOWI_MSG_TYPE_LIRR);

		towi1->control_1 = VXGE_HAL_TOWI_PAYLOAD_BYTES(length);

	}

	towi1->control_0 |=
	    VXGE_HAL_TOWI_TYPE(VXGE_HAL_TOWI_TYPE_OD) |
	    VXGE_HAL_TOWI_TCP_PUSH(((push)? VXGE_HAL_TOWI_PUSH_INCLUDED :
	    VXGE_HAL_TOWI_TCP_PUSH_NOT_INCLUDED)) |
	    VXGE_HAL_TOWI_COMPLETION(VXGE_HAL_TOWI_COMPLETION_UPDATE_HW_CQ_SQ) |
	    VXGE_HAL_TOWI_READ_COMPLETION(
	    VXGE_HAL_TOWI_READ_COMPLETION_UPDATE_HW_CQ_SQ);

	towi1->ddp_info_0 =
	    VXGE_HAL_TOWI_DDP_RDMA_OPCODE(VXGE_HAL_TOWI_DDP_RDMA_OPCODE_READ) |
	    VXGE_HAL_TOWI_DDP_LAST_SEGMENT_BIT(
	    VXGE_HAL_TOWI_DDP_LAST_SEGMENT_SET);

	towi1->ddp_info_1 =
	    VXGE_HAL_TOWI_DDP_START_MSG_OFFSET(0) |
	    VXGE_HAL_TOWI_DDP_MSG_SEQ_NO(msn);

	towi1->payload_locator_0 =
	    VXGE_HAL_TOWI_LOCAL_STAG(local_stag) |
	    VXGE_HAL_TOWI_LENGTH(length);

	towi1->payload_locator_1 = ptr_to;

	towi1->host_control_1 = opaque_handle;

	if (local_fencing == 1) {

		status = __hal_channel_dtr_alloc(&sq->channel, &towih2);

		if (status != VXGE_HAL_OK) {
			vxge_hal_err_log_sq(hldev, ((__hal_vpath_handle_t *)
			    sq->channel.vph)->vpath->vp_id,
			    "sq channel dtr allocation failed %s : %d",
			    __func__, __LINE__);
			__hal_channel_dtr_restore(&sq->channel, towih1);
			goto done;
		}

		towi2 = (__hal_towi_t *)towih2;

		if (invalidate_local_stag == 1) {

			towi2->control_0 = VXGE_HAL_TOWI_MSG_TYPE(
			    VXGE_HAL_TOWI_MSG_TYPE_INVALID_LOCAL_STAG_FENCE);

			towi2->control_1 =
			    VXGE_HAL_TOWI_LOCAL_STAG_TO_INVALID(local_stag);

		} else {

			towi2->control_0 = VXGE_HAL_TOWI_MSG_TYPE(
			    VXGE_HAL_TOWI_MSG_TYPE_BLOCK_FOR_LIRR);

			towi2->control_1 = 0;

		}

		towi2->control_0 |= VXGE_HAL_TOWI_COMPLETION(
		    VXGE_HAL_TOWI_COMPLETION_UPDATE_HW_CQ_SQ);

		towi2->ddp_info_0 = 0;

		towi2->ddp_info_1 = 0;

		towi2->payload_locator_0 = 0;

		towi2->payload_locator_1 = 0;

		towi2->host_control_1 = opaque_handle;

	}

	__hal_channel_dtr_post(&sq->channel, towih1);
	INCREMENT_HIGH_TOWI_NUMBER_POSTED(sq);

	if (towih2 != NULL) {
		__hal_channel_dtr_post(&sq->channel, towih2);
		INCREMENT_HIGH_TOWI_NUMBER_POSTED(sq);
	}

	__hal_offload_db_post(sq,
	    VXGE_HAL_RNIC_OID_DB_SID_GET(sq->session_id),
	    VXGE_HAL_RNIC_OID_DB_SIN_GET(sq->session_id),
	    sq->high_towi_number_posted);

	status = VXGE_HAL_OK;

done:

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_unlock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_unlock_irq(&sq->channel.post_lock, flags);
#endif
	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);

	return (status);
}

/*
 * vxge_hal_bind_mw_towi_post - Post Bind Memory Window TOWI
 * @vpath_handle: Handle of Virtual path to which this queue belongs.
 * @sq_id: Handle of sq
 * @opaque_handle: RNICPI private handle.
 * @mw_stag_index: The existing Memory Window's STag Index.
 * @mw_stag_key: The STag Key to use for the MW.
 * @mr_stag: The existing Memory Region's STAG index.
 *		Least significant 8 bits is the key.
 * @mw_ctrl: Control flags for Memory Window.
 * @mw_length: Memory window length in bytes.
 * @tagged_offset: Memory window tagged offset in bytes.
 *
 * This function reserves, sets and posts Bind Memory Window TOWI to SQ.
 *
 */
vxge_hal_status_e
vxge_hal_bind_mw_towi_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 mw_stag_index,
	u32 mw_stag_key,
	u32 mr_stag,
	u32 mw_ctrl,
	u64 mw_length,
	u64 tagged_offset)
{
	__hal_device_t *hldev;
	__hal_sq_t *sq = (__hal_sq_t *)(ptr_t)sq_id;
	__hal_towi_t *towi;
	vxge_hal_towi_h towih;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (sq_id != 0));

	hldev = (__hal_device_t *)sq->channel.devh;

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", sq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", mw_stag_index = %d, "
	    "mw_stag_key = %d, mr_stag = %d, mw_ctrl = %d, "
	    "mw_length = 0x"VXGE_OS_LLDFMT", tagged_offset = 0x"VXGE_OS_LLDFMT,
	    (ptr_t)vpath_handle, (ptr_t)sq_id, (ptr_t)opaque_handle,
	    mw_stag_index, mw_stag_key, mr_stag, mw_ctrl, mw_length,
	    tagged_offset);

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_lock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_lock_irq(&sq->channel.post_lock, flags);
#endif

	status = __hal_channel_dtr_alloc(&sq->channel, &towih);

	if (status != VXGE_HAL_OK) {
		vxge_hal_err_log_sq(hldev,
		    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
		    "sq channel dtr allocation failed %s : %d",
		    __func__, __LINE__);
		goto done;
	}

	towi = (__hal_towi_t *)towih;

	towi->control_0 =
	    VXGE_HAL_TOWI_MSG_TYPE(VXGE_HAL_TOWI_MSG_TYPE_BIND_MEM_WINDOW) |
	    VXGE_HAL_TOWI_COMPLETION(VXGE_HAL_TOWI_COMPLETION_UPDATE_HW_CQ_SQ);

	towi->control_1 = 0;
	towi->ddp_info_0 =
	    VXGE_HAL_TOWI_DDP_MW_STAG_INDEX(mw_stag_index) |
	    VXGE_HAL_TOWI_DDP_MW_STAG_KEY(mw_stag_key) |
	    VXGE_HAL_TOWI_DDP_MR_STAG(mr_stag);

	towi->ddp_info_1 = VXGE_HAL_TOWI_DDP_MW_CONTROL(mw_ctrl);

	towi->payload_locator_0 = mw_length;

	towi->payload_locator_1 = tagged_offset;

	towi->host_control_1 = opaque_handle;

	__hal_channel_dtr_post(&sq->channel, towih);

	INCREMENT_HIGH_TOWI_NUMBER_POSTED(sq);

	__hal_offload_db_post(sq,
	    VXGE_HAL_RNIC_OID_DB_SID_GET(sq->session_id),
	    VXGE_HAL_RNIC_OID_DB_SIN_GET(sq->session_id),
	    sq->high_towi_number_posted);

	status = VXGE_HAL_OK;

done:

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_unlock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_unlock_irq(&sq->channel.post_lock, flags);
#endif
	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);

	return (status);
}

/*
 * vxge_hal_fmr_towi_post - Post Fast Memory Registration TOWI
 * @vpath_handle: Handle of Virtual path to which this queue belongs.
 * @sq_id: Handle of sq
 * @opaque_handle: RNICPI private handle.
 * @stag_index: Stag Index
 * @stag_key: Stag Key
 * @pb_size: Size of each physical buffer
 * @pble_list: pble list allocated previously
 * @local_write_flag: Flag to enable local write
 * @remote_write_flag: Flag to enable remote write
 * @local_read_flag: Flag to enable local read
 * @remote_read_flag: Flag to enable remote read
 * @enable_mw_binding_flag: Flag to enable MW binding
 * @first_byte_offset: Byte offset of first byte in the first physical buffer
 * @mr_length: Length of memory region
 * @base_to: Base To of virtual Address
 *
 * This function reserves, sets and posts Fast Memory Registration TOWI to SQ.
 *
 */
vxge_hal_status_e
vxge_hal_fmr_towi_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 stag_index,
	u32 stag_key,
	u32 pb_size,
	u64 pble_list,
	u32 local_write_flag,
	u32 remote_write_flag,
	u32 local_read_flag,
	u32 remote_read_flag,
	u32 enable_mw_binding_flag,
	u32 first_byte_offset,
	u64 mr_length,
	u64 base_to)
{
	__hal_device_t *hldev;
	__hal_sq_t *sq = (__hal_sq_t *)(ptr_t)sq_id;
	__hal_towi_t *towi;
	vxge_hal_towi_h towih;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (sq_id != 0));

	hldev = (__hal_device_t *)sq->channel.devh;

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", sq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", stag_index = %d, "
	    "stag_key = %d, pb_size = %d, pble_list = 0x"VXGE_OS_LLXFMT", "
	    "local_write_flag = %d, remote_write_flag = %d, "
	    "local_read_flag = %d, remote_read_flag = %d, "
	    "enable_mw_binding_flag = %d, first_byte_offset = %d, "
	    "mr_length = 0x"VXGE_OS_LLDFMT", base_to = 0x"VXGE_OS_LLXFMT,
	    (ptr_t)vpath_handle, (ptr_t)sq_id, (ptr_t)opaque_handle,
	    stag_index, stag_key, pb_size, pble_list, local_write_flag,
	    remote_write_flag, local_read_flag, remote_read_flag,
	    enable_mw_binding_flag, first_byte_offset, mr_length, base_to);

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_lock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_lock_irq(&sq->channel.post_lock, flags);
#endif

	status = __hal_channel_dtr_alloc(&sq->channel, &towih);

	if (status != VXGE_HAL_OK) {
		vxge_hal_err_log_sq(hldev,
		    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
		    "sq channel dtr allocation failed %s : %d",
		    __func__, __LINE__);
		goto done;
	}

	towi = (__hal_towi_t *)towih;

	towi->control_0 =
	    VXGE_HAL_TOWI_MSG_TYPE(VXGE_HAL_TOWI_MSG_TYPE_FAST_REG_NSMR) |
	    VXGE_HAL_TOWI_COMPLETION(VXGE_HAL_TOWI_COMPLETION_UPDATE_HW_CQ_SQ);

	towi->control_1 = 0;
	towi->ddp_info_0 = 0;
	towi->ddp_info_1 = 0;

	__hal_towi_nsmr_ctrl_block_set(
	    sq,
	    towi,
	    stag_index,
	    stag_key,
	    pb_size,
	    pble_list,
	    local_write_flag,
	    remote_write_flag,
	    local_read_flag,
	    remote_read_flag,
	    enable_mw_binding_flag,
	    first_byte_offset,
	    mr_length,
	    base_to);

	towi->host_control_1 = opaque_handle;

	__hal_channel_dtr_post(&sq->channel, towih);

	INCREMENT_HIGH_TOWI_NUMBER_POSTED(sq);

	__hal_offload_db_post(sq,
	    VXGE_HAL_RNIC_OID_DB_SID_GET(sq->session_id),
	    VXGE_HAL_RNIC_OID_DB_SIN_GET(sq->session_id),
	    sq->high_towi_number_posted);

	status = VXGE_HAL_OK;

done:

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_unlock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_unlock_irq(&sq->channel.post_lock, flags);
#endif
	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);

	return (status);
}

/*
 * vxge_hal_local_invalidate_towi_post - Post Local Invalidate MR or MW TOWI
 * @vpath_handle: Handle of Virtual path to which this queue belongs.
 * @sq_id: Handle of sq
 * @opaque_handle: RNICPI private handle.
 * @stag: Stag to be invalidated.
 * @local_fencing_flag: Falg to enable local fencing.
 *
 * This function reserves, sets and posts Local Invalidate MR or MW TOWI to SQ.
 *
 */
vxge_hal_status_e
vxge_hal_local_invalidate_towi_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 stag,
	u32 local_fencing_flag)
{
	__hal_device_t *hldev;
	__hal_sq_t *sq = (__hal_sq_t *)(ptr_t)sq_id;
	__hal_towi_t *towi;
	vxge_hal_towi_h towih;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (sq_id != 0));

	hldev = (__hal_device_t *)sq->channel.devh;

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", sq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", stag = %d, "
	    "local_fencing_flag = %d", (ptr_t)vpath_handle, (ptr_t)sq_id,
	    (ptr_t)opaque_handle, stag, local_fencing_flag);

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_lock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_lock_irq(&sq->channel.post_lock, flags);
#endif

	status = __hal_channel_dtr_alloc(&sq->channel, &towih);

	if (status != VXGE_HAL_OK) {
		vxge_hal_err_log_sq(hldev,
		    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
		    "sq channel dtr allocation failed %s : %d",
		    __func__, __LINE__);
		goto done;
	}

	towi = (__hal_towi_t *)towih;

	if (local_fencing_flag) {
		towi->control_0 = VXGE_HAL_TOWI_MSG_TYPE(
		    VXGE_HAL_TOWI_MSG_TYPE_INVALID_LOCAL_STAG_FENCE);
	} else {
		towi->control_0 = VXGE_HAL_TOWI_MSG_TYPE(
		    VXGE_HAL_TOWI_MSG_TYPE_INVALID_LOCAL_STAG);
	}

	towi->control_0 |=
	    VXGE_HAL_TOWI_COMPLETION(VXGE_HAL_TOWI_COMPLETION_UPDATE_HW_CQ_SQ);

	towi->control_1 = VXGE_HAL_TOWI_LOCAL_STAG_TO_INVALID(stag);

	towi->ddp_info_0 = 0;
	towi->ddp_info_1 = 0;
	towi->payload_locator_0 = 0;
	towi->payload_locator_1 = 0;

	towi->host_control_1 = opaque_handle;

	__hal_channel_dtr_post(&sq->channel, towih);

	INCREMENT_HIGH_TOWI_NUMBER_POSTED(sq);

	__hal_offload_db_post(sq,
	    VXGE_HAL_RNIC_OID_DB_SID_GET(sq->session_id),
	    VXGE_HAL_RNIC_OID_DB_SIN_GET(sq->session_id),
	    sq->high_towi_number_posted);

	status = VXGE_HAL_OK;

done:

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_unlock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_unlock_irq(&sq->channel.post_lock, flags);
#endif
	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);

	return (status);
}

/*
 * vxge_hal_terminate_towi_post - Post Terminate TOWIs
 * @vpath_handle: Handle of Virtual path to which this queue belongs.
 * @sq_id: Handle of sq
 * @opaque_handle: RNICPI private handle.
 * @push: Flag to include push in TOWI.
 * @msn: Message sequence number of message.
 * @local_stag: Local stag where TERMINATE Request is located.
 * @length: Length of Physical buffer containing TERMINATE Request
 * @ptr_to: Physical buffer containing TERMINATE Request(if
 *		local_stag = 0) or tagged offset(if local_stag > 0)
 *
 * This function reserves, sets and posts Terminate TOWI to SQ.
 *
 */
vxge_hal_status_e
vxge_hal_terminate_towi_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 push,
	u32 msn,
	u32 local_stag,
	u32 length,
	u64 ptr_to)
{
	__hal_device_t *hldev;
	__hal_sq_t *sq = (__hal_sq_t *)(ptr_t)sq_id;
	__hal_towi_t *towi;
	vxge_hal_towi_h towih;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (sq_id != 0));

	hldev = (__hal_device_t *)sq->channel.devh;

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", sq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", push = %d, msn = %d, "
	    "local_stag = %d, length = %d, ptr_to = 0x"VXGE_OS_LLDFMT,
	    (ptr_t)vpath_handle, (ptr_t)sq_id, (ptr_t)opaque_handle,
	    push, msn, local_stag, length, ptr_to);

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_lock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_lock_irq(&sq->channel.post_lock, flags);
#endif

	status = __hal_channel_dtr_alloc(&sq->channel, &towih);

	if (status != VXGE_HAL_OK) {
		vxge_hal_err_log_sq(hldev,
		    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
		    "sq channel dtr allocation failed %s : %d",
		    __func__, __LINE__);
		goto done;
	}

	towi = (__hal_towi_t *)towih;

	towi->control_0 =
	    VXGE_HAL_TOWI_TYPE(VXGE_HAL_TOWI_TYPE_OD) |
	    VXGE_HAL_TOWI_TCP_PUSH(((push)? VXGE_HAL_TOWI_PUSH_INCLUDED :
	    VXGE_HAL_TOWI_TCP_PUSH_NOT_INCLUDED)) |
	    VXGE_HAL_TOWI_MSG_TYPE(VXGE_HAL_TOWI_MSG_TYPE_TERMINATE) |
	    VXGE_HAL_TOWI_COMPLETION(VXGE_HAL_TOWI_COMPLETION_UPDATE_HW_CQ_SQ);

	towi->control_1 = VXGE_HAL_TOWI_PAYLOAD_BYTES(length);

	towi->ddp_info_0 =
	    VXGE_HAL_TOWI_DDP_RDMA_OPCODE(
	    VXGE_HAL_TOWI_DDP_RDMA_OPCODE_TERMINATE) |
	    VXGE_HAL_TOWI_DDP_LAST_SEGMENT_BIT(
	    VXGE_HAL_TOWI_DDP_LAST_SEGMENT_SET);

	towi->ddp_info_1 =
	    VXGE_HAL_TOWI_DDP_START_MSG_OFFSET(0) |
	    VXGE_HAL_TOWI_DDP_MSG_SEQ_NO(msn);

	towi->payload_locator_0 =
	    VXGE_HAL_TOWI_LOCAL_STAG(local_stag) |
	    VXGE_HAL_TOWI_LENGTH(length);

	towi->payload_locator_1 = ptr_to;

	towi->host_control_1 = opaque_handle;

	__hal_channel_dtr_post(&sq->channel, towih);

	INCREMENT_HIGH_TOWI_NUMBER_POSTED(sq);

	__hal_offload_db_post(sq,
	    VXGE_HAL_RNIC_OID_DB_SID_GET(sq->session_id),
	    VXGE_HAL_RNIC_OID_DB_SIN_GET(sq->session_id),
	    sq->high_towi_number_posted);

	status = VXGE_HAL_OK;

done:

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_unlock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_unlock_irq(&sq->channel.post_lock, flags);
#endif
	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);

	return (status);
}

/*
 * vxge_hal_bytestream_towi_post - Post Bytestream TOWI from buffer pointer
 * @vpath_handle: Handle of Virtual path to which this queue belongs.
 * @sq_id: Handle of sq
 * @opaque_handle: RNICPI private handle.
 * @push: Flag to include push in TOWI.
 * @ptr: Physical address of buffer
 * @length: Length of buffer
 *
 * This function reserves, sets and posts bytestream TOWI to SQ.
 *
 */
vxge_hal_status_e
vxge_hal_bytestream_towi_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 push,
	u64 ptr,
	u32 length)
{
	__hal_device_t *hldev;
	__hal_sq_t *sq = (__hal_sq_t *)(ptr_t)sq_id;
	__hal_towi_t *towi;
	vxge_hal_towi_h towih;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (sq_id != 0));

	hldev = (__hal_device_t *)sq->channel.devh;

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", sq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", push = %d, "
	    "ptr = 0x"VXGE_OS_LLXFMT", length = %d",
	    (ptr_t)vpath_handle, (ptr_t)sq_id, (ptr_t)opaque_handle,
	    push, ptr, length);

	vxge_assert(sq != NULL);

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_lock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_lock_irq(&sq->channel.post_lock, flags);
#endif

	status = __hal_channel_dtr_alloc(&sq->channel, &towih);

	if (status != VXGE_HAL_OK) {

		vxge_hal_err_log_sq(hldev,
		    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
		    "sq channel dtr allocation failed %s : %d",
		    __func__, __LINE__);
		goto done;
	}

	towi = (__hal_towi_t *)towih;

	towi->control_0 =
	    VXGE_HAL_TOWI_TYPE(VXGE_HAL_TOWI_TYPE_OD) |
	    VXGE_HAL_TOWI_TCP_PUSH(((push)? VXGE_HAL_TOWI_PUSH_INCLUDED :
	    VXGE_HAL_TOWI_TCP_PUSH_NOT_INCLUDED)) |
	    VXGE_HAL_TOWI_MSG_TYPE(VXGE_HAL_TOWI_MSG_TYPE_BYTE_STREAM) |
	    VXGE_HAL_TOWI_COMPLETION(VXGE_HAL_TOWI_COMPLETION_UPDATE_HW_CQ_SQ);

	towi->control_1 = VXGE_HAL_TOWI_PAYLOAD_BYTES(length);

	towi->ddp_info_0 = 0;

	towi->ddp_info_1 = 0;

	towi->payload_locator_0 = VXGE_HAL_TOWI_LENGTH(length);

	towi->payload_locator_1	= ptr;

	towi->host_control_1 = opaque_handle;

	__hal_channel_dtr_post(&sq->channel, towih);

	INCREMENT_HIGH_TOWI_NUMBER_POSTED(sq);

	__hal_offload_db_post(sq,
	    VXGE_HAL_RNIC_OID_DB_SID_GET(sq->session_id),
	    VXGE_HAL_RNIC_OID_DB_SIN_GET(sq->session_id),
	    sq->high_towi_number_posted);

	status = VXGE_HAL_OK;

done:

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_unlock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_unlock_irq(&sq->channel.post_lock, flags);
#endif
	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);

	return (status);
}

/*
 * vxge_hal_bytestream_towi_odlist_post - Post Bytestream TOWI from od list
 * @vpath_handle: Handle of Virtual path to which this queue belongs.
 * @sq_id: Handle of sq
 * @opaque_handle: RNICPI private handle.
 * @push: Flag to include push in TOWI.
 * @odlist: OD List prepared earlier
 *
 * This function reserves, sets and posts bytestream TOWI to SQ.
 *
 */
vxge_hal_status_e
vxge_hal_bytestream_towi_odlist_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 push,
	vxge_hal_odlist_t *odlist)
{
	__hal_device_t *hldev;
	__hal_sq_t *sq = (__hal_sq_t *)(ptr_t)sq_id;
	__hal_towi_t *towi;
	vxge_hal_towi_h towih;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (sq_id != 0));

	hldev = (__hal_device_t *)sq->channel.devh;

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", sq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_LLXFMT", push = %d, "
	    "odlist = 0x"VXGE_OS_STXFMT, (ptr_t)vpath_handle,
	    (ptr_t)sq_id, opaque_handle, push, (ptr_t)odlist);

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_lock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_lock_irq(&sq->channel.post_lock, flags);
#endif

	status = __hal_channel_dtr_alloc(&sq->channel, &towih);

	if (status != VXGE_HAL_OK) {

		vxge_hal_err_log_sq(hldev,
		    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
		    "sq channel dtr allocation failed %s : %d",
		    __func__, __LINE__);
		goto done;
	}

	towi = (__hal_towi_t *)towih;

	towi->control_0 =
	    VXGE_HAL_TOWI_TYPE(VXGE_HAL_TOWI_TYPE_OD_LB_WRAPPER) |
	    VXGE_HAL_TOWI_TCP_PUSH(((push)? VXGE_HAL_TOWI_PUSH_INCLUDED :
	    VXGE_HAL_TOWI_TCP_PUSH_NOT_INCLUDED)) |
	    VXGE_HAL_TOWI_MSG_TYPE(VXGE_HAL_TOWI_MSG_TYPE_BYTE_STREAM) |
	    VXGE_HAL_TOWI_COMPLETION(VXGE_HAL_TOWI_COMPLETION_UPDATE_HW_CQ_SQ);

	towi->control_1 = VXGE_HAL_TOWI_PAYLOAD_BYTES(odlist->total_bytes);

	towi->ddp_info_0 = 0;

	towi->ddp_info_1 = 0;

	towi->payload_locator_0 =
	    VXGE_HAL_TOWI_LENGTH(__hal_odlist_length_get(odlist));

	towi->payload_locator_1	= __hal_odlist_dma_addr_get(odlist);

	towi->host_control_1 = opaque_handle;

	__hal_channel_dtr_post(&sq->channel, towih);

	INCREMENT_HIGH_TOWI_NUMBER_POSTED(sq);

	__hal_offload_db_post(sq,
	    VXGE_HAL_RNIC_OID_DB_SID_GET(sq->session_id),
	    VXGE_HAL_RNIC_OID_DB_SIN_GET(sq->session_id),
	    sq->high_towi_number_posted);

	status = VXGE_HAL_OK;

done:

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_unlock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_unlock_irq(&sq->channel.post_lock, flags);
#endif
	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);

	return (status);
}

/*
 * vxge_hal_tcp_fin_towi_post - Post TCP Fin TOWI
 * @vpath_handle: Handle of Virtual path to which this queue belongs.
 * @sq_id: Handle of sq
 * @opaque_handle: RNICPI private handle.
 * @push: Flag to include push in TOWI.
 *
 * This function reserves, sets and posts TCP Fin TOWI to SQ.
 *
 */
vxge_hal_status_e
vxge_hal_tcp_fin_towi_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 push)
{
	__hal_device_t *hldev;
	__hal_sq_t *sq = (__hal_sq_t *)(ptr_t)sq_id;
	__hal_towi_t *towi;
	vxge_hal_towi_h towih;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (sq_id != 0));

	hldev = (__hal_device_t *)sq->channel.devh;

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", sq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", push = %d",
	    (ptr_t)vpath_handle, (ptr_t)sq_id, (ptr_t)opaque_handle, push);

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_lock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_lock_irq(&sq->channel.post_lock, flags);
#endif

	status = __hal_channel_dtr_alloc(&sq->channel, &towih);

	if (status != VXGE_HAL_OK) {
		vxge_hal_err_log_sq(hldev,
		    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
		    "sq channel dtr allocation failed %s : %d",
		    __func__, __LINE__);
		goto done;
	}

	towi = (__hal_towi_t *)towih;

	towi->control_0 =
	    VXGE_HAL_TOWI_TCP_PUSH(((push)? VXGE_HAL_TOWI_PUSH_INCLUDED :
	    VXGE_HAL_TOWI_TCP_PUSH_NOT_INCLUDED)) |
	    VXGE_HAL_TOWI_MSG_TYPE(VXGE_HAL_TOWI_MSG_TYPE_FIN) |
	    VXGE_HAL_TOWI_COMPLETION(VXGE_HAL_TOWI_COMPLETION_UPDATE_HW_CQ_SQ);

	towi->control_1 = 0;
	towi->ddp_info_0 = 0;
	towi->ddp_info_1 = 0;
	towi->payload_locator_0 = 0;
	towi->payload_locator_1 = 0;

	towi->host_control_1 = opaque_handle;

	__hal_channel_dtr_post(&sq->channel, towih);

	INCREMENT_HIGH_TOWI_NUMBER_POSTED(sq);

	__hal_offload_db_post(sq,
	    VXGE_HAL_RNIC_OID_DB_SID_GET(sq->session_id),
	    VXGE_HAL_RNIC_OID_DB_SIN_GET(sq->session_id),
	    sq->high_towi_number_posted);

	status = VXGE_HAL_OK;

done:

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_unlock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_unlock_irq(&sq->channel.post_lock, flags);
#endif
	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);

	return (status);
}

/*
 * vxge_hal_sq_completed_towi_check - Check the HW_CQ_SQ_IR for
 *		 completed TOWIs
 * @vpath_handle: Handle of Virtual path to which this queue belongs.
 * @sq_id: Handle of sq
 *
 * This function return (VXGE_HAL_OK) if a new TOWI completed.
 */
__HAL_STATIC_SQ __HAL_INLINE_SQ vxge_hal_status_e
vxge_hal_sq_completed_towi_check(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id)
{
	__hal_device_t *hldev;
	__hal_sq_t *sq = (__hal_sq_t *)(ptr_t)sq_id;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (sq_id != 0));

	hldev = (__hal_device_t *)sq->channel.devh;

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", sq_id = 0x"VXGE_OS_STXFMT,
	    (ptr_t)vpath_handle, (ptr_t)sq_id);

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_lock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_lock_irq(&sq->channel.post_lock, flags);
#endif

	sq->high_towi_number_completed = (u32)
	    VXGE_HAL_HW_CQ_SQ_IR_GET_TOWI_COMPLETED(sq->cq_sq_ir->control_0);

	vxge_hal_info_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "high_towi_number_completed = %d",
	    sq->high_towi_number_completed);

	if ((sq->high_towi_number_completed % sq->sq_length) !=
	    sq->channel.compl_index) {

		status = VXGE_HAL_OK;

	} else {

		status = VXGE_HAL_INF_NO_MORE_COMPLETED_DESCRIPTORS;

	}

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_unlock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_unlock_irq(&sq->channel.post_lock, flags);
#endif

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);

	return (status);
}

/*
 * vxge_hal_sq_next_completed_towi_get - Return the next completed TOWI
 * @vpath_handle: Handle of Virtual path to which this queue belongs.
 * @sq_id: Handle of sq
 * @opaque_handle: Buffer to return the opaque handle passed in towi post
 *
 * This function returns the opaque handle from the next completed TOWI
 * and frees it.
 */
__HAL_STATIC_SQ __HAL_INLINE_SQ vxge_hal_status_e
vxge_hal_sq_next_completed_towi_get(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 *opaque_handle)
{
	__hal_sq_t *sq = (__hal_sq_t *)(ptr_t)sq_id;
	__hal_towi_t *towi;
	__hal_towi_priv_t *towi_priv;
	vxge_hal_towi_h towih;
	__hal_device_t *hldev;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (sq_id != 0) &&
	    (opaque_handle != 0));

	hldev = (__hal_device_t *)sq->channel.devh;

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", sq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT,
	    (ptr_t)vpath_handle, (ptr_t)sq_id, (ptr_t)opaque_handle);

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_lock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_lock_irq(&sq->channel.post_lock, flags);
#endif

	if ((sq->high_towi_number_completed % sq->sq_length) ==
	    sq->channel.compl_index) {

		status = VXGE_HAL_INF_NO_MORE_COMPLETED_DESCRIPTORS;

	} else {
		__hal_channel_dtr_try_complete(&sq->channel, &towih);

		__hal_channel_dtr_complete(&sq->channel);

		towi = (__hal_towi_t *)towih;

		*opaque_handle = towi->host_control_1;

		towi_priv = (__hal_towi_priv_t *)(ptr_t)towi->host_control_0;

		if (VXGE_HAL_TOWI_GET_MSG_TYPE(towi->control_0) ==
		    VXGE_HAL_TOWI_MSG_TYPE_FAST_REG_NSMR) {

			__hal_regpool_free(&hldev->reg_pool,
			    towi_priv->nsmr_ctrl_block_reg_entry);
		}

		__hal_channel_dtr_free(&sq->channel, towih);

		status = VXGE_HAL_OK;
	}

#if defined(VXGE_HAL_SQ_POST)
	vxge_os_spin_unlock(&sq->channel.post_lock);
#elif defined(VXGE_HAL_SQ_POST_IRQ)
	vxge_os_spin_unlock_irq(&sq->channel.post_lock, flags);
#endif

	vxge_hal_trace_log_sq(hldev,
	    ((__hal_vpath_handle_t *)sq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);

	return (status);
}
