/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */
/*
 * Copyright (c) 2002-2010 Neterion, Inc.
 * All right Reserved.
 *
 * FileName :   vxgehal-sq.h
 *
 * Description:  HAL HW Full offload towi list structures
 *
 * Created:       15 Dec 2005
 */


#ifndef	VXGE_HAL_SQ_H
#define	VXGE_HAL_SQ_H

#define	__HAL_STATIC_SQ
#define	__HAL_INLINE_SQ

__EXTERN_BEGIN_DECLS

/*
 * struct __hal_sq_t - Towi list - SQ channel
 * @channel: Channel "base" of this Queue, the common part of all HAL
 *	  channels.
 * @callback: Callback routine to be called when TOWIs are completed
 * @opaque_handle: Opaque handle passed by the client
 * @mempool: Memory pool from which towi list blocks get allocated.
 * @doorbell: Address of the register to which TOWI doorbell is posted
 * @cq_sq_ir: The adapter will overwrite and update this location as TOWIs
 *		are Completed.
 * @cq_sq_ir_reg_entry: Reg entry of cq_sq_ir
 * @cq_rr_ir: The adapter will overwrite and update this location as RDMA READs
 *		are Completed.
 * @cq_rr_ir_reg_entry: Reg Entry of cq_rr_ir
 * @first_block_address: The DMA Address of the first block.
 * @first_block_length: The Length of the first block.
 * @high_towi_number_posted: The high towi number posted to adapter.
 *		   This is used to specify the high towi number.
 *		   (always modulo 2^24).
 * @high_towi_number_completed: The high towi number completed by the
 *		   adapter. This is used to get the high towi
 *		   number.(always modulo 2^24).
 * @session_id: The object id of the session.
 * @state: State of the session using this SQ
 *		- VXGEHAL_SQ_STATE_UNKNOWN-Unknown
 *		- VXGEHAL_SQ_STATE_INITIALIZED-Initialized
 *		- VXGEHAL_SQ_STATE_SESS_OPEN_PENDING-Open req sent
 *		- VXGEHAL_SQ_STATE_SESS_ESTED-Open resp received
 *		- VXGEHAL_SQ_STATE_SESS_CLOSE_PENDING-Close req sent
 *		- VXGEHAL_SQ_STATE_SESS_CLOSED-Close response received
 * @msn: Sequence number of messages send on this queue.
 * @sq_length: The length of the SQ/TOWI List.
 * @towi_completion: The indication type when a TOWI is completed.
 * @rdma_read_completion: The indication type when RDMA READ complete.
 * @towi_size: size of TOWI.
 * @towi_priv_size: Per TOWI size reserved (by HAL) for ULD to keep per-towi
 *		data (e.g., DMA handle for Solaris).
 * @towis_per_block: Number of towis per hardware-defined RxD
 *		 block.
 * @towiblock_priv_size: Reserved at the end of each list block. HAL internal
 *			 usage. Not to confuse with @towi_priv_size.
 * @stats: Statistics of this sq
 *
 * TOWI List or SQ channel. This channel keeps track of the TOWI list to which
 * TOWIs are posted by host for the adapter. This channel provides methods to
 * free the completed TOWIs and recycle them.
 *
 */
typedef struct __hal_sq_t {
	__hal_channel_t				channel;
	vxge_hal_towi_callback_f			callback;
	vxge_hal_opaque_handle_t			*opaque_handle;
	vxge_hal_mempool_t			*mempool;
	u64					doorbell;
	__hal_hw_cq_sq_ir_t			*cq_sq_ir;
	__hal_reg_entry_t			*cq_sq_ir_reg_entry;
	__hal_hw_cq_rr_ir_t			*cq_rr_ir;
	__hal_reg_entry_t			*cq_rr_ir_reg_entry;
	u64					first_block_address;
	u32					first_block_length;
	u32					high_towi_number_posted;
	u32					high_towi_number_completed;
	u32					session_id;
	u32					state;
#define	VXGEHAL_SQ_STATE_UNKNOWN			0
#define	VXGEHAL_SQ_STATE_INITIALIZED		1
#define	VXGEHAL_SQ_STATE_SESS_OPEN_PENDING	2
#define	VXGEHAL_SQ_STATE_SESS_ESTED	3
#define	VXGEHAL_SQ_STATE_SESS_CLOSE_PENDING	4
#define	VXGEHAL_SQ_STATE_SESS_CLOSED		5
	u32					msn;
	u32					sq_length;
	u32					towi_completion;
	u32					rdma_read_completion;
	u32					towi_size;
	u32					towi_priv_size;
	u32					towis_per_block;
	u32					towiblock_priv_size;
	vxge_hal_vpath_stats_sw_sq_info_t	stats;
} __hal_sq_t;

#define	VXGE_HAL_SQ_STATE_SET(sq, s)	((sq)->state = s)

#define	VXGE_HAL_SQ_STATE_GET(sq)	(sq)->state

#define	INCREMENT_HIGH_TOWI_NUMBER_POSTED(sq)				  \
	(sq)->high_towi_number_posted = ((sq)->high_towi_number_posted + 1) ^  \
			0x1000000;

/*
 * Private API
 */
vxge_hal_status_e
__hal_sq_create(
	vxge_hal_vpath_h vpath_handle,
	__hal_sq_t *sq);

vxge_hal_status_e
__hal_sq_delete(
	vxge_hal_vpath_h vpath_handle,
	__hal_sq_t *sq);


__HAL_STATIC_SQ __HAL_INLINE_SQ vxge_hal_status_e
__hal_towi_nsmr_ctrl_block_set(
	__hal_sq_t *sq,
	__hal_towi_t *towi,
	u32 stag_index,
	u32 stag_key,
	u32 pb_size,
	u64 pble_list,
	u32 local_write_flag,
	u32 remote_write_flag,
	u32 local_read_flag,
	u32 remote_read_flag,
	u32 enable_mw_binding_flag,
	u32 first_byte_offset,
	u64 mr_length,
	u64 base_to);

/*
 * Public API
 */
vxge_hal_status_e
vxge_hal_sq_create(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_sq_attr_t *attr,
	vxge_hal_opaque_handle_t *opaque_handle);

vxge_hal_status_e
vxge_hal_sq_delete(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	vxge_hal_opaque_handle_t *opaque_handle);

vxge_hal_status_e
vxge_hal_send_towi_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 push,
	u32 msn,
	u32 last,
	u32 start_msg_offset,
	u32 read_fencing,
	u32 remote_invalidate,
	u32 solicited_event,
	u32 local_stag,
	u32 ptr_to,
	u32 length,
	u32 remote_stag);

vxge_hal_status_e
vxge_hal_send_towi_odlist_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 push,
	u32 msn,
	u32 last,
	u32 start_msg_offset,
	u32 read_fencing,
	u32 remote_invalidate,
	u32 solicited_event,
	u32 remote_stag,
	vxge_hal_odlist_t *odlist);

vxge_hal_status_e
vxge_hal_write_towi_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 push,
	u32 last,
	u32 read_fencing,
	u32 local_stag,
	u64 ptr_to,
	u32 length,
	u32 remote_stag,
	u64 remote_tagged_offset);

vxge_hal_status_e
vxge_hal_write_towi_odlist_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 push,
	u32 last,
	u32 read_fencing,
	u32 remote_stag,
	u64 remote_tagged_offset,
	vxge_hal_odlist_t *odlist);

vxge_hal_status_e
vxge_hal_read_towi_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 push,
	u32 msn,
	u32 invalidate_local_stag,
	u32 local_fencing,
	u32 local_stag,
	u32 length,
	u64 ptr_to);

vxge_hal_status_e
vxge_hal_bind_mw_towi_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 mw_stag_index,
	u32 mw_stag_key,
	u32 mr_stag,
	u32 mw_ctrl,
	u64 mw_length,
	u64 tagged_offset);

vxge_hal_status_e
vxge_hal_fmr_towi_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 stag_index,
	u32 stag_key,
	u32 pb_size,
	u64 pble_list,
	u32 local_write_flag,
	u32 remote_write_flag,
	u32 local_read_flag,
	u32 remote_read_flag,
	u32 enable_mw_binding_flag,
	u32 first_byte_offset,
	u64 mr_length,
	u64 base_to);

vxge_hal_status_e
vxge_hal_local_invalidate_towi_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 stag,
	u32 local_fencing);

vxge_hal_status_e
vxge_hal_terminate_towi_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 push,
	u32 msn,
	u32 local_stag,
	u32 length,
	u64 ptr_to);

vxge_hal_status_e
vxge_hal_bytestream_towi_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 push,
	u64 ptr,
	u32 length);

vxge_hal_status_e
vxge_hal_bytestream_towi_odlist_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 push,
	vxge_hal_odlist_t *odlist);

vxge_hal_status_e
vxge_hal_tcp_fin_towi_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 opaque_handle,
	u32 push);

vxge_hal_status_e
vxge_hal_sq_completed_towi_check(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id);

vxge_hal_status_e
vxge_hal_sq_next_completed_towi_get(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t sq_id,
	u64 *opaque_handle);

__EXTERN_END_DECLS

#endif /* VXGE_HAL_SQ_H */
