/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */
/*
 * Copyright (c) 2002-2010 Neterion, Inc.
 * All right Reserved.
 *
 * FileName :   vxgehal-srq.c
 *
 * Description:  HAL offload (Shared) receive queue functionality
 *
 * Created:       14 December 2005
 */
#include "vxgehal.h"

/*
 * __hal_wqe_mempool_item_alloc - Allocate List blocks for WQE list
 * callback
 * @mempoolh: Handle to memory pool
 * @memblock: Address of this memory block
 * @memblock_index: Index of this memory block
 * @dma_object: dma object for this block
 * @item: Pointer to this item
 * @index: Index of this item in memory block
 * @is_last: If this is last item in the block
 * @userdata: Specific data of user
 *
 * This function is callback passed to __hal_mempool_create to create memory
 * pool for WQE list
 */
static vxge_hal_status_e
__hal_wqe_mempool_item_alloc(
	vxge_hal_mempool_h mempoolh,
	void *memblock,
	u32 memblock_index,
	vxge_hal_mempool_dma_t *dma_object,
	void *item,
	u32 index,
	u32 is_last,
	void *userdata)
{
	u32 memblock_item_idx;
	__hal_srq_t *srqdata = (__hal_srq_t *)userdata;
	__hal_list_block_t *prev_block, *next_block;
	__hal_list_block_wrapper_t *prev_wrapper, *wrapper;
	__hal_srq_blk_priv_t *srq_blk_priv;
	u32 wqes_block_size =
	    srqdata->lro_wqe_size * srqdata->lro_wqes_per_block;
	__hal_device_t *hldev = (__hal_device_t *)srqdata->channel.devh;

	vxge_assert(srqdata != NULL);

	vxge_hal_trace_log_pool(hldev,
	    ((__hal_vpath_handle_t *)srqdata->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_pool(hldev,
	    ((__hal_vpath_handle_t *)srqdata->channel.vph)->vpath->vp_id,
	    "mempoolh = 0x"VXGE_OS_STXFMT", memblock = 0x"VXGE_OS_STXFMT", "
	    "memblock_index = %d, dma_object = 0x"VXGE_OS_STXFMT", "
	    "item = 0x"VXGE_OS_STXFMT", index = %d, is_last = %d, "
	    "userdata = 0x"VXGE_OS_STXFMT, (ptr_t)mempoolh, (ptr_t)memblock,
	    memblock_index, (ptr_t)dma_object, (ptr_t)item, index, is_last,
	    (ptr_t)userdata);

	srq_blk_priv = (__hal_srq_blk_priv_t *)__hal_mempool_item_priv(
	    mempoolh, memblock_index, item, &memblock_item_idx);

	vxge_assert(srq_blk_priv != NULL);

	if (srqdata->srq_type == VXGE_HAL_SRQ_TYPE_BYTE_STREAM_IWARP) {

		wqes_block_size = VXGE_HAL_LIST_BLOCK_ITEMS_SIZE;

		srqdata->channel.dtr_arr[index].dtr = item;

		srq_blk_priv->dma_offset = 0;
		srq_blk_priv->dma_addr = dma_object->addr;
		srq_blk_priv->dma_handle = dma_object->handle;
		srq_blk_priv->memblock = memblock;
		srq_blk_priv->wrapper = (__hal_list_block_wrapper_t *)
		    (((char *)memblock) + wqes_block_size);
		srq_blk_priv->srqh = (vxge_hal_srq_h)userdata;
		srq_blk_priv->compl_offset = wqes_block_size;
		srq_blk_priv->post_offset = 0;

		__hal_list_block_memblock_priv_data_set(memblock,
		    (u64)(ptr_t)srq_blk_priv);

	} else {
		u32 i;
		wqes_block_size =
		    srqdata->lro_wqe_size * srqdata->lro_wqes_per_block;

		for (i = 0; i < srqdata->lro_wqes_per_block; i++) {

			__hal_lro_wqe_priv_t *lro_wqe_priv;
			__hal_lro_wqe_t *lro_wqe;

			int reserve_index =
			    srqdata->channel.reserve_ptr - 1 -
			    (index * srqdata->lro_wqes_per_block + i);

			lro_wqe = (__hal_lro_wqe_t *)
			    ((char *)item + i * srqdata->lro_wqe_size);

			srqdata->channel.reserve_arr[reserve_index] = lro_wqe;

			lro_wqe_priv = (__hal_lro_wqe_priv_t *)(void *)
			    ((char *)srq_blk_priv +
			    srqdata->lro_wqe_priv_size * i);

			lro_wqe->wqe_host_control = (u64)(ptr_t)lro_wqe_priv;

			lro_wqe_priv->dma_offset =
			    (char *)lro_wqe - (char *)memblock;
			lro_wqe_priv->dma_addr =
			    dma_object->addr +  lro_wqe_priv->dma_offset;
			lro_wqe_priv->dma_handle = dma_object->handle;
			lro_wqe_priv->memblock = memblock;
			lro_wqe_priv->wrapper = (__hal_list_block_wrapper_t *)
			    (((char *)memblock) + wqes_block_size);
			lro_wqe_priv->srqh = (vxge_hal_srq_h)userdata;
			lro_wqe_priv->opaque_handle = 0;
			lro_wqe_priv->owner =
			    VXGE_HAL_LRO_WQE_PRIV_OWNER_ADAPTER;
			lro_wqe_priv->first = (i == 0);
		}

	}

	__hal_list_block_memblock_idx_set(memblock, memblock_index);

	wrapper = (__hal_list_block_wrapper_t *)
	    (((char *)memblock) + wqes_block_size);

	if (index > 0) {
		prev_block = __hal_mempool_item(mempoolh, index-1);
		next_block = __hal_mempool_item(mempoolh, 0);
		prev_wrapper = (__hal_list_block_wrapper_t *)
		    (((char *)prev_block) + wqes_block_size);
		__hal_wrapper_link(prev_wrapper, wrapper,
		    dma_object->addr, wqes_block_size);

	} else {
		prev_block = memblock;
		next_block = memblock;
		__hal_wrapper_link(wrapper, wrapper,
		    dma_object->addr, wqes_block_size);
	}

	if (srqdata->first_block == NULL) {
		srqdata->first_block = memblock;
		srqdata->last_block = memblock;
	}

	__hal_list_block_link(memblock, prev_block, next_block);

	vxge_hal_trace_log_pool(hldev,
	    ((__hal_vpath_handle_t *)srqdata->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * __hal_wqe_mempool_item_free - Free List blocks for wqe list callback
 * @mempoolh: Handle to memory pool
 * @memblock: Address of this memory block
 * @memblock_index: Index of this memory block
 * @dma_object: dma object for this block
 * @item: Pointer to this item
 * @index: Index of this item in memory block
 * @is_last: If this is last item in the block
 * @userdata: Specific data of user
 *
 * This function is callback passed to __hal_mempool_free to destroy memory
 * pool for wqe list
 */
static vxge_hal_status_e
__hal_wqe_mempool_item_free(
	vxge_hal_mempool_h mempoolh,
	void *memblock,
	u32 memblock_index,
	vxge_hal_mempool_dma_t *dma_object,
	void *item,
	u32 index,
	u32 is_last,
	void *userdata)
{
	u32 memblock_item_idx;
	__hal_srq_t *srqdata = (__hal_srq_t *)userdata;
	__hal_list_block_t *prev_block;
	__hal_list_block_wrapper_t *prev_wrapper, *wrapper;
	__hal_srq_blk_priv_t *srq_blk_priv;
	u32 wqes_block_size;
	__hal_device_t *hldev = (__hal_device_t *)srqdata->channel.devh;

	vxge_assert(srqdata != NULL);

	vxge_hal_trace_log_pool(hldev,
	    ((__hal_vpath_handle_t *)srqdata->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_pool(hldev,
	    ((__hal_vpath_handle_t *)srqdata->channel.vph)->vpath->vp_id,
	    "mempoolh = 0x"VXGE_OS_STXFMT", memblock = 0x"VXGE_OS_STXFMT", "
	    "memblock_index = %d, dma_object = 0x"VXGE_OS_STXFMT", "
	    "item = 0x"VXGE_OS_STXFMT", index = %d, is_last = %d, "
	    "userdata = 0x"VXGE_OS_STXFMT, (ptr_t)mempoolh, (ptr_t)memblock,
	    memblock_index, (ptr_t)dma_object, (ptr_t)item, index, is_last,
	    (ptr_t)userdata);

	srq_blk_priv = (__hal_srq_blk_priv_t *)__hal_mempool_item_priv(
	    mempoolh, memblock_index, item, &memblock_item_idx);

	vxge_assert(srq_blk_priv != NULL);

	prev_block = __hal_list_block_prev_pointer(memblock);

	vxge_assert(prev_block != NULL);

	if (srqdata->srq_type == VXGE_HAL_SRQ_TYPE_BYTE_STREAM_IWARP) {
		wqes_block_size = VXGE_HAL_LIST_BLOCK_ITEMS_SIZE;
	} else {
		wqes_block_size =
		    srqdata->lro_wqe_size * srqdata->lro_wqes_per_block;
	}

	wrapper = (__hal_list_block_wrapper_t *)
	    (((char *)memblock) + wqes_block_size);

	prev_wrapper = (__hal_list_block_wrapper_t *)
	    (((char *)prev_block) + wqes_block_size);

	__hal_wrapper_unlink(prev_wrapper, wrapper);

	__hal_list_block_unlink(memblock);

	vxge_hal_trace_log_pool(hldev,
	    ((__hal_vpath_handle_t *)srqdata->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * __hal_srq_create - Create Receive Queue
 * @vpath_handle: Virtual path handle to which this queue belongs
 * @srq_: SRQ
 *
 * This function creates receive queue with the attributes passed
 */
vxge_hal_status_e
__hal_srq_create(
	vxge_hal_vpath_h vpath_handle,
	__hal_srq_t *srq)
{
	vxge_hal_status_e status;
	__hal_list_block_t *compl_block;
	__hal_srq_blk_priv_t *compl_priv;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	__hal_device_t *hldev;
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	vxge_hal_mempool_dma_t *dma_object;
	u32 event_notification_mask = 0;
	u32 nblocks;

	/*
	 * Note: at this point we have channel.devh and channel.pdev
	 *	   pre-set only!
	 */

	vxge_assert((vpath_handle != NULL) && (srq != 0));

	hldev = vp->vpath->hldev;

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", srq = 0x"VXGE_OS_STXFMT,
	    (ptr_t)vpath_handle, (ptr_t)srq);

	if ((vpath_handle == NULL) || (srq == NULL)) {
		vxge_hal_info_log_srq(hldev, vp->vpath->vp_id,
		    "null pointer passed attr = null == > %s : %d",
		    __func__, __LINE__);
		vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
		    __LINE__, VXGE_HAL_ERR_INVALID_HANDLE);
		return (VXGE_HAL_ERR_INVALID_HANDLE);
	}

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	if (srq->srq_type == VXGE_HAL_SRQ_TYPE_BYTE_STREAM_IWARP) {
		nblocks = srq->srq_length;
	} else {
		nblocks = srq->srq_length / srq->lro_wqes_per_block;
	}

	srq->mempool = __hal_mempool_create(
	    (vxge_hal_device_h)vp->vpath->hldev,
	    VXGE_OS_HOST_PAGE_SIZE,
	    VXGE_OS_HOST_PAGE_SIZE,
	    srq->wqeblock_priv_size,
	    nblocks,
	    nblocks,
	    __hal_wqe_mempool_item_alloc,
	    __hal_wqe_mempool_item_free,
	    srq);
	if (srq->mempool == NULL) {
		vxge_hal_info_log_srq(hldev, vp->vpath->vp_id,
		    "Memory allocation for srq failed %s : %d",
		    __func__, __LINE__);
		vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
		    __LINE__, VXGE_HAL_ERR_OUT_OF_MEMORY);
		return (VXGE_HAL_ERR_OUT_OF_MEMORY);
	}

	status = __hal_channel_initialize(&srq->channel);

	if (status != VXGE_HAL_OK) {
		vxge_hal_info_log_srq(hldev, vp->vpath->vp_id,
		    "srq channel initialization failed %s : %d",
		    __func__, __LINE__);
		goto srq_create_error;
	}

	if (srq->srq_type == VXGE_HAL_SRQ_TYPE_BYTE_STREAM_IWARP) {

		srq->channel.compl_index = srq->channel.length - 1;

		compl_block =
		    srq->channel.dtr_arr[srq->channel.compl_index].dtr;
		compl_priv = (__hal_srq_blk_priv_t *)(ptr_t)
		    __hal_list_block_priv_data(compl_block);

		vxge_assert(compl_priv != NULL);

		srq->high_wqe_address =
		    compl_priv->dma_addr + compl_priv->compl_offset;

		srq->bytes_used = 0;

		srq->bytes_free =
		    srq->srq_length * VXGE_HAL_LIST_BLOCK_ITEMS_SIZE;

	}

	srq->rq_ir->control_0 = 0;

	dma_object = __hal_mempool_memblock_dma(srq->mempool, 0);

	vxge_assert(dma_object != NULL);

	srq->first_block_address = dma_object->addr;

	srq->first_block_length = VXGE_OS_HOST_PAGE_SIZE;

	srq->first_block_status = VXGE_HAL_SRQ_FWQEBW_STATUS_UNKNOWN;

	srq->wsn = 0;

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(srq->opaque_handle,
	    vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(srq->opaque_handle,
	    (ptr_t)srq);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(srq->srq_opaque_handle,
	    vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(srq->srq_opaque_handle,
	    (ptr_t)srq);

	__hal_srq_create_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    srq->srq_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    srq->opaque_handle,
	    srq->srq_opaque_handle,
	    vp->vpath->hldev->header.config.max_num_wqe_od_groups,
	    vp->vpath->hldev->header.config.refill_threshold_high,
	    vp->vpath->hldev->header.config.refill_threshold_low,
	    srq->low_threshold_notify,
	    ((srq->srq_type == VXGE_HAL_SRQ_TYPE_LRO) ?
	    VXGE_HAL_SRQ_CREATE_REQ_LRO_BS_LRO :
	    VXGE_HAL_SRQ_CREATE_REQ_LRO_BS_BS),
	    vp->vpath->hldev->header.config.poll_or_doorbell,
	    ((srq->srq_limit_value > 0) ? 0 : 1),
	    event_notification_mask,
	    srq->srq_limit_value,
	    vp->vpath->bmap_intr_num,
	    VXGE_HAL_RNIC_OID_DB_SID_GET(srq->srq_id),
	    srq->rq_ir_reg_entry->dma_addr);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK) {
		vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);
		return (status);
	}

	vxge_hal_info_log_srq(hldev, vp->vpath->vp_id,
	    "srq channel dmq post failed == > %s : %d", __func__, __LINE__);

srq_create_error:

	__hal_srq_delete(vpath_handle, srq);

	vp->cb_fn(vp->client_handle,
	    NULL,
	    VXGE_HAL_MSG_TYPE_SRQ_CREATE_RESP,
	    0,
	    status,
	    srq->opaque_handle);

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * __hal_srq_delete - Removes the Receive Queue
 * @vpath_handle: Virtual path handle to which this queue belongs
 * @srq: SRQ to be deleted
 *
 * This function freeup the memory pool and removes the
 * hw receive queue.
 */

vxge_hal_status_e
__hal_srq_delete(
	vxge_hal_vpath_h vpath_handle,
	__hal_srq_t *srq)
{
	__hal_device_t *hldev;
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;

	vxge_assert((vpath_handle != NULL) && (srq != NULL));

	hldev = vp->vpath->hldev;

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", srq = 0x"VXGE_OS_STXFMT,
	    (ptr_t)vpath_handle, (ptr_t)srq);

	if (srq->mempool) {
		__hal_mempool_destroy(srq->mempool);
	}

	if (srq->rq_ir_reg_entry) {
		__hal_regpool_free(&vp->vpath->hldev->reg_pool,
		    srq->rq_ir_reg_entry);
	}

	__hal_vp_oid_free(vp->vpath,
	    &vp->vpath->srq_oid_db,
	    srq->srq_id);

	__hal_channel_terminate(&srq->channel);

#if defined(VXGE_HAL_SRQ_POST)
	vxge_os_spin_lock_destroy(&srq->channel.post_lock,
	    vp->vpath->hldev->header.pdev);
#elif defined(VXGE_HAL_SRQ_POST_IRQ)
	vxge_os_spin_lock_destroy_irq(&srq->channel.post_lock,
	    vp->vpath->hldev->header.pdev);
#endif

#if defined(VXGE_HAL_SRQ_FREE)
	vxge_os_spin_lock_destroy(&srq->channel.free_lock,
	    vp->vpath->hldev->header.pdev);
#elif defined(VXGE_HAL_SRQ_FREE_IRQ)
	vxge_os_spin_lock_destroy_irq(&srq->channel.free_lock,
	    vp->vpath->hldev->header.pdev);
#endif

#if defined(VXGE_HAL_SRQ_RESERVE)
	vxge_os_spin_lock_destroy(&srq->channel.reserve_lock,
	    vp->vpath->hldev->header.pdev);
#elif defined(VXGE_HAL_SRQ_RESERVE_IRQ)
	vxge_os_spin_lock_destroy_irq(&srq->channel.reserve_lock,
	    vp->vpath->hldev->header.pdev);
#endif

	__hal_channel_free(&srq->channel);

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * __hal_hw_wqes_complete - Process completed HW WQEs in the srq
 * @srq: srq
 *
 * This function reads IR register and adjusts the compl_index and
 * compl_offset so that the space can be reused.
 */

vxge_hal_status_e
__hal_hw_wqes_complete(
	__hal_srq_t *srq)
{
	__hal_device_t *hldev;
	__hal_list_block_t *compl_block;
	__hal_srq_blk_priv_t *compl_priv;
	u64 prev_offset;

	vxge_assert(srq != NULL);

	hldev = (__hal_device_t *)srq->channel.devh;

	vxge_hal_trace_log_srq(hldev,
	    ((__hal_vpath_handle_t *)srq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_srq(hldev,
	    ((__hal_vpath_handle_t *)srq->channel.vph)->vpath->vp_id,
	    "srq = 0x"VXGE_OS_STXFMT, (ptr_t)srq);

	compl_block = srq->channel.dtr_arr[srq->channel.compl_index].dtr;
	compl_priv = (__hal_srq_blk_priv_t *)(ptr_t)
	    __hal_list_block_priv_data(compl_block);

	vxge_assert(compl_priv != NULL);

	srq->high_wqe_address = srq->rq_ir->control_0;

	vxge_hal_info_log_srq(hldev,
	    ((__hal_vpath_handle_t *)srq->channel.vph)->vpath->vp_id,
	    "high_wqe_address = "VXGE_OS_LLXFMT, srq->high_wqe_address);

	if (srq->high_wqe_address ==
	    compl_priv->dma_addr + compl_priv->compl_offset) {
		vxge_hal_info_log_srq(hldev,
		    ((__hal_vpath_handle_t *)srq->channel.vph)->vpath->vp_id,
		    "No new completed WQEs %s : %d", __func__, __LINE__);
		vxge_hal_trace_log_srq(hldev,
		    ((__hal_vpath_handle_t *)srq->channel.vph)->vpath->vp_id,
		    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);
		return (VXGE_HAL_OK);
	}

	vxge_hal_info_log_srq(hldev,
	    ((__hal_vpath_handle_t *)srq->channel.vph)->vpath->vp_id,
	    "compl_index = %d, block_addr = 0x"VXGE_OS_STXFMT
	    ", compl_offset = 0x"VXGE_OS_LLXFMT, srq->channel.compl_index,
	    compl_priv->dma_addr, compl_priv->compl_offset);

	for (;;) {

		if ((srq->high_wqe_address >= compl_priv->dma_addr) &&
		    (srq->high_wqe_address <= (compl_priv->dma_addr +
		    VXGE_HAL_LIST_BLOCK_ITEMS_SIZE))) {

			prev_offset = compl_priv->compl_offset;
			compl_priv->compl_offset = srq->high_wqe_address -
			    compl_priv->dma_addr;

			if (compl_priv->compl_offset ==
			    VXGE_HAL_LIST_BLOCK_ITEMS_SIZE) {

				compl_priv->post_offset = 0;

				srq->channel.compl_index++;
				if (srq->channel.compl_index ==
				    srq->channel.length)
					srq->channel.compl_index = 0;
			}

			srq->bytes_used -=
			    (compl_priv->compl_offset - prev_offset);
			srq->bytes_free +=
			    (compl_priv->compl_offset - prev_offset);

			break;

		} else {

			srq->bytes_used -= (VXGE_HAL_LIST_BLOCK_ITEMS_SIZE -
			    compl_priv->compl_offset);
			srq->bytes_free +=
			    (VXGE_HAL_LIST_BLOCK_ITEMS_SIZE -
			    compl_priv->compl_offset);

			compl_priv->compl_offset =
			    VXGE_HAL_LIST_BLOCK_ITEMS_SIZE;
			compl_priv->post_offset = 0;

			srq->channel.compl_index++;
			if (srq->channel.compl_index == srq->channel.length)
				srq->channel.compl_index = 0;

			compl_block =
			    srq->channel.dtr_arr[srq->channel.compl_index].dtr;
			compl_priv = (__hal_srq_blk_priv_t *)(ptr_t)
			    __hal_list_block_priv_data(compl_block);

			vxge_hal_info_log_srq(hldev, ((__hal_vpath_handle_t *)
			    srq->channel.vph)->vpath->vp_id,
			    "compl_index = %d, block_addr = 0x"VXGE_OS_STXFMT","
			    " compl_offset = 0x"VXGE_OS_LLXFMT,
			    srq->channel.compl_index,
			    compl_priv->dma_addr, compl_priv->compl_offset);

		}

	}

	vxge_hal_info_log_srq(hldev,
	    ((__hal_vpath_handle_t *)srq->channel.vph)->vpath->vp_id,
	    "compl_index = %d, block_addr = 0x"VXGE_OS_STXFMT
	    ", compl_offset = 0x"VXGE_OS_LLXFMT, srq->channel.compl_index,
	    compl_priv->dma_addr, compl_priv->compl_offset);

	vxge_hal_info_log_srq(hldev,
	    ((__hal_vpath_handle_t *)srq->channel.vph)->vpath->vp_id,
	    "completed WQEs processed. free 0x"VXGE_OS_LLXFMT
	    "used 0x"VXGE_OS_LLXFMT,
	    srq->bytes_free, srq->bytes_used);

	vxge_hal_trace_log_srq(hldev,
	    ((__hal_vpath_handle_t *)srq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * __hal_lro_wqes_complete - Process completed LRO WQEs in the srq
 * @srq: srq
 *
 * This function processes the work array and returns the completed wqes to
 * free array until it finds an wqe that is not completed. It also adjusts
 * block length in the wrapper of the previous block.
 */

vxge_hal_status_e
__hal_lro_wqes_complete(
	__hal_srq_t *srq)
{
	void *prev_block;
	vxge_hal_lro_wqe_h wqeh;
	__hal_device_t *hldev;
	__hal_list_block_wrapper_t *wrapper;

	vxge_assert(srq != NULL);

	hldev = (__hal_device_t *)srq->channel.devh;

	vxge_hal_trace_log_srq(hldev,
	    ((__hal_vpath_handle_t *)srq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_srq(hldev,
	    ((__hal_vpath_handle_t *)srq->channel.vph)->vpath->vp_id,
	    "srq = 0x"VXGE_OS_STXFMT, (ptr_t)srq);

	for (;;) {
		__hal_channel_dtr_try_complete(&srq->channel,
		    (__hal_dtr_h *)&wqeh);

		if (wqeh == NULL)
			break;

		if (!__hal_srq_lro_wqe_owner_host_check((u64)(ptr_t)wqeh))
			break;

		__hal_channel_dtr_complete(&srq->channel);

		if (__hal_srq_lro_wqe_first_check((u64)(ptr_t)wqeh)) {

			prev_block = __hal_list_block_prev_pointer(
			    __hal_srq_lro_wqe_memblock_get((u64)(ptr_t)wqeh));

			wrapper = (__hal_list_block_wrapper_t *)
			    (((char *)prev_block) +
			    srq->lro_wqes_per_block * srq->lro_wqe_size);

			wrapper->control_0 = VXGE_HAL_LBW_NEXT_BLOCK_LENGTH(0);

		}

		__hal_channel_dtr_free(&srq->channel, (__hal_dtr_h)wqeh);

	}

	vxge_hal_trace_log_srq(hldev,
	    ((__hal_vpath_handle_t *)srq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * __vxge_hal_hw_wqe_set - Set WQE from work request
 * @srq: SRQ queue.
 * @opaque_handle: Opaque handle passed by RNICPI
 * @buffer: The memory for WQE
 * @length: Length of memory buffer
 * @offset: Offset in WQE from where to set into memory buffer
 * @sg_list: List of physical buffer to be set.
 * @num_sgle: Number of elements in sgle.
 *
 * This function initializes WQE.
 */
vxge_hal_status_e
__vxge_hal_hw_wqe_set(
	__hal_srq_t *srq,
	u64 opaque_handle,
	void *buffer,
	u32 length,
	u32 offset,
	vxge_hal_sg_list_t *sg_list,
	u32 num_sgle)
{
	__hal_od_t *odlist;
	__hal_hw_wqe_t wqe;
	u32 capacity = 0;
	u32 i, start, end;
	__hal_device_t *hldev;

	vxge_assert((srq != NULL) && (buffer != NULL) && (sg_list != NULL));

	hldev = (__hal_device_t *)srq->channel.devh;

	vxge_hal_trace_log_srq(hldev,
	    ((__hal_vpath_handle_t *)srq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_srq(hldev,
	    ((__hal_vpath_handle_t *)srq->channel.vph)->vpath->vp_id,
	    "srq = 0x"VXGE_OS_STXFMT", opaque_handle = 0x"VXGE_OS_STXFMT", "
	    "buffer = 0x"VXGE_OS_STXFMT", length = %d, offset = %d, "
	    "sg_list = 0x"VXGE_OS_STXFMT", num_sgle = %d", (ptr_t)srq,
	    (ptr_t)opaque_handle, (ptr_t)buffer, length, offset,
	    (ptr_t)sg_list, num_sgle);

	if (offset < sizeof (__hal_hw_wqe_t)) {

		wqe.host_control = 0;

		wqe.opaque_handle = opaque_handle;

		wqe.control_0 = 0;

		if (srq->interrupt_control)
			wqe.control_0 |= VXGE_HAL_HW_WQE_INTERRUPT_CONTROL;

		if (srq->low_latency)
			wqe.control_0 |= VXGE_HAL_HW_WQE_LOW_LATENCY;

		for (i = 0; i < num_sgle; i++) {

			capacity += sg_list[i].length;

		}

		wqe.control_0 |= VXGE_HAL_HW_WQE_CAPACITY(capacity);

		wqe.control_1 = VXGE_HAL_HW_WQE_NUM_OID(num_sgle) |
		    VXGE_HAL_HW_WQE_WSN(++(srq->wsn));

		if (length > (sizeof (__hal_hw_wqe_t) - offset)) {

			vxge_os_memcpy(buffer, (void*)(((char *)&wqe)+offset),
			    sizeof (__hal_hw_wqe_t)-offset);
			buffer =
			    ((char *)buffer) + sizeof (__hal_hw_wqe_t)-offset;
			length -= sizeof (__hal_hw_wqe_t)-offset;
			offset = sizeof (__hal_hw_wqe_t);

		} else {

			vxge_os_memcpy(buffer, (void*)(((char *)&wqe)+offset),
			    length);

			vxge_hal_trace_log_srq(hldev, ((__hal_vpath_handle_t *)
			    srq->channel.vph)->vpath->vp_id,
			    "<==  %s:%s:%d  Result: 0",
			    __FILE__, __func__, __LINE__);

			return (VXGE_HAL_OK);
		}

	}

	start = ((offset - sizeof (__hal_hw_wqe_t)) / sizeof (__hal_od_t));

	if (num_sgle <= (start + (length / sizeof (__hal_od_t)))) {

		end = num_sgle;

	} else {

		end = start + (length / sizeof (__hal_od_t));

	}

	odlist = (__hal_od_t *)buffer;

	for (i = start; i < end; i++) {

		odlist[i-start].ptr_to = sg_list[i].ptr_to;
		odlist[i-start].control_0 =
		    VXGE_HAL_OD_STAG(sg_list[i].stag) |
		    VXGE_HAL_OD_LENGTH(sg_list[i].length);

	}

	vxge_hal_trace_log_srq(hldev,
	    ((__hal_vpath_handle_t *)srq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0", __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);

}

/*
 * vxge_hal_srq_create - Create SRQ
 *
 * @vpath_handle: Virtual path handle to which this queue belongs
 * @attr: SRQ configuration parameters structure
 * @opaque_handle: HAL Opaque handle
 * @srq_opaque_handle: An HAL opaque handle used by the host to reference data
 *		structures in host memory
 *
 * Creates a SRQ on the adapter.
 */
vxge_hal_status_e
vxge_hal_srq_create(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_srq_attr_t *attr,
	vxge_hal_opaque_handle_t *opaque_handle,
	vxge_hal_opaque_handle_t *srq_opaque_handle)
{
	__hal_device_t *hldev;
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_srq_t *srq;
	u32 wqes_per_block = 0;
	u32 srq_length;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (attr != NULL) &&
	    (opaque_handle != NULL) && (srq_opaque_handle != NULL));

	hldev = vp->vpath->hldev;

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", attr = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", "
	    "srq_opaque_handle = 0x"VXGE_OS_STXFMT, (ptr_t)vpath_handle,
	    (ptr_t)attr, (ptr_t)opaque_handle, (ptr_t)srq_opaque_handle);

	srq_length = attr->srq_length;

	if (attr->srq_type == VXGE_HAL_SRQ_TYPE_LRO) {

		wqes_per_block = VXGE_HAL_LIST_BLOCK_ITEMS_SIZE /
		    sizeof (__hal_lro_wqe_t);

		srq_length = ((attr->srq_length + wqes_per_block - 1) /
		    wqes_per_block) * wqes_per_block;

	}

	srq = (__hal_srq_t *)__hal_channel_allocate(
	    (vxge_hal_device_h)vp->vpath->hldev,
	    vpath_handle,
	    VXGE_HAL_CHANNEL_TYPE_RECEIVE_QUEUE,
	    srq_length,
	    TRUE,
	    ((attr->srq_type == VXGE_HAL_SRQ_TYPE_LRO)?
	    TRUE:FALSE),
	    ((attr->srq_type == VXGE_HAL_SRQ_TYPE_LRO)?
	    TRUE:FALSE),
	    0,
	    NULL);

	if (srq == NULL) {
		vxge_hal_info_log_srq(hldev, vp->vpath->vp_id,
		    "Memory allocation failed  %s : %d",
		    __func__, __LINE__);
		vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
		    __LINE__, VXGE_HAL_ERR_OUT_OF_MEMORY);
		return (VXGE_HAL_ERR_OUT_OF_MEMORY);
	}

	srq->srq_type = ((vxge_hal_srq_attr_t *)attr)->srq_type;
	srq->srq_length = srq_length;
	srq->interrupt_control =
	    ((vxge_hal_srq_attr_t *)attr)->interrupt_control;
	srq->low_latency = ((vxge_hal_srq_attr_t *)attr)->low_latency;
	srq->is_shared = ((vxge_hal_srq_attr_t *)attr)->is_shared;
	srq->low_threshold_notify =
	    ((vxge_hal_srq_attr_t *)attr)->low_threshold_notify;
	srq->srq_limit_value = ((vxge_hal_srq_attr_t *)attr)->srq_limit_value;

#if defined(VXGE_HAL_SRQ_POST)
	vxge_os_spin_lock_init(&srq->channel.post_lock, hldev->pdev);
#elif defined(VXGE_HAL_SRQ_POST_IRQ)
	vxge_os_spin_lock_init_irq(&srq->channel.post_lock, hldev->irqh);
#endif

#if defined(VXGE_HAL_SRQ_FREE)
	vxge_os_spin_lock_init(&srq->channel.free_lock, hldev->pdev);
#elif defined(VXGE_HAL_SRQ_FREE_IRQ)
	vxge_os_spin_lock_init_irq(&srq->channel.free_lock, hldev->irqh);
#endif

#if defined(VXGE_HAL_SRQ_RESERVE)
	vxge_os_spin_lock_init(&srq->channel.reserve_lock, hldev->pdev);
#elif defined(VXGE_HAL_SRQ_RESERVE_IRQ)
	vxge_os_spin_lock_init_irq(&srq->channel.reserve_lock, hldev->irqh);
#endif

	srq->first_block = NULL;
	srq->last_block = NULL;

	if (attr->srq_type == VXGE_HAL_SRQ_TYPE_LRO) {
		srq->lro_wqe_size = sizeof (__hal_lro_wqe_t);
		srq->lro_wqes_per_block = wqes_per_block;
		srq->lro_wqe_priv_size = sizeof (__hal_lro_wqe_priv_t);
	}

	srq->wqeblock_priv_size = sizeof (__hal_srq_blk_priv_t);

	status = __hal_vp_oid_allocate(
	    vp->vpath,
	    &vp->vpath->srq_oid_db,
	    &srq->srq_id);

	if (status != VXGE_HAL_OK) {
		vxge_hal_info_log_srq(hldev, vp->vpath->vp_id,
		    "Memory allocation for srq oid db failed == > %s : %d",
		    __func__, __LINE__);
		__hal_srq_delete(vpath_handle, srq);
		vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
		    __LINE__, VXGE_HAL_ERR_TOO_MANY);
		return (VXGE_HAL_ERR_TOO_MANY);
	}

	status = __hal_regpool_allocate(&vp->vpath->hldev->reg_pool,
	    &srq->rq_ir_reg_entry);

	if (status != VXGE_HAL_OK) {
		vxge_hal_info_log_srq(hldev, vp->vpath->vp_id,
		    "Memory allocation for rq_ir failed %s : %d",
		    __func__, __LINE__);
		__hal_srq_delete(vpath_handle, srq);
		vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
		    __LINE__, status);
		return (status);
	}

	srq->rq_ir = srq->rq_ir_reg_entry->reg_addr;

	srq->srq_opaque_handle = srq_opaque_handle;

	srq->opaque_handle = opaque_handle;

	__hal_channel_insert_pending_list(&srq->channel);

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_srq_delete - Deletes a SRQ
 * @vpath_handle: Virtual path handle to which this queue belongs
 * @srq_id: SRQ to be deleted
 * @opaque_handle: HAL Opaque handle
 *
 * This function deletes SRQ.
 *
 */
vxge_hal_status_e
vxge_hal_srq_delete(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t srq_id,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_device_t *hldev;
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	__hal_srq_t *srq;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (srq_id != 0) &&
	    (opaque_handle != NULL));

	hldev = vp->vpath->hldev;

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "srq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT,
	    (ptr_t)vpath_handle, (ptr_t)srq_id,
	    (ptr_t)opaque_handle);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	srq = (__hal_srq_t *)(ptr_t)srq_id;

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, srq_id);

	__hal_srq_delete_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    srq->srq_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_srq_modify - Modify the attributes of a SRQ
 *
 * @vpath_handle: Virtual path handle to which this queue belongs
 * @srq_id: SRQ to be modified
 * @opaque_handle: HAL Opaque handle
 *
 * Modify the attributes of a SRQ.
 */
vxge_hal_status_e
vxge_hal_srq_modify(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t srq_id,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_device_t *hldev;
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	u32 event_notification_mask = 0;
	__hal_srq_t *srq;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (srq_id != 0) &&
	    (opaque_handle != NULL));

	hldev = vp->vpath->hldev;

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "srq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT,
	    (ptr_t)vpath_handle, (ptr_t)srq_id,
	    (ptr_t)opaque_handle);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	srq = (__hal_srq_t *)(ptr_t)srq_id;

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, srq_id);

	__hal_srq_modify_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    srq->srq_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    vp->vpath->hldev->header.config.max_num_wqe_od_groups,
	    vp->vpath->hldev->header.config.refill_threshold_high,
	    vp->vpath->hldev->header.config.refill_threshold_low,
	    srq->low_threshold_notify,
	    ((srq->srq_limit_value > 0) ? 0 : 1),
	    event_notification_mask,
	    srq->srq_limit_value);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_srq_query - Query the information from a SRQ
 *
 * @vpath_handle: Virtual path handle to which this queue belongs
 * @srq_id: SRQ to be modified
 * @opaque_handle: HAL Opaque handle
 *
 * Query Information from a SRQ
 */
vxge_hal_status_e
vxge_hal_srq_query(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t srq_id,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_device_t *hldev;
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	__hal_srq_t *srq;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (srq_id != 0) &&
	    (opaque_handle != NULL));

	hldev = vp->vpath->hldev;

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "srq_id = 0x"VXGE_OS_STXFMT", opaque_handle = 0x"VXGE_OS_STXFMT,
	    (ptr_t)vpath_handle, (ptr_t)srq_id, (ptr_t)opaque_handle);


	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	srq = (__hal_srq_t *)(ptr_t)srq_id;

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, srq_id);

	__hal_srq_query_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    srq->srq_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_srq_arm - Arm an SRQ
 *
 * @vpath_handle: Virtual path handle to which this queue belongs
 * @srq_id: SRQ to be armed
 * @opaque_handle: HAL Opaque handle
 *
 * Arms an SRQ
 */
vxge_hal_status_e
vxge_hal_srq_arm(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t srq_id,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_device_t *hldev;
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	__hal_srq_t *srq;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (srq_id != 0) &&
	    (opaque_handle != NULL));

	hldev = vp->vpath->hldev;

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "srq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT,
	    (ptr_t)vpath_handle, (ptr_t)srq_id,
	    (ptr_t)opaque_handle);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	srq = (__hal_srq_t *)(ptr_t)srq_id;

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, (ptr_t)srq_id);

	__hal_srq_arm_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    srq->srq_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    srq->srq_limit_value);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * __hal_srq_fwqebw - Post First WQE block wrapper to SRQ
 *
 * @vpath_handle: Virtual path handle to which this queue belongs
 * @srq_id: SRQ to be posted to
 * @opaque_handle: HAL Opaque handle
 *
 * Posts First WQE block wrapper to SRQ
 */
vxge_hal_status_e
__hal_srq_fwqebw(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t srq_id,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_device_t *hldev;
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	__hal_srq_t *srq;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (srq_id != 0) &&
	    (opaque_handle != NULL));

	hldev = vp->vpath->hldev;

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "srq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT,
	    (ptr_t)vpath_handle, (ptr_t)srq_id,
	    (ptr_t)opaque_handle);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	srq = (__hal_srq_t *)(ptr_t)srq_id;

	vxge_os_xchg(&srq->first_block_status,
	    VXGE_HAL_SRQ_FWQEBW_STATUS_POST_SUCCESS);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, srq_id);

	__hal_srq_fwqebw_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    srq->srq_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    srq->first_block_address,
	    srq->first_block_length,
	    srq->wsn);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;
	else
		vxge_os_xchg(&srq->first_block_status,
		    VXGE_HAL_SRQ_FWQEBW_STATUS_UNKNOWN);


	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * __hal_srq_fwqebw_resp - rResponse for Post First WQE block wrapper to SRQ
 *
 * @vpath_handle: Virtual path handle to which this queue belongs
 * @srq_id: SRQ to be posted to
 * @result: Result of the operation
 *
 * Called by UMQ processor when it received the response from adapter for
 * Post First WQE block wrapper to SRQ.
 */
vxge_hal_status_e
__hal_srq_fwqebw_resp(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t srq_id,
	vxge_hal_result_e result)
{
	__hal_device_t *hldev;
	__hal_srq_t *srq = (__hal_srq_t *)(ptr_t)srq_id;

	vxge_assert((vpath_handle != NULL) && (srq_id != 0));

	hldev = (__hal_device_t *)srq->channel.devh;

	vxge_hal_trace_log_srq(hldev,
	    ((__hal_vpath_handle_t *)srq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_srq(hldev,
	    ((__hal_vpath_handle_t *)srq->channel.vph)->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "srq_id = 0x"VXGE_OS_STXFMT", "
	    "result = %d", (ptr_t)vpath_handle, (ptr_t)srq_id,
	    result);

	if (result == VXGE_HAL_OK) {
		vxge_os_xchg(&srq->first_block_status,
		    VXGE_HAL_SRQ_FWQEBW_STATUS_POST_SUCCESS);
	} else {
		vxge_os_xchg(&srq->first_block_status,
		    VXGE_HAL_SRQ_FWQEBW_STATUS_POST_FAIL);
	}

	vxge_hal_trace_log_srq(hldev,
	    ((__hal_vpath_handle_t *)srq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: 0",
	    __FILE__, __func__, __LINE__);

	return (VXGE_HAL_OK);
}

/*
 * __hal_srq_lro_wqebadd - Add a filled LRO wqe block
 * @srq: SRQ
 * @wqeh: Last wqe
 *
 * This function checks the wqe posted is the last wqe of the block and
 * updates the next block length of the previous block
 */
vxge_hal_status_e
__hal_srq_lro_wqebadd(
	__hal_srq_t *srq,
	vxge_hal_wqe_h wqeh)
{
	void *first, *last, *prev_block;
	vxge_hal_status_e status = VXGE_HAL_OK;
	__hal_device_t *hldev;
	__hal_list_block_wrapper_t *wrapper;

	vxge_assert((srq != NULL) && (wqeh != NULL));

	hldev = (__hal_device_t *)srq->channel.devh;

	vxge_hal_trace_log_srq(hldev,
	    ((__hal_vpath_handle_t *)srq->channel.vph)->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_srq(hldev,
	    ((__hal_vpath_handle_t *)srq->channel.vph)->vpath->vp_id,
	    "srq = 0x"VXGE_OS_STXFMT", wqeh = 0x"VXGE_OS_STXFMT,
	    (ptr_t)srq, (ptr_t)wqeh);

	first = srq->last_block;

	last =
	    ((char *)first) + (srq->lro_wqes_per_block-1) * srq->lro_wqe_size;

	vxge_assert((first <= wqeh)&& (wqeh <= last));

	if (wqeh != last)
		return (VXGE_HAL_OK);

	wrapper = (__hal_list_block_wrapper_t *)(((char *)first) +
	    srq->lro_wqes_per_block *  srq->lro_wqe_size);

	wrapper->control_0 = 0;

	if (srq->first_block_status == VXGE_HAL_SRQ_FWQEBW_STATUS_UNKNOWN) {

		status = __hal_srq_fwqebw(srq->channel.vph,
		    (vxge_hal_obj_id_t)(ptr_t)srq,
		    srq->srq_opaque_handle);
	} else {

		prev_block = __hal_list_block_prev_pointer(srq->last_block);

		wrapper = (__hal_list_block_wrapper_t *)
		    (((char *)prev_block) +
		    srq->lro_wqes_per_block * srq->lro_wqe_size);

		wrapper->control_0 = VXGE_HAL_LBW_NEXT_BLOCK_LENGTH(
		    srq->lro_wqes_per_block * srq->lro_wqe_size);

	}

	srq->last_block = __hal_list_block_next_pointer(srq->last_block);

	vxge_hal_trace_log_srq(hldev,
	    ((__hal_vpath_handle_t *)srq->channel.vph)->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);

	return (status);
}

/*
 * vxge_hal_srq_wqebadd - Post additional WQE Blocks to a SRQ
 *
 * @vpath_handle: Virtual path handle to which this queue belongs
 * @srq_id: SRQ to be posted to
 * @opaque_handle: HAL Opaque handle
 * @nblocks: Number of additional blocks
 *
 * Posts additional WQE Blocks to a SRQ
 */
vxge_hal_status_e
vxge_hal_srq_wqebadd(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t srq_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 nblocks)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	__hal_srq_t *srq;
	__hal_device_t *hldev;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (srq_id != 0) &&
	    (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", srq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", nblocks = %d",
	    (ptr_t)vpath_handle, (ptr_t)srq_id,
	    (ptr_t)opaque_handle, nblocks);

	// TODO: Actually add additional blocks into SRQ

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	srq = (__hal_srq_t *)(ptr_t)srq_id;

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, srq_id);

	__hal_srq_wqebadd_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    srq->srq_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    srq->wsn);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_srq_return_unused_wqes - Request return of unused WQEs
 *
 * @vpath_handle: Virtual path handle to which this queue belongs
 * @srq_id: SRQ to be posted to
 * @opaque_handle: HAL Opaque handle
 * @number_of_wqes: Number of WQEs requested
 *
 * Requests the adapter return of unused WQEs
 */
vxge_hal_status_e
vxge_hal_srq_return_unused_wqes(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t srq_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 number_of_wqes)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	__hal_srq_t *srq;
	__hal_device_t *hldev;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (srq_id != 0) &&
	    (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", srq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", number_of_wqes = %d",
	    (ptr_t)vpath_handle, (ptr_t)srq_id,
	    (ptr_t)opaque_handle, number_of_wqes);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	srq = (__hal_srq_t *)(ptr_t)srq_id;

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, srq_id);

	__hal_srq_return_unused_wqes_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    srq->srq_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    number_of_wqes);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_srq_hw_wqe_post - Post a WQE
 * @vpath_handle: Handle returned by virtual path open
 * @srq_id: Object Id of SRQ
 * @opaque_handle: Opaque handle passed by RNICPI
 * @sg_list: List of physical buffer to be set.
 * @num_sgle: Number of elements in sgle.
 *
 * This function posts a WQE for work request.
 */
vxge_hal_status_e
vxge_hal_srq_hw_wqe_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t srq_id,
	u64 opaque_handle,
	vxge_hal_sg_list_t *sg_list,
	u32 num_sgle)
{
	vxge_hal_status_e status = VXGE_HAL_OK;
	__hal_srq_t *srq = (__hal_srq_t *)(ptr_t)srq_id;
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_list_block_t *post_block;
	__hal_srq_blk_priv_t *post_priv;
	__hal_device_t *hldev;
	u32 wqe_length, pad_length, offset;
	char *tmp_ptr;

	vxge_assert((vpath_handle != NULL) &&
	    (srq != NULL) && (sg_list != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", srq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", sg_list = 0x"VXGE_OS_STXFMT
	    ", num_sgle = %d", (ptr_t)vpath_handle, (ptr_t)srq_id,
	    (ptr_t)opaque_handle, (ptr_t)sg_list, num_sgle);

#if defined(VXGE_HAL_SRQ_POST)
	vxge_os_spin_lock(&srq->channel.post_lock);
#elif defined(VXGE_HAL_SRQ_POST_IRQ)
	vxge_os_spin_lock_irq(&srq->channel.post_lock, flags);
#endif

	__hal_hw_wqes_complete(srq);

	wqe_length = sizeof (__hal_hw_wqe_t) + num_sgle * sizeof (__hal_od_t);

	if (wqe_length % __vxge_os_cacheline_size) {
		pad_length = __vxge_os_cacheline_size -
		    (wqe_length % __vxge_os_cacheline_size);
	} else {
		pad_length = 0;
	}

	vxge_hal_info_log_srq(hldev, vp->vpath->vp_id,
	    "wqe_length = %d, pad_length = %d", wqe_length, num_sgle);

	if ((wqe_length + pad_length) > srq->bytes_free) {

		vxge_hal_info_log_srq(hldev, vp->vpath->vp_id,
		    "No more space in the queue == > %s : %d",
		    __func__, __LINE__);

		vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d", __FILE__, __func__,
		    __LINE__, VXGE_HAL_ERR_OUT_OF_SPACE);

		return (VXGE_HAL_ERR_OUT_OF_SPACE);

	}

	post_block = srq->channel.dtr_arr[srq->channel.post_index].dtr;

	post_priv = (__hal_srq_blk_priv_t *)(ptr_t)
	    __hal_list_block_priv_data(post_block);

	tmp_ptr = ((char *)post_block) + post_priv->post_offset;

	offset = 0;

	vxge_hal_info_log_srq(hldev, vp->vpath->vp_id,
	    "post_index = %d, block_addr = 0x"VXGE_OS_STXFMT
	    ", post_offset = 0x"VXGE_OS_LLXFMT, srq->channel.post_index,
	    post_priv->dma_addr, post_priv->post_offset);

	for (;;) {

		if (post_priv->post_offset + wqe_length + pad_length <=
		    VXGE_HAL_LIST_BLOCK_ITEMS_SIZE) {

			if (wqe_length > 0) {
				__vxge_hal_hw_wqe_set(srq, opaque_handle,
				    tmp_ptr, wqe_length, offset,
				    sg_list, num_sgle);
			}

			post_priv->post_offset += (wqe_length + pad_length);

			srq->bytes_used += (wqe_length + pad_length);
			srq->bytes_free -= (wqe_length + pad_length);

			if (post_priv->post_offset ==
			    VXGE_HAL_LIST_BLOCK_ITEMS_SIZE) {

				srq->channel.post_index++;
				if (srq->channel.post_index ==
				    srq->channel.length)
					srq->channel.post_index = 0;
			}

			break;
		} else {
			u32 length;

			length = (u32)(VXGE_HAL_LIST_BLOCK_ITEMS_SIZE -
			    post_priv->post_offset);

			if (length >= wqe_length) {

				__vxge_hal_hw_wqe_set(srq, opaque_handle,
				    tmp_ptr, wqe_length, offset,
				    sg_list, num_sgle);

				pad_length -= (length - wqe_length);

				wqe_length = 0;

			} else {
				__vxge_hal_hw_wqe_set(srq, opaque_handle,
				    tmp_ptr, length, offset,
				    sg_list, num_sgle);

				wqe_length -= length;

				offset += length;

			}

			post_priv->post_offset = VXGE_HAL_LIST_BLOCK_ITEMS_SIZE;

			srq->bytes_used += length;
			srq->bytes_free -= length;


			srq->channel.post_index++;
			if (srq->channel.post_index == srq->channel.length)
				srq->channel.post_index = 0;

			if ((wqe_length == 0) && (pad_length == 0))
				break;

			post_block =
			    srq->channel.dtr_arr[srq->channel.post_index].dtr;

			post_priv = (__hal_srq_blk_priv_t *)(ptr_t)
			    __hal_list_block_priv_data(post_block);

			tmp_ptr = (char *)post_block;

			vxge_hal_info_log_srq(hldev, vp->vpath->vp_id,
			    "post_index = %d, block_addr = 0x"VXGE_OS_LLXFMT", "
			    "post_offset = "VXGE_OS_LLXFMT,
			    srq->channel.post_index, post_priv->dma_addr,
			    post_priv->post_offset);

		}

	}

	vxge_hal_info_log_srq(hldev, vp->vpath->vp_id,
	    "post_index = %d, block_addr = 0x"VXGE_OS_STXFMT", "
	    "post_offset = 0x"VXGE_OS_LLXFMT,
	    srq->channel.post_index, post_priv->dma_addr,
	    post_priv->post_offset);

	if (srq->first_block_status == VXGE_HAL_SRQ_FWQEBW_STATUS_UNKNOWN) {

		status = __hal_srq_fwqebw(vpath_handle,
		    srq_id,
		    srq->srq_opaque_handle);

#if 0
		if (status == VXGE_HAL_PENDING) {

			while (srq->first_block_status ==
			    VXGE_HAL_SRQ_FWQEBW_STATUS_UNKNOWN) {
				vxge_os_udelay(20);
			}

			if (srq->first_block_status ==
			    VXGE_HAL_SRQ_FWQEBW_STATUS_POST_SUCCESS)
				status = VXGE_HAL_OK;
		}
#else
		status = VXGE_HAL_OK;
#endif

	}

	if (status == VXGE_HAL_OK) {
		// TODO : post doorbell
	}

#if defined(VXGE_HAL_SRQ_POST)
	vxge_os_spin_unlock(&srq->channel.post_lock);
#elif defined(VXGE_HAL_SRQ_POST_IRQ)
	vxge_os_spin_unlock_irq(&srq->channel.post_lock, flags);
#endif

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);

	return (status);
}


/*
 * vxge_hal_lro_wqe_post - Post a WQE
 * @vpath_handle: Handle returned by virtual path open
 * @srq_id: Object Id of SRQ
 * @opaque_handle: Opaque handle passed by RNICPI
 * @intr_override: Interrupt override. 1 to force interrupt, 0 no interrupt
 * @buf0_ptr: Buffer 0 pointer.
 * @buf1_ptr: Buffer 1 pointer.
 * @buf2_ptr: Buffer 2 pointer.
 * @buf3_ptr: Buffer 3 pointer.
 * @buf4_ptr: Buffer 4 pointer.
 *
 * This function posts a LRO-WQE for work request. This function
 * returns synchronously.
 */
vxge_hal_status_e
vxge_hal_lro_wqe_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t srq_id,
	u64 opaque_handle,
	u32 intr_override,
	u64 buf0_ptr,
	u64 buf1_ptr,
	u64 buf2_ptr,
	u64 buf3_ptr,
	u64 buf4_ptr)
{
	vxge_hal_status_e status = VXGE_HAL_OK;
	vxge_hal_wqe_h wqeh;
	__hal_lro_wqe_t *lro_wqe;
	__hal_device_t *hldev;
	__hal_srq_t *srq = (__hal_srq_t *)(ptr_t)srq_id;
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;

	vxge_assert((vpath_handle != NULL) && (srq != NULL) && (buf0_ptr != 0));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", srq_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", intr_override = %d, "
	    "buf0_ptr = 0x"VXGE_OS_LLXFMT", buf1_ptr = 0x"VXGE_OS_LLXFMT", "
	    "buf2_ptr = 0x"VXGE_OS_LLXFMT", buf3_ptr = 0x"VXGE_OS_LLXFMT", "
	    "buf4_ptr = 0x"VXGE_OS_LLXFMT, (ptr_t)vpath_handle,
	    (ptr_t)srq_id, (ptr_t)opaque_handle, intr_override,
	    buf0_ptr, buf1_ptr, buf2_ptr, buf3_ptr, buf4_ptr);

#if defined(VXGE_HAL_SRQ_POST)
	vxge_os_spin_lock(&srq->channel.post_lock);
#elif defined(VXGE_HAL_SRQ_POST_IRQ)
	vxge_os_spin_lock_irq(&srq->channel.post_lock, flags);
#endif

	__hal_lro_wqes_complete(srq);

	if (__hal_channel_dtr_alloc(&srq->channel, &wqeh) == VXGE_HAL_OK) {

		lro_wqe = (__hal_lro_wqe_t *)wqeh;

		((__hal_lro_wqe_priv_t *)(ptr_t)
		    lro_wqe->wqe_host_control)->opaque_handle = opaque_handle;
		__hal_srq_lro_wqe_owner_adapter_set((u64)(ptr_t)lro_wqe);
		lro_wqe->wqe_opaque_handle = (u64)(ptr_t)lro_wqe;
		lro_wqe->control_0 =
		    VXGE_HAL_LRO_WQE_INTERRUPT_OVERRIDE(intr_override);
		lro_wqe->buffer0_ptr = buf0_ptr;
		lro_wqe->buffer1_ptr = buf1_ptr;
		lro_wqe->buffer2_ptr = buf2_ptr;
		lro_wqe->buffer3_ptr = buf3_ptr;
		lro_wqe->buffer4_ptr = buf4_ptr;

		__hal_channel_dtr_post(&srq->channel, wqeh);
	}

	status = __hal_srq_lro_wqebadd(srq, wqeh);

#if defined(VXGE_HAL_SRQ_POST)
	vxge_os_spin_unlock(&srq->channel.post_lock);
#elif defined(VXGE_HAL_SRQ_POST_IRQ)
	vxge_os_spin_unlock_irq(&srq->channel.post_lock, flags);
#endif

	vxge_hal_trace_log_srq(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);

	return (status);
}
