/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */
/*
 * Copyright (c) 2002-2010 Neterion, Inc.
 * All right Reserved.
 *
 * FileName :   vxgehal-srq.h
 *
 * Description:  HAL offload SRQ list structures
 *
 * Created:       15 Dec 2005
 */

#ifndef	VXGE_HAL_SRQ_H
#define	VXGE_HAL_SRQ_H

__EXTERN_BEGIN_DECLS

/*
 * struct __hal_srq_blk_priv_t - SRQ Block HAL-private data
 * @dma_addr: DMA (mapped) address of _this_ WQE.
 * @dma_handle: DMA handle used to map the WQE onto device.
 * @dma_offset: WQE's offset in the memory block.
 * @memblock: Pointer to the WQE LIST memory block or memory page.
 * @wrapper: Address of the List Block wrapper for this list.
 * @srqh: SRQ for this structure.
 * @compl_offset: Number of bytes from the beginning of the block processed.
 * @post_offset: Number of bytes used from the beginning of the block for the
 *		 messages so far.
 * @pad: Padding to make it cache-aligned.
 *
 * Per-SRQ block HAL-private data. HAL uses the space to keep DMA
 * information associated with the block.
 *
 *
 */
typedef struct __hal_srq_blk_priv_t {
	dma_addr_t			dma_addr;
	pci_dma_h			dma_handle;
	ptrdiff_t			dma_offset;
	void				*memblock;
	__hal_list_block_wrapper_t	*wrapper;
	vxge_hal_srq_h			srqh;
	u64				compl_offset;
	u64				post_offset;
	u8				pad[(((63+__vxge_os_cacheline_size)/
						__vxge_os_cacheline_size)*
						__vxge_os_cacheline_size)-64];
} __hal_srq_blk_priv_t;

/*
 * struct __hal_srq_t - SRQ channel
 * @channel: Channel "base" of this Queue, the common part of all HAL
 *	  channels.
 * @high_wqe_address: The DMA address upto which the adapter read the WQEs.
 * @srq_opaque_handle: Opaque handle for this SRQ.(Allocated by the ULD)
 * @opaque_handle: Opaque handle passed for creation of SRQ(Allocated by ULD)
 * @mempool: Memory pool, the pool from which wqe lsit blocks get allocated.
 * @rq_ir: The adapter will overwrite and update this location as WQEs
 *		are consumed.
 * @rq_ir_reg_entry: Reg entry of rq_ir_t
 * @bytes_used: The number of bytes used in SRQ.
 * @bytes_free: The number of bytes free in SRQ.
 * @first_block: The address of first block in the SRQ.
 * @last_block: The address of last block in the SRQ.
 * @first_block_address: The DMA Address of the first block.
 * @first_block_length: The Length of the first block.
 * @first_block_posted: Flag is set when first block address in successfully
 *		posted.
 * @srq_id: The object Id of the SRQ channel
 * @srq_type: SRQ type.
 * @srq_length: The length of the SRQ.
 * @is_shared: If this channel is shared.
 * @interrupt_control: Enable/Disable Interrupt generation.
 * @low_latency: Enable/Disable placing message in CQE as immediate message.
 * @lro_wqe_size: size of WQE.
 * @lro_wqe_priv_size: Per WQE size reserved (by HAL) for ULD to keep per-wqe
 *		data (e.g., DMA handle for Solaris).
 * @lro_wqes_per_block: Number of wqes per hardware-defined RxD
 *		 block.
 * @wqeblock_priv_size: Reserved at the end of each list block. HAL internal
 *			 usage. Not to confuse with @wqe_priv_size.
 * @low_threshold_notify: WQE Low Threshold to generate interrupt/message.
 * @srq_limit_value: If the number of WQEs in the SRQ (both on-adapter and
 *		on-host) falls below this value, an SRQ Limit violation is
 *		asserted - if armed.
 * @wsn: WQE Sequence number incremeted each time a WQE is posted.
 * @stats: Statistics of this srq
 *
 * WQE List or SRQ channel. Adapter takes the buffers from the WQEs
 * posted into this channel to place the received data.
 *
 */
typedef struct __hal_srq_t {
	__hal_channel_t				channel;
	u64 __vxge_os_attr_cacheline_aligned	high_wqe_address;
	vxge_hal_opaque_handle_t			*srq_opaque_handle;
	vxge_hal_opaque_handle_t			*opaque_handle;
	vxge_hal_mempool_t			*mempool;
	__hal_rq_ir_t				*rq_ir;
	__hal_reg_entry_t			*rq_ir_reg_entry;
	u64					bytes_used;
	u64					bytes_free;
	void					*first_block;
	void					*last_block;
	u64					first_block_address;
	u32					first_block_length;
	u32					first_block_status;
#define	VXGE_HAL_SRQ_FWQEBW_STATUS_UNKNOWN	0
#define	VXGE_HAL_SRQ_FWQEBW_STATUS_POST_PENDING	1
#define	VXGE_HAL_SRQ_FWQEBW_STATUS_POST_SUCCESS	2
#define	VXGE_HAL_SRQ_FWQEBW_STATUS_POST_FAIL	3
	u32					srq_id;
	u32					srq_type;
	u32					srq_length;
	u32					is_shared;
	u32					interrupt_control;
	u32					low_latency;
	u32					lro_wqe_size;
	u32					lro_wqe_priv_size;
	u32					lro_wqes_per_block;
	u32					wqeblock_priv_size;
	u32					low_threshold_notify;
	u32					srq_limit_value;
	u32					wsn;
	vxge_hal_vpath_stats_sw_srq_info_t	stats;
} __hal_srq_t;

static inline u64
__hal_srq_lro_wqe_opaque_handle_get(u64 lro_wqe)
{
	return (((__hal_lro_wqe_priv_t *)(ptr_t)((__hal_lro_wqe_t *)
	    (ptr_t)lro_wqe)->wqe_host_control)->opaque_handle);
}

static inline void
__hal_srq_lro_wqe_owner_unknown_set(u64 lro_wqe)
{
	((__hal_lro_wqe_priv_t *)(ptr_t)((__hal_lro_wqe_t *)
	    (ptr_t)lro_wqe)->wqe_host_control)->owner =
	    VXGE_HAL_LRO_WQE_PRIV_OWNER_UNKNOWN;
}

static inline void
__hal_srq_lro_wqe_owner_adapter_set(u64 lro_wqe)
{
	((__hal_lro_wqe_priv_t *)(ptr_t)((__hal_lro_wqe_t *)
	    (ptr_t)lro_wqe)->wqe_host_control)->owner =
	    VXGE_HAL_LRO_WQE_PRIV_OWNER_ADAPTER;
}

static inline void
__hal_srq_lro_wqe_owner_host_set(u64 lro_wqe)
{
	((__hal_lro_wqe_priv_t *)(ptr_t)((__hal_lro_wqe_t *)
	    (ptr_t)lro_wqe)->wqe_host_control)->owner =
	    VXGE_HAL_LRO_WQE_PRIV_OWNER_HOST;
}

static inline u32
__hal_srq_lro_wqe_owner_host_check(u64 lro_wqe)
{
	return (((__hal_lro_wqe_priv_t *)(ptr_t)((__hal_lro_wqe_t *)
	    (ptr_t)lro_wqe)->wqe_host_control)->owner ==
	    VXGE_HAL_LRO_WQE_PRIV_OWNER_HOST);
}

static inline u32
__hal_srq_lro_wqe_first_check(u64 lro_wqe)
{
	return (((__hal_lro_wqe_priv_t *)(ptr_t)((__hal_lro_wqe_t *)
	    (ptr_t)lro_wqe)->wqe_host_control)->first);
}

static inline void *
__hal_srq_lro_wqe_memblock_get(u64 lro_wqe)
{
	return (((__hal_lro_wqe_priv_t *)(ptr_t)((__hal_lro_wqe_t *)
	    (ptr_t)lro_wqe)->wqe_host_control)->memblock);
}

vxge_hal_status_e
__hal_srq_create(
	vxge_hal_vpath_h vpath_handle,
	__hal_srq_t *srq);

vxge_hal_status_e
__hal_srq_delete(
	vxge_hal_vpath_h vpath_handle,
	__hal_srq_t *srq);

vxge_hal_status_e
__hal_hw_wqes_complete(
	__hal_srq_t *srq);

vxge_hal_status_e
__hal_lro_wqes_complete(
	__hal_srq_t *srq);

vxge_hal_status_e
__vxge_hal_hw_wqe_set(
	__hal_srq_t *srq,
	u64 opaque_handle,
	void *buffer,
	u32 length,
	u32 offset,
	vxge_hal_sg_list_t *sg_list,
	u32 num_sgle);

vxge_hal_status_e
__hal_srq_fwqebw(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t srq_id,
	vxge_hal_opaque_handle_t *opaque_handle);

vxge_hal_status_e
__hal_srq_fwqebw_resp(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t srq_id,
	vxge_hal_result_e result);

vxge_hal_status_e
__hal_srq_lro_wqebadd(
	__hal_srq_t *srq,
	vxge_hal_wqe_h wqeh);

vxge_hal_status_e
vxge_hal_srq_create(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_srq_attr_t *attr,
	vxge_hal_opaque_handle_t *opaque_handle,
	vxge_hal_opaque_handle_t *srq_opaque_handle);

vxge_hal_status_e
vxge_hal_srq_delete(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t srq_id,
	vxge_hal_opaque_handle_t *opaque_handle);

vxge_hal_status_e
vxge_hal_srq_hw_wqe_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t srq_id,
	u64 opaque_handle,
	vxge_hal_sg_list_t *sg_list,
	u32 num_sgle);

vxge_hal_status_e
vxge_hal_lro_wqe_post(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t srq_id,
	u64 opaque_handle,
	u32 intr_override,
	u64 buf0_ptr,
	u64 buf1_ptr,
	u64 buf2_ptr,
	u64 buf3_ptr,
	u64 buf4_ptr);

__EXTERN_END_DECLS

#endif /* VXGE_HAL_SRQ_H */
