/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */
/*
 * Copyright (c) 2002-2010 Neterion, Inc.
 * All right Reserved.
 *
 * FileName :   vxgehal-stag.c
 *
 * Description:  Stag related routines
 *
 * Created:       11 January 2006
 */
#include "vxgehal.h"

/*
 * vxge_hal_stag_allocate_nsmr - Allocate non shared memory region
 *
 * @vpath_handle: Virtual path handle
 * @obj_id: Object Id of STAG
 * @opaque_handle: Opaque handle
 * @pd_id: Protection domain to which the new NSMR belongs
 * @num_of_pbles: The number of PBLEs that the new NSMR must be able
 *		to accommodate.
 * @stag_index: The STAG Index of the NSMR to be created.
 * @remote_access_flag: Flag to Specify if this NSMR will be allowed to be used
 *		for any remote access operations (read, write, invalidate) when
 *		it is subsequently registered.
 *
 * Allocates non shared memory region on adapter
 */
vxge_hal_status_e
vxge_hal_stag_allocate_nsmr(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t obj_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	u64 pd_id,
	u32 num_of_pbles,
	u32 stag_index,
	u32 remote_access_flag)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "obj_id = 0x"VXGE_OS_STXFMT", opaque_handle = 0x"VXGE_OS_STXFMT", "
	    "pd_id = 0x"VXGE_OS_LLDFMT", num_of_pbles = %d, "
	    "stag_index = %d, remote_access_flag = %d",
	    (ptr_t)vpath_handle, (ptr_t)obj_id, (ptr_t)opaque_handle,
	    pd_id, num_of_pbles, stag_index, remote_access_flag);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, obj_id);

	__hal_stag_allocate_nsmr_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    (u32)obj_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    pd_id,
	    num_of_pbles,
	    stag_index,
	    remote_access_flag);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_stag_allocate_mw - Allocate memory window
 *
 * @vpath_handle: Virtual path handle
 * @obj_id: Object Id of STAG
 * @opaque_handle: HAL Opaque handle
 * @stag_index: The host assigned STag index of the MW being allocated on
 *		the adapter.
 * @pd_id: Protection domain to which the new NSMR belongs
 *
 * Allocates memory window on a memory on the adapter
 */
vxge_hal_status_e
vxge_hal_stag_allocate_mw(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t obj_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 stag_index,
	u64 pd_id)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "obj_id = 0x"VXGE_OS_STDFMT", opaque_handle = 0x"VXGE_OS_STXFMT", "
	    "stag_index = %d, pd_id = 0x"VXGE_OS_LLDFMT, (ptr_t)vpath_handle,
	    (ptr_t)obj_id, (ptr_t)opaque_handle, stag_index, pd_id);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, obj_id);

	__hal_stag_allocate_mw_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    (u32)obj_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    stag_index,
	    pd_id);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_stag_deallocate - De-Allocate stag
 *
 * @vpath_handle: Virtual path handle
 * @obj_id: Object Id of STAG
 * @opaque_handle: HAL Opaque handle
 * @stag_index: The index of the STag being de-allocated on the adapter.
 * @dealloc_pbl: Flag to Specify if the PBL that underlies this STag is
 *		also to be deallocated with the STag.
 *
 * De-Allocates stag from the adapter
 */
vxge_hal_status_e
vxge_hal_stag_deallocate(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t obj_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 stag_index,
	u32 dealloc_pbl)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "obj_id = 0x"VXGE_OS_STXFMT", opaque_handle = 0x"VXGE_OS_STXFMT", "
	    "stag_index = %d, dealloc_pbl = %d", (ptr_t)vpath_handle,
	    (ptr_t)obj_id, (ptr_t)opaque_handle, stag_index, dealloc_pbl);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, obj_id);

	__hal_stag_deallocate_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    (u32)obj_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    stag_index,
	    dealloc_pbl);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_stag_register_nsmr - Register Non shared memory region
 *
 * @vpath_handle: Virtual path handle
 * @obj_id: Object Id of STAG
 * @opaque_handle: HAL Opaque handle
 * @params: Parameters
 *
 * Registers non shared memory region with the adapter
 */
vxge_hal_status_e
vxge_hal_stag_register_nsmr(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t obj_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	vxge_hal_stag_register_nsmr_params_t *params)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	vxge_hal_pblist_t *pblist;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (opaque_handle != NULL) &&
	    (params != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", obj_id = 0x"VXGE_OS_STDFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", "
	    "params = 0x"VXGE_OS_STXFMT, (ptr_t)vpath_handle,
	    (ptr_t)obj_id, (ptr_t)opaque_handle, (ptr_t)params);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, obj_id);

	pblist = (vxge_hal_pblist_t *)(ptr_t)params->pble_list;

	vxge_assert(pblist != NULL);

	__hal_stag_register_nsmr_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    (u32)obj_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    params->stag_index,
	    params->stag_key,
	    params->pble_size,
	    params->pd_id,
	    __hal_pblist_dma_addr_get(pblist),
	    __hal_pblist_length_get(pblist, pblist->num_pbles),
	    pblist->num_pbles,
	    params->va_base_flag,
	    params->remote_access_flag,
	    params->local_write_flag,
	    params->local_read_flag,
	    params->remote_write_flag,
	    params->remote_read_flag,
	    params->enable_mw_binding_flag,
	    params->first_byte_offset,
	    params->mr_length,
	    params->base_to);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_stag_re_register_nsmr - Re-register non shared memory region
 *
 * @vpath_handle: Virtual path handle
 * @obj_id: Object Id of STAG
 * @opaque_handle: HAL Opaque handle
 * @params: Parameters.
 *
 * Re-registers non shared memory region with adapter
 */
vxge_hal_status_e
vxge_hal_stag_re_register_nsmr(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t obj_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	vxge_hal_stag_register_nsmr_params_t *params)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	vxge_hal_pblist_t *pblist;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (opaque_handle != NULL) &&
	    (params != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", obj_id = 0x"VXGE_OS_STDFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", "
	    "params = 0x"VXGE_OS_STXFMT, (ptr_t)vpath_handle,
	    (ptr_t)obj_id, (ptr_t)opaque_handle, (ptr_t)params);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, obj_id);

	pblist = (vxge_hal_pblist_t *)(ptr_t)params->pble_list;

	vxge_assert(pblist != NULL);

	__hal_stag_re_register_nsmr_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    (u32)obj_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    params->stag_index,
	    params->stag_key,
	    params->pble_size,
	    params->pd_id,
	    __hal_pblist_dma_addr_get(pblist),
	    __hal_pblist_length_get(pblist, pblist->num_pbles),
	    pblist->num_pbles,
	    params->va_base_flag,
	    params->remote_access_flag,
	    params->local_write_flag,
	    params->local_read_flag,
	    params->remote_write_flag,
	    params->remote_read_flag,
	    params->enable_mw_binding_flag,
	    params->first_byte_offset,
	    params->mr_length,
	    params->base_to);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_stag_register_smr - Register shared memory region
 *
 * @vpath_handle: Virtual path handle
 * @obj_id: Object Id of STAG
 * @opaque_handle: HAL Opaque handle
 * @base_mr_stag_index: Unique identifier for the underlying MR on which this
 *		SMR will be created.
 * @va_base_flag: Flag to Specify if the SMR will use virtual addressing
 * @remote_access_flag: Flag to Specify if this SMR will be allowed to be used
 *		for any remote access operations (read, write, invalidate) on
 *		this and subsequent registrations.
 * @local_write_flag: Flag to Specify if the adapter can write to this SMR.
 * @local_read_flag: Flag to Specify if the adapter can read from this SMR.
 * @remote_write_flag: Flag to Specify if remote peers are allowed to write into
 *		this SMR
 * @remote_read_flag: Flag to Specify if remote peers are allowed to read from
 *		this SMR.
 * @enable_mw_binding_flag: Flag to Specify if this SMR may have MWs bound to
 *		it by subsequent PostSQ.BindMW work requests.
 * @stag_index: STag Index of the new STag record that is to be created.
 * @stag_key: Instance identifier for this validation of this SMR
 * @pd_id: Protection domain to which the new SMR belongs
 * @base_to: Virtual address of the first byte of the SMR.
 *
 * Registers a shared memory region with adapter
 */
vxge_hal_status_e
vxge_hal_stag_register_smr(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t obj_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 base_mr_stag_index,
	u32 va_base_flag,
	u32 remote_access_flag,
	u32 local_write_flag,
	u32 local_read_flag,
	u32 remote_write_flag,
	u32 remote_read_flag,
	u32 enable_mw_binding_flag,
	u32 stag_index,
	u32 stag_key,
	u64 pd_id,
	u64 base_to)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);
	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", obj_id = 0x"VXGE_OS_STDFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", "
	    "base_mr_stag_index = %d, va_base_flag = %d, "
	    "remote_access_flag = %d, local_write_flag = %d, "
	    "local_read_flag = %d, remote_write_flag = %d, "
	    "remote_read_flag = %d, enable_mw_binding_flag = %d, "
	    "stag_index = %d, stag_key = %d, pd_id = 0x"VXGE_OS_LLDFMT", "
	    "base_to = 0x"VXGE_OS_LLDFMT, (ptr_t)vpath_handle,
	    (ptr_t)obj_id, (ptr_t)opaque_handle, base_mr_stag_index,
	    va_base_flag, remote_access_flag, local_write_flag,
	    local_read_flag, remote_write_flag, remote_read_flag,
	    enable_mw_binding_flag, stag_index, stag_key, pd_id,
	    base_to);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, obj_id);

	__hal_stag_register_smr_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    (u32)obj_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    base_mr_stag_index,
	    va_base_flag,
	    remote_access_flag,
	    local_write_flag,
	    local_read_flag,
	    remote_write_flag,
	    remote_read_flag,
	    enable_mw_binding_flag,
	    stag_index,
	    stag_key,
	    pd_id,
	    base_to);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_stag_query - Query of the stag information
 *
 * @vpath_handle: Virtual path handle
 * @obj_id: Object Id of STAG
 * @opaque_handle: HAL Opaque handle
 * @stag_index: The STag index of the existing on-adapter STag that is being
 *		queried
 *
 * Queries the information of a stag from adapter.
 */
vxge_hal_status_e
vxge_hal_stag_query(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t obj_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 stag_index)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", obj_id = 0x"VXGE_OS_STDFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", "
	    "stag_index = %d", (ptr_t)vpath_handle, (ptr_t)obj_id,
	    (ptr_t)opaque_handle, stag_index);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, obj_id);

	__hal_stag_query_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    (u32)obj_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    stag_index);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_stag_validate - Validates a stag
 *
 * @vpath_handle: Virtual path handle
 * @obj_id: Object Id of STAG
 * @opaque_handle: HAL Opaque handle
 * @stag_index: The index of the existing on-adapter STag that is being
 *		validated.
 * @stag_key: The key that to assign to the STag when it is validated.
 *
 * Validates an stag on adapter
 */
vxge_hal_status_e
vxge_hal_stag_validate(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t obj_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 stag_index,
	u32 stag_key)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", obj_id = 0x"VXGE_OS_STDFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", "
	    "stag_index = %d, stag_key = %d", (ptr_t)vpath_handle,
	    (ptr_t)obj_id, (ptr_t)opaque_handle, stag_index, stag_key);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, obj_id);

	__hal_stag_validate_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    (u32)obj_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    stag_index,
	    stag_key);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_stag_invalidate - Invalidates a stag
 *
 * @vpath_handle: Virtual path handle
 * @obj_id: Object Id of STAG
 * @opaque_handle: HAL Opaque handle
 * @stag_index: The index of the existing on-adapter STag that is being
 *		invalidated.
 * @stag_key: The key that is currently assigned to the STag.
 * @pd_id: The PD ID of the session or QP.
 *
 * Invalidates a stag on the adapter
 */
vxge_hal_status_e
vxge_hal_stag_invalidate(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t obj_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 stag_index,
	u32 stag_key,
	u64 pd_id)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", obj_id = 0x"VXGE_OS_STDFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", stag_index = %d, "
	    "stag_key = %d, pd_id = 0x"VXGE_OS_LLXFMT,
	    (ptr_t)vpath_handle, (ptr_t)obj_id,
	    (ptr_t)opaque_handle, stag_index, stag_key, pd_id);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, obj_id);

	__hal_stag_invalidate_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    (u32)obj_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    stag_index,
	    stag_key,
	    pd_id);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_stag_bind_mw - Binds a memory windows to a memory region
 *
 * @vpath_handle: Virtual path handle
 * @obj_id: Object Id of STAG
 * @opaque_handle: HAL Opaque handle
 * @mr_stag_index: The STag Index of the underlying MR to bind to.
 * @mr_stag_key: The STag Key of the underlying MR to bind to.
 * @mw_stag_index: The existing Memory Window's STag Index.
 * @mw_stag_key: The STag Key to use for the MW.
 * @mr_stag: The existing Memory Region's STAG index.
 *		Least significant 8 bits is the key.
 * @mw_ctrl: Control flags for Memory Window.
 * @mw_length: Memory window length in bytes.
 * @tagged_offset: Memory window tagged offset in bytes.
 *
 * Binds a memory windows to a memory region on the adapter
 */
vxge_hal_status_e
vxge_hal_stag_bind_mw(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t obj_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 mr_stag_index,
	u32 mr_stag_key,
	u32 mw_stag_index,
	u32 mw_stag_key,
	u32 mr_stag,
	u32 mw_ctrl,
	u64 mw_length,
	u64 tagged_offset)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "obj_id = 0x"VXGE_OS_STDFMT", opaque_handle = 0x"VXGE_OS_STXFMT", "
	    "mr_stag_index = %d, mr_stag_key = %d, mw_stag_index = %d, "
	    "mw_stag_key = %d, mr_stag = %d, mw_ctrl = %d, "
	    "mw_length = 0x"VXGE_OS_LLXFMT", "
	    "tagged_offset = 0x"VXGE_OS_LLXFMT, (ptr_t)vpath_handle,
	    (ptr_t)obj_id, (ptr_t)opaque_handle, mr_stag_index,
	    mr_stag_key, mw_stag_index, mw_stag_key, mr_stag,
	    mw_ctrl, mw_length, tagged_offset);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, obj_id);

	__hal_stag_bind_mw_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    (u32)obj_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    mr_stag_index,
	    mr_stag_key,
	    mw_stag_index,
	    mw_stag_key,
	    mr_stag,
	    mw_ctrl,
	    mw_length,
	    tagged_offset);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_stag_fast_register_nsmr - Fast register a non shared memory region
 *
 * @vpath_handle: Virtual path handle
 * @obj_id: Object Id of STAG
 * @opaque_handle: HAL Opaque handle
 * @mr_control_size: Length in bytes of the Memory Region control structure.
 * @mr_control_ptr: Pointer to Memory Region control structure in host memory.
 *
 * Fast registera a non shared memory region with the adapter
 */
vxge_hal_status_e
vxge_hal_stag_fast_register_nsmr(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t obj_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 mr_control_size,
	u64 mr_control_ptr)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "obj_id = 0x"VXGE_OS_STDFMT", opaque_handle = 0x"VXGE_OS_STXFMT", "
	    "mr_control_size = %d, mr_control_ptr = 0x"VXGE_OS_LLXFMT,
	    (ptr_t)vpath_handle, (ptr_t)obj_id, (ptr_t)opaque_handle,
	    mr_control_size, mr_control_ptr);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	vxge_assert(dmq != NULL);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, obj_id);

	__hal_stag_fast_register_nsmr_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    (u32)obj_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    mr_control_size,
	    mr_control_ptr);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_stag(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}
