/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */
/*
 * Copyright (c) 2002-2010 Neterion, Inc.
 * All right Reserved.
 *
 * FileName :   vxgehal-tcp.c
 *
 * Description:  TCP offload related routines
 *
 * Created:       11 January 2006
 */
#include "vxgehal.h"

#ifdef	VXGE_HAL_INTERNAL_COMPILER_ERROR

#pragma optimize("", off)

#endif

/*
 * vxge_hal_tcp_open - Open TCP Session
 *
 * @vpath_handle: Virtual path handle
 * @opaque_handle: Opaque handle
 * @session_opaque_handle: Opaque handle used by the host to reference data
 *		structures in host memory related to this session
 * @params: Parameters for the TCP session
 *
 * Open a TCP sesson on the adapter.
 */
vxge_hal_status_e
vxge_hal_tcp_open(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_opaque_handle_t *opaque_handle,
	vxge_hal_opaque_handle_t *session_opaque_handle,
	vxge_hal_tcp_open_params_t *params)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_sq_t *sq;
	__hal_srq_t *srq;
	__hal_cqrq_t *cqrq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (opaque_handle != 0) &&
	    (session_opaque_handle != NULL) && (params != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", "
	    "session_opaque_handle = 0x"VXGE_OS_STXFMT", "
	    "params = 0x"VXGE_OS_STXFMT, (ptr_t)vpath_handle,
	    (ptr_t)opaque_handle, (ptr_t)session_opaque_handle,
	    (ptr_t)params);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	sq = (__hal_sq_t *)(ptr_t)params->sq_id;

	srq = (__hal_srq_t *)(ptr_t)params->srq_id;

	cqrq = (__hal_cqrq_t *)(ptr_t)params->cqrq_id;

	vxge_assert((dmq != NULL) && (sq != NULL) &&
	    (srq != NULL) && (cqrq != NULL));

	if (VXGE_HAL_SQ_STATE_GET(sq) != VXGEHAL_SQ_STATE_INITIALIZED) {

		vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__,
		    VXGE_HAL_ERR_TOO_MANY);
		return (VXGE_HAL_ERR_TOO_MANY);
	}

	VXGE_HAL_SQ_STATE_SET(
	    sq, VXGEHAL_SQ_STATE_SESS_OPEN_PENDING);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(
	    opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(
	    opaque_handle, params->sq_id);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(
	    session_opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(
	    session_opaque_handle, params->sq_id);

	__hal_down_msg_header_set(
	    (vxge_hal_down_msg_h)&dmsg, sq->session_id,
	    VXGE_HAL_MSG_TYPE_TCP_OPEN_REQ,
	    dmq->intr_ctrl, dmq->gen_compl, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_tcp_open_req_dmq_t));

	dmsg.m.tcp_open.session_opaque_handle = vxge_os_htonll(
	    (u64)(ptr_t)session_opaque_handle);

	dmsg.m.tcp_open.control_0 =
	    VXGE_HAL_TCP_OPEN_REQ_TCP_SRC_PORT(params->tcp_src_port) |
	    VXGE_HAL_TCP_OPEN_REQ_TCP_DEST_PORT(params->tcp_dest_port) |
	    VXGE_HAL_TCP_OPEN_REQ_VLAN_ID(params->vlan_id);

	if (params->active_opener)
		dmsg.m.tcp_open.control_0 |=
		    VXGE_HAL_TCP_OPEN_REQ_ACTIVE_OPENER;

	if (params->use_ipv6)
		dmsg.m.tcp_open.control_0 |= VXGE_HAL_TCP_OPEN_REQ_USE_IPV6;

	if (params->use_vlan)
		dmsg.m.tcp_open.control_0 |= VXGE_HAL_TCP_OPEN_REQ_USE_VLAN;

	dmsg.m.tcp_open.control_0 = vxge_os_htonll(dmsg.m.tcp_open.control_0);

	dmsg.m.tcp_open.ip_sa.high = vxge_os_htonll(params->ip_sa_high);

	dmsg.m.tcp_open.ip_sa.low = vxge_os_htonll(params->ip_sa_low);

	dmsg.m.tcp_open.ip_da.high = vxge_os_htonll(params->ip_da_high);

	dmsg.m.tcp_open.ip_da.low = vxge_os_htonll(params->ip_da_low);

	dmsg.m.tcp_open.control_1 = vxge_os_htonll(
	    VXGE_HAL_TCP_OPEN_REQ_BS_CQRQ_INST_NUM(
	    VXGE_HAL_RNIC_OID_DB_SIN_GET(cqrq->cqrq_id)) |
	    VXGE_HAL_TCP_OPEN_REQ_BS_CQRQ_ID(
	    VXGE_HAL_RNIC_OID_DB_SID_GET(cqrq->cqrq_id)) |
	    VXGE_HAL_TCP_OPEN_REQ_BS_SRQ_INST_NUM(
	    VXGE_HAL_RNIC_OID_DB_SIN_GET(srq->srq_id)) |
	    VXGE_HAL_TCP_OPEN_REQ_BS_SRQ_ID(
	    VXGE_HAL_RNIC_OID_DB_SID_GET(srq->srq_id)));

	dmsg.m.tcp_open.control_2 =
	    VXGE_HAL_TCP_OPEN_REQ_NO_WQE_TH(
	    vp->vpath->hldev->header.config.no_wqe_threshold) |
	    VXGE_HAL_TCP_OPEN_REQ_CQRQ_LOW_LAT(
	    ((cqrq->immed_data_length == 0)? 0 : 1));

	if (params->break_merge_on_psh)
		dmsg.m.tcp_open.control_2 |=
		    VXGE_HAL_TCP_OPEN_REQ_BREAK_MERGE_EN;

	if (params->data_merging_enabled)
		dmsg.m.tcp_open.control_2 |=
		    VXGE_HAL_TCP_OPEN_REQ_DATA_MERGE_EN;

	dmsg.m.tcp_open.control_2 = vxge_os_htonll(dmsg.m.tcp_open.control_2);

	dmsg.m.tcp_open.control_3 =
	    VXGE_HAL_TCP_OPEN_REQ_INT_NUM(vp->vpath->bmap_intr_num) |
	    VXGE_HAL_TCP_OPEN_REQ_INT_BITMAP(
	    VXGE_HAL_RNIC_OID_DB_SID_GET(sq->session_id)) |
	    VXGE_HAL_TCP_OPEN_REQ_NCE_ID(params->nce_id);

	if (params->send_vlan_up)
		dmsg.m.tcp_open.control_3 |= VXGE_HAL_TCP_OPEN_REQ_SEND_VLAN_UP;

	if (params->wscale_ok)
		dmsg.m.tcp_open.control_3 |= VXGE_HAL_TCP_OPEN_REQ_WSCALE_OK;

	if (params->sack_ok)
		dmsg.m.tcp_open.control_3 |= VXGE_HAL_TCP_OPEN_REQ_SACK_OK;

	if (params->ecn_ok)
		dmsg.m.tcp_open.control_3 |= VXGE_HAL_TCP_OPEN_REQ_ECN_OK;

	if (params->tstamp_ok)
		dmsg.m.tcp_open.control_3 |= VXGE_HAL_TCP_OPEN_REQ_TSTAMP_OK;

	if (params->kpalive_ok)
		dmsg.m.tcp_open.control_3 |= VXGE_HAL_TCP_OPEN_REQ_KEEPALIVE_OK;

	dmsg.m.tcp_open.control_3 = vxge_os_htonll(dmsg.m.tcp_open.control_3);

	dmsg.m.tcp_open.control_4 = vxge_os_htonll(
	    VXGE_HAL_TCP_OPEN_REQ_VLAN_USER_PRI(params->vlan_user_priority) |
	    VXGE_HAL_TCP_OPEN_REQ_DUP_ACK_TH(params->dup_ack_thres) |
	    VXGE_HAL_TCP_OPEN_REQ_IP_TOS(params->ip_tos) |
	    VXGE_HAL_TCP_OPEN_REQ_IP_FLAGS(params->ip_flags) |
	    VXGE_HAL_TCP_OPEN_REQ_IP_FLOW_LABEL(params->ip_flowlabel));

	dmsg.m.tcp_open.control_5 = vxge_os_htonll(
	    VXGE_HAL_TCP_OPEN_REQ_KPA_IDLE_TIME(params->kpa_idle_time) |
	    VXGE_HAL_TCP_OPEN_REQ_KPA_RETRY_TIME(params->kpa_retry_time));

	dmsg.m.tcp_open.control_6 = vxge_os_htonll(
	    VXGE_HAL_TCP_OPEN_REQ_RETX_TIME_MAX(params->retx_time_max) |
	    VXGE_HAL_TCP_OPEN_REQ_SND_WSCALE(params->snd_wscale) |
	    VXGE_HAL_TCP_OPEN_REQ_KPA_PROBES_MAX(params->kpa_probes_max));

	dmsg.m.tcp_open.control_7 = vxge_os_htonll(
	    VXGE_HAL_TCP_OPEN_REQ_SND_NXT(params->snd_nxt) |
	    VXGE_HAL_TCP_OPEN_REQ_SND_MSS(params->snd_mss) |
	    VXGE_HAL_TCP_OPEN_REQ_SND_MSS_LIMIT(params->snd_mss_limit));

	dmsg.m.tcp_open.control_8 = vxge_os_htonll(
	    VXGE_HAL_TCP_OPEN_REQ_RCV_NXT(params->rcv_nxt) |
	    VXGE_HAL_TCP_OPEN_REQ_RCV_WSCALE(params->rcv_wscale) |
	    VXGE_HAL_TCP_OPEN_REQ_RCV_MSS(params->rcv_mss));

	dmsg.m.tcp_open.control_9 = vxge_os_htonll(
	    VXGE_HAL_TCP_OPEN_REQ_RCV_BUF_SIZE(params->rcv_buf_size) |
	    VXGE_HAL_TCP_OPEN_REQ_RCV_MAX_WND(params->rcv_max_wnd));

	dmsg.m.tcp_open.control_10 = vxge_os_htonll(
	    VXGE_HAL_TCP_OPEN_REQ_RCV_BUF(params->rcv_buf) |
	    VXGE_HAL_TCP_OPEN_REQ_RTT_MIN(params->rttmin));

	dmsg.m.tcp_open.control_11 = vxge_os_htonll(
	    VXGE_HAL_TCP_OPEN_REQ_TS_RECENT(params->ts_recent));

	dmsg.m.tcp_open.nxt_towi_ptr = vxge_os_htonll(sq->first_block_address);

	dmsg.m.tcp_open.control_12 = vxge_os_htonll(
	    VXGE_HAL_TCP_OPEN_REQ_NXT_TOWI_CNT(sq->towis_per_block) |
	    VXGE_HAL_TCP_OPEN_REQ_ACK_BLK_LIMIT(
	    vp->vpath->hldev->header.config.ack_blk_limit));

	dmsg.m.tcp_open.hw_cq_sq =
	    vxge_os_htonll(sq->cq_sq_ir_reg_entry->dma_addr);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;
	else
		VXGE_HAL_SQ_STATE_SET(sq, VXGEHAL_SQ_STATE_INITIALIZED);

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_tcp_promote_to_iwarp - Promote a TCP session to iWARP
 *
 * @vpath_handle: Virtual path handle
 * @session_id: Object Id of Session
 * @opaque_handle: HAL Opaque handle
 * @params: Parameters for promote
 *
 * Promotes a TCP session established on adapter to iWARP session
 */
vxge_hal_status_e
vxge_hal_tcp_promote_to_iwarp(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t session_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	vxge_hal_tcp_promote_to_iwarp_params_t *params)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_sq_t *sq;
	__hal_srq_t *srq;
	__hal_cqrq_t *cqrq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (session_id != 0) &&
	    (opaque_handle != NULL) && (params != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "session_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", params = 0x"VXGE_OS_STXFMT,
	    (ptr_t)vpath_handle, (ptr_t)session_id,
	    (ptr_t)opaque_handle, (ptr_t)params);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	sq = (__hal_sq_t *)(ptr_t)session_id;

	srq = (__hal_srq_t *)(ptr_t)params->srq_id;

	cqrq = (__hal_cqrq_t *)(ptr_t)params->cqrq_id;

	vxge_assert((dmq != NULL) && (srq != NULL) && (cqrq != NULL));

	if (VXGE_HAL_SQ_STATE_GET(sq) != VXGEHAL_SQ_STATE_SESS_ESTED) {

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__,
	    VXGE_HAL_INF_QUEUE_IS_NOT_READY);
		return (VXGE_HAL_INF_QUEUE_IS_NOT_READY);
	}

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, session_id);

	__hal_tcp_promote_to_iwarp_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    sq->session_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    params->mpa_crc_valid,
	    params->tx_marker_status,
	    params->rx_marker_status,
	    params->iwarp_local_stag_zero_allowed,
	    params->iwarp_wire_stag_zero_allowed,
	    params->inbound_write_readresp_enable,
	    params->inbound_readreq_enable,
	    params->srq_pd_id,
	    params->iwarp_qp_pd_id,
	    params->cqrq_ll_send_frame_data_offset,
	    params->iwarp_qp_limit_value,
	    params->rq_max_length,
	    params->iwarp_qp_limit_armed,
	    VXGE_HAL_RNIC_OID_DB_SIN_GET(cqrq->cqrq_id),
	    VXGE_HAL_RNIC_OID_DB_SID_GET(cqrq->cqrq_id),
	    VXGE_HAL_RNIC_OID_DB_SIN_GET(srq->srq_id),
	    VXGE_HAL_RNIC_OID_DB_SID_GET(srq->srq_id),
	    params->rdmap_version,
	    params->ddp_version,
	    params->mpa_reply_frame_size,
	    params->is_mpa_initiator,
	    params->block_permissions,
	    params->ird_limit,
	    sq->cq_rr_ir_reg_entry->dma_addr);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_tcp_modify - Modify a TCP session
 *
 * @vpath_handle: Virtual path handle
 * @session_id: Object Id of Session
 * @opaque_handle: Opaque handle
 * @mpa_crc_valid: If set then MPA CRCs are expected to be valid on received
 *		frames. The adapter always calculates MPA CRCs in the transmit
 *		direction. This field is ignored for a bytestream session.
 * @tx_marker_mode: Whether MPA markers exist in the transmit direction or not.
 *		-- 1 - enable MPA markers in transmit frames
 *		-- 0 - disable MPA markers in transmit frames
 *		This field is ignored for a bytestream session.
 * @rx_marker_mode: Whether markers exist in the receive direction or not.
 *		-- 1 - enable MPA markers in receive frames
 *		-- 0 - disable MPA markers in receive frames
 *		If this is set to zero for an RDMA session, the RxPE drops all
 *		out of order receive segments. This field is ignored for a
 *		bytestream session.
 * @rdmap_version: The RDMA version number
 *		  -- 01 - for IETF RNICs
 *		  -- 00 - for RDMAC RNICs
 * @ddp_version: Direct Data Placement Protocol Version
 *		  -- 01 - for protocol described in
 *		http://ietf.org/internet-drafts/draft-ietf-rddp-ddp-06.txt
 * @iwarp_qp_limit_armed: If set, the Queue Pair limit is armed at the
 *		iwarp_qp_limit_value. This field is ignored for a bytestream
 *		session.
 * @iwarp_qp_limit_value: If the number of in-progress WQEs reaches this
 *		value and the Queue Pair limit is armed, the session produces
 *		an asynchronous event. This field is ignored for a bytestream
 *		session.
 * @rq_max_length: The maximum number of WQEs that this session is
 *		allowed to have in progress on this SRQ.
 * @block_permissions: Used to restrict session's ability to use special
 *		TOWI atoms. It is used to enforce checks such as preventing the
 *		QP from performing a Fast Register NSMR operation.
 * @ird_limit: Host supplied IRD limit for this session.
 *
 * Modifies a TCP session
 */
vxge_hal_status_e
vxge_hal_tcp_modify(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t session_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 mpa_crc_valid,
	u32 tx_marker_mode,
	u32 rx_marker_mode,
	u32 rdmap_version,
	u32 ddp_version,
	u32 iwarp_qp_limit_armed,
	u32 iwarp_qp_limit_value,
	u32 rq_max_length,
	u32 block_permissions,
	u32 ird_limit)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_sq_t *sq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (session_id != 0) &&
	    (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "session_id = 0x"VXGE_OS_STXFMT", opaque_handle = 0x"VXGE_OS_STXFMT
	    ", mpa_crc_valid = %d, tx_marker_mode = %d, "
	    "rx_marker_mode = %d, rdmap_version = %d, "
	    "ddp_version = %d, iwarp_qp_limit_armed = %d, "
	    "iwarp_qp_limit_value = %d, rq_max_length = %d, "
	    "block_permissions = %d, ird_limit = %d", (ptr_t)vpath_handle,
	    (ptr_t)session_id, (ptr_t)opaque_handle, mpa_crc_valid,
	    tx_marker_mode, rx_marker_mode, rdmap_version, ddp_version,
	    iwarp_qp_limit_armed, iwarp_qp_limit_value, rq_max_length,
	    block_permissions, ird_limit);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	sq = (__hal_sq_t *)(ptr_t)session_id;

	vxge_assert((dmq != NULL) && (sq != NULL));

	if (VXGE_HAL_SQ_STATE_GET(sq) != VXGEHAL_SQ_STATE_SESS_ESTED) {

		vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d",
		    __FILE__, __func__, __LINE__,
		    VXGE_HAL_INF_QUEUE_IS_NOT_READY);
		return (VXGE_HAL_INF_QUEUE_IS_NOT_READY);
	}

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, session_id);

	__hal_tcp_modify_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    sq->session_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    mpa_crc_valid,
	    tx_marker_mode,
	    rx_marker_mode,
	    rdmap_version,
	    ddp_version,
	    iwarp_qp_limit_armed,
	    iwarp_qp_limit_value,
	    rq_max_length,
	    block_permissions,
	    ird_limit);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_tcp_delete - Deletes a TCP session
 *
 * @vpath_handle: Virtual path handle
 * @session_id: Object Id of Session
 * @opaque_handle: HAL Opaque handle
 *
 * Deletes a TCP session from adapter.
 */
vxge_hal_status_e
vxge_hal_tcp_delete(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t session_id,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_sq_t *sq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;
	u32 prev_state;

	vxge_assert((vpath_handle != NULL) && (session_id != 0) &&
	    (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "session_id = 0x"VXGE_OS_STDFMT", opaque_handle = 0x"VXGE_OS_STXFMT,
	    (ptr_t)vpath_handle, (ptr_t)session_id, (ptr_t)opaque_handle);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	sq = (__hal_sq_t *)(ptr_t)session_id;

	vxge_assert((dmq != NULL) && (sq != NULL));

	if ((prev_state = VXGE_HAL_SQ_STATE_GET(sq)) !=
	    VXGEHAL_SQ_STATE_SESS_ESTED) {

		vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d",
		    __FILE__, __func__, __LINE__,
		    VXGE_HAL_INF_QUEUE_IS_NOT_READY);
		return (VXGE_HAL_INF_QUEUE_IS_NOT_READY);
	}

	VXGE_HAL_SQ_STATE_SET(sq, VXGEHAL_SQ_STATE_SESS_CLOSE_PENDING);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, session_id);

	__hal_tcp_delete_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    sq->session_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;
	else
		VXGE_HAL_SQ_STATE_SET(sq, prev_state);

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_tcp_abort - Abort a TCP Session
 *
 * @vpath_handle: Virtual path handle
 * @session_id: Object Id of Session
 * @opaque_handle: HAL Opaque handle
 *
 * Aborts a TCP session on the adapter
 */
vxge_hal_status_e
vxge_hal_tcp_abort(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t session_id,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_sq_t *sq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;
	u32 prev_state;

	vxge_assert((vpath_handle != NULL) && (session_id != 0) &&
	    (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "session_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT, (ptr_t)vpath_handle,
	    (ptr_t)session_id, (ptr_t)opaque_handle);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	sq = (__hal_sq_t *)(ptr_t)session_id;

	vxge_assert((dmq != NULL) && (sq != NULL));

	if ((prev_state = VXGE_HAL_SQ_STATE_GET(sq)) !=
	    VXGEHAL_SQ_STATE_SESS_ESTED) {

		vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d",
		    __FILE__, __func__, __LINE__,
		    VXGE_HAL_INF_QUEUE_IS_NOT_READY);
		return (VXGE_HAL_INF_QUEUE_IS_NOT_READY);
	}

	VXGE_HAL_SQ_STATE_SET(sq, VXGEHAL_SQ_STATE_SESS_CLOSE_PENDING);

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, session_id);

	__hal_tcp_abort_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    sq->session_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;
	else
		VXGE_HAL_SQ_STATE_SET(sq, prev_state);

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_tcp_update_rxwin - Set rxwin of a TCP Session
 *
 * @vpath_handle: Virtual path handle
 * @session_id: Object Id of Session
 * @opaque_handle: HAL Opaque handle
 * @rcv_buf: For bytestream sessions: lowest TCP sequence number buffered in
 *		the connection's receive buffer.  This is updated by the host
 *		as data is consumed by the ULP/application.  Must increase
 *		monotonically to avoid window shrinks (either check in PE when
 *		host updates it, or just make it a requirement of the host).
 *		For iWARP sessions: the DDP mainland RHS TCP sequence number.
 *		Represents the last aligned DDP mainland placement.  The DDP
 *		mainland edge differs from the TCP mainland edge (rcv_nxt)
 *		if MPA misaligned frames are arriving (ie session is ""DDP OOO""
 *		and iwarp_align_state != aligned).  This field is only updated
 *		when misaligned.  When aligned, including in the normal fast
 *		path DDP in order case, this is not updated to save fast path
 *		instructions; instead rcv_nxt defines both the TCP and DDP
 *		mainland edge.
 *		Supplied by host for passive open.  Initialized by adapter for
 *              active open on receipt of SYN/ACK or SYN segment.
 *		In both cases the proper initial value is IRS + 1
 *              (same as rcv_nxt).
 *		This is set equal to rcv_nxt on promote to iWARP"
 * @rcv_buf_size: Maximum receive buffer size that the connection is allowed
 *		to consume.  Units are TCP payload bytes (ie sequence number
 *		bytes).
 * @rcv_max_wnd: Maximum value for receive window to be advertised.  This can
 *		be less than rcv_buf_size, so that in cases when the transmit
 *		rate is not limited by the receive application, the advertised
 *		window will remain constant. This is allows less frequent buffer
 *		usage updates from host without causing bursty behaviour from
 *		the sender.
 *
 * Set rxwin of a TCP Session on the adapter
 */
vxge_hal_status_e
vxge_hal_tcp_update_rxwin(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t session_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 rcv_buf,
	u32 rcv_buf_size,
	u32 rcv_max_wnd)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_sq_t *sq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (session_id != 0) &&
	    (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "session_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", rcv_buf = %d, "
	    "rcv_buf_size = %d, rcv_max_wnd = %d",
	    (ptr_t)vpath_handle, (ptr_t)session_id,
	    (ptr_t)opaque_handle, rcv_buf, rcv_buf_size, rcv_max_wnd);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	sq = (__hal_sq_t *)(ptr_t)session_id;

	vxge_assert((dmq != NULL) && (sq != NULL));

	if (VXGE_HAL_SQ_STATE_GET(sq) != VXGEHAL_SQ_STATE_SESS_ESTED) {

		vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d",
		    __FILE__, __func__, __LINE__,
		    VXGE_HAL_INF_QUEUE_IS_NOT_READY);
		return (VXGE_HAL_INF_QUEUE_IS_NOT_READY);
	}

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, session_id);

	__hal_tcp_update_rxwin_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    sq->session_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    rcv_buf,
	    rcv_buf_size,
	    rcv_max_wnd,
	    rcv_buf,
	    rcv_buf_size,
	    rcv_max_wnd);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);


	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_tcp_update_mss - Set MSS of a TCP Session
 *
 * @vpath_handle: Virtual path handle
 * @session_id: Object Id of Session
 * @opaque_handle: HAL Opaque handle
 * @snd_mss: Max segment size for transmit.This is defined as the maximum
 *	     permitted TCP payload size, accounting for the maximum TCP option
 *	     size (40B if both SACK and timestamps are enabled)
 *	     - snd_mss = MSS - max_tcp_options max_tcp_options is 40B
 *	     assuming both SACK and timestamps are enabled (note that the
 *	     firmware might choose to use a fixed value of 40 even if SACK or
 *	     timestamps are not enabled - this is TBD). The definition is
 *	     slightly different from most TCP stacks which store the true MSS
 *	     value (max payload assuming no IP or TCP options).  It allows the
 *	     same amount of payload whether SACK is sent or not - important
 *	     for sending aligned DDP segments.  Most stacks typically pack the
 *	     max possible data into a segment, and then resegment on retransmit
 *	     if the options size has grown. When sending the MSS option on SYN
 *	     or SYN/ACK segments, the PE firmware calculates the true MSS
 *	     value based on this field (ie MSS = snd_mss + max_tcp_options).
 *	     Supplied by host for both active and passive open. In active open
 *	     case, adapter may adjust value to the minimum of the host supplied
 *	     value and the value from the received MSS option.
 * @snd_mss_limit: Value of MSS option received from far end on the
 *	     SYN or SYN/ACK segment. The PE applies this as a limit on snd_mss
 *	     when the host signals path MTU changes as follows -
 *	     snd_mss = min(new_mss_from_host, snd_mss_limit) - max_tcp_options
 *	     Supplied by host for passive open.
 *	     Set by adapter for active open based on the MSS option value on
 *	     the received SYN or SYN/ACK segment.
 *
 * Set MSS of a TCP Session on the adapter
 */
vxge_hal_status_e
vxge_hal_tcp_update_mss(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t session_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 snd_mss,
	u32 snd_mss_limit)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_sq_t *sq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (session_id != 0) &&
	    (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "session_id = 0x"VXGE_OS_STXFMT", opaque_handle = 0x"VXGE_OS_STXFMT
	    ", snd_mss = %d, snd_mss_limit = %d", (ptr_t)vpath_handle,
	    (ptr_t)session_id, (ptr_t)opaque_handle, snd_mss,
	    snd_mss_limit);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	sq = (__hal_sq_t *)(ptr_t)session_id;

	vxge_assert((dmq != NULL) && (sq != NULL));

	if (VXGE_HAL_SQ_STATE_GET(sq) != VXGEHAL_SQ_STATE_SESS_ESTED) {

		vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d",
		    __FILE__, __func__, __LINE__,
		    VXGE_HAL_INF_QUEUE_IS_NOT_READY);
		return (VXGE_HAL_INF_QUEUE_IS_NOT_READY);
	}

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, session_id);

	__hal_tcp_update_mss_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    sq->session_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    snd_mss,
	    snd_mss_limit,
	    snd_mss,
	    snd_mss_limit);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_tcp_update_ip_header - Update IP Header
 *
 * @vpath_handle: Virtual path handle
 * @session_id: Object Id of Session
 * @opaque_handle: HAL Opaque handle
 * @ip_tos: Value to place in the IPv4 TOS field or the Ipv6 Traffic Class
 *		field.  The 2 lower bits are overwritten if ECN is enabled.
 * @ip_flowlabel: Value to place in the IPv6 Flow Label field.  This might
 *		be moved to the H2L context.
 *
 * Updates IP Header of a TCP session on the adapter
 */
vxge_hal_status_e
vxge_hal_tcp_update_ip_header(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t session_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 ip_tos,
	u32 ip_flowlabel)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_sq_t *sq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (session_id != 0) &&
	    (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "session_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", ip_tos = %d, "
	    "ip_flowlabel = %d", (ptr_t)vpath_handle,
	    (ptr_t)session_id, (ptr_t)opaque_handle,
	    ip_tos, ip_flowlabel);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	sq = (__hal_sq_t *)(ptr_t)session_id;

	vxge_assert((dmq != NULL) && (sq != NULL));

	if (VXGE_HAL_SQ_STATE_GET(sq) != VXGEHAL_SQ_STATE_SESS_ESTED) {

		vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d",
		    __FILE__, __func__, __LINE__,
		    VXGE_HAL_INF_QUEUE_IS_NOT_READY);
		return (VXGE_HAL_INF_QUEUE_IS_NOT_READY);
	}

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, session_id);

	__hal_tcp_update_ip_header_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    sq->session_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    ip_tos,
	    ip_flowlabel,
	    ip_tos,
	    ip_flowlabel);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d",
	    __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_tcp_update_keepalive - Update Keepalive of TCP Session
 *
 * @vpath_handle: Virtual path handle
 * @session_id: Object Id of Session
 * @opaque_handle: HAL Opaque handle
 * @kpa_ok: Enable for keepalive probes for this connection.
 * @probes_max: Keepalive maximum probe count: maximum number of unanswered
 *		keepalive probes before giving up on session.
 * @idle_time: Keepalive idle timer: the idle time before sending the first
 *		keepalive probe (default is 2 hours).
 * @retry_time: Keepalive retry time interval: idle time before resending an
 *		unanswered keepalive probe.
 *
 * Updates Keepalive of TCP Session on adapter
 */
vxge_hal_status_e
vxge_hal_tcp_update_keepalive(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t session_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 kpa_ok,
	u32 probes_max,
	u32 idle_time,
	u32 retry_time)
{
	__hal_vpath_handle_t *vp = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_sq_t *sq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (session_id != 0) &&
	    (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "session_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", kpa_ok = %d, "
	    "probes_max = %d, idle_time = %d, retry_time = %d",
	    (ptr_t)vpath_handle, (ptr_t)session_id,
	    (ptr_t)opaque_handle, kpa_ok, probes_max, idle_time,
	    retry_time);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	sq = (__hal_sq_t *)(ptr_t)session_id;

	vxge_assert((dmq != NULL) && (sq != NULL));

	if (VXGE_HAL_SQ_STATE_GET(sq) != VXGEHAL_SQ_STATE_SESS_ESTED) {

		vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d",
		    __FILE__, __func__, __LINE__,
		    VXGE_HAL_INF_QUEUE_IS_NOT_READY);
		return (VXGE_HAL_INF_QUEUE_IS_NOT_READY);
	}

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, session_id);

	__hal_tcp_update_keepalive_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    sq->session_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    kpa_ok,
	    probes_max,
	    idle_time,
	    retry_time);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);
	return (status);
}

/*
 * vxge_hal_tcp_relink_to_nce - Rlink a NCE to TCP Session
 *
 * @vpath_handle: Virtual path handle
 * @session_id: Object Id of Session
 * @opaque_handle: HAL Opaque handle
 * @nce_id: NCE Id.
 *
 * Rlinks a NCE to TCP Session on the adapter
 */
vxge_hal_status_e
vxge_hal_tcp_relink_to_nce(
	vxge_hal_vpath_h vpath_handle,
	vxge_hal_obj_id_t session_id,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 nce_id)
{
	__hal_vpath_handle_t *vp  = (__hal_vpath_handle_t *)vpath_handle;
	__hal_device_t *hldev;
	__hal_dmq_t *dmq;
	__hal_sq_t *sq;
	__hal_down_msg_t dmsg;
	vxge_hal_status_e status;

	vxge_assert((vpath_handle != NULL) && (session_id != 0) &&
	    (opaque_handle != NULL));

	hldev = (__hal_device_t *)vp->vpath->hldev;

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "==>  %s:%s:%d", __FILE__, __func__, __LINE__);

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "vpath_handle = 0x"VXGE_OS_STXFMT", "
	    "session_id = 0x"VXGE_OS_STXFMT", "
	    "opaque_handle = 0x"VXGE_OS_STXFMT", nce_id = %d",
	    (ptr_t)vpath_handle, (ptr_t)session_id,
	    (ptr_t)opaque_handle, nce_id);

	dmq = (__hal_dmq_t *)vp->vpath->dmqh;

	sq = (__hal_sq_t *)(ptr_t)session_id;

	vxge_assert((dmq != NULL) && (sq != NULL));

	if (VXGE_HAL_SQ_STATE_GET(sq) != VXGEHAL_SQ_STATE_SESS_ESTED) {

		vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
		    "<==  %s:%s:%d  Result: %d",
		    __FILE__, __func__, __LINE__,
		    VXGE_HAL_INF_QUEUE_IS_NOT_READY);
		return (VXGE_HAL_INF_QUEUE_IS_NOT_READY);
	}

	VXGE_HAL_OPAQUE_HANDLE_VPATH_HANDLE(opaque_handle, vpath_handle);
	VXGE_HAL_OPAQUE_HANDLE_HAL_PRIV(opaque_handle, session_id);

	__hal_tcp_relink_to_nce_req_dmsg_set(
	    (vxge_hal_down_msg_h)&dmsg,
	    sq->session_id,
	    dmq->intr_ctrl,
	    dmq->gen_compl,
	    opaque_handle,
	    nce_id);

	status =  __hal_dmq_post(dmq, (vxge_hal_down_msg_h)&dmsg);

	if (status == VXGE_HAL_OK)
		status = VXGE_HAL_PENDING;

	vxge_hal_trace_log_tcp(hldev, vp->vpath->vp_id,
	    "<==  %s:%s:%d  Result: %d", __FILE__, __func__, __LINE__, status);
	return (status);
}
