/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */
/*
 * Copyright (c) 2002-2010 Neterion, Inc.
 * All right Reserved.
 *
 * FileName :   vxgehal-umq.h
 *
 * Description:  HAL Up Message Queue list structures
 *
 * Created:       15 Dec 2005
 */

#ifndef	VXGE_HAL_UMQ_H
#define	VXGE_HAL_UMQ_H

#define	VXGE_HAL_UMQ_MSG_BUFFER_LENGTH	240
#define	VXGE_HAL_UMQ_MAX_MESSAGES	1024

__EXTERN_BEGIN_DECLS

/*
 * struct __hal_umq_blk_priv_t - Up Message Block HAL-private data.
 * @dma_addr: DMA (mapped) address of _this_ Message block.
 * @dma_handle: DMA handle used to map the Message block to the device.
 * @dma_offset: Message block's offset in the memory block(always 0).
 * @memblock: Virtual address of the message block.
 * @wrapper: Address of the List Block wrapper for this list.
 * @channel: Channel identifier for this structure.
 * @compl_offset: Number of bytes from the beginning of the block processed.
 * @post_offset: Number of bytes used from the beginning of the block for the
 *		 messages so far.
 * @pad: Padding to make it cache-aligned.
 *
 * Per-Message block HAL-private data. HAL uses the space to keep DMA
 * information associated with the Message block.
 *
 */
typedef struct __hal_umq_blk_priv_t {
	dma_addr_t			dma_addr;
	pci_dma_h			dma_handle;
	ptrdiff_t			dma_offset;
	void				*memblock;
	__hal_list_block_wrapper_t	*wrapper;
	vxge_hal_umq_h			umqh;
	u64				compl_offset;
	u64				post_offset;
	u8				pad[(((63+__vxge_os_cacheline_size)/
						__vxge_os_cacheline_size)*
						__vxge_os_cacheline_size)-64];
} __hal_umq_blk_priv_t;

/*
 * struct __hal_umq_t - Upward Message Queue channel.
 * @channel: Channel "base" of this Queue, the common part of all HAL
 *	  channels.
 * @mempool: Memory pool, the pool from which message blocks get allocated.
 * @free_msg_list: List of free bocks to indicate messages to clients.
 * @alloc_msg_list: List of blocks used to indicate messages to clients.
 * @first_block_address: The DMA Address of the first block.
 * @first_block_length: The Length of the first block.
 * @umq_id: The object id of the UMQ.
 * @umq_length: Length of the Message queue.
 * @bytes_processed: The total number of bytes processed.
 * @intr_ctrl: Enable/Disable Interrupt generation.
 * @gen_compl: Enable/Disable completion message.
 * @msgblock_priv_size: Reserved at the end of each list block for HAL internal
 *			 usage.
 * @stats: Statistics of this umq
 *
 * Upward Message Queue or Message Queue channel.
 */
typedef struct __hal_umq_t {
	__hal_channel_t				channel;
	vxge_hal_mempool_t			*mempool;
	vxge_list_t				free_msg_list;
	vxge_list_t				alloc_msg_list;
	u64					first_block_address;
	u32					first_block_length;
	u32					umq_id;
	u32					umq_length;
	u32 __vxge_os_attr_cacheline_aligned	bytes_processed;
	u32					interrupt_control;
	u32					intr_ctrl;
	u32					gen_compl;
	u32					msgblock_priv_size;
	vxge_hal_vpath_stats_sw_umq_info_t	*stats;
} __hal_umq_t;

u64
__hal_umq_first_block_address_get(
	vxge_hal_umq_h umqh);

u32
__hal_umq_first_block_length_get(
	vxge_hal_umq_h umqh);

vxge_hal_status_e
__hal_umq_create(
	vxge_hal_vpath_h vpath_handle);

vxge_hal_status_e
__hal_umq_reset(
	vxge_hal_umq_h umqh);

vxge_hal_status_e
__hal_umq_delete(
	vxge_hal_vpath_h vpath_handle,
	__hal_umq_t *umq);

vxge_hal_status_e
__hal_umq_complete(
	__hal_umq_t *umq,
	vxge_hal_up_msg_h umsgh);

vxge_hal_status_e
__hal_umq_ready_msg_check(
	__hal_umq_t *umq);

vxge_hal_status_e
__hal_umq_next_ready_msg_get(
	__hal_umq_t *umq,
	vxge_hal_up_msg_h *umsgh);

vxge_hal_status_e
__hal_umq_is_empty(
	__hal_umq_t *umq);

vxge_hal_status_e
__hal_process_umq(
	__hal_device_t *hldev,
	u32 vp_id);

vxge_hal_status_e
__hal_umq_poll(
	__hal_umq_t *umq,
	u8 *buffer,
	u32 length);

__EXTERN_END_DECLS

#endif /* VXGE_HAL_UMQ_H */
