/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * Copyright (c) 2002-2010 Neterion, Inc.
 * All right Reserved.
 *
 * FileName :   vxgehal-message.c
 *
 * Description:  HAL Message functionality
 *
 * Created:       14 December 2005
 */
#include "vxgehal.h"

#ifdef	VXGE_HAL_INTERNAL_COMPILER_ERROR

#pragma optimize("", off)

#endif

/*
 * __hal_down_msg_header_set - Set the header fields of dmsg
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @msg_type: Message Type
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @length: Length of message
 *
 * This function is used to set the header fields of dmsg.
 */
void
__hal_down_msg_header_set(vxge_hal_down_msg_h dmsgh, u32 msg_id,
			u32 msg_type, u32 intr_ctrl, u32 compl_umq,
			vxge_hal_opaque_handle_t *opaque_handle, u32 length)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	dmsg->control_0 = vxge_os_htonll(
	    VXGE_HAL_DOWN_MSG_ID_NUMBER(msg_id) |
	    VXGE_HAL_DOWN_MSG_MESSAGE_TYPE(msg_type) |
	    VXGE_HAL_DOWN_MSG_INTERRUPT_CONTROL(intr_ctrl) |
	    VXGE_HAL_DOWN_MSG_GEN_COMPL_UMQ(compl_umq));

	dmsg->control_1 = vxge_os_htonll((u64)(ptr_t)opaque_handle);

	dmsg->control_2 = vxge_os_htonll(
	    VXGE_HAL_DOWN_MSG_PROTOCOL_VERSION(0) |
	    VXGE_HAL_DOWN_MSG_LENGTH(length));

}

/*
 * vxge_hal_up_msg_header_get - Get the header fields of umsg
 * @umsgh: Handle to umsg.
 * @msg_id: Buffer to return message Id
 * @msg_type: Buffer to return message Type
 * @compl_dmq: Buffer to return Indication that a DMQ message from the host is
 *		required upon completion of message processing.
 * @opaque_handle: Buffer to return opaque handle
 * @length: Buffer to return length of message
 * @result: Buffer to return the result code of the operation requested in
 *		corresponding DMQ message.
 *
 * This function is used to get the header fields of umsg.
 */
void
vxge_hal_up_msg_header_get(vxge_hal_up_msg_h umsgh, u32 *msg_id, u32 *msg_type,
		u32 *compl_dmq, vxge_hal_opaque_handle_t **opaque_handle,
		u32 *length, u32 *result)
{
	u64 val64;
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (msg_id != NULL) && (msg_type != NULL) &&
	    (compl_dmq != NULL) && (opaque_handle != NULL) &&
	    (length != NULL) && (result != NULL));

	val64 = vxge_os_ntohll(umsg->control_0);

	*msg_id = (u32)VXGE_HAL_UP_MSG_GET_ID_NUMBER(val64);
	*msg_type = (u32)VXGE_HAL_UP_MSG_GET_MESSAGE_TYPE(val64);
	*compl_dmq = (u32)VXGE_HAL_UP_MSG_GET_GEN_COMPL_DMQ(val64);

	*opaque_handle = (vxge_hal_opaque_handle_t *)(ptr_t)
	    vxge_os_ntohll(umsg->control_1);

	val64 = vxge_os_ntohll(umsg->control_2);

	*length = (u32)VXGE_HAL_UP_MSG_GET_LENGTH(val64);
	*result = (u32)VXGE_HAL_UP_MSG_GET_RESULT(val64);
}

/* =========================== NCE MESSAGES ============================ */
/*
 * __hal_nce_create_req_dmsg_set - Set NCE Create Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @nce_opaque_handle: NCE opaque handle
 * @src_mac_addr: Source Mac Address
 * @rem_mac_addr: Remote Mac Address
 * @vlan_id: Valn Id
 * @host_rch_time: Host Reach Time
 *
 * Sets the NCE Create Request DMQ message used by the host to create an NCE
 * on the adapter.
 */
void
__hal_nce_create_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	vxge_hal_opaque_handle_t *nce_opaque_handle,
	u64 src_mac_addr,
	u64 rem_mac_addr,
	u32 vlan_id,
	u32 host_rch_time)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_NCE_CREATE_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_nce_create_req_dmq_t));

	dmsg->m.nce_create.nce_opaque_handle =
	    vxge_os_htonll((u64)(ptr_t)nce_opaque_handle);

	dmsg->m.nce_create.control_0 =
	    vxge_os_htonll(VXGE_HAL_NCE_CREATE_REQ_SRC_MAC(src_mac_addr));

	dmsg->m.nce_create.control_1 =
	    vxge_os_htonll(VXGE_HAL_NCE_CREATE_REQ_RMT_MAC(rem_mac_addr));

	dmsg->m.nce_create.control_2 =
	    vxge_os_htonll(VXGE_HAL_NCE_CREATE_REQ_VLAN_ID(vlan_id) |
	    VXGE_HAL_NCE_CREATE_REQ_RCH_TIME(host_rch_time));
}

/*
 * __hal_nce_delete_req_dmsg_set - Set NCE Delete Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 *
 * Sets the NCE Delete Request DMQ message used by the host to delete an NCE
 * on the adapter.
 */
void
__hal_nce_delete_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_NCE_DELETE_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE);

}

/*
 * vxge_hal_nce_delete_resp_umsg_get - Get values from NCE Delete Response
 *					Up Message
 *
 * @umsgh: Handle to umsg.
 * @nce_opaque_handle: Buffer to return NCE Opaque handle
 *
 * Gets the values from NCE Delete Response Up Message.
 */
void
vxge_hal_nce_delete_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	vxge_hal_opaque_handle_t **nce_opaque_handle)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (nce_opaque_handle != NULL));

	*nce_opaque_handle = (vxge_hal_opaque_handle_t *)(ptr_t)
	    vxge_os_ntohll(umsg->m.nce_delete.nce_opaque_handle);
}

/*
 * __hal_nce_update_mac_addr_req_dmsg_set - Set NCE Update MAC Request DMQ
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @mac_addr: Mac Address
 *
 * Set the message to modify the MAC address for an NCE
 * on the adapter.
 */
void
__hal_nce_update_mac_addr_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u64 mac_addr)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_NCE_UPDATE_MAC_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_nce_update_mac_addr_req_dmq_t));

	dmsg->m.nce_update_mac.control_0 =
	    vxge_os_htonll(VXGE_HAL_NCE_UPDATE_MAC_REQ_MAC_ADDR(mac_addr));
}

/*
 * __hal_nce_update_rch_time_req_dmsg_set - Set NCE Update RCH Time Request DMQ
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @rch_time: Host reachability time
 *
 * Set the message to modify the RCH Time for an NCE on the adapter.
 */
void
__hal_nce_update_rch_time_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u64 rch_time)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_NCE_UPDATE_RCH_TIME_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_nce_update_rch_time_req_dmq_t));

	dmsg->m.nce_update_rch.control_0 =
	    vxge_os_htonll(VXGE_HAL_NCE_UPDATE_RCH_REQ_RCH_TIME(rch_time));
}

/*
 * __hal_nce_query_req_dmsg_set - Set NCE Query Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 *
 * Sets the NCE Query Request DMQ message used by the host to query a NCE
 * on the adapter.
 */
void
__hal_nce_query_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_NCE_QUERY_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE);
}

/*
 * vxge_hal_nce_query_resp_umsg_get - Get values from NCE Query Response
 *
 * @umsgh: Handle to umsg.
 * @nce_opaque_handle: Buffer to return NCE opaque handle
 * @src_mac_addr: Buffer to return Source Mac Address
 * @rem_mac_addr: Buffer to return Remote Mac Address
 * @host_rch_time: Buffer to return Host Reachability Time
 * @adapter_rch_time: Buffer to return Adapter Reachability Time
 * @valid_flag: Buffer to return valid flag for NCE
 * @vlan_id: Buffer to return Valn Id
 *
 * Gets the values from NCE Query Response Up Message.
 */
void
vxge_hal_nce_query_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	vxge_hal_opaque_handle_t **nce_opaque_handle,
	u64 *src_mac_addr,
	u64 *rem_mac_addr,
	u32 *host_rch_time,
	u32 *adapter_rch_time,
	u32 *valid_flag,
	u32 *vlan_id)
{
	u64 val64;
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (nce_opaque_handle != NULL) &&
	    (src_mac_addr != NULL) && (rem_mac_addr != NULL) &&
	    (host_rch_time != NULL) && (adapter_rch_time != NULL) &&
	    (valid_flag != NULL) && (vlan_id != NULL));

	*nce_opaque_handle = (vxge_hal_opaque_handle_t *)(ptr_t)
	    vxge_os_ntohll(umsg->m.nce_query.nce_opaque_handle);

	*src_mac_addr = (u64)VXGE_HAL_NCE_QUERY_RESP_GET_SRC_MAC(
	    vxge_os_ntohll(umsg->m.nce_query.control_0));

	*rem_mac_addr = (u64)VXGE_HAL_NCE_QUERY_RESP_GET_RMT_MAC(
	    vxge_os_ntohll(umsg->m.nce_query.control_1));

	val64 = vxge_os_ntohll(umsg->m.nce_query.control_2);

	*host_rch_time = (u32)VXGE_HAL_NCE_QUERY_RESP_GET_HOST_RCH_TIME(val64);

	*adapter_rch_time =
	    (u32)VXGE_HAL_NCE_QUERY_RESP_GET_ADPT_RCH_TIME(val64);

	val64 = vxge_os_ntohll(umsg->m.nce_query.control_3);

	*valid_flag = (u32)VXGE_HAL_NCE_QUERY_RESP_GET_VALID_FLAG(val64);

	*vlan_id = (u32)VXGE_HAL_NCE_QUERY_RESP_GET_VLAN_ID(val64);
}

/* ============================ CQRQ MESSAGES =============================== */

/*
 * __hal_cqrq_create_req_dmsg_set - Set CQRQ Create Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @cq_opaque_handle: CQ Opaque Handle
 * @interrupt_number: Interrupt Number used
 * @bitmap_id: Bitmap position
 * @max_cqe_groups: Maximum number of CQE groups
 * @refill_threshold_high: Upperbound for automatic adapter refill of CQEs
 * @refill_threshold_low: Lowerbound for automatic adapter refill of CQEs
 * @low_threshold_notify: Generate Interrupt or notification if number of CQEs
 *		falls below the Lower Threshhold
 * @event_notification_mask: Event notifications mask
 * @cqe_period: CQE size
 * @cqrq_ir: This is CQRQ IR (Indication Register) which is the host system
 *		address to which this CQRQ reports the number of returned CQEs.
 *
 * Sets the CQRQ Create Request DMQ message used by the host to create an CQRQ
 * on the adapter.
 */
void
__hal_cqrq_create_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	vxge_hal_opaque_handle_t *cq_opaque_handle,
	u32 interrupt_number,
	u32 bitmap_id,
	u32 max_cqe_groups,
	u32 refill_threshold_high,
	u32 refill_threshold_low,
	u32 low_threshold_notify,
	u32 event_notification_mask,
	u32 cqe_period,
	u64 cqrq_ir)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_CQRQ_CREATE_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_cqrq_create_req_dmq_t));

	dmsg->m.cqrq_create.cq_opaque_handle = (u64)(ptr_t)cq_opaque_handle;

	dmsg->m.cqrq_create.control_0 = vxge_os_htonll(
	    VXGE_HAL_CQRQ_CREATE_REQ_INT_NUM(interrupt_number) |
	    VXGE_HAL_CQRQ_CREATE_REQ_BITMAP_ID(bitmap_id) |
	    VXGE_HAL_CQRQ_CREATE_REQ_MAX_CQE_GRPS(max_cqe_groups) |
	    VXGE_HAL_CQRQ_CREATE_REQ_REF_TH_HIGH(refill_threshold_high));

	dmsg->m.cqrq_create.control_1 = vxge_os_htonll(
	    VXGE_HAL_CQRQ_CREATE_REQ_REF_TH_LOW(refill_threshold_low) |
	    VXGE_HAL_CQRQ_CREATE_REQ_LOW_TH_NOTIF(low_threshold_notify) |
	    VXGE_HAL_CQRQ_CREATE_REQ_EVENT_MASK(event_notification_mask) |
	    VXGE_HAL_CQRQ_CREATE_REQ_CQE_PERIOD(cqe_period));

	dmsg->m.cqrq_create.cqrq_ir = vxge_os_htonll(cqrq_ir);
}

/*
 * __hal_cqrq_delete_req_dmsg_set - Set CQRQ Delete Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 *
 * Sets the CQRQ Delete Request DMQ message used by the host to delete an CQRQ
 * on the adapter.
 */
void
__hal_cqrq_delete_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_CQRQ_DELETE_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE);

}

/*
 * vxge_hal_cqrq_delete_resp_umsg_get - Get values from CQRQ Delete Response
 *				Up Message
 *
 * @umsgh: Handle to umsg.
 * @cq_opaque_handle: Buffer to return CQ Opaque handle
 *
 * Gets the values from CQRQ Delete Response Up Message.
 */
void
vxge_hal_cqrq_delete_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	vxge_hal_opaque_handle_t **cq_opaque_handle)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (cq_opaque_handle != NULL));

	*cq_opaque_handle = (vxge_hal_opaque_handle_t *)(ptr_t)
	    vxge_os_ntohll(umsg->m.cqrq_delete.cq_opaque_handle);
}

/*
 * __hal_cqrq_modify_req_dmsg_set - Set CQRQ Modify Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @max_cqe_groups: Maximum number of CQE groups
 * @refill_threshold_high: Upperbound for automatic adapter refill of CQEs
 * @refill_threshold_low: Lowerbound for automatic adapter refill of CQEs
 * @low_threshold_notify: Generate Interrupt or notification if number of CQEs
 *		falls below the Lower Threshhold
 * @event_notification_mask: Event notifications mask
 *
 * Sets the CQRQ Modify Request DMQ message used by the host to modify a CQRQ
 * on the adapter.
 */
void
__hal_cqrq_modify_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 max_cqe_groups,
	u32 refill_threshold_high,
	u32 refill_threshold_low,
	u32 low_threshold_notify,
	u32 event_notification_mask)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_CQRQ_MODIFY_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_cqrq_modify_req_dmq_t));

	dmsg->m.cqrq_modify.control_0 = vxge_os_htonll(
	    VXGE_HAL_CQRQ_MODIFY_REQ_MAX_CQE_GRPS(max_cqe_groups) |
	    VXGE_HAL_CQRQ_MODIFY_REQ_REF_TH_HIGH(refill_threshold_high) |
	    VXGE_HAL_CQRQ_MODIFY_REQ_REF_TH_LOW(refill_threshold_low) |
	    VXGE_HAL_CQRQ_MODIFY_REQ_LOW_TH_NOTIF(low_threshold_notify));

	dmsg->m.cqrq_modify.control_1 = vxge_os_htonll(
	    VXGE_HAL_CQRQ_MODIFY_REQ_EVENT_MASK(event_notification_mask));
}

/*
 * __hal_cqrq_query_req_dmsg_set - Set CQRQ Query Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 *
 * Sets the CQRQ Query Request DMQ message used by the host to query an CQRQ
 * on the adapter.
 */
void
__hal_cqrq_query_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_CQRQ_QUERY_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE);
}

/*
 * vxge_hal_cqrq_query_resp_umsg_get - Get the values from CQRQ Query Response
 *				    Up Message
 *
 * @umsgh: Handle to umsg.
 * @cqrq_info: Pointer to structure to return values from CQRQ query response
 *
 * Gets the values from CQRQ Query Response Up Message.
 */
void
vxge_hal_cqrq_query_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	vxge_hal_cqrq_info_t *cqrq_info)
{
	u64 val64;
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (cqrq_info != NULL));

	cqrq_info->cq_opaque_handle = vxge_os_ntohll(
	    umsg->m.cqrq_query.cq_opaque_handle);

	val64 = vxge_os_ntohll(umsg->m.cqrq_query.control_0);

	cqrq_info->cqe_period =
	    (u32)VXGE_HAL_CQRQ_QUERY_RESP_GET_CQE_PERIOD(val64);

	cqrq_info->bitmap_id =
	    (u32)VXGE_HAL_CQRQ_QUERY_RESP_GET_BITMAP_ID(val64);

	cqrq_info->max_cqe_groups =
	    (u32)VXGE_HAL_CQRQ_QUERY_RESP_GET_MAX_CQE_GRPS(val64);

	cqrq_info->refill_threshold_high =
	    (u32)VXGE_HAL_CQRQ_QUERY_RESP_GET_REF_TH_HIGH(val64);

	val64 = vxge_os_ntohll(umsg->m.cqrq_query.control_1);

	cqrq_info->low_threshold_notify =
	    (u32)VXGE_HAL_CQRQ_QUERY_RESP_GET_REF_TH_LOW(val64);

	cqrq_info->low_threshold_notify =
	    (u32)VXGE_HAL_CQRQ_QUERY_RESP_GET_LOW_TH_NOTIF(val64);

	cqrq_info->umq = (u32)VXGE_HAL_CQRQ_QUERY_RESP_GET_UMQ(val64);

	cqrq_info->event_notification_mask =
	    (u32)VXGE_HAL_CQRQ_QUERY_RESP_GET_EVENT_MASK(val64);

	cqrq_info->interrupt_number =
	    (u32)VXGE_HAL_CQRQ_QUERY_RESP_GET_INT_NUM(val64);

	cqrq_info->enable =
	    (u32)VXGE_HAL_CQRQ_QUERY_RESP_GET_ENABLE(val64);

	cqrq_info->tombstoned =
	    (u32)VXGE_HAL_CQRQ_QUERY_RESP_GET_TOMBSTONED(val64);

	cqrq_info->eol =
	    (u32)VXGE_HAL_CQRQ_QUERY_RESP_GET_EOL(val64);

	cqrq_info->drained =
	    (u32)VXGE_HAL_CQRQ_QUERY_RESP_GET_DRAINED(val64);

	cqrq_info->refill =
	    (u32)VXGE_HAL_CQRQ_QUERY_RESP_GET_REFILL(val64);

	cqrq_info->cqe_block_wrapper_hsa = vxge_os_ntohll(
	    umsg->m.cqrq_query.cqe_block_wrapper_hsa);

	val64 = vxge_os_ntohll(umsg->m.cqrq_query.control_2);

	cqrq_info->cqe_block_wrapper_num_bytes =
	    (u32)VXGE_HAL_CQRQ_QUERY_RESP_GET_CBW_BYTES(val64);

	cqrq_info->cq_indication_cqes_reported =
	    (u32)VXGE_HAL_CQRQ_QUERY_RESP_GET_CBW_CQES(val64);

	cqrq_info->cqrq_ir = vxge_os_ntohll(umsg->m.cqrq_query.cqrq_ir);
}

/*
 * __hal_cqrq_arm_req_dmsg_set - Set CQRQ Arm Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @arm_type: The armed state of the CQRQ.
 *
 * Sets the CQRQ Arm Request DMQ message used by the host to changed the armed
 * state of CQRQ on the adapter.
 */
void
__hal_cqrq_arm_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 arm_type)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_CQRQ_ARM_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_cqrq_arm_req_dmq_t));

	dmsg->m.cqrq_arm.control_0 = vxge_os_htonll(
	    VXGE_HAL_CQRQ_ARM_REQ_ARM_TYPE(arm_type));
}

/*
 * vxge_hal_cqrq_event_notification_umsg_get - Get values from CQRQ Event
 *						Notification Response
 *
 * @umsgh: Handle to umsg.
 * @event_type : Buffer to return Event Type
 *
 * Gets the values from CQRQ Event Notification Up Message.
 */
void
vxge_hal_cqrq_event_notification_umsg_get(
	vxge_hal_up_msg_h umsgh,
	u32 *event_type)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (event_type != NULL));

	*event_type = (u32)VXGE_HAL_CQRQ_EVENT_NOTIF_GET_EVENT_TYPE(
	    vxge_os_ntohll(umsg->m.cqrq_event_notif.control_0));
}

/*
 * __hal_cqrq_fcqebw_req_dmsg_set - Set CQRQ CQRQ First CQE BW Notification
 *				   Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @cqe_block_wrapper_hsa: Address of the first block wrapper in host memory.
 * @cqe_block_wrapper_bytes: Length of the first block wrapper.
 *
 * Sets the CQRQ CQRQ First CQE BW Notification Request DMQ message used by the
 * host to notify the First CQE Block to the the adapter.
 */
void
__hal_cqrq_fcqebw_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u64 cqe_block_wrapper_hsa,
	u32 cqe_block_wrapper_bytes)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_CQRQ_FIRST_CQE_BW_NOTIF_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_cqrq_fcqebw_req_dmq_t));

	dmsg->m.cqrq_fcqebw.cqe_block_wrapper_hsa =
	    vxge_os_htonll(cqe_block_wrapper_hsa);

	dmsg->m.cqrq_fcqebw.control_0 = vxge_os_htonll(
	    VXGE_HAL_CQRQ_FCQEBW_LENGTH(cqe_block_wrapper_bytes));
}

/* ========================= SRQ MESSAGES ============================= */

/*
 * __hal_srq_create_req_dmsg_set - Set SRQ Create Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @srq_opaque_handle: An opaque handle used by the host to reference data
 *		structures in host memory
 * @max_wqe_od_groups: Maximum number of WQE Headers/OD Groups that this S-RQ
 *		can own at any one time
 * @refill_threshold_high: Upperbound for automatic adapter refill of WQEs
 * @refill_threshold_low: Lowerbound for automatic adapter refill of WQEs
 * @low_threshold_notify: Generate Interrupt or notification if number of WQEs
 *		falls below the Lower Threshhold
 * @lro_bytestream: Indicates whether the S-RQ is being used for LRO or
 *		Bytestream/iWarp.
 * @poll_doorbell: This slects the mode for the EOL operation of this S-RQ.
 *		If set to 1'b1 the mode is poll and WQEs are polling will be
 *		done if the EOL is reached. If set to 1'b0 then the mode is
 *	    "Directed Reading" where a user doorbell message is used to
 *		indicate that more WQEs have been added to the end of the queue
 * @srq_arm: Arm the SRQ Limit
 *		1 - Arm the SRQ limit
 *		0 - Disarm the SRQ limit
 * @event_notification_mask: Event notifications mask
 * @srq_limit_value: If the number of WQEs in the (S-)RQ (both on-adapter and
 *		on-host) falls below this value, an (S-)RQ Limit violation is
 *		asserted - if armed
 * @interrupt_number: Interrupt Number used
 * @bitmap_id: Bitmap Identifier of the bit used
 * @srq_ir: The SRQ indication register.
 *
 * Sets the CQRQ Create Request DMQ message used by the host to create an CQRQ
 * on the adapter.
 */
void
__hal_srq_create_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	vxge_hal_opaque_handle_t *srq_opaque_handle,
	u32 max_wqe_od_groups,
	u32 refill_threshold_high,
	u32 refill_threshold_low,
	u32 low_threshold_notify,
	u32 lro_bytestream,
	u32 poll_doorbell,
	u32 srq_arm,
	u32 event_notification_mask,
	u32 srq_limit_value,
	u32 interrupt_number,
	u32 bitmap_id,
	u64 srq_ir)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SRQ_CREATE_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_srq_create_req_dmq_t));

	dmsg->m.srq_create.srq_opaque_handle = vxge_os_htonll(
	    (u64)(ptr_t)srq_opaque_handle);

	dmsg->m.srq_create.control_0 = vxge_os_htonll(
	    VXGE_HAL_SRQ_CREATE_REQ_MAX_WQE_OD_GRPS(max_wqe_od_groups) |
	    VXGE_HAL_SRQ_CREATE_REQ_REF_TH_HIGH(refill_threshold_high) |
	    VXGE_HAL_SRQ_CREATE_REQ_REF_TH_LOW(refill_threshold_low) |
	    VXGE_HAL_SRQ_CREATE_REQ_LOW_TH_NOTIF(low_threshold_notify));

	dmsg->m.srq_create.control_1 =
	    VXGE_HAL_SRQ_CREATE_REQ_EVENT_MASK(event_notification_mask) |
	    VXGE_HAL_SRQ_CREATE_REQ_SRQ_LIMIT(srq_limit_value);

	if (lro_bytestream)
		dmsg->m.srq_create.control_1 |=
		    VXGE_HAL_SRQ_CREATE_REQ_LRO_BS;

	if (poll_doorbell)
		dmsg->m.srq_create.control_1 |=
		    VXGE_HAL_SRQ_CREATE_REQ_POLL_DOORBELL;

	if (srq_arm)
		dmsg->m.srq_create.control_1 |=
		    VXGE_HAL_SRQ_CREATE_REQ_SRQ_ARM;

	dmsg->m.srq_create.control_1 =
	    vxge_os_htonll(dmsg->m.srq_create.control_1);

	dmsg->m.srq_create.control_2 = vxge_os_htonll(
	    VXGE_HAL_SRQ_CREATE_REQ_INT_NUM(interrupt_number) |
	    VXGE_HAL_SRQ_CREATE_REQ_BITMAP_ID(bitmap_id));

	dmsg->m.srq_create.srq_ir = vxge_os_htonll(srq_ir);
}

/*
 * __hal_srq_delete_req_dmsg_set - Set SRQ Delete Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 *
 * Sets the SRQ Delete Request DMQ message used by the host to delete an SRQ
 * on the adapter.
 */
void
__hal_srq_delete_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SRQ_DELETE_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE);
}

/*
 * vxge_hal_srq_delete_resp_umsg_get - Get values from SRQ Delete Response
 *			Up Message
 *
 * @umsgh: Handle to umsg.
 * @srq_opaque_handle: Buffer to return SRQ Opaque handle
 *
 * Gets the values from SRQ Delete Response Up Message.
 */
void
vxge_hal_srq_delete_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	vxge_hal_opaque_handle_t **srq_opaque_handle)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (srq_opaque_handle != NULL));

	*srq_opaque_handle = (vxge_hal_opaque_handle_t *)(ptr_t)
	    vxge_os_ntohll(umsg->m.srq_delete.srq_opaque_handle);
}

/*
 * __hal_srq_modify_req_dmsg_set - Set SRQ Modify Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @max_wqe_od_groups: Maximum number of OD groups
 * @refill_threshold_high: Upperbound for automatic adapter refill of WQEs
 * @refill_threshold_low: Lowerbound for automatic adapter refill of WQEs
 * @low_threshold_notify: Generate Interrupt or notification if number of WQEs
 *		falls below the Lower Threshhold
 * @srq_arm: Arm the SRQ Limit
 *		1 - Arm the SRQ limit
 *		0 - Disarm the SRQ limit
 * @event_notification_mask: Event notifications mask
 * @srq_limit_value: number of WQEs in the (S-)RQ (both on-adapter and on-host)
 *		falls below this value, an (S-)RQ Limit violation is asserted
 *		- if armed
 *
 * Sets the SRQ Modify Request DMQ message used by the host to modify a SRQ
 * on the adapter.
 */
void
__hal_srq_modify_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 max_wqe_od_groups,
	u32 refill_threshold_high,
	u32 refill_threshold_low,
	u32 low_threshold_notify,
	u32 srq_arm,
	u32 event_notification_mask,
	u32 srq_limit_value)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SRQ_MODIFY_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_srq_modify_req_dmq_t));

	dmsg->m.srq_modify.control_0 = vxge_os_htonll(
	    VXGE_HAL_SRQ_MODIFY_REQ_MAX_WQE_OD_GRPS(max_wqe_od_groups) |
	    VXGE_HAL_SRQ_MODIFY_REQ_REF_TH_HIGH(refill_threshold_high)|
	    VXGE_HAL_SRQ_MODIFY_REQ_REF_TH_LOW(refill_threshold_low) |
	    VXGE_HAL_SRQ_MODIFY_REQ_LOW_TH_NOTIF(low_threshold_notify));

	dmsg->m.srq_modify.control_1 =
	    VXGE_HAL_SRQ_MODIFY_REQ_EVENT_MASK(event_notification_mask) |
	    VXGE_HAL_SRQ_MODIFY_REQ_SRQ_LIMIT(srq_limit_value);

	if (srq_arm)
		dmsg->m.srq_modify.control_1 |= VXGE_HAL_SRQ_MODIFY_REQ_SRQ_ARM;

	dmsg->m.srq_modify.control_1 =
	    vxge_os_htonll(dmsg->m.srq_modify.control_1);
}

/*
 * __hal_srq_query_req_dmsg_set - Set SRQ Query Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 *
 * Sets the SRQ Query Request DMQ message used by the host to query an SRQ
 * on the adapter.
 */
void
__hal_srq_query_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SRQ_QUERY_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE);
}

/*
 * vxge_hal_srq_query_resp_umsg_get - Get the values from SRQ Query Response Up
 *				  Message
 *
 * @umsgh: Handle to umsg.
 * @srq_info: Pointer to structure to return information in SRQ Query response
 *
 * Set the SRQ Query Response UMQ message used by the adapter to return the
 * configuration of an SRQ on the adapter.
 */
void
vxge_hal_srq_query_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	vxge_hal_srq_info_t *srq_info)
{
	u64 val64;
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (srq_info != NULL));

	srq_info->srq_opaque_handle =
	    vxge_os_ntohll(umsg->m.srq_query.srq_opaque_handle);

	val64 = vxge_os_ntohll(umsg->m.srq_query.control_0);

	srq_info->virtual_path =
	    (u32)VXGE_HAL_SRQ_QUERY_RESP_GET_VIRTUAL_PATH(val64);

	srq_info->max_num_wqe_od_grps =
	    (u32)VXGE_HAL_SRQ_QUERY_RESP_GET_MAX_WQE_OD_GRPS(val64);

	srq_info->refill_threshold_high =
	    (u32)VXGE_HAL_SRQ_QUERY_RESP_GET_REF_TH_HIGH(val64);

	srq_info->refill_threshold_low =
	    (u32)VXGE_HAL_SRQ_QUERY_RESP_GET_REF_TH_LOW(val64);

	val64 = vxge_os_ntohll(umsg->m.srq_query.control_1);

	srq_info->low_threshold_notify =
	    (u32)VXGE_HAL_SRQ_QUERY_RESP_GET_LOW_TH_NOTIF(val64);

	srq_info->event_notification_mask =
	    (u32)VXGE_HAL_SRQ_QUERY_RESP_GET_EVENT_MASK(val64);

	srq_info->srq_limit_armed =
	    (u32)VXGE_HAL_SRQ_QUERY_RESP_GET_SRQ_LIM_ARMED(val64);

	srq_info->lro_bytestream =
	    (u32)VXGE_HAL_SRQ_QUERY_RESP_GET_LRO_BS(val64);

	srq_info->srq_state =
	    (u32)VXGE_HAL_SRQ_QUERY_RESP_GET_SRQ_STATE(val64);

	val64 = vxge_os_ntohll(umsg->m.srq_query.control_2);

	srq_info->srq_limit_value =
	    (u32)VXGE_HAL_SRQ_QUERY_RESP_GET_SRQ_LIM_VALUE(val64);

	val64 = vxge_os_ntohll(umsg->m.srq_query.control_2);

	srq_info->interrupt_number =
	    (u32)VXGE_HAL_SRQ_QUERY_RESP_GET_INT_NUM(val64);

	srq_info->bitmap_id = (u32)VXGE_HAL_SRQ_QUERY_RESP_GET_BITMAP_ID(val64);

	srq_info->wb_bytes = (u32)VXGE_HAL_SRQ_QUERY_RESP_GET_WB_BYTES(val64);

	srq_info->wb_host_address =
	    vxge_os_ntohll(umsg->m.srq_query.wb_host_address);

	srq_info->srq_ir = vxge_os_ntohll(umsg->m.srq_query.srq_ir);
}

/*
 * __hal_srq_arm_req_dmsg_set - Set SRQ Arm Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @srq_limit: The SRQ limit to set if 'srq_limit_set' is set.
 *
 * Sets the SRQ Arm Request DMQ message used by the host to changed the armed
 * state of SRQ on the adapter.
 */
void
__hal_srq_arm_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 srq_limit)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SRQ_ARM_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_srq_arm_req_dmq_t));

	dmsg->m.srq_arm.control_0 = vxge_os_htonll(
	    VXGE_HAL_SRQ_ARM_REQ_LIMIT(srq_limit));
}

/*
 * vxge_hal_srq_event_notification_umsg_get - Get values from SRQ Event
 *						Notification Response
 *
 * @umsgh: Handle to umsg.
 * @event_type : Buffer to return Event Type
 *
 * Gets the values from SRQ Event Notification Up Message.
 */
void
vxge_hal_srq_event_notification_umsg_get(
	vxge_hal_up_msg_h umsgh,
	u32 *event_type)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (event_type != NULL));

	*event_type =
	    (u32)VXGE_HAL_SRQ_EVENT_NOTIF_GET_EVENT_TYPE(
	    vxge_os_ntohll(umsg->m.srq_event_notif.control_0));
}

/*
 * __hal_srq_fwqebw_req_dmsg_set - Set SRQ First WQE BW Notification
 *				   Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @wqe_block_host_address: The host address of the first WQE block wrapper in
 *		host memory.
 * @wqe_block_bytes: The number of bytes in the first WQE block wrapper in host
 *		memory.
 * @on_host_tail_wsn: The WQE sequence number (WSN) of the tail or last WQE that
 *		was added to the list.
 *
 * Sets the SRQ First WQE BW Notification Request DMQ message used by the
 * host to notify the First WQE Block to the the adapter.
 */
void
__hal_srq_fwqebw_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u64 wqe_block_host_address,
	u32 wqe_block_bytes,
	u32 on_host_tail_wsn)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SRQ_FIRST_WQE_BW_NOTIF_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_srq_fwqebw_req_dmq_t));

	dmsg->m.srq_fwqebw.wqe_block_host_address =
	    vxge_os_htonll(wqe_block_host_address);

	dmsg->m.srq_fwqebw.control_0 = vxge_os_htonll(
	    VXGE_HAL_SRQ_FWQEBW_REQ_WB_BYTES(wqe_block_bytes) |
	    VXGE_HAL_SRQ_FWQEBW_REQ_ON_TAIL_WSN(on_host_tail_wsn));
}

/*
 * __hal_srq_wqebadd_req_dmsg_set - Set SRQ WQE Blocks Added Notification
 *				   Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @on_host_tail_wsn: The sequence number of the last or tail WQE added to the
 *		list of WQEs in host memory.
 *
 * Sets the SRQ WQE Blocks Added Notification Request DMQ message used by the
 * host to notify the adapter about the added WQE blocks.
 */
void
__hal_srq_wqebadd_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 on_host_tail_wsn)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SRQ_WQE_BLOCKS_ADDED_NOTIF_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_srq_wqebadd_req_dmq_t));

	dmsg->m.srq_wqebadd.control_0 = vxge_os_htonll(
	    VXGE_HAL_SRQ_WQEBADD_REQ_ON_TAIL_WSN(on_host_tail_wsn));
}

/*
 * __hal_srq_return_unused_wqes_req_dmsg_set - Set the SRQ Return WQEs
 *				request DMQ message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @number_of_wqes: Number of Wqes to be returned
 *
 * Sets the values in SRQ return unused WQEs Request DMQ message
 * used by the host to notify the adapter to return WQES.
 */
void
__hal_srq_return_unused_wqes_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 number_of_wqes)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SRQ_RETURN_UNUSED_WQES_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_srq_return_unused_wqes_req_dmq_t));

	dmsg->m.srq_return_wqes.control_0 = vxge_os_htonll(
	    VXGE_HAL_SRQ_RETURN_WQES_REQ_NUM_WQES(number_of_wqes));
}

/*
 * vxge_hal_srq_return_wqes_umsg_get - Get the values from SRQ Return WQEs UMQ
 *
 * @umsgh: Handle to umsg.
 * @srq_opaque_handle: Buffer to return SRQ opaque handle echoed in response
 * @number_of_wqes_in_msg: Buffer to return The number of WQE opaque handles
 *		in this message
 * @number_of_wqes_remaining: Buffer to return the number of WQE opaque handles
 *		left to send
 * @wqe_opaque_handles: Array to return the WQE Opaque Handles
 *
 * Get the values from SRQ Return WQEs UMQ used when deleting a TCP session or
 * tearing down an SRQ to return any WQEs that might still be in play for the
 * SRQ.
 */
void
vxge_hal_srq_return_unused_wqes_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	vxge_hal_opaque_handle_t **srq_opaque_handle,
	u32 *number_of_wqes_in_msg,
	u32 *number_of_wqes_remaining,
	u64 *wqe_opaque_handles)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;
	u64 val64;
	u32 i;
	u64 *wqe_handles;

	vxge_assert((umsg != NULL) && (srq_opaque_handle != NULL) &&
	    (number_of_wqes_in_msg != NULL) &&
	    (number_of_wqes_remaining != NULL) &&
	    (wqe_opaque_handles != NULL));

	*srq_opaque_handle = (vxge_hal_opaque_handle_t *)(ptr_t)
	    vxge_os_ntohll(umsg->m.srq_return_wqes.srq_opaque_handle);

	val64 = vxge_os_ntohll(umsg->m.srq_return_wqes.control_0);

	*number_of_wqes_in_msg =
	    (u32)VXGE_HAL_SRQ_RETURN_WQES_RESP_GET_NUM_IN_MSG(val64);

	*number_of_wqes_remaining =
	    (u32)VXGE_HAL_SRQ_RETURN_WQES_RESP_GET_NUM_REMAIN(val64);

	wqe_handles = umsg->m.srq_return_wqes.wqe_opaque_handles;

	for (i = 0; i < (*number_of_wqes_in_msg); i++) {
		*wqe_opaque_handles++ = vxge_os_ntohll(*wqe_handles++);
	}
}

/* ========================== STAG MESSAGES ================================= */

/*
 * __hal_stag_allocate_nsmr_req_dmsg_set - Set STAG Allocate NSMR Request Down
 *					  Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @pd_id: Protection domain to which the new NSMR belongs
 * @num_of_pbles: The number of PBLEs that the new NSMR must be able
 *		to accommodate.
 * @stag_index: The STAG Index of the NSMR to be created.
 * @remote_access_flag: Flag to Specify if this NSMR will be allowed to be used
 *		for any remote access operations (read, write, invalidate) when
 *		it is subsequently registered.
 *
 * Sets the STAG Allocate NSMR  Request DMQ message used to implement the
 * Allocate Non-Shared Memory Region memory management verb.
 */
void
__hal_stag_allocate_nsmr_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u64 pd_id,
	u32 num_of_pbles,
	u32 stag_index,
	u32 remote_access_flag)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_STAG_ALLOCATE_NSMR_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_stag_allocate_nsmr_req_dmq_t));

	dmsg->m.stag_allocate_nsmr.pd_id = vxge_os_htonll(pd_id);

	dmsg->m.stag_allocate_nsmr.control_0 =
	    VXGE_HAL_STAG_ALLOC_NSMR_REQ_PBL_LEN(num_of_pbles) |
	    VXGE_HAL_STAG_ALLOC_NSMR_REQ_STAG_IDX(stag_index);

	if (remote_access_flag)
		dmsg->m.stag_allocate_nsmr.control_0 |=
		    VXGE_HAL_STAG_ALLOC_NSMR_REQ_RAC_FLAG;

	dmsg->m.stag_allocate_nsmr.control_0 =
	    vxge_os_htonll(dmsg->m.stag_allocate_nsmr.control_0);
}

/*
 * vxge_hal_stag_allocate_nsmr_resp_umsg_get - Get the values from STAG Allocate
 *					NSMR Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @stag_index: Buffer to return the STAG Index of the NSMR to be created.
 * @num_of_pbles: Buffer to return the number of PBLEs that the newly-allocated
 *		NSMR is capable of accommodating.
 *
 * Get the valueas from STAG Allocate NSMR Response UMQ used to communicate to
 * the host the success or failure of the corresponding STAG Allocate NSMR
 * Request DMQ message
 */
void
vxge_hal_stag_allocate_nsmr_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	u32 *stag_index,
	u32 *num_of_pbles)
{
	u64 val64;
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (stag_index != NULL) &&
	    (num_of_pbles != NULL));

	val64 = vxge_os_ntohll(umsg->m.stag_allocate_nsmr.control_0);

	*stag_index = (u32)VXGE_HAL_STAG_ALLOC_NSMR_RESP_GET_STAG_IDX(val64);

	*num_of_pbles = (u32)VXGE_HAL_STAG_ALLOC_NSMR_RESP_GET_PBL_LEN(val64);
}

/*
 * __hal_stag_allocate_mw_req_dmsg_set - Set STAG Allocate MW Request Down
 *					  Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @stag_index: The host assigned STag index of the MW being allocated on
 *		the adapter.
 * @pd_id: Protection domain to which the new NSMR belongs
 *
 * Sets the STAG Allocate MW Request DMQ message used to implement the
 * Allocate Memory Window memory management verb.
 */
void
__hal_stag_allocate_mw_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 stag_index,
	u64 pd_id)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_STAG_ALLOCATE_MW_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_stag_allocate_mw_req_dmq_t));

	dmsg->m.stag_allocate_mw.control_0 = vxge_os_htonll(
	    VXGE_HAL_STAG_ALLOC_MW_REQ_STAG_IDX(stag_index));

	dmsg->m.stag_allocate_mw.pd_id = vxge_os_htonll(pd_id);
}

/*
 * vxge_hal_stag_allocate_mw_resp_umsg_get - Get the values from STAG Allocate
 *				  MW Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @stag_index: Buffer to return the host assigned STag index of the MW being
 *		allocated on the adapter.
 *
 * Get the values from STAG Allocate MW Response UMQ used to communicate to the
 * host the success or failure of the corresponding STAG Allocate MW Request
 * DMQ message
 */
void
vxge_hal_stag_allocate_mw_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	u32 *stag_index)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (stag_index != NULL));

	*stag_index = (u32)VXGE_HAL_STAG_ALLOC_MW_RESP_STAG_IDX(
	    vxge_os_ntohll(umsg->m.stag_allocate_mw.control_0));
}

/*
 * __hal_stag_deallocate_req_dmsg_set - Set STAG De-Allocate Request Down
 *					  Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @stag_index: The index of the STag being de-allocated on the adapter.
 * @dealloc_pbl: Flag to Specify if the PBL that underlies this STag is
 *		also to be deallocated with the STag.
 *
 * Sets the STAG De-Allocate Request DMQ message used to deallocate the
 * specified STag, freeing up any on-adapter resources
 */
void
__hal_stag_deallocate_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 stag_index,
	u32 dealloc_pbl)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_STAG_DE_ALLOCATE_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_stag_deallocate_req_dmq_t));

	dmsg->m.stag_deallocate.control_0 =
	    VXGE_HAL_STAG_DEALLOC_REQ_STAG_IDX(stag_index);

	if (dealloc_pbl)
		dmsg->m.stag_deallocate.control_0 |=
		    VXGE_HAL_STAG_DEALLOC_REQ_PBL_FLAG;

	dmsg->m.stag_deallocate.control_0 =
	    vxge_os_htonll(dmsg->m.stag_deallocate.control_0);
}

/*
 * vxge_hal_stag_deallocate_resp_umsg_get - Get the values from STAG deallocate
 *				  Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @stag_index: Buffer to return The host assigned STag index of the
 *		 de-allocated STag
 *
 * Get the values from STAG Adeallocate Response UMQ used to communicate to the
 * host the success or failure of the corresponding STAG deallocate Request
 * DMQ message
 */
void
vxge_hal_stag_deallocate_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	u32 *stag_index)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (stag_index != NULL));

	*stag_index = (u32)VXGE_HAL_STAG_DEALLOC_RESP_STAG_IDX(
	    vxge_os_ntohll(umsg->m.stag_deallocate.control_0));
}

/*
 * __hal_stag_register_nsmr_req_dmsg_set - Set STAG Register NSMR Request Down
 *					  Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @stag_index: The host assigned STag index to associate with the NSMR being
 *		registered on the adapter.
 * @stag_key: Instance identifier for this validation of this NSMR
 * @pble_size: Size in bytes of each physical buffer in the PBL.
 * @pd_id: Protection domain to which the new NSMR belongs
 * @first_pbl_block_ptr: Starting address in host memory of the 1st PBL block
 *		for the NSMRs PBL
 * @first_pbl_block_length: Number of PBLEs in the 1st PBL block
 * @pbl_length: Total number of PBLEs in the PBL.
 * @va_base_flag: Flag to Specify if the NSMR will use virtual addressing
 * @remote_access_flag: Flag to Specify if this NSMR will be allowed to be used
 *		for any remote access operations (read, write, invalidate) on
 *		this and subsequent registrations.
 * @local_write_flag: Flag to Specify if the adapter can write to this NSMR.
 * @local_read_flag: Flag to Specify if the adapter can read from this NSMR.
 * @remote_write_flag: Flag to Specify if remote peers are allowed to write into
 *		this NSMR
 * @remote_read_flag: Flag to Specify if remote peers are allowed to read from
 *		this NSMR.
 * @enable_mw_binding_flag: Flag to Specify if this NSMR may have MWs bound to
 *		it by subsequent PostSQ.BindMW work requests.
 * @first_byte_offset: Number of bytes into the first PBLE where the MR begins.
 * @mr_length: Size in bytes of the NSMR.
 * @base_to: Virtual address of the first byte of the NSMR.
 *
 * Sets the STAG Register NSMR  Request DMQ message used to implement the
 * Register Non-Shared Memory Region memory management verb.
 */
void
__hal_stag_register_nsmr_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 stag_index,
	u32 stag_key,
	u32 pble_size,
	u64 pd_id,
	u64 first_pbl_block_ptr,
	u32 first_pbl_block_length,
	u32 pbl_length,
	u32 va_base_flag,
	u32 remote_access_flag,
	u32 local_write_flag,
	u32 local_read_flag,
	u32 remote_write_flag,
	u32 remote_read_flag,
	u32 enable_mw_binding_flag,
	u32 first_byte_offset,
	u64 mr_length,
	u64 base_to)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_STAG_REGISTER_NSMR_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_stag_register_nsmr_req_dmq_t));

	dmsg->m.stag_reg_nsmr.control_0 = vxge_os_htonll(
	    VXGE_HAL_STAG_REG_NSMR_REQ_STAG_IDX(stag_index) |
	    VXGE_HAL_STAG_REG_NSMR_REQ_STAG_KEY(stag_key) |
	    VXGE_HAL_STAG_REG_NSMR_REQ_PBLE_SIZE(pble_size));

	dmsg->m.stag_reg_nsmr.pd_id = vxge_os_htonll(pd_id);

	dmsg->m.stag_reg_nsmr.first_pbl_block_ptr =
	    vxge_os_htonll(first_pbl_block_ptr);

	dmsg->m.stag_reg_nsmr.control_1 = vxge_os_htonll(
	    VXGE_HAL_STAG_REG_NSMR_REQ_FPBLB_LEN(first_pbl_block_length) |
	    VXGE_HAL_STAG_REG_NSMR_REQ_NUM_PBLS(pbl_length));

	dmsg->m.stag_reg_nsmr.control_2 =
	    VXGE_HAL_STAG_REG_NSMR_REQ_FB_OFFSET(first_byte_offset);

	if (va_base_flag)
		dmsg->m.stag_reg_nsmr.control_2 |=
		    VXGE_HAL_STAG_REG_NSMR_REQ_VA_BASE;

	if (remote_access_flag)
		dmsg->m.stag_reg_nsmr.control_2 |=
		    VXGE_HAL_STAG_REG_NSMR_REQ_RAC_FLAG;

	if (local_write_flag)
		dmsg->m.stag_reg_nsmr.control_2 |=
		    VXGE_HAL_STAG_REG_NSMR_REQ_LW_FLAG;

	if (local_read_flag)
		dmsg->m.stag_reg_nsmr.control_2 |=
		    VXGE_HAL_STAG_REG_NSMR_REQ_LR_FLAG;

	if (remote_write_flag)
		dmsg->m.stag_reg_nsmr.control_2 |=
		    VXGE_HAL_STAG_REG_NSMR_REQ_RW_FLAG;

	if (remote_read_flag)
		dmsg->m.stag_reg_nsmr.control_2 |=
		    VXGE_HAL_STAG_REG_NSMR_REQ_RR_FLAG;

	if (enable_mw_binding_flag)
		dmsg->m.stag_reg_nsmr.control_2 |=
		    VXGE_HAL_STAG_REG_NSMR_REQ_MW_BIND_EN;

	dmsg->m.stag_reg_nsmr.control_2 =
	    vxge_os_htonll(dmsg->m.stag_reg_nsmr.control_2);

	dmsg->m.stag_reg_nsmr.mr_length = vxge_os_htonll(mr_length);

	dmsg->m.stag_reg_nsmr.base_to = vxge_os_htonll(base_to);
}

/*
 * vxge_hal_stag_register_nsmr_resp_umsg_get - Get the values from STAG Register
 *					NSMR Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @stag_index: Buffer to return the host assigned STag index of the NSMR that
 *		was registered via the original DMQ. This is returned unmodified
 *		from the DMQ for convenience.
 * @stag_key: Buffer to return the host assigned STag key of the NSMR that was
 *		registered via the original DMQ. This is returned unmodified
 *		from the DMQ for convenience.
 * @num_of_pbles: Buffer to return the number of PBLEs that the newly-allocated
 *		NSMR is capable of accommodating.
 *
 * Get the valueas from STAG Register NSMR Response UMQ used to communicate to
 * the host the success or failure of the corresponding STAG Register NSMR
 * Request DMQ message
 */
void
vxge_hal_stag_register_nsmr_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	u32 *stag_index,
	u32 *stag_key,
	u32 *num_of_pbles)
{
	u64 val64;
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (stag_index != NULL) &&
	    (stag_key != NULL) && (num_of_pbles != NULL));

	val64 = vxge_os_ntohll(umsg->m.stag_reg_nsmr.control_0);

	*stag_index =
	    (u32)VXGE_HAL_STAG_REG_NSMR_RESP_GET_STAG_IDX(val64);

	*stag_key =
	    (u32)VXGE_HAL_STAG_REG_NSMR_RESP_GET_STAG_KEY(val64);

	*num_of_pbles =
	    (u32)VXGE_HAL_STAG_REG_NSMR_RESP_GET_PBL_LEN(val64);
}

/*
 * __hal_stag_re_register_nsmr_req_dmsg_set - Set STAG Re-register NSMR
 *					  Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @stag_index: The STag index of the original NSMR that is being re-registered.
 *		The adapter will use the same STag index for the re-registered
 *		NSMR.
 * @stag_key: Instance identifier for this validation of this NSMR
 * @pble_size: Size in bytes of each physical buffer in the PBL.
 * @pd_id: Protection domain to which the new NSMR belongs
 * @first_pbl_block_ptr: Starting address in host memory of the 1st PBL block
 *		for the NSMRs PBL
 * @first_pbl_block_length: Number of PBLEs in the 1st PBL block
 * @pbl_length: Total number of PBLEs in the PBL.
 * @va_base_flag: Flag to Specify if the NSMR will use virtual addressing
 * @remote_access_flag: Flag to Specify if this NSMR will be allowed to be used
 *		for any remote access operations (read, write, invalidate) on
 *		this and subsequent registrations.
 * @local_write_flag: Flag to Specify if the adapter can write to this NSMR.
 * @local_read_flag: Flag to Specify if the adapter can read from this NSMR.
 * @remote_write_flag: Flag to Specify if remote peers are allowed to write into
 *		this NSMR
 * @remote_read_flag: Flag to Specify if remote peers are allowed to read from
 *		this NSMR.
 * @enable_mw_binding_flag: Flag to Specify if this NSMR may have MWs bound to
 *		it by subsequent PostSQ.BindMW work requests.
 * @first_byte_offset: Number of bytes into the first PBLE where the MR begins.
 * @mr_length: Size in bytes of the NSMR.
 * @base_to: Virtual address of the first byte of the NSMR.
 *
 * Sets the STAG Re-register NSMR  Request DMQ message used to implement the
 * Re-register Non-Shared Memory Region memory management verb.
 */
void
__hal_stag_re_register_nsmr_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 stag_index,
	u32 stag_key,
	u32 pble_size,
	u64 pd_id,
	u64 first_pbl_block_ptr,
	u32 first_pbl_block_length,
	u32 pbl_length,
	u32 va_base_flag,
	u32 remote_access_flag,
	u32 local_write_flag,
	u32 local_read_flag,
	u32 remote_write_flag,
	u32 remote_read_flag,
	u32 enable_mw_binding_flag,
	u32 first_byte_offset,
	u64 mr_length,
	u64 base_to)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_STAG_RE_REGISTER_NSMR_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_stag_re_register_nsmr_req_dmq_t));

	dmsg->m.stag_rereg_nsmr.control_0 = vxge_os_htonll(
	    VXGE_HAL_STAG_RREG_NSMR_REQ_STAG_IDX(stag_index) |
	    VXGE_HAL_STAG_RREG_NSMR_REQ_STAG_KEY(stag_key) |
	    VXGE_HAL_STAG_RREG_NSMR_REQ_PBLE_SIZE(pble_size));

	dmsg->m.stag_rereg_nsmr.pd_id = vxge_os_htonll(pd_id);

	dmsg->m.stag_rereg_nsmr.first_pbl_block_ptr =
	    vxge_os_htonll(first_pbl_block_ptr);

	dmsg->m.stag_rereg_nsmr.control_1 = vxge_os_htonll(
	    VXGE_HAL_STAG_RREG_NSMR_REQ_FPBLB_LEN(first_pbl_block_length) |
	    VXGE_HAL_STAG_RREG_NSMR_REQ_NUM_PBLS(pbl_length));

	dmsg->m.stag_rereg_nsmr.control_2 =
	    VXGE_HAL_STAG_RREG_NSMR_REQ_FB_OFFSET(first_byte_offset);

	if (va_base_flag)
		dmsg->m.stag_rereg_nsmr.control_2 |=
		    VXGE_HAL_STAG_RREG_NSMR_REQ_VA_BASE;

	if (remote_access_flag)
		dmsg->m.stag_rereg_nsmr.control_2 |=
		    VXGE_HAL_STAG_RREG_NSMR_REQ_RAC_FLAG;

	if (local_write_flag)
		dmsg->m.stag_rereg_nsmr.control_2 |=
		    VXGE_HAL_STAG_RREG_NSMR_REQ_LW_FLAG;

	if (local_read_flag)
		dmsg->m.stag_rereg_nsmr.control_2 |=
		    VXGE_HAL_STAG_RREG_NSMR_REQ_LR_FLAG;

	if (remote_write_flag)
		dmsg->m.stag_rereg_nsmr.control_2 |=
		    VXGE_HAL_STAG_RREG_NSMR_REQ_RW_FLAG;

	if (remote_read_flag)
		dmsg->m.stag_rereg_nsmr.control_2 |=
		    VXGE_HAL_STAG_RREG_NSMR_REQ_RR_FLAG;

	if (enable_mw_binding_flag)
		dmsg->m.stag_rereg_nsmr.control_2 |=
		    VXGE_HAL_STAG_RREG_NSMR_REQ_MW_BIND;

	dmsg->m.stag_rereg_nsmr.control_2 =
	    vxge_os_htonll(dmsg->m.stag_rereg_nsmr.control_2);

	dmsg->m.stag_rereg_nsmr.mr_length = vxge_os_htonll(mr_length);

	dmsg->m.stag_rereg_nsmr.base_to = vxge_os_htonll(base_to);
}

/*
 * vxge_hal_stag_re_register_nsmr_resp_umsg_get - Get the values from STAG
 *					  Re_register NSMR Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @stag_index: Buffer to return the STag index of the original NSMR that is
 *		being re-registered.
 * @stag_key: Buffer to return the Instance identifier for this validation of
 *		this NSMR
 * @num_of_pbles: Buffer to return the number of PBLEs that the newly-allocated
 *		NSMR is capable of accommodating.
 *
 * Get the valueas from STAG Re_register NSMR Response UMQ used to communicate
 * to the host the success or failure of the corresponding STAG Re_register
 * NSMR Request DMQ message
 */
void
vxge_hal_stag_re_register_nsmr_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	u32 *stag_index,
	u32 *stag_key,
	u32 *num_of_pbles)
{
	u64 val64;
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (stag_index != NULL) &&
	    (stag_key != NULL) && (num_of_pbles != NULL));

	val64 = vxge_os_ntohll(umsg->m.stag_rereg_nsmr.control_0);

	*stag_index = (u32)VXGE_HAL_STAG_RREG_NSMR_RESP_GET_STAG_IDX(val64);

	*stag_key = (u32)VXGE_HAL_STAG_RREG_NSMR_RESP_GET_STAG_KEY(val64);

	*num_of_pbles = (u32)VXGE_HAL_STAG_RREG_NSMR_RESP_GET_PBL_LEN(val64);
}

/*
 * __hal_stag_register_smr_req_dmsg_set - Set STAG Register SMR Request Down
 *					  Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @base_mr_stag_index: Unique identifier for the underlying MR on which this
 *		SMR will be created.
 * @va_base_flag: Flag to Specify if the SMR will use virtual addressing
 * @remote_access_flag: Flag to Specify if this SMR will be allowed to be used
 *		for any remote access operations (read, write, invalidate) on
 *		this and subsequent registrations.
 * @local_write_flag: Flag to Specify if the adapter can write to this SMR.
 * @local_read_flag: Flag to Specify if the adapter can read from this SMR.
 * @remote_write_flag: Flag to Specify if remote peers are allowed to write into
 *		this SMR
 * @remote_read_flag: Flag to Specify if remote peers are allowed to read from
 *		this SMR.
 * @enable_mw_binding_flag: Flag to Specify if this SMR may have MWs bound to
 *		it by subsequent PostSQ.BindMW work requests.
 * @stag_index: STag Index of the new STag record that is to be created.
 * @stag_key: Instance identifier for this validation of this SMR
 * @pd_id: Protection domain to which the new SMR belongs
 * @base_to: Virtual address of the first byte of the SMR.
 *
 * Sets the STAG Register SMR  Request DMQ message used to implement the
 * Register Shared Memory Region memory management verb.
 */
void
__hal_stag_register_smr_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 base_mr_stag_index,
	u32 va_base_flag,
	u32 remote_access_flag,
	u32 local_write_flag,
	u32 local_read_flag,
	u32 remote_write_flag,
	u32 remote_read_flag,
	u32 enable_mw_binding_flag,
	u32 stag_index,
	u32 stag_key,
	u64 pd_id,
	u64 base_to)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_STAG_REGISTER_SMR_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_stag_register_smr_req_dmq_t));

	dmsg->m.stag_reg_smr.control_0 =
	    VXGE_HAL_STAG_REG_SMR_REQ_MR_STAG_IDX(base_mr_stag_index) |
	    VXGE_HAL_STAG_REG_SMR_REQ_STAG_IDX(stag_index) |
	    VXGE_HAL_STAG_REG_SMR_REQ_STAG_KEY(stag_key);

	if (va_base_flag)
		dmsg->m.stag_reg_smr.control_0 |=
		    VXGE_HAL_STAG_REG_SMR_REQ_VA_BASE;

	if (remote_access_flag)
		dmsg->m.stag_reg_smr.control_0 |=
		    VXGE_HAL_STAG_REG_SMR_REQ_RAC_FLAG;

	if (local_write_flag)
		dmsg->m.stag_reg_smr.control_0 |=
		    VXGE_HAL_STAG_REG_SMR_REQ_LW_FLAG;

	if (local_read_flag)
		dmsg->m.stag_reg_smr.control_0 |=
		    VXGE_HAL_STAG_REG_SMR_REQ_LR_FLAG;

	if (remote_write_flag)
		dmsg->m.stag_reg_smr.control_0 |=
		    VXGE_HAL_STAG_REG_SMR_REQ_RW_FLAG;

	if (remote_read_flag)
		dmsg->m.stag_reg_smr.control_0 |=
		    VXGE_HAL_STAG_REG_SMR_REQ_RR_FLAG;

	if (enable_mw_binding_flag)
		dmsg->m.stag_reg_smr.control_0 |=
		    VXGE_HAL_STAG_REG_SMR_REQ_MW_BIND;

	dmsg->m.stag_reg_smr.control_0 =
	    vxge_os_htonll(dmsg->m.stag_reg_smr.control_0);

	dmsg->m.stag_reg_smr.pd_id = vxge_os_htonll(pd_id);

	dmsg->m.stag_reg_smr.base_to = vxge_os_htonll(base_to);
}

/*
 * vxge_hal_stag_register_smr_resp_umsg_get - Get the values from STAG Register
 *					 SMR Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @stag_index: Buffer to return the host assigned STag index of the NSMR that
 *		was registered via the original DMQ.
 * @stag_key: Buffer to return the host assigned STag key of the NSMR that was
 *		registered via the original DMQ.
 *
 * Get the values from STAG Register SMR Response UMQ used to communicate to the
 * host the success or failure of the corresponding STAG Register SMR Request
 * DMQ message
 */
void
vxge_hal_stag_register_smr_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	u32 *stag_index,
	u32 *stag_key)
{
	u64 val64;
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) &&
	    (stag_index != NULL) && (stag_key != NULL));

	val64 = vxge_os_ntohll(umsg->m.stag_reg_smr.control_0);

	*stag_index = (u32)VXGE_HAL_STAG_REG_SMR_RESP_GET_STAG_IDX(val64);

	*stag_key = (u32)VXGE_HAL_STAG_REG_SMR_RESP_GET_STAG_KEY(val64);
}

/*
 * __hal_stag_query_req_dmsg_set - Set STAG Query Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @stag_index: The STag index of the existing on-adapter STag that is
 *		being queried
 *
 * Sets the STAG Query Request DMQ message used by the host to query an STAG
 * on the adapter.
 */
void
__hal_stag_query_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 stag_index)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_STAG_QUERY_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_stag_query_req_dmq_t));

	dmsg->m.stag_query.control_0 = vxge_os_htonll(
	    VXGE_HAL_STAG_QUERY_REQ_STAG_IDX(stag_index));
}

/*
 * vxge_hal_stag_query_resp_umsg_get - Get the values from STAG Query Response
 *				   Up Message
 *
 * @umsgh: Handle to umsg.
 * @stag_info: Poniter to structure to return the query reposne values
 *
 * Get the values from STAG Query Response UMQ message used by the adapter to
 * return the configuration of an STAG on the adapter.
 */
void
vxge_hal_stag_query_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	vxge_hal_stag_info_t *stag_info)
{
	u64 val64;
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (stag_info != NULL));

	val64 = vxge_os_ntohll(umsg->m.stag_query.control_0);

	stag_info->stag_index = (u32)VXGE_HAL_STAG_QUERY_RESP_STAG_IDX(val64);

	stag_info->stag_key = (u32)VXGE_HAL_STAG_QUERY_RESP_STAG_KEY(val64);

	stag_info->is_allocated =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_IS_ALLOC(val64);

	stag_info->is_window =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_IS_WINDOW(val64);

	stag_info->is_zero_based =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_IS_ZERO_BASE(val64);

	stag_info->is_valid =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_IS_VALID(val64);

	stag_info->is_shared =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_IS_SHARED(val64);

	stag_info->allow_bind =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_ALLOW_BIND(val64);

	stag_info->allow_remote_access =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_ALLOW_RAC(val64);

	stag_info->allow_remote_read =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_ALLOW_RR(val64);

	stag_info->allow_remote_write =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_ALLOW_RW(val64);

	stag_info->allow_local_read =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_ALLOW_LR(val64);

	stag_info->allow_local_write =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_ALLOW_LW(val64);

	stag_info->pd_id = vxge_os_ntohll(umsg->m.stag_query.pd_id);

	val64 = vxge_os_ntohll(umsg->m.stag_query.control_1);

	stag_info->sid = (u32)VXGE_HAL_STAG_QUERY_RESP_GET_SID(val64);

	stag_info->sin = (u32)VXGE_HAL_STAG_QUERY_RESP_GET_SIN(val64);

	stag_info->length = vxge_os_ntohll(umsg->m.stag_query.length);

	stag_info->mr_base_to = vxge_os_ntohll(umsg->m.stag_query.mr_base_to);

	stag_info->mw_base_to = vxge_os_ntohll(umsg->m.stag_query.mw_base_to);

	val64 = vxge_os_ntohll(umsg->m.stag_query.control_2);

	stag_info->fbo = (u32)VXGE_HAL_STAG_QUERY_RESP_GET_FBO(val64);

	stag_info->pble_size =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_PBLE_SIZE(val64);

	val64 = vxge_os_ntohll(umsg->m.stag_query.control_3);

	stag_info->root_pbl_record_id =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_ROOT_PBL_ID(val64);

	stag_info->pbl_tree_depth =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_PBL_DEPTH(val64);

	stag_info->parent_stag_record_id =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_PSTAG_ID(val64);

	stag_info->parent_stag_valid =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_PSTAG_VALID(val64);

	stag_info->left_subtree_valid =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_LEFT_VALID(val64);

	stag_info->right_subtree_valid =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_RIGHT_VALID(val64);

	val64 = vxge_os_ntohll(umsg->m.stag_query.control_4);

	stag_info->left_subtree_record_id =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_LEFT_ID(val64);

	stag_info->right_subtree_record_id =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_RIGHT_ID(val64);

	val64 = vxge_os_ntohll(umsg->m.stag_query.control_5);

	stag_info->num_mws_bound =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_NUM_MW_BOUND(val64);

	stag_info->subtree_imbalance =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_ST_IMBALANCE(val64);

	stag_info->max_pbl_length =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_MAX_PBL_LEN(val64);

	val64 = vxge_os_ntohll(umsg->m.stag_query.control_6);

	stag_info->actual_pbl_length =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_PBL_LEN(val64);

	stag_info->underlying_mr_record_id =
	    (u32)VXGE_HAL_STAG_QUERY_RESP_GET_MR_ID(val64);
}

/*
 * __hal_stag_validate_req_dmsg_set - Set STAG Validate Request Down
 *					  Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @stag_index: The index of the existing on-adapter STag that is being
 *		validated.
 * @stag_key: The key that to assign to the STag when it is validated.
 *
 * Sets the STAG Validate Request DMQ message.
 */
void
__hal_stag_validate_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 stag_index,
	u32 stag_key)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_STAG_VALID_LOCAL_TAG_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_stag_validate_req_dmq_t));

	dmsg->m.stag_validate.control_0 = vxge_os_htonll(
	    VXGE_HAL_STAG_VALID_LOCAL_REQ_STAG_IDX(stag_index) |
	    VXGE_HAL_STAG_VALID_LOCAL_REQ_STAG_KEY(stag_key));
}

/*
 * vxge_hal_stag_validate_resp_umsg_get - Get the values from STAG Validate
 *				 Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @stag_index: Buffer to return the host assigned STag index of the STag
 *		that was validated via the original DMQ.
 * @stag_key: Buffer to return the host assigned STag key of the STag that
 *		was validated via the original DMQ.
 *
 * Get the values from the STAG Validate Response UMQ used to communicate to
 * the host the success or failure of the corresponding STAG Validate Request
 * DMQ message
 */
void
vxge_hal_stag_validate_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	u32 *stag_index,
	u32 *stag_key)
{
	u64 val64;
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (stag_index != NULL) &&
	    (stag_key != NULL));

	val64 = vxge_os_ntohll(umsg->m.stag_validate.control_0);

	*stag_index = (u32)VXGE_HAL_STAG_VALID_LOCAL_RESP_GET_STAG_IDX(val64);

	*stag_key = (u32)VXGE_HAL_STAG_VALID_LOCAL_RESP_GET_STAG_KEY(val64);
}

/*
 * __hal_stag_invalidate_req_dmsg_set - Set STAG InValidate Request Down
 *					  Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @stag_index: The index of the existing on-adapter STag that is being
 *		invalidated.
 * @stag_key: The key that is currently assigned to the STag.
 * @pd_id: The PD ID of the session or QP.
 *
 * Sets the STAG Validate Request DMQ message.
 */
void
__hal_stag_invalidate_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 stag_index,
	u32 stag_key,
	u64 pd_id)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_STAG_INVALID_LOCAL_TAG_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_stag_invalidate_req_dmq_t));

	dmsg->m.stag_invalidate.control_0 = vxge_os_htonll(
	    VXGE_HAL_STAG_INVALID_REQ_STAG_IDX(stag_index) |
	    VXGE_HAL_STAG_INVALID_REQ_STAG_KEY(stag_key));

	dmsg->m.stag_invalidate.pd_id = vxge_os_htonll(pd_id);
}

/*
 * vxge_hal_stag_invalidate_resp_umsg_get - Get the values from STAG InValidate
 *				  Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @stag_index: Buffer to return he host assigned STag index of the STag that
 *		was invalidated.
 * @stag_key: Buffer to return the host assigned STag key of the STag that was
 *		invalidated.
 *
 * Get the values from the STAG InValidate Response UMQ used to communicate to
 * the host the success or failure of the corresponding STAG InValidate Request
 * DMQ message
 */
void
vxge_hal_stag_invalidate_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	u32 *stag_index,
	u32 *stag_key)
{
	u64 val64;
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (stag_index != NULL) &&
	    (stag_key != NULL));

	val64 = vxge_os_ntohll(umsg->m.stag_invalidate.control_0);

	*stag_index = (u32)VXGE_HAL_STAG_INVALID_RESP_GET_STAG_IDX(val64);

	*stag_key = (u32)VXGE_HAL_STAG_INVALID_RESP_GET_STAG_KEY(val64);
}

/*
 * __hal_stag_bind_mw_req_dmsg_set - Set STAG Bind MW Request Down
 *					  Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @mr_stag_index: The STag Index of the underlying MR to bind to.
 * @mr_stag_key: The STag Key of the underlying MR to bind to.
 * @mw_stag_index: The existing Memory Window's STag Index.
 * @mw_stag_key: The STag Key to use for the MW.
 * @mr_stag: The existing Memory Region's STAG index.
 *		Least significant 8 bits is the key.
 * @mw_ctrl: Control flags for Memory Window.
 * @mw_length: Memory window length in bytes.
 * @tagged_offset: Memory window tagged offset in bytes.
 *
 * Sets the STAG Bind MW Request DMQ message.
 */
void
__hal_stag_bind_mw_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 mr_stag_index,
	u32 mr_stag_key,
	u32 mw_stag_index,
	u32 mw_stag_key,
	u32 mr_stag,
	u32 mw_ctrl,
	u64 mw_length,
	u64 tagged_offset)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_STAG_BIND_MW_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_stag_bind_mw_req_dmq_t));

	dmsg->m.stag_bind_mw.control_0 = vxge_os_htonll(
	    VXGE_HAL_STAG_BIND_MW_REQ_MR_STAG_IDX(mr_stag_index) |
	    VXGE_HAL_STAG_BIND_MW_REQ_MR_STAG_KEY(mr_stag_key) |
	    VXGE_HAL_STAG_BIND_MW_REQ_MW_STAG_IDX(mw_stag_index) |
	    VXGE_HAL_STAG_BIND_MW_REQ_MW_STAG_KEY(mw_stag_key));

	dmsg->m.stag_bind_mw.control_1 = vxge_os_htonll(
	    VXGE_HAL_STAG_BIND_MW_REQ_MR_STAG(mr_stag) |
	    VXGE_HAL_STAG_BIND_MW_REQ_MW_CTRL(mw_ctrl));

	dmsg->m.stag_bind_mw.mw_length = vxge_os_htonll(mw_length);

	dmsg->m.stag_bind_mw.tagged_offset = vxge_os_htonll(tagged_offset);
}

/*
 * vxge_hal_stag_bind_mw_resp_umsg_Get - Get the values from STAG Bind MW
 *				Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @stag_index: Buffer to return the host assigned STag index MW that was
 *		bound.
 * @stag_key: Buffer to return the host assigned STag key of the MW that
 *		was bound.
 * @result: Result code
 *
 * Get the values from STAG Bind MW Response UMQ used to communicate to the host
 * the success or failure of the corresponding STAG Bind MW Request DMQ message
 */
void
vxge_hal_stag_bind_mw_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	u32 *stag_index,
	u32 *stag_key)
{
	u64 val64;
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (stag_index != NULL) &&
	    (stag_key != NULL));

	val64 = vxge_os_ntohll(umsg->m.stag_bind_mw.control_0);

	*stag_index = (u32)VXGE_HAL_STAG_BIND_MW_RESP_GET_STAG_IDX(val64);

	*stag_key = (u32)VXGE_HAL_STAG_BIND_MW_RESP_GET_STAG_KEY(val64);
}

/*
 * __hal_stag_fast_register_nsmr_req_dmsg_set - Set STAG Fast Register NSMR
 *					  Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @mr_control_size: Length in bytes of the Memory Region control structure.
 * @mr_control_ptr: Pointer to Memory Region control structure in host memory.
 *
 * Sets the STAG Fast Register Request DMQ message.
 */
void
__hal_stag_fast_register_nsmr_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 mr_control_size,
	u64 mr_control_ptr)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_STAG_FAST_REGISTER_NSMR_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_stag_fast_register_nsmr_req_dmq_t));

	dmsg->m.stag_frnsmr.control_0 = vxge_os_htonll(
	    VXGE_HAL_STAG_FRNSMR_REQ_MR_CTRL_SIZE(mr_control_size));

	dmsg->m.stag_frnsmr.mr_control_ptr = vxge_os_htonll(mr_control_ptr);
}

/*
 * vxge_hal_stag_fast_register_nsmr_resp_umsg_get - Get values from the
 *				    STAG Fast Register NSMR Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @stag_index: Buffer to return the host assigned STag index of the NSMR that
 *		was fast-registered.
 * @stag_key: Buffer to return the host assigned STag key of the NSMR that was
 *		fast-registerd.
 *
 * Get the values from STAG Fast Register Response UMQ used to communicate to
 * the host the success or failure of the corresponding STAG Fast Register
 * Request DMQ message
 */
void
vxge_hal_stag_fast_register_nsmr_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	u32 *stag_index,
	u32 *stag_key)
{
	u64 val64;
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (stag_index != NULL) &&
	    (stag_key != NULL));

	val64 = vxge_os_ntohll(umsg->m.stag_frnsmr.control_0);

	*stag_index = (u32)VXGE_HAL_STAG_FRNSMR_RESP_GET_STAG_IDX(val64);

	*stag_key = (u32)VXGE_HAL_STAG_FRNSMR_RESP_GET_STAG_KEY(val64);
}

/* ======================== SESSION MESSAGES ============================= */

/*
 * __hal_tcp_promote_to_iwarp_req_dmsg_set - Set TCP Promote to iWARP
 *					 Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @mpa_crc_valid: Flag to specify MPA CRCs are expected to be valid on
 *		received frames.
 * @tx_marker_status: Flag to specify markers exist in the transmit direction.
 * @rx_marker_status: Flag to specify markers exist in the receive direction.
 * @iwarp_local_stag_zero_allowed: Flag to specify the session is allowed to
 *		use STag = 0 in Rx WQEs.
 * @iwarp_wire_stag_zero_allowed: Flag to specify session to accept STag = 0 in
 *		receive frames.
 * @inbound_write_readresp_enable: Flag to specify the session has remote write
 *		permission, incoming RDMA WRITE or RDMA READ RESPONSE messages
 *		can be placed
 * @inbound_readreq_enable: Flag to specify when a SEND* WITH INVALIDATE message
 *		is received, the STag can only be invalidated if the session has
 *		remote write and/or remote read access.  Also used to ensure
 *		that RIRRs are allowed on this session
 * @srq_pd_id: Protection domain to which the associated iWARP SRQ belongs.
 *		This is needed to do access checking.
 * @iwarp_qp_pd_id: Protection domain to which this session belongs.
 * @cqrq_ll_send_frame_data_offset: Ooffset from the CQE base address.
 * @iwarp_qp_limit_value: Ifn-progress WQEs limit value.
 * @rq_max_length: maximum number of WQEs that this session is allowed to have
 *		in progress on this SRQ.
 * @iwarp_qp_limit_armed: Flag to specify the Queue Pair limit is armed at the
 *		iwarp_qp_limit_value
 * @iwarp_cqrq_sin: Instance number of the CQRQ for the iWARP session.
 * @iwarp_cqrq_sid: Adapter ID of the CQRQ for the iWARP session.
 * @iwarp_srq_sin: Instance number of the SRQ for the iWARP session.
 * @iwarp_srq_sid: Adapter ID of the SRQ for the iWARP session.
 * @rdmap_version:The RDMA version number 01 - IETF RNICs  00 - RDMAC RNICs
 * @ddp_version: Direct Data Placement Protocol Version
 *		01 - for protocol described in
 *		http://ietf.org/internet-drafts/draft-ietf-rddp-ddp-06.txt
 * @mpa_reply_frame_size: This field represents the size of TCP payload of MPA
 *		reply frame (in bytes) on the Responder. This value is relevant
 *		only for the MPA Responder. It allows the Tx PE to determine the
 *		TCP SN at which iWarp will begin (ie the TxIMSN).
 * @is_mpa_initiator: This allows the PE to take distinquish whether it is the
 *		MPA Initiator or the MPA Responder.
 * @block_permissions: Used to restrict session's ability to use special TOWI
 *		atoms. It is used to enforce checks such as preventing the QP
 *		from performing a Fast Register NSMR operation.
 * @ird_limit: Host supplied IRD limit for this session.
 * @hw_cq_rr_ir: Pointer to the "hw-cq-rr" location on the host.
 *
 * Sets the TCP Promote to iWARP Request DMQ message sent from the host to the
 * adapter in order to migrate an existing bytestream session to iWARP mode.
 */
void
__hal_tcp_promote_to_iwarp_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 mpa_crc_valid,
	u32 tx_marker_status,
	u32 rx_marker_status,
	u32 iwarp_local_stag_zero_allowed,
	u32 iwarp_wire_stag_zero_allowed,
	u32 inbound_write_readresp_enable,
	u32 inbound_readreq_enable,
	u64 srq_pd_id,
	u64 iwarp_qp_pd_id,
	u32 cqrq_ll_send_frame_data_offset,
	u32 iwarp_qp_limit_value,
	u32 rq_max_length,
	u32 iwarp_qp_limit_armed,
	u32 iwarp_cqrq_sin,
	u32 iwarp_cqrq_sid,
	u32 iwarp_srq_sin,
	u32 iwarp_srq_sid,
	u32 rdmap_version,
	u32 ddp_version,
	u32 mpa_reply_frame_size,
	u32 is_mpa_initiator,
	u32 block_permissions,
	u32 ird_limit,
	u64 hw_cq_rr_ir)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_TCP_PROMOTE_TO_IWARP_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_tcp_promote_to_iwarp_req_dmq_t));

	dmsg->m.tcp_promote_iwarp.control_0 = 0;

	if (mpa_crc_valid)
		dmsg->m.tcp_promote_iwarp.control_0 |=
		    VXGE_HAL_TCP_PROMO_REQ_MPA_CRC_VALID;

	if (tx_marker_status)
		dmsg->m.tcp_promote_iwarp.control_0 |=
		    VXGE_HAL_TCP_PROMO_REQ_TX_MARK_STAT;

	if (rx_marker_status)
		dmsg->m.tcp_promote_iwarp.control_0 |=
		    VXGE_HAL_TCP_PROMO_REQ_RX_MARK_STAT;

	if (iwarp_local_stag_zero_allowed)
		dmsg->m.tcp_promote_iwarp.control_0 |=
		    VXGE_HAL_TCP_PROMO_REQ_IWLS_ZERO_OK;

	if (iwarp_wire_stag_zero_allowed)
		dmsg->m.tcp_promote_iwarp.control_0 |=
		    VXGE_HAL_TCP_PROMO_REQ_IWWS_ZERO_OK;

	if (inbound_write_readresp_enable)
		dmsg->m.tcp_promote_iwarp.control_0 |=
		    VXGE_HAL_TCP_PROMO_REQ_IBWRR_EN;

	if (inbound_readreq_enable)
		dmsg->m.tcp_promote_iwarp.control_0 |=
		    VXGE_HAL_TCP_PROMO_REQ_IBRR_EN;

	dmsg->m.tcp_promote_iwarp.control_0 =
	    vxge_os_htonll(dmsg->m.tcp_promote_iwarp.control_0);

	dmsg->m.tcp_promote_iwarp.srq_pd_id = vxge_os_htonll(srq_pd_id);

	dmsg->m.tcp_promote_iwarp.iwarp_qp_pd_id =
	    vxge_os_htonll(iwarp_qp_pd_id);

	dmsg->m.tcp_promote_iwarp.control_1 = vxge_os_htonll(
	    VXGE_HAL_TCP_PROMO_REQ_CQRQ_LLSFD_OFF(
	    cqrq_ll_send_frame_data_offset)|
	    VXGE_HAL_TCP_PROMO_REQ_QP_LIMIT_VALUE(iwarp_qp_limit_value) |
	    VXGE_HAL_TCP_PROMO_REQ_RQ_MAX_LENGTH(rq_max_length) |
	    VXGE_HAL_TCP_PROMO_REQ_QP_LIMIT_ARMED(iwarp_qp_limit_armed));

	dmsg->m.tcp_promote_iwarp.control_2 = vxge_os_htonll(
	    VXGE_HAL_TCP_PROMO_REQ_CQRQ_INST_NUM(iwarp_cqrq_sin) |
	    VXGE_HAL_TCP_PROMO_REQ_CQRQ_ID(iwarp_cqrq_sid) |
	    VXGE_HAL_TCP_PROMO_REQ_SRQ_INST_NUM(iwarp_srq_sin) |
	    VXGE_HAL_TCP_PROMO_REQ_SRQ_ID(iwarp_srq_sid));

	dmsg->m.tcp_promote_iwarp.control_3 =
	    VXGE_HAL_TCP_PROMO_REQ_RDMAP_VERSION(rdmap_version) |
	    VXGE_HAL_TCP_PROMO_REQ_DDP_VERSION(ddp_version) |
	    VXGE_HAL_TCP_PROMO_REQ_MPA_FRAME_SIZE(mpa_reply_frame_size) |
	    VXGE_HAL_TCP_PROMO_REQ_BLOCK_PERMISSIONS(block_permissions);

	if (is_mpa_initiator)
		dmsg->m.tcp_promote_iwarp.control_3 |=
		    VXGE_HAL_TCP_PROMO_REQ_MPA_INITIATOR;

	dmsg->m.tcp_promote_iwarp.control_3 =
	    vxge_os_htonll(dmsg->m.tcp_promote_iwarp.control_3);

	dmsg->m.tcp_promote_iwarp.control_4 = vxge_os_htonll(
	    VXGE_HAL_TCP_PROMO_REQ_IRD_LIMIT(ird_limit));

	dmsg->m.tcp_promote_iwarp.hw_cq_rr_ir = vxge_os_htonll(hw_cq_rr_ir);
}

/*
 * vxge_hal_tcp_promote_to_iwarp_resp_umsg_get - Get the values from TCP Promote
 *				 to iWARP Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @rx_imsn: Buffer to return 32b per-session constant that identifies TCP SN
 *
 * Get the valueas from TCP Promote to iWARP Response UMQ used to communicate to
 * the host the success or failure of the corresponding TCP Promote to iWARP
 * Request DMQ message
 */
void
vxge_hal_tcp_promote_to_iwarp_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	u32 *rx_imsn)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (rx_imsn != NULL));

	*rx_imsn = (u32)VXGE_HAL_TCP_PROMO_RESP_GET_RX_IMSN(
	    vxge_os_ntohll(umsg->m.tcp_promote_iwarp.control_0));
}

/*
 * __hal_tcp_modify_req_dmsg_set - Sets the TCP Modify Request DMQ
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @mpa_crc_valid: If set then MPA CRCs are expected to be valid on received
 *		frames. The adapter always calculates MPA CRCs in the transmit
 *		direction. This field is ignored for a bytestream session.
 * @tx_marker_mode: Whether MPA markers exist in the transmit direction or not.
 *		-- 1 - enable MPA markers in transmit frames
 *		-- 0 - disable MPA markers in transmit frames
 *		This field is ignored for a bytestream session.
 * @rx_marker_mode: Whether markers exist in the receive direction or not.
 *		-- 1 - enable MPA markers in receive frames
 *		-- 0 - disable MPA markers in receive frames
 *		If this is set to zero for an RDMA session, the RxPE drops all
 *		out of order receive segments. This field is ignored for
 *		a bytestream session.
 * @rdmap_version: The RDMA version number
 *		  -- 01 - for IETF RNICs
 *		  -- 00 - for RDMAC RNICs
 * @ddp_version: Direct Data Placement Protocol Version
 *		  -- 01 - for protocol described in
 *		http://ietf.org/internet-drafts/draft-ietf-rddp-ddp-06.txt
 * @iwarp_qp_limit_armed: If set, the Queue Pair limit is armed at the
 *		iwarp_qp_limit_value. This field is ignored for a bytestream
 *		session.
 * @iwarp_qp_limit_value: If the number of in-progress WQEs reaches this
 *		value and the Queue Pair limit is armed, the session produces
 *		an asynchronous event. This field is ignored for a bytestream
 *		session.
 * @rq_max_length: The maximum number of WQEs that this session is
 *		allowed to have in progress on this SRQ.
 * @block_permissions: Used to restrict session's ability to use special
 *		TOWI atoms. It is used to enforce checks such as preventing the
 *		QP from performing a Fast Register NSMR operation.
 * @ird_limit: Host supplied IRD limit for this session.
 *
 * This routine sets TCP Modify Request DMQ message is sent by the host to
 * modify the attributes associated with a bytestream or iWARP session.
 */
void
__hal_tcp_modify_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 mpa_crc_valid,
	u32 tx_marker_mode,
	u32 rx_marker_mode,
	u32 rdmap_version,
	u32 ddp_version,
	u32 iwarp_qp_limit_armed,
	u32 iwarp_qp_limit_value,
	u32 rq_max_length,
	u32 block_permissions,
	u32 ird_limit)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_TCP_MODIFY_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_tcp_modify_req_dmq_t));

	dmsg->m.tcp_modify.control_0 =
	    VXGE_HAL_TCP_MODIFY_REQ_RDMAP_VERSION(rdmap_version) |
	    VXGE_HAL_TCP_MODIFY_REQ_DDP_VERSION(ddp_version) |
	    VXGE_HAL_TCP_MODIFY_REQ_QP_LIMIT_ARMED(iwarp_qp_limit_armed) |
	    VXGE_HAL_TCP_MODIFY_REQ_QP_LIMIT_VALUE(iwarp_qp_limit_value);

	if (mpa_crc_valid)
		dmsg->m.tcp_modify.control_0 |=
		    VXGE_HAL_TCP_MODIFY_REQ_MPA_CRC_VALID;

	if (tx_marker_mode)
		dmsg->m.tcp_modify.control_0 |=
		    VXGE_HAL_TCP_MODIFY_REQ_TX_MARK_MODE;

	if (rx_marker_mode)
		dmsg->m.tcp_modify.control_0 |=
		    VXGE_HAL_TCP_MODIFY_REQ_RX_MARK_MODE;

	dmsg->m.tcp_modify.control_0 =
	    vxge_os_htonll(dmsg->m.tcp_modify.control_0);

	dmsg->m.tcp_modify.control_1 = vxge_os_htonll(
	    VXGE_HAL_TCP_MODIFY_REQ_RQ_MAX_LEN(rq_max_length) |
	    VXGE_HAL_TCP_MODIFY_REQ_BLOCK_PERMISSIONS(block_permissions) |
	    VXGE_HAL_TCP_MODIFY_REQ_IRD_LIMIT(ird_limit));
}

/*
 * __hal_tcp_delete_req_dmsg_set - Set TCP Delete Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 *
 * Sets the TCP Delete Request DMQ message used by the host to delete an TCP
 * session on the adapter.
 */
void
__hal_tcp_delete_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_TCP_DELETE_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE);
}

/*
 * vxge_hal_tcp_delete_resp_umsg_get - Get values from TCP Delete Response
 *				Up Message
 *
 * @umsgh: Handle to umsg.
 * @session_opaque_handle: Buffer to return session Opaque handle
 *
 * Gets the values from TCP Delete Response Up Message.
 */
void
vxge_hal_tcp_delete_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	vxge_hal_opaque_handle_t **session_opaque_handle)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (session_opaque_handle != NULL));

	*session_opaque_handle = (vxge_hal_opaque_handle_t *)(ptr_t)
	    vxge_os_ntohll(umsg->m.tcp_delete.session_opaque_handle);
}

/*
 * __hal_tcp_abort_req_dmsg_set - Set TCP Abort Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 *
 * Sets the TCP Abort Request DMQ message used by the host to abort a TCP
 * session on the adapter.
 */
void
__hal_tcp_abort_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_TCP_ABORT_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE);
}

/*
 * __hal_tcp_update_rxwin_req_dmsg_set - Set TCP Update Rxwin Request
 *					Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @rcv_buf_valid: Does the rcv_buf field contain a value to be updated?
 * @rcv_buf_size_valid: Does the rcv_buf_size field contain a value to be
 *		updated?
 * @rcv_max_wnd_valid: Does the rcv_max_win field contain a value to be updated?
 * @rcv_buf: For bytestream sessions: lowest TCP sequence number buffered in
 *		the connection's receive buffer.  This is updated by the host
 *		as data is consumed by the ULP/application.  Must increase
 *		monotonically to avoid window shrinks (either check in PE when
 *		host updates it, or just make it a requirement of the host).
 *		For iWARP sessions: the DDP mainland RHS TCP sequence number.
 *		Represents the last aligned DDP mainland placement.  The DDP
 *		mainland edge differs from the TCP mainland edge (rcv_nxt)
 *		if MPA misaligned frames are arriving (ie session is ""DDP OOO""
 *		and iwarp_align_state != aligned).  This field is only updated
 *		when misaligned.  When aligned, including in the normal fast
 *		path DDP in order case, this is not updated to save fast path
 *		instructions; instead rcv_nxt defines both the TCP and DDP
 *		mainland edge.
 *		Supplied by host for passive open.  Initialized by adapter for
 *		active open on receipt of SYN/ACK or SYN segment.
 *		In both cases the proper initial value is IRS + 1 (same as
 *              rcv_nxt).
 *		This is set equal to rcv_nxt on promote to iWARP"
 * @rcv_buf_size: Maximum receive buffer size that the connection is allowed
 *		to consume.  Units are TCP payload bytes (ie sequence number
 *		bytes).
 * @rcv_max_wnd: Maximum value for receive window to be advertised.  This can
 *		be less than rcv_buf_size, so that in cases when the transmit
 *		rate is not limited by the receive application, the advertised
 *		window will remain constant. This is allows less frequent buffer
 *		usage updates from host without causing bursty behaviour from
 *		the sender.
 *
 * This message is used for receive window updates, both for rx window flow
 * control updates (updates to rcv_buf as data is consumed by the application
 * on the host) and for maximum receive window size updates (when the receive
 * buffer size changes on the host)
 * If the update is successful there will be no response from the adapter.
 * If an error occurred the adapter will generate a TCP Update Failure UMQ
 * message with the reason for the failure.
 */
void
__hal_tcp_update_rxwin_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 rcv_buf_valid,
	u32 rcv_buf_size_valid,
	u32 rcv_max_wnd_valid,
	u32 rcv_buf,
	u32 rcv_buf_size,
	u32 rcv_max_wnd)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_TCP_UPDATE_RXWIN,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_tcp_update_rxwin_req_dmq_t));

	dmsg->m.tcp_rxwin.control_0 =
	    VXGE_HAL_TCP_UPDATE_RXWIN_RBUF(rcv_buf);

	if (rcv_buf_valid)
		dmsg->m.tcp_rxwin.control_0 |=
		    VXGE_HAL_TCP_UPDATE_RXWIN_RB_VALID;

	if (rcv_buf_size_valid)
		dmsg->m.tcp_rxwin.control_0 |=
		    VXGE_HAL_TCP_UPDATE_RXWIN_RBSZ_VALID;

	if (rcv_max_wnd_valid)
		dmsg->m.tcp_rxwin.control_0 |=
		    VXGE_HAL_TCP_UPDATE_RXWIN_RW_VALID;

	dmsg->m.tcp_rxwin.control_0 =
	    vxge_os_htonll(dmsg->m.tcp_rxwin.control_0);

	dmsg->m.tcp_rxwin.control_1 = vxge_os_htonll(
	    VXGE_HAL_TCP_UPDATE_RXWIN_RBUF_SIZE(rcv_buf_size) |
	    VXGE_HAL_TCP_UPDATE_RXWIN_RWND_MAX(rcv_max_wnd));
}

/*
 * __hal_tcp_update_mss_req_dmsg_set - Set TCP Update Mss Request
 *					Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @snd_mss_valid: Does the send_mss field contain a value to be updated?
 *		  1 = snd_mss field is valid
 *		  0 = snd_mss field is not valid
 * @snd_mss_limit_valid: Does the send_mss_limit field contain a value to be
 *		updated?
 *		  1 = snd_mss_limit field is valid
 *		  0 = snd_mss_limit field is not valid
 * @snd_mss: Max segment size for transmit.This is defined as the maximum
 *		permitted TCP payload size, accounting for the maximum TCP
 *		option size (40B if both SACK and timestamps are enabled)
 *		  - snd_mss = MSS - max_tcp_options max_tcp_options is 40B
 *		assuming both SACK and timestamps are enabled (note that the
 *		firmware might choose to use a fixed value of 40 even if SACK or
 *		timestamps are not enabled - this is TBD). The definition is
 *		slightly different from most TCP stacks which store the true MSS
 *		value (max payload assuming no IP or TCP options). It allows the
 *		same amount of payload whether SACK is sent or not - important
 *		for sending aligned DDP segments. Most stacks typically pack the
 *		max possible data into a segment, and resegment on retransmit
 *		if the options size has grown.When sending the MSS option on SYN
 *		or SYN/ACK segments, the PE firmware calculates the true MSS
 *		value based on this field (ie MSS = snd_mss + max_tcp_options).
 *		Supplied by host for both active and passive open.In active open
 *		case, adapter may adjust value to the minimum of the host
 *		supplied value and the value from the received MSS option.
 * @snd_mss_limit: Value of MSS option received from far end on the
 *		SYN or SYN/ACK segment.The PE applies this as a limit on snd_mss
 *		when the host signals path MTU changes as follows -
 *		snd_mss = min(new_mss_from_host,snd_mss_limit) - max_tcp_options
 *		Supplied by host for passive open.
 *		Set by adapter for active open based on the MSS option value on
 *		the received SYN or SYN/ACK segment.
 *
 * This DMQ is sent by the host to the adapter to update the MSS for the session
 * If the update is successful there will be no response from the adapter. If an
 * error occurred the adapter will generate a TCP Update Failure UMQ message
 * with the reason for the failure.
 */
void
__hal_tcp_update_mss_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 snd_mss_valid,
	u32 snd_mss_limit_valid,
	u32 snd_mss,
	u32 snd_msss_limit)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_TCP_UPDATE_MSS,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_tcp_update_mss_req_dmq_t));

	dmsg->m.tcp_mss.control_0 =
	    VXGE_HAL_TCP_UPDATE_MSS_MSS(snd_mss) |
	    VXGE_HAL_TCP_UPDATE_MSS_MSS_LIMIT(snd_msss_limit);

	if (snd_mss_valid)
		dmsg->m.tcp_mss.control_0 |=
		    VXGE_HAL_TCP_UPDATE_RXWIN_MSS_VALID;

	if (snd_msss_limit)
		dmsg->m.tcp_mss.control_0 |=
		    VXGE_HAL_TCP_UPDATE_MSS_MSS_LIM_VALID;

	dmsg->m.tcp_mss.control_0 = vxge_os_htonll(dmsg->m.tcp_mss.control_0);
}

/*
 * __hal_tcp_update_ip_header_req_dmsg_set - Set TCP Update IP Header
 *					Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @ip_tos_valid: Does the ip_tos field contain a value to be updated?
 *		1 = ip_tos valid
 *		1 = ip_tos not valid
 * @ip_flowlabel_valid: Does the ip_flowlabel field contain a value to be
 *		updated?
 *		1 = ip_flowlabel valid
 *		0 = ip_flowlabel not valid
 * @ip_tos: Value to place in the IPv4 TOS field or the Ipv6 Traffic Class
 *		field.  The 2 lower bits are overwritten if ECN is enabled.
 * @ip_flowlabel: Value to place in the IPv6 Flow Label field.  This might
 *		be moved to the H2L context.
 *
 * The TCP Update IP Header DMQ is used to update the IP TOS and IP flow label
 * in the IP header. If the update is successful there will be no response from
 * the adapter. If an error occurred the adapter will generate a TCP Update
 * Failure UMQ message with the reason for the failure.
 */
void
__hal_tcp_update_ip_header_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 ip_tos_valid,
	u32 ip_flowlabel_valid,
	u32 ip_tos,
	u32 ip_flowlabel)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_TCP_UPDATE_IP_HEADER,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_tcp_update_ip_header_req_dmq_t));

	dmsg->m.tcp_ipheader.control_0 =
	    VXGE_HAL_TCP_UPDATE_IPHEAD_TOS(ip_tos) |
	    VXGE_HAL_TCP_UPDATE_IPHEAD_FLOW_LABEL(ip_flowlabel);

	if (ip_tos_valid)
		dmsg->m.tcp_ipheader.control_0 |=
		    VXGE_HAL_TCP_UPDATE_IPHEAD_TOS_VALID;

	if (ip_flowlabel_valid)
		dmsg->m.tcp_ipheader.control_0 |=
		    VXGE_HAL_TCP_UPDATE_IPHEAD_FL_VALID;

	dmsg->m.tcp_ipheader.control_0 =
	    vxge_os_htonll(dmsg->m.tcp_ipheader.control_0);
}

/*
 * __hal_tcp_update_keepalive_req_dmsg_set - Set TCP Update Keepalive
 *					Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @kpa_ok: Enable for keepalive probes for this connection.
 * @probes_max: Keepalive maximum probe count: maximum number of unanswered
 *		keepalive probes before giving up on session.
 * @idle_time: Keepalive idle timer: the idle time before sending the first
 *		keepalive probe (default is 2 hours).
 * @retry_time: Keepalive retry time interval: idle time before resending an
 *		unanswered keepalive probe.
 *
 * The TCP Update Keepalive DMQ message is sent from the host to the adapter to
 * update the keep-alive timer for the session.If the update is successful there
 * will be no response from the adapter. If an error occurred the adapter will
 * generate a TCP Update Failure UMQ message with the reason for the failure.
 */
void
__hal_tcp_update_keepalive_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 kpa_ok,
	u32 probes_max,
	u32 idle_time,
	u32 retry_time)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_TCP_UPDATE_KEEPALIVE,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_tcp_update_keepalive_req_dmq_t));

	dmsg->m.tcp_keepalive.control_0 =
	    VXGE_HAL_TCP_UPDATE_KPA_PROBES_MAX(probes_max);

	if (kpa_ok)
		dmsg->m.tcp_keepalive.control_0 |=
		    VXGE_HAL_TCP_UPDATE_KPA_KPA_OK;

	dmsg->m.tcp_keepalive.control_0 =
	    vxge_os_htonll(dmsg->m.tcp_keepalive.control_0);

	dmsg->m.tcp_keepalive.control_1 = vxge_os_htonll(
	    VXGE_HAL_TCP_UPDATE_KPA_IDLE_TIME(idle_time) |
	    VXGE_HAL_TCP_UPDATE_KPA_RETRY_TIME(retry_time));
}

/*
 * __hal_tcp_relink_to_nce_req_dmsg_set - Set TCP Relink to NCE Request DMQ
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @nce_id : NCE Id
 *
 * Sets the TCP Relink to NCE Request downp Message.
 */
void
__hal_tcp_relink_to_nce_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 nce_id)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_TCP_RELINK_TO_NCE_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_tcp_relink_to_nce_req_dmq_t));

	dmsg->m.tcp_relink.control_0 = vxge_os_htonll(
	    VXGE_HAL_TCP_RLINK_NCE_REQ_NCE_ID(nce_id));
}

/*
 * __hal_lro_end_classification_req_dmsg_set - Set LRO End Classification
 *						  Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 *
 * Sets the LRO End Classification Request DMQ is sent by the host before the
 * LRO Delete Request DMQ to tell the adapter to stop steering Rx frames from
 * that session into the LRO path.
 */
void
__hal_lro_end_classification_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_LRO_END_CLASSIF_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE);
}

/*
 * __hal_lro_delete_req_dmsg_set - Set LRO Delete Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 *
 * Sets the LRO Delete Request DMQ message used by the host to delete an LRO
 * session on the adapter.
 */
void
__hal_lro_delete_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_LRO_DELETE_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE);
}

/*
 * vxge_hal_lro_delete_resp_umsg_get - Get values from LRO Delete Response
 *				     Up Message
 *
 * @umsgh: Handle to umsg.
 * @session_opaque_handle: Buffer to return session Opaque handle
 *
 * Gets the values from LRO Delete Response Up Message.
 */
void
vxge_hal_lro_delete_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	vxge_hal_opaque_handle_t **session_opaque_handle)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (session_opaque_handle != NULL));

	*session_opaque_handle = (vxge_hal_opaque_handle_t *)(ptr_t)
	    vxge_os_ntohll(umsg->m.lro_delete.session_opaque_handle);
}

/*
 * vxge_hal_lro_session_candidate_umsg_get - Get values from LRO Delete Response
 *
 * @umsgh: Handle to umsg.
 * @l4sp: Buffer to return layer 4 source port
 * @l4dp: Buffer to return layer 4 destination port
 * @ipv4_not_ipv6: Buffer to return flag to specify IP Address type
 * @ip_sa_low: Buffer to return lower 64 bits of IPv4 or IPv6 source address
 * @ip_sa_high: Buffer to return higher 64 bits of IPv4 or IPv6 source address
 * @ip_da_low: Buffer to return lower 64 bits of IPv4 or IPv6
 *		destination address
 * @ip_da_high: Buffer to return higher 64 bits of IPv4 or IPv6
 *		destination address
 *
 * Get the values from the msg that indicates to the host that the adapter's
 * autoLRO feature has identified a candidate LRO session.
 */
void
vxge_hal_lro_session_candidate_umsg_get(
	vxge_hal_up_msg_h umsgh,
	u32 *l4sp,
	u32 *l4dp,
	u32 *ipv4_not_ipv6,
	u64 *ip_sa_low,
	u64 *ip_sa_high,
	u64 *ip_da_low,
	u64 *ip_da_high)
{
	u64 val64;
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (l4sp != NULL) &&
	    (l4dp != NULL) && (ipv4_not_ipv6 != NULL) &&
	    (ip_sa_low != NULL) && (ip_sa_high != NULL) &&
	    (ip_da_low != NULL) && (ip_da_high != NULL));

	val64 = vxge_os_ntohll(umsg->m.lro_sess_cand.control_0);

	*l4sp = (u32)VXGE_HAL_LRO_SESS_CAND_GET_L4_SRC_PORT(val64);

	*l4dp = (u32)VXGE_HAL_LRO_SESS_CAND_GET_L4_DEST_PORT(val64);

	*ipv4_not_ipv6 = (u32)VXGE_HAL_LRO_SESS_CAND_GET_IPV4_NOT_IPV6(val64);

	*ip_sa_low = vxge_os_ntohll(umsg->m.lro_sess_cand.ip_sa.low);

	*ip_sa_high = vxge_os_ntohll(umsg->m.lro_sess_cand.ip_sa.high);

	*ip_da_low = vxge_os_ntohll(umsg->m.lro_sess_cand.ip_da.low);

	*ip_da_high = vxge_os_ntohll(umsg->m.lro_sess_cand.ip_da.high);
}

/*
 * __hal_spdm_open_req_dmsg_set - Set SPDM Open Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 * @session_opaque_handle: Opaque handle used by the host to reference data
 *		structures in host memory related to this session
 * @l4sp: Layer 4 source port
 * @l4dp: Layer 4 destination port
 * @use_ipv6: Flag to specify if IPv6 is used
 * @use_vlan: Flag to specify if vlan is used
 * @vlan_id: vlan id
 * @ip_sa_low: Lower 64 bits of IPv4 or IPv6 source address
 * @ip_sa_high: Higher 64 bits of IPv4 or IPv6 source address
 * @ip_da_low: Lower 64 bits of IPv4 or IPv6 destination address
 * @ip_da_high: Higher 64 bits of IPv4 or IPv6 destination address
 *
 * Sets the SPDM Open Request DMQ message sent by the host to open
 * a SPDM connection on the adapter.
 */
void
__hal_spdm_open_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle,
	vxge_hal_opaque_handle_t *session_opaque_handle,
	u32 l4sp,
	u32 l4dp,
	u32 use_ipv6,
	u32 use_vlan,
	u32 vlan_id,
	u64 ip_sa_high,
	u64 ip_sa_low,
	u64 ip_da_high,
	u64 ip_da_low)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert(dmsg != NULL);

	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SPDM_OPEN_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE +
	    sizeof (__hal_spdm_open_req_dmq_t));

	dmsg->m.spdm_open.session_opaque_handle = vxge_os_htonll(
	    (u64)(ptr_t)session_opaque_handle);

	dmsg->m.spdm_open.control_0 =
	    VXGE_HAL_SPDM_OPEN_REQ_L4_SRC_PORT(l4sp) |
	    VXGE_HAL_SPDM_OPEN_REQ_L4_DEST_PORT(l4dp) |
	    VXGE_HAL_SPDM_OPEN_REQ_VLAN_ID(vlan_id);

	if (use_ipv6)
		dmsg->m.spdm_open.control_0 |= VXGE_HAL_SPDM_OPEN_REQ_USE_IPV6;

	if (use_vlan)
		dmsg->m.spdm_open.control_0 |= VXGE_HAL_SPDM_OPEN_REQ_USE_VLAN;

	dmsg->m.spdm_open.control_0 =
	    vxge_os_htonll(dmsg->m.spdm_open.control_0);

	dmsg->m.spdm_open.ip_sa.high = vxge_os_htonll(ip_sa_high);

	dmsg->m.spdm_open.ip_sa.low = vxge_os_htonll(ip_sa_low);

	dmsg->m.spdm_open.ip_da.high = vxge_os_htonll(ip_da_high);

	dmsg->m.spdm_open.ip_da.low = vxge_os_htonll(ip_da_low);
}

/*
 * vxge_hal_spdm_open_resp_umsg_get - Get the values from SPDM Open Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @session_hash: Buffer to return classification hash used to classify the
 *		session context.
 *
 * Get the valueas from SPDM Open Response UMQ used to communicate to
 * the host the success or failure of the corresponding SPDM Open
 * Request DMQ message
 */
void
vxge_hal_spdm_open_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	u32 *session_hash)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (session_hash != NULL));

	*session_hash = (u32)VXGE_HAL_SPDM_OPEN_RESP_GET_SESS_HASH(
	    vxge_os_ntohll(umsg->m.spdm_open.control_0));
}

/*
 * __hal_spdm_delete_req_dmsg_set - Set SPDM Delete Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 *
 * Sets the SPDM Delete Request DMQ message used by the host to delete an SPDM
 * session on the adapter.
 */
void
__hal_spdm_delete_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SPDM_DELETE_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE);
}

/*
 * vxge_hal_spdm_delete_resp_umsg_get - Get values from SPDM Delete Response
 *				    Up Message
 *
 * @umsgh: Handle to umsg.
 * @session_opaque_handle: Buffer to return session Opaque handle
 *
 * Gets the values from SPDM Delete Response Up Message.
 */
void
vxge_hal_spdm_delete_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	vxge_hal_opaque_handle_t **session_opaque_handle)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (session_opaque_handle != NULL));

	*session_opaque_handle = (vxge_hal_opaque_handle_t *)(ptr_t)
	    vxge_os_ntohll(umsg->m.spdm_delete.session_opaque_handle);
}

/*
 * __hal_session_query_req_dmsg_set - Set session query Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @msg_id: Message Id
 * @intr_ctrl: Indicates that an interrupt is required upon completion of
 *		message processing
 * @compl_umq: Indicates that a UMQ message to the host is required upon
 *		completion of message processing
 * @opaque_handle: Opaque handle
 *
 * Sets the session query Request DMQ message used by the host to query a
 * session on the adapter.
 */
void
__hal_session_query_req_dmsg_set(
	vxge_hal_down_msg_h dmsgh,
	u32 msg_id,
	u32 intr_ctrl,
	u32 compl_umq,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_down_msg_header_set(dmsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SESSION_QUERY_REQ,
	    intr_ctrl, compl_umq, opaque_handle,
	    VXGE_HAL_DOWN_MSG_HEADER_SIZE);
}

/*
 * vxge_hal_session_query_resp_umsg_get - Get the values from a Session query
 *				 response UMQ message
 *
 * @umsgh: Handle to umsg.
 * @sess_info: Pointer to structure to return the session attributes
 *
 * Get the values from the Session Query Response UMQ message is sent in
 * response to a Session Query Request DMQ message to return the attributes
 * associated with the specified session
 */
void
vxge_hal_session_query_resp_umsg_get(
	vxge_hal_up_msg_h umsgh,
	vxge_hal_session_info_t *sess_info)
{
	u64 val64;
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (sess_info != NULL));

	val64 = vxge_os_ntohll(umsg->m.session_query.control_0);

	sess_info->sess_type =
	    (u32)VXGE_HAL_SESSION_QUERY_RESP_GET_SESS_TYPE(val64);

	sess_info->vp_id =
	    (u32)VXGE_HAL_SESSION_QUERY_RESP_GET_VP_ID(val64);

	sess_info->sess_inst_num =
	    (u32)VXGE_HAL_SESSION_QUERY_RESP_GET_INST_NUM(val64);

	sess_info->is_spdm =
	    (u32)VXGE_HAL_SESSION_QUERY_RESP_GET_IS_SPDM(val64);

	sess_info->use_ipv6 =
	    (u32)VXGE_HAL_SESSION_QUERY_RESP_GET_USE_IPV6(val64);

	sess_info->use_vlan =
	    (u32)VXGE_HAL_SESSION_QUERY_RESP_GET_USE_VLAN(val64);

	sess_info->vlan_id =
	    (u32)VXGE_HAL_SESSION_QUERY_RESP_GET_VLAN_ID(val64);

	sess_info->rx_imsn =
	    (u32)VXGE_HAL_SESSION_QUERY_RESP_GET_RX_IMSN(val64);

	val64 = vxge_os_ntohll(umsg->m.session_query.control_1);

	sess_info->fw_ctrl =
	    (u32)VXGE_HAL_SESSION_QUERY_RESP_GET_FW_CTRL(val64);

	sess_info->src_port =
	    (u32)VXGE_HAL_SESSION_QUERY_RESP_GET_SRC_PORT(val64);

	sess_info->dest_port =
	    (u32)VXGE_HAL_SESSION_QUERY_RESP_GET_DEST_PORT(val64);

	sess_info->ip_sa_high =
	    vxge_os_ntohll(umsg->m.session_query.ip_sa.high);

	sess_info->ip_sa_low =
	    vxge_os_ntohll(umsg->m.session_query.ip_sa.low);

	sess_info->ip_da_high =
	    vxge_os_ntohll(umsg->m.session_query.ip_da.high);

	sess_info->ip_da_low =
	    vxge_os_ntohll(umsg->m.session_query.ip_da.low);

	val64 = vxge_os_ntohll(umsg->m.session_query.control_2);

	sess_info->use_mpa_crc232 =
	    (u32)VXGE_HAL_SESSION_QUERY_RESP_GET_USE_MPA_CRC32C(val64);
	sess_info->use_mpa_markers =
	    (u32)VXGE_HAL_SESSION_QUERY_RESP_GET_USE_MPA_MARKERS(val64);
}


/*
 * vxge_hal_session_return_in_progress_wqes_umsg_get - Get the values from
 *			 Session Return in progress WQEs UMQ Message
 *
 * @umsgh: Handle to umsg.
 * @number_of_wqes_in_msg: Buffer to return The number of WQE opaque handles
 *		in this message
 * @number_of_wqes_remaining: Buffer to return the number of WQE opaque handles
 *		left to send
 * @wqe_opaque_handles: Array to return the WQE Opaque Handles
 *
 * Get the values from Session Return in progress WQEs UMQ used when deleting a
 * session to return any WQEs that might still be in play for the SRQ.
 */
void
vxge_hal_session_return_in_progress_wqes_umsg_get(
	vxge_hal_up_msg_h umsgh,
	u32 *number_of_wqes_in_msg,
	u32 *number_of_wqes_remaining,
	u64 *wqe_opaque_handles)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;
	u64 val64;
	u32 i;
	u64 *wqe_handles;

	vxge_assert((umsg != NULL) && (number_of_wqes_in_msg != NULL) &&
	    (number_of_wqes_remaining != NULL) &&
	    (wqe_opaque_handles != NULL));

	val64 = vxge_os_ntohll(umsg->m.session_return_wqes.control_0);

	*number_of_wqes_in_msg =
	    (u32)VXGE_HAL_SESSION_RETURN_WQES_GET_NUM_IN_MSG(val64);

	*number_of_wqes_remaining =
	    (u32)VXGE_HAL_SESSION_RETURN_WQES_GET_NUM_REMAIN(val64);

	wqe_handles = umsg->m.session_return_wqes.wqe_opaque_handles;

	for (i = 0; i < (*number_of_wqes_in_msg); i++) {
		*wqe_opaque_handles++ = vxge_os_ntohll(*wqe_handles++);
	}

}

/*
 * vxge_hal_session_frame_write_umsg_get - Get the values from Session Frame
 *				 Write UMQ Message
 *
 * @umsgh: Handle to umsg.
 * @umq_opaque_data: Buffer to return a code provided to the RPE by the RxPE
 *		firmware.
 * @num_markers: Buffer to return the number of markers contained in the frame,
 *		or portion of the frame, that will be passed up the UMQ.
 * @first_marker_index: Buffer to return the first byte of frame data passed
 *		to the host is considered to be Byte 0.
 * @fdata: Buffer to return the pointer to variable length data of the frame
 *		pulled from the frame buffer by the ROCRC.
 * @fdata_length: Length of fdata
 *
 * Get the values from Session Frame Write UMQ used by adapter to send
 * certain frames to the host via the UMQ path instead of the normal path..
 */
void vxge_hal_session_frame_write_umsg_get(
	vxge_hal_up_msg_h umsgh,
	u32 *umq_opaque_data,
	u32 *num_markers,
	u32 *first_marker_index,
	u64 *fdata,
	u32 *fdata_length)
{
	u64 val64;
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert((umsg != NULL) && (umq_opaque_data != NULL) &&
	    (num_markers != NULL) && (first_marker_index != NULL) &&
	    (fdata != NULL));

	val64 = vxge_os_ntohll(umsg->m.session_frame_write.control_0);

	*umq_opaque_data =
	    (u32)VXGE_HAL_SESSION_FRAME_WRITE_GET_UMQ_OPAQUE_DATA(val64);

	*num_markers =
	    (u32)VXGE_HAL_SESSION_FRAME_WRITE_GET_NUM_MARKERS(val64);

	*first_marker_index =
	    (u32)VXGE_HAL_SESSION_FRAME_WRITE_GET_FIRST_MARKER_IDX(val64);

	*fdata = (u64)(ptr_t)&umsg->m.session_frame_write.fdata[0];

	*fdata_length =
	    (u32)(VXGE_HAL_UP_MSG_GET_LENGTH(vxge_os_ntohll(umsg->control_2)) -
	    ((u64)(*fdata) - ((u64)(ptr_t)umsg)));
}


#ifdef	RNIC_LOOPBACK

/*
 * __hal_down_msg_header_get - Get the header fields of dmsg
 * @dmsgh: Handle to dmsg.
 * @msg_id: Buffer to return message Id
 * @msg_type: Buffer to return message Type
 * @intr_ctrl: Buffer to return Interrupt control flag
 * @compl_umq: Buffer to return if UMQ should be posted up on completion.
 * @opaque_handle: Buffer to return opaque handle
 * @length: Buffer to return length of message
 *
 * This function is used to get the header fields of umsg.
 */
void
__hal_down_msg_header_get(vxge_hal_down_msg_h dmsgh, u32 *msg_id,
		u32 *msg_type, u32 *intr_ctrl,
		u32 *compl_umq, vxge_hal_opaque_handle_t **opaque_handle,
		u32 *length)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (msg_id != NULL) && (msg_type != NULL) &&
	    (intr_ctrl != NULL) && (compl_umq != NULL) &&
	    (opaque_handle != NULL) && (length != NULL));

	val64 = vxge_os_ntohll(dmsg->control_0);
	*msg_id = (u32)VXGE_HAL_DOWN_MSG_GET_ID_NUMBER(val64);
	*msg_type = (u32)VXGE_HAL_DOWN_MSG_GET_MESSAGE_TYPE(val64);
	*intr_ctrl = (u32)VXGE_HAL_DOWN_MSG_GET_INTERRUPT_CONTROL(val64);
	*compl_umq = (u32)VXGE_HAL_DOWN_MSG_GET_GEN_COMPL_UMQ(val64);

	*opaque_handle =
	    (vxge_hal_opaque_handle_t *)(ptr_t)vxge_os_ntohll(dmsg->control_1);

	*length =
	    (u32)VXGE_HAL_DOWN_MSG_GET_LENGTH(vxge_os_ntohll(dmsg->control_2));
}

/*
 * __hal_up_msg_header_set - Set the header fields of umsg
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @msg_type: Message Type
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @length: Length of message
 * @result: Result code of the operation requested in corresponding DMQ
 *		message.
 *
 * This function is used to set the header fields of umsg.
 */
void
__hal_up_msg_header_set(vxge_hal_up_msg_h umsgh, u32 msg_id, u32 msg_type,
		    u32 compl_dmq, vxge_hal_opaque_handle_t *opaque_handle,
		    u32 length, u32 result)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	umsg->control_0 = vxge_os_htonll(
	    VXGE_HAL_UP_MSG_ID_NUMBER(msg_id) |
	    VXGE_HAL_UP_MSG_MESSAGE_TYPE(msg_type) |
	    VXGE_HAL_UP_MSG_GEN_COMPL_DMQ(compl_dmq));
	umsg->control_1 = vxge_os_htonll((u64)(ptr_t)opaque_handle);
	umsg->control_2 = vxge_os_htonll(
	    VXGE_HAL_UP_MSG_LENGTH(length) |
	    VXGE_HAL_UP_MSG_RESULT(result));
}

/* =========================== NCE MESSAGES ============================ */
/*
 * __hal_nce_create_req_dmsg_get - Get the values from NCE Create Request
 *
 * @dmsgh: Handle to dmsg.
 * @nce_opaque_handle: Buffer to return NCE opaque handle
 * @src_mac_addr: Buffer to return Source Mac Address
 * @rem_mac_addr: Buffer to return Remote Mac Address
 * @vlan_id: Buffer to return Valn Id
 * @host_rch_time: Buffer to return Host Reach Time
 *
 * Gets the values from NCE Create Request DMQ message
 */
void
__hal_nce_create_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	vxge_hal_opaque_handle_t **nce_opaque_handle,
	u64 *src_mac_addr,
	u64 *rem_mac_addr,
	u32 *vlanId,
	u32 *host_rch_time)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (nce_opaque_handle != NULL) &&
	    (src_mac_addr != NULL) && (rem_mac_addr != NULL) &&
	    (vlanId != NULL) && (host_rch_time != NULL));

	*nce_opaque_handle = (vxge_hal_opaque_handle_t *)(ptr_t)
	    vxge_os_ntohll(dmsg->m.nce_create.nce_opaque_handle);

	*src_mac_addr = (u64)VXGE_HAL_NCE_CREATE_REQ_GET_SRC_MAC(
	    vxge_os_ntohll(dmsg->m.nce_create.control_0));

	*rem_mac_addr = (u64)VXGE_HAL_NCE_CREATE_REQ_GET_RMT_MAC(
	    vxge_os_ntohll(dmsg->m.nce_create.control_1));

	val64 = vxge_os_ntohll(dmsg->m.nce_create.control_2);

	*vlanId = (u32)VXGE_HAL_NCE_CREATE_REQ_GET_VLAN_ID(val64);

	*host_rch_time = (u32)VXGE_HAL_NCE_CREATE_REQ_GET_RCH_TIME(val64);
}

/*
 * __hal_nce_create_resp_umsg_set - Set NCE Create Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result : Result
 *
 * Sets the NCE Create Response Up Message.
 */
void
__hal_nce_create_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_NCE_CREATE_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    result);
}

/*
 * __hal_nce_delete_resp_umsg_set - Set NCE Delete Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @nce_opaque_handle: NCE Opaque handle
 * @result : Result
 *
 * Sets the NCE Delete Response Up Message.
 */
void
__hal_nce_delete_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	vxge_hal_opaque_handle_t *nce_opaque_handle,
	u32 result)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_NCE_DELETE_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE + sizeof (__hal_nce_delete_resp_umq_t),
	    result);

	umsg->m.nce_delete.nce_opaque_handle = vxge_os_htonll(
	    (u64)(ptr_t)nce_opaque_handle);
}

/*
 * __hal_nce_update_mac_addr_req_dmsg_get - Get the values from NCE
 *						Update MAC Request DMQ
 *
 * @dmsgh: Handle to dmsg.
 * @mac_addr: Buffer to return Mac Address
 *
 * Get the values from the message to modify the MAC address for an NCE
 * on the adapter.
 */
void
__hal_nce_update_mac_addr_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u64 *mac_addr)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (mac_addr != NULL));

	*mac_addr = (u64)VXGE_HAL_NCE_UPDATE_MAC_REQ_GET_MAC_ADDR(
	    vxge_os_ntohll(dmsg->m.nce_update_mac.control_0));
}

/*
 * __hal_nce_update_mac_addr_resp_umsg_set - Set the NCE Update MAC Response
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result : Result
 *
 * Sets the NCE Update MAC Response Up Message.
 */
void
__hal_nce_update_mac_addr_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_NCE_UPDATE_MAC_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    result);
}

/*
 * __hal_nce_update_rch_time_req_dmsg_get - Get the values from NCE
 *						Update RCH Time Request DMQ
 *
 * @dmsgh: Handle to dmsg.
 * @rch_time: Buffer to return Host reachability time
 *
 * Get the values from the message to modify the RCH Time for an NCE
 * on the adapter.
 */
void
__hal_nce_update_rch_time_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u64 *rch_time)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (rch_time != NULL));

	*rch_time = VXGE_HAL_NCE_UPDATE_RCH_REQ_GET_RCH_TIME(
	    vxge_os_ntohll(dmsg->m.nce_update_rch.control_0));
}

/*
 * __hal_nce_update_rch_time_resp_umsg_set - Set the NCE Update
 *				RCH Time Response
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result : Result
 *
 * Sets the NCE Update RCH Time Response Up Message.
 */
void
__hal_nce_update_rch_time_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_NCE_UPDATE_RCH_TIME_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    result);
}

/*
 * __hal_nce_query_resp_umsg_set - Set NCE Query Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result : Result
 * @nce_opaque_handle: NCE opaque handle
 * @src_mac_addr: Source Mac Address
 * @rem_mac_addr: Remote Mac Address
 * @host_rch_time: Host Reachability Time
 * @adapter_rch_time: Adapter Reachability Time
 * @valid_flag: valid flag for NCE
 * @vlan_id: Valn Id
 *
 * Sets the NCE Query Response Up Message.
 */
void
__hal_nce_query_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result,
	vxge_hal_opaque_handle_t *nce_opaque_handle,
	u64 src_mac_addr,
	u64 rem_mac_addr,
	u64 host_rch_time,
	u64 adapter_rch_time,
	u32 valid_flag,
	u32 vlan_id)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_NCE_QUERY_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE +
	    sizeof (__hal_nce_query_resp_umq_t),
	    result);

	umsg->m.nce_query.nce_opaque_handle = vxge_os_htonll(
	    (u64)(ptr_t)nce_opaque_handle);

	umsg->m.nce_query.control_0 = vxge_os_htonll(
	    VXGE_HAL_NCE_QUERY_RESP_SRC_MAC(src_mac_addr));

	umsg->m.nce_query.control_1 = vxge_os_htonll(
	    VXGE_HAL_NCE_QUERY_RESP_RMT_MAC(rem_mac_addr));

	umsg->m.nce_query.control_2 = vxge_os_htonll(
	    VXGE_HAL_NCE_QUERY_RESP_HOST_RCH_TIME(host_rch_time) |
	    VXGE_HAL_NCE_QUERY_RESP_ADPT_RCH_TIME(adapter_rch_time));

	umsg->m.nce_query.control_3 = vxge_os_htonll(
	    VXGE_HAL_NCE_QUERY_RESP_VALID_FLAG(valid_flag) |
	    VXGE_HAL_NCE_QUERY_RESP_VLAN_ID(vlan_id));
}



/* ============================ CQRQ MESSAGES =============================== */
/*
 * __hal_cqrq_create_req_dmsg_get - Get the values from CQRQ Create Request
 *
 * @dmsgh: Handle to dmsg.
 * @cq_opaque_handle: Buffer to return CQ Opaque Handle
 * @interrupt_number: Buffer to return Interrupt Number used
 * @bitmap_id: Buffer to return Bitmap position
 * @max_cqe_groups: Buffer to return Maximum number of CQE groups
 * @refill_threshold_high: Buffer to return Upperbound for automatic adapter
 *		refill of CQEs
 * @refill_threshold_low: Buffer to return Lowerbound for automatic adapter
 *		refill of CQEs
 * @low_threshold_notify: Buffer to return Generate Interrupt or notification
 *		if number of CQEs falls below the Lower Threshhold
 * @event_notification_mask: Buffer to return Event notifications mask
 * @cqe_period: Buffer to return CQE size
 * @cqrq_ir: Buffer to return host system address to which this CQRQ
 *		reports the number of returned CQEs
 *
 * Gets the values from CQRQ Create Request DMQ message
 */
void
__hal_cqrq_create_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	vxge_hal_opaque_handle_t **cq_opaque_handle,
	u32 *interrupt_number,
	u32 *bitmap_id,
	u32 *max_cqe_groups,
	u32 *refill_threshold_high,
	u32 *refill_threshold_low,
	u32 *low_threshold_notify,
	u32 *event_notification_mask,
	u32 *cqe_period,
	u64 *cqrq_ir)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (cq_opaque_handle != NULL) &&
	    (interrupt_number != NULL) && (bitmap_id != NULL) &&
	    (max_cqe_groups != NULL) && (refill_threshold_high != NULL) &&
	    (refill_threshold_low != NULL) &&
	    (low_threshold_notify != NULL) &&
	    (event_notification_mask != NULL) && (cqe_period != NULL) &&
	    (cqrq_ir != NULL));

	*cq_opaque_handle = (vxge_hal_opaque_handle_t *)(ptr_t)
	    vxge_os_ntohll(dmsg->m.cqrq_create.cq_opaque_handle);

	val64 = vxge_os_ntohll(dmsg->m.cqrq_create.control_0);

	*interrupt_number =
	    (u32)VXGE_HAL_CQRQ_CREATE_REQ_GET_INT_NUM(val64);

	*bitmap_id = (u32)VXGE_HAL_CQRQ_CREATE_REQ_GET_BITMAP_ID(val64);

	*max_cqe_groups =
	    (u32)VXGE_HAL_CQRQ_CREATE_REQ_GET_MAX_CQE_GRPS(val64);

	*refill_threshold_high =
	    (u32)VXGE_HAL_CQRQ_CREATE_REQ_GET_REF_TH_HIGH(val64);

	val64 = vxge_os_ntohll(dmsg->m.cqrq_create.control_1);

	*refill_threshold_low =
	    (u32)VXGE_HAL_CQRQ_CREATE_REQ_GET_REF_TH_LOW(val64);

	*low_threshold_notify =
	    (u32)VXGE_HAL_CQRQ_CREATE_REQ_GET_LOW_TH_NOTIF(val64);

	*event_notification_mask =
	    (u32)VXGE_HAL_CQRQ_CREATE_REQ_GET_EVENT_MASK(val64);

	*cqe_period = (u32)VXGE_HAL_CQRQ_CREATE_REQ_GET_CQE_PERIOD(val64);

	*cqrq_ir = vxge_os_ntohll(dmsg->m.cqrq_create.cqrq_ir);
}

/*
 * __hal_cqrq_create_resp_umsg_set - Set CQRQ Create Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result : Result
 *
 * Sets the CQRQ Create Response Up Message.
 */
void
__hal_cqrq_create_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_CQRQ_CREATE_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    result);
}

/*
 * __hal_cqrq_delete_resp_umsg_set - Set CQRQ Delete Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @cq_opaque_handle: CQ Opaque handle
 * @result : Result
 *
 * Sets the CQRQ Delete Response Up Message.
 */
void
__hal_cqrq_delete_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	vxge_hal_opaque_handle_t *cq_opaque_handle,
	u32 result)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_CQRQ_DELETE_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE+sizeof (__hal_cqrq_delete_resp_umq_t),
	    result);

	umsg->m.cqrq_delete.cq_opaque_handle = vxge_os_htonll(
	    (u64)(ptr_t)cq_opaque_handle);
}

/*
 * __hal_cqrq_modify_req_dmsg_get - Get values from CQRQ Modify Request Down
 *				   Message
 *
 * @dmsgh: Handle to dmsg.
 * @max_cqe_groups: Buffer to return Maximum number of CQE groups
 * @refill_threshold_high: Buffer to return Upperbound for automatic adapter
 *		refill of CQEs
 * @refill_threshold_low: Buffer to return Lowerbound for automatic adapter
 *		refill of CQEs
 * @low_threshold_notify: Buffer to return Generate Interrupt or notification
 *		if number of CQEs falls below the Lower Threshhold
 * @event_notification_mask: Buffer to return Event notifications mask
 *
 * Gets the values from CQRQ Modify Request DMQ message used by the host to
 * modify a CQRQ on the adapter.
 */
void
__hal_cqrq_modify_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *max_cqe_groups,
	u32 *refill_threshold_high,
	u32 *refill_threshold_low,
	u32 *low_threshold_notify,
	u32 *event_notification_mask)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (max_cqe_groups != NULL) &&
	    (refill_threshold_high != NULL) &&
	    (refill_threshold_low != NULL) &&
	    (low_threshold_notify != NULL) &&
	    (event_notification_mask != NULL));

	val64 = vxge_os_ntohll(dmsg->m.cqrq_modify.control_0);

	*max_cqe_groups = (u32)VXGE_HAL_CQRQ_MODIFY_REQ_GET_MAX_CQE_GRPS(val64);

	*refill_threshold_high =
	    (u32)VXGE_HAL_CQRQ_MODIFY_REQ_GET_REF_TH_HIGH(val64);

	*refill_threshold_low =
	    (u32)VXGE_HAL_CQRQ_MODIFY_REQ_GET_REF_TH_LOW(val64);

	*low_threshold_notify =
	    (u32)VXGE_HAL_CQRQ_MODIFY_REQ_GET_LOW_TH_NOTIF(val64);

	*event_notification_mask = (u32)VXGE_HAL_CQRQ_MODIFY_REQ_GET_EVENT_MASK(
	    vxge_os_ntohll(dmsg->m.cqrq_modify.control_1));

}

/*
 * __hal_cqrq_modify_resp_umsg_set - Set CQRQ Modify Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result : Result
 *
 * Sets the CQRQ Modify Response Up Message.
 */
void
__hal_cqrq_modify_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_CQRQ_MODIFY_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    result);
}

/*
 * __hal_cqrq_query_resp_umsg_set - Set CQRQ Query Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result: Result of query
 * @cq_opaque_handle: CQ Opaque Handle
 * @cqe_period: CQE size
 * @bitmap_id: Bitmap position
 * @max_cqe_groups: Maximum number of CQE groups
 * @refill_threshold_high: Upperbound for automatic adapter refill of CQEs
 * @refill_threshold_low: Lowerbound for automatic adapter refill of CQEs
 * @low_threshold_notify: Generate Interrupt or notification if number of CQEs
 *		falls below the Lower Threshhold
 * @umq: UMQ through which this CQRQ communicates messages
 * @event_notification_mask: Event notifications mask
 * @interrupt_number: Interrupt Number used
 * @enable: Enable CQRQ opeartions
 * @tombstoned: Tombstone operation.
 * @eol: Flag specifies if the CQRQ is in end of list operation
 * @drained: Drained Condition
 * @refill: Flag specifies if refill operation is in progress
 * @cqe_block_wrapper_hsa: Host system address within the wrapper of the last
 *		 CQE block DMAed to the adapter.
 * @cqe_block_wrapper_num_bytes: Number of bytess (in next host CQE block)
 *		 specified by the wrapper of the last CQE block DMAed to the
 *		 adapter.
 * @cq_indication_cqes_reported: Number of CQEs provided to the CQRQ.
 * @cqrq_ir: Host system address to which this CQRQ reports the number
 *		 of returned CQEs
 *
 * Sets the CQRQ Query Response Up Message.
 */
void
__hal_cqrq_query_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result,
	vxge_hal_opaque_handle_t *cq_opaque_handle,
	u32 cqe_period,
	u32 bitmap_id,
	u32 max_cqe_groups,
	u32 refill_threshold_high,
	u32 refill_threshold_low,
	u32 low_threshold_notify,
	u32 umq,
	u32 event_notification_mask,
	u32 interrupt_number,
	u32 enable,
	u32 tombstoned,
	u32 eol,
	u32 drained,
	u32 refill,
	u64 cqe_block_wrapper_hsa,
	u32 cqe_block_wrapper_num_bytes,
	u32 cq_indication_cqes_reported,
	u64 cqrq_ir)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_CQRQ_QUERY_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE +
	    sizeof (__hal_cqrq_query_resp_umq_t),
	    result);

	umsg->m.cqrq_query.cq_opaque_handle = vxge_os_htonll(
	    (u64)(ptr_t)cq_opaque_handle);

	umsg->m.cqrq_query.control_0 = vxge_os_htonll(
	    VXGE_HAL_CQRQ_QUERY_RESP_CQE_PERIOD(cqe_period) |
	    VXGE_HAL_CQRQ_QUERY_RESP_BITMAP_ID(bitmap_id) |
	    VXGE_HAL_CQRQ_QUERY_RESP_MAX_CQE_GRPS(max_cqe_groups) |
	    VXGE_HAL_CQRQ_QUERY_RESP_REF_TH_HIGH(refill_threshold_high));

	umsg->m.cqrq_query.control_1 =
	    VXGE_HAL_CQRQ_QUERY_RESP_REF_TH_LOW(low_threshold_notify) |
	    VXGE_HAL_CQRQ_QUERY_RESP_LOW_TH_NOTIF(low_threshold_notify) |
	    VXGE_HAL_CQRQ_QUERY_RESP_UMQ(umq) |
	    VXGE_HAL_CQRQ_QUERY_RESP_EVENT_MASK(event_notification_mask) |
	    VXGE_HAL_CQRQ_QUERY_RESP_INT_NUM(interrupt_number);

	if (enable)
		umsg->m.cqrq_query.control_1 |= VXGE_HAL_CQRQ_QUERY_RESP_ENABLE;

	if (tombstoned)
		umsg->m.cqrq_query.control_1 |=
		    VXGE_HAL_CQRQ_QUERY_RESP_TOMBSTONED;

	if (eol)
		umsg->m.cqrq_query.control_1 |= VXGE_HAL_CQRQ_QUERY_RESP_EOL;

	if (drained)
		umsg->m.cqrq_query.control_1 |=
		    VXGE_HAL_CQRQ_QUERY_RESP_DRAINED;

	if (refill)
		umsg->m.cqrq_query.control_1 |= VXGE_HAL_CQRQ_QUERY_RESP_REFILL;

	umsg->m.cqrq_query.control_1 =
	    vxge_os_htonll(umsg->m.cqrq_query.control_1);

	umsg->m.cqrq_query.cqe_block_wrapper_hsa =
	    vxge_os_htonll(cqe_block_wrapper_hsa);

	umsg->m.cqrq_query.control_2 = vxge_os_htonll(
	    VXGE_HAL_CQRQ_QUERY_RESP_CBW_BYTES(cqe_block_wrapper_num_bytes) |
	    VXGE_HAL_CQRQ_QUERY_RESP_CBW_CQES(cq_indication_cqes_reported));

	umsg->m.cqrq_query.cqrq_ir = vxge_os_htonll(cqrq_ir);


}

/*
 * __hal_cqrq_arm_req_dmsg_get - Get the values from CQRQ Arm Request
 *				Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @arm_type: Buffer to return The armed state of the CQRQ.
 *
 * Gets the values from CQRQ Arm Request DMQ message used by the host to
 * changed the armed state of CQRQ on the adapter.
 */
void
__hal_cqrq_arm_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *arm_type)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (arm_type != NULL));

	*arm_type = (u32)VXGE_HAL_CQRQ_ARM_REQ_GET_ARM_TYPE(
	    vxge_os_ntohll(dmsg->m.cqrq_arm.control_0));

}

/*
 * __hal_cqrq_arm_resp_umsg_set - Set CQRQ Arm Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result : Result code
 *
 * Sets the CQRQ Arm Response Up Message.
 */
void
__hal_cqrq_arm_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_CQRQ_ARM_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    result);
}

/*
 * __hal_cqrq_event_notification_umsg_set - Set CQRQ Event Notification
 *					   Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @event_type : Event Type
 *
 * Sets the CQRQ Event Notification Up Message.
 */
void
__hal_cqrq_event_notification_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 event_type)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_CQRQ_EVENT_NOTIF,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE +
	    sizeof (__hal_cqrq_event_notification_umq_t),
	    0);

	umsg->m.cqrq_event_notif.control_0 = vxge_os_htonll(
	    VXGE_HAL_CQRQ_EVENT_NOTIF_EVENT_TYPE(event_type));
}

/*
 * __hal_cqrq_fcqebw_req_dmsg_get - Get the values from CQRQ CQRQ First
 *				  CQE BW Notification Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @cqe_block_wrapper_hsa: Buffer to return Address of the first block wrapper
 *		in host memory.
 * @cqe_block_wrapper_bytes: Buffer to return Length of the first block wrapper.
 *
 * Gets the values from CQRQ CQRQ First CQE BW Notification Request DMQ
 * message used by the host to notify the first cqe block to the adapter.
 */
void
__hal_cqrq_fcqebw_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u64 *cqe_block_wrapper_hsa,
	u32 *cqe_block_wrapper_bytes)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (cqe_block_wrapper_hsa != NULL) &&
	    (cqe_block_wrapper_bytes != NULL));

	*cqe_block_wrapper_hsa =
	    vxge_os_ntohll(dmsg->m.cqrq_fcqebw.cqe_block_wrapper_hsa);

	*cqe_block_wrapper_bytes = (u32)VXGE_HAL_CQRQ_FCQEBW_GET_LENGTH(
	    vxge_os_ntohll(dmsg->m.cqrq_fcqebw.control_0));

}

/*
 * __hal_cqrq_fcqebw_resp_umsg_set - Set CQRQ CQRQ First CQE BW Notification
 *					Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @cq_opaque_handle: CQRQ Opaque Handle
 * @result : Result code
 *
 * Sets the CQRQ CQRQ First CQE BW Notification Response Up Message.
 */
void
__hal_cqrq_fcqebw_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_CQRQ_FIRST_CQE_BW_NOTIF_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    result);
}

/* ========================= SRQ MESSAGES ========================== === */
/*
 * __hal_srq_create_req_dmsg_get - Get the values from SRQ Create Request
 *
 * @dmsgh: Handle to dmsg.
 * @srq_opaque_handle: Buffer to return SRQ Opaque Handle
 * @max_wqe_od_groups: Buffer to return maximum number of WQE Headers/OD Groups
 *		that this S-RQ can own at any one time
 * @refill_threshold_high: Buffer to return Upperbound for automatic adapter
 *		refill of WQEs
 * @refill_threshold_low: Buffer to return Lowerbound for automatic adapter
 *		refill of WQEs
 * @low_threshold_notify: Buffer to return Generate Interrupt or notification
 *		if number of WQEs falls below the Lower Threshhold
 * @lro_bytestream: Buffer to return the flag that Indicates whether the S-RQ
 *		is being used for LRO or Bytestream/iWarp.
 * @poll_doorbell: Buffer to return the mode for the EOL operation of this S-RQ.
 *		If set to 1'b1 the mode is poll and WQEs are polling will be
 *		done if the EOL is reached. If set to 1'b0 then the mode is
 *	    "Directed Reading" where a user doorbell message is used to
 *		indicate that more WQEs have been added to the end of the queue
 * @srq_arm: Buffer to return if SRQ limit is armed
 * @event_notification_mask: Buffer to return Event notifications mask
 * @srq_limit_value: Buffer to return the value, if the number of WQEs in the
 *		(S-)RQ (both on-adapter and on-host) falls below which,an (S-)RQ
 *		Limit violation is asserted - if armed
 * @interrupt_number: Buffer to return Interrupt Number used
 * @bitmap_id: Buffer to returnBitmap Identifier
 * @srq_ir: Buffer to return the SRQ indication register.
 *
 * Gets the values from SRQ Create Request DMQ message
 */
void
__hal_srq_create_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	vxge_hal_opaque_handle_t **srq_opaque_handle,
	u32 *max_wqe_od_groups,
	u32 *refill_threshold_high,
	u32 *refill_threshold_low,
	u32 *low_threshold_notify,
	u32 *lro_bytestream,
	u32 *poll_doorbell,
	u32 *srq_arm,
	u32 *event_notification_mask,
	u32 *srq_limit_value,
	u32 *interrupt_number,
	u32 *bitmap_id,
	u64 *srq_ir)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (srq_opaque_handle != NULL) &&
	    (max_wqe_od_groups != NULL) &&
	    (refill_threshold_high != NULL) &&
	    (refill_threshold_low != NULL) &&
	    (low_threshold_notify != NULL) && (lro_bytestream != NULL) &&
	    (poll_doorbell != NULL) && (srq_arm != NULL) &&
	    (event_notification_mask != NULL) &&
	    (srq_limit_value != NULL) && (interrupt_number != NULL) &&
	    (bitmap_id != NULL) && (srq_ir != NULL));

	*srq_opaque_handle = (vxge_hal_opaque_handle_t *)(ptr_t)
	    vxge_os_ntohll(dmsg->m.srq_create.srq_opaque_handle);

	val64 = vxge_os_ntohll(dmsg->m.srq_create.control_0);

	*max_wqe_od_groups =
	    (u32)VXGE_HAL_SRQ_CREATE_REQ_GET_MAX_WQE_OD_GRPS(val64);

	*refill_threshold_high =
	    (u32)VXGE_HAL_SRQ_CREATE_REQ_GET_REF_TH_HIGH(val64);

	*refill_threshold_low =
	    (u32)VXGE_HAL_SRQ_CREATE_REQ_GET_REF_TH_LOW(val64);

	*low_threshold_notify =
	    (u32)VXGE_HAL_SRQ_CREATE_REQ_GET_LOW_TH_NOTIF(val64);

	val64 = vxge_os_ntohll(dmsg->m.srq_create.control_1);

	*lro_bytestream =
	    (u32)VXGE_HAL_SRQ_CREATE_REQ_GET_LRO_BS(val64);

	*poll_doorbell =
	    (u32)VXGE_HAL_SRQ_CREATE_REQ_GET_POLL_DOORBELL(val64);

	*srq_arm =
	    (u32)VXGE_HAL_SRQ_CREATE_REQ_GET_SRQ_ARM(val64);

	*event_notification_mask =
	    (u32)VXGE_HAL_SRQ_CREATE_REQ_GET_EVENT_MASK(val64);

	*srq_limit_value =
	    (u32)VXGE_HAL_SRQ_CREATE_REQ_GET_SRQ_LIMIT(val64);

	val64 = vxge_os_ntohll(dmsg->m.srq_create.control_2);

	*interrupt_number =
	    (u32)VXGE_HAL_SRQ_CREATE_REQ_GET_INT_NUM(val64);

	*bitmap_id =
	    (u32)VXGE_HAL_SRQ_CREATE_REQ_GET_BITMAP_ID(val64);

	*srq_ir = vxge_os_ntohll(dmsg->m.srq_create.srq_ir);

}

/*
 * __hal_srq_create_resp_umsg_set - Set SRQ Create Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result : Result code
 *
 * Sets the SRQ Create Response Up Message.
 */
void
__hal_srq_create_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SRQ_CREATE_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    result);
}

/*
 * __hal_srq_delete_resp_umsg_set - Set SRQ Delete Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @srq_opaque_handle: SRQ Opaque handle
 * @result : result
 *
 * Sets the SRQ Delete Response Up Message.
 */
void
__hal_srq_delete_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	vxge_hal_opaque_handle_t *srq_opaque_handle,
	u32 result)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SRQ_DELETE_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE+sizeof (__hal_srq_delete_resp_umq_t),
	    result);

	umsg->m.srq_delete.srq_opaque_handle = vxge_os_htonll(
	    (u64)(ptr_t)srq_opaque_handle);
}

/*
 * __hal_srq_modify_req_dmsg_get - Get values from SRQ Modify Request Down
 *				   Message
 *
 * @dmsgh: Handle to dmsg.
 * @max_wqe_od_groups: Buffer to return Maximum number of OD groups
 * @refill_threshold_high: Buffer to return Upperbound for automatic adapter
 *		refill of WQEs
 * @refill_threshold_low: Buffer to return Lowerbound for automatic adapter
 *		refill of WQEs
 * @low_threshold_notify: Buffer to return Generate Interrupt or notification
 *		if number of WQEs falls below the Lower Threshhold
 * @srq_arm: Buffer to return if SRQ limit is armed
 * @event_notification_mask: Buffer to return Event notifications mask
 * @srq_limit_value: Buffer to return the number of WQEs in the (S-)RQ (both
 *		on-adapter and on-host) falls below this value, an (S-)RQ Limit
 *		violation is asserted - if armed
 *
 * Gets the values from SRQ Modify Request DMQ message used by the host to
 * modify a SRQ on the adapter.
 */
void
__hal_srq_modify_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *max_wqe_od_groups,
	u32 *refill_threshold_high,
	u32 *refill_threshold_low,
	u32 *low_threshold_notify,
	u32 *srq_arm,
	u32 *event_notification_mask,
	u32 *srq_limit_value)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (max_wqe_od_groups != NULL) &&
	    (refill_threshold_high != NULL) &&
	    (refill_threshold_low != NULL) &&
	    (low_threshold_notify != NULL) &&
	    (srq_arm != NULL) &&
	    (event_notification_mask != NULL) &&
	    (srq_limit_value != NULL));

	val64 = vxge_os_ntohll(dmsg->m.srq_modify.control_0);

	*max_wqe_od_groups =
	    (u32)VXGE_HAL_SRQ_MODIFY_REQ_GET_MAX_WQE_OD_GRPS(val64);

	*refill_threshold_high =
	    (u32)VXGE_HAL_SRQ_MODIFY_REQ_GET_REF_TH_HIGH(val64);

	*refill_threshold_low =
	    (u32)VXGE_HAL_SRQ_MODIFY_REQ_GET_REF_TH_LOW(val64);

	*low_threshold_notify =
	    (u32)VXGE_HAL_SRQ_MODIFY_REQ_GET_LOW_TH_NOTIF(val64);

	val64 = vxge_os_ntohll(dmsg->m.srq_modify.control_1);

	*srq_arm = (u32)VXGE_HAL_SRQ_MODIFY_REQ_GET_SRQ_ARM(val64);

	*event_notification_mask =
	    (u32)VXGE_HAL_SRQ_MODIFY_REQ_GET_EVENT_MASK(val64);

	*srq_limit_value = (u32)VXGE_HAL_SRQ_MODIFY_REQ_GET_SRQ_LIMIT(val64);

}

/*
 * __hal_srq_modify_resp_umsg_set - Set SRQ Modify Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result : Result
 *
 * Sets the SRQ Modify Response Up Message.
 */
void
__hal_srq_modify_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SRQ_MODIFY_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    result);
}

/*
 * __hal_srq_query_resp_umsg_set - Set SRQ Query Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result: Status of the operation
 * @srq_opaque_handle: Host-specified constant allowing the (S-)RQ to identify
 *		itself to the host within any UMQ message generated on behalf of
 *		the (S-)RQ.
 * @virtual_path:  host-specified constant allowing the (S-)RQ to identify
 *		itself to the host within any UMQ message generated on behalf of
 *		the (S-)RQ
 * @max_num_wqe_od_grps: Maximum number of WQE Headers/OD Groups that this S-RQ
 *		can own at any one time.
 * @refill_threshold_high: Hysteresis upper bound for automatic adapter (S-)RQ
 *		WQE replenishment/refill operations.
 * @refill_threshold_low: Hysteresis lower bound for automatic adapter (S-)RQ
 *		WQEs replenishment/refill operations.
 * @low_threshold_notify: If enabled, the host is notified (by interrupt or by
 *		UMQ message) whenever the number of WQEs in the adapter portion
 *		of the SRQ falls below this number.
 * @event_notification_mask: The event notification mask is a mask used to
 *		determine what events the adapter will send a notification to
 *		the host about.
 * @srq_limit_armed: Flag indicates whether the (S-)RQ Limit is armed.
 * @lro_bytestream: Indicates whether the S-RQ is being used for LRO or
 *		Bytestream/iWarp.
 * @srq_state: Current state of the (S-)RQ.
 * @srq_limit_value: If the number of WQEs in the (S-)RQ (both on-adapter and
 *		on-host) falls below this value, an (S-)RQ Limit violation is
 *		asserted - if armed
 * @interrupt_number: Identifies the interrupt used by this (S-)RQ.
 * @bitmap_id: Identifies the bimap bit used by this (S-)RQ.
 * @wb_bytes: This field holds the number of bytess (in next host WQE block)
 *		specified by the wrapper of the last WQE block DMAed to the
 *		adapter.
 * @wb_host_address: This field holds the host system address within the
 *		wrapper of the last WQE block DMAed to the adapter.
 * @srq_ir: This is the host system address to which this SRQ reports the number
 *		of DMAed WQEs from host memory.
 *
 * Set the SRQ Query Response UMQ message used by the adapter to return the
 * configuration of an SRQ on the adapter.
 */
void
__hal_srq_query_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result,
	vxge_hal_opaque_handle_t *srq_opaque_handle,
	u32 virtual_path,
	u32 max_num_wqe_od_grps,
	u32 refill_threshold_high,
	u32 refill_threshold_low,
	u32 low_threshold_notify,
	u32 event_notification_mask,
	u32 srq_limit_armed,
	u32 lro_bytestream,
	u32 srq_state,
	u32 srq_limit_value,
	u32 interrupt_number,
	u32 bitmap_id,
	u32 wb_bytes,
	u64 wb_host_address,
	u64 srq_ir)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SRQ_QUERY_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE +
	    sizeof (__hal_srq_query_resp_umq_t),
	    result);

	umsg->m.srq_query.srq_opaque_handle = vxge_os_htonll(
	    (u64)(ptr_t)srq_opaque_handle);

	umsg->m.srq_query.control_0 = vxge_os_htonll(
	    VXGE_HAL_SRQ_QUERY_RESP_VIRTUAL_PATH(virtual_path) |
	    VXGE_HAL_SRQ_QUERY_RESP_MAX_WQE_OD_GRPS(max_num_wqe_od_grps) |
	    VXGE_HAL_SRQ_QUERY_RESP_REF_TH_HIGH(refill_threshold_high) |
	    VXGE_HAL_SRQ_QUERY_RESP_REF_TH_LOW(refill_threshold_low));

	umsg->m.srq_query.control_1 =
	    VXGE_HAL_SRQ_QUERY_RESP_LOW_TH_NOTIF(low_threshold_notify) |
	    VXGE_HAL_SRQ_QUERY_RESP_EVENT_MASK(event_notification_mask) |
	    VXGE_HAL_SRQ_QUERY_RESP_SRQ_STATE(srq_state);

	if (srq_limit_armed)
		umsg->m.srq_query.control_1 |=
		    VXGE_HAL_SRQ_QUERY_RESP_SRQ_LIM_ARMED;

	if (lro_bytestream)
		umsg->m.srq_query.control_1 |= VXGE_HAL_SRQ_QUERY_RESP_LRO_BS;

	umsg->m.srq_query.control_1 =
	    vxge_os_htonll(umsg->m.srq_query.control_1);

	umsg->m.srq_query.control_2 = vxge_os_htonll(
	    VXGE_HAL_SRQ_QUERY_RESP_SRQ_LIM_VALUE(srq_limit_value));

	umsg->m.srq_query.control_3 = vxge_os_htonll(
	    VXGE_HAL_SRQ_QUERY_RESP_INT_NUM(interrupt_number) |
	    VXGE_HAL_SRQ_QUERY_RESP_BITMAP_ID(bitmap_id) |
	    VXGE_HAL_SRQ_QUERY_RESP_WB_BYTES(wb_bytes));

	umsg->m.srq_query.wb_host_address = vxge_os_htonll(wb_host_address);

	umsg->m.srq_query.srq_ir = vxge_os_htonll(srq_ir);
}

/*
 * __hal_srq_arm_req_dmsg_get - Get the values from SRQ Arm Request
 *				Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @srq_limit: Buffer to return the SRQ limit to set if 'srq_limit_set' is set.
 *
 * Gets the values from SRQ Arm Request DMQ message used by the host to
 * changed the armed state of SRQ on the adapter.
 */
void
__hal_srq_arm_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *srq_limit)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (srq_limit != NULL));

	*srq_limit = (u32)VXGE_HAL_SRQ_ARM_REQ_GET_LIMIT(
	    vxge_os_ntohll(dmsg->m.srq_arm.control_0));

}

/*
 * __hal_srq_arm_resp_umsg_set - Set SRQ Arm Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result : result
 *
 * Sets the SRQ Arm Response Up Message.
 */
void
__hal_srq_arm_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SRQ_ARM_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    result);
}

/*
 * __hal_srq_event_notification_umsg_set - Set SRQ Event Notification
 *					   Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @event_type : Event Type
 *
 * Sets the SRQ Event Notification Up Message.
 */
void
__hal_srq_event_notification_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 event_type)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SRQ_EVENT_NOTIF,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE +
	    sizeof (__hal_srq_event_notification_umq_t),
	    0);

	umsg->m.srq_event_notif.control_0 = vxge_os_htonll(
	    VXGE_HAL_SRQ_EVENT_NOTIF_EVENT_TYPE(event_type));
}

/*
 * __hal_srq_fwqebw_req_dmsg_get - Get the values from SRQ First WQE BW
 *					 Notification Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @wqe_block_host_address: Buffer to return the host address of the first WQE
 *		block wrapper in host memory.
 * @wqe_block_bytes: Buffer to return the number of bytes in the first WQE block
 *		wrapper in host memory.
 * @on_host_tail_wsn: Buffer to return the WQE sequence number (WSN) of the tail
 *		or last WQE that was added to the list.
 *
 * Gets the values from SRQ First WQE BW Notification Request DMQ
 * message used by the host to notify the first wqe block to the adapter.
 */
void
__hal_srq_fwqebw_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u64 *wqe_block_host_address,
	u32 *wqe_block_bytes,
	u32 *on_host_tail_wsn)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (wqe_block_host_address != NULL) &&
	    (wqe_block_bytes != NULL) && (on_host_tail_wsn != NULL));

	*wqe_block_host_address =
	    vxge_os_ntohll(dmsg->m.srq_fwqebw.wqe_block_host_address);

	val64 = vxge_os_ntohll(dmsg->m.srq_fwqebw.control_0);

	*wqe_block_bytes = (u32)VXGE_HAL_SRQ_FWQEBW_REQ_GET_WB_BYTES(val64);

	*on_host_tail_wsn = (u32)VXGE_HAL_SRQ_FWQEBW_REQ_GET_ON_TAIL_WSN(val64);
}

/*
 * __hal_srq_fwqebw_resp_umsg_set - Set SRQ First WQE BW Notification
 *					Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result : Result
 *
 * Sets the SRQ First CQE BW Notification Response Up Message.
 */
void
__hal_srq_fwqebw_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SRQ_FIRST_WQE_BW_NOTIF_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    result);
}

/*
 * __hal_srq_wqebadd_req_dmsg_get - Get values from SRQ WQE Blocks Added
 *				   Notification Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @on_host_tail_wsn: Buffer to return the sequence number of the last or tail
 *		WQE added to the list of WQEs in host memory.
 *
 * Gets the values from SRQ WQE Blocks Added Notification Request DMQ message
 * used by the host to notify the adapter about the added WQE blocks.
 */
void
__hal_srq_wqebadd_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *on_host_tail_wsn)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (on_host_tail_wsn != NULL));

	*on_host_tail_wsn = (u32)VXGE_HAL_SRQ_WQEBADD_REQ_ON_TAIL_WSN(
	    vxge_os_ntohll(dmsg->m.srq_wqebadd.control_0));

}

/*
 * __hal_srq_wqebadd_resp_umsg_set - Set SRQ WQE Blocks Added Notification
 *						Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result : Result
 *
 * Sets the SRQ WQE Blocks Added Notification Response Up Message.
 */
void
__hal_srq_wqebadd_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SRQ_WQE_BLOCKS_ADDED_NOTIF_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    result);
}

/*
 * __hal_srq_return_unused_wqes_req_dmsg_get - Get the values from SRQ Return
 *				WQEs request DMQ message
 *
 * @dmsgh: Handle to dmsg.
 * @number_of_wqes: Buffer to return number of Wqes
 *
 * Gets the values from SRQ return unused WQEs Request DMQ message
 * used by the host to notify the adapter to return WQES.
 */
void
__hal_srq_return_unused_wqes_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *number_of_wqes)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (number_of_wqes != NULL));

	*number_of_wqes = (u32)VXGE_HAL_SRQ_RETURN_WQES_REQ_GET_NUM_WQES(
	    vxge_os_ntohll(dmsg->m.srq_return_wqes.control_0));
}

/*
 * __hal_srq_return_unused_wqes_resp_umsg_set - Set the SRQ Return WQEs
 *				response UMQ message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @srq_opaque_handle: SRQ opaque handle echoed in response
 * @number_of_wqes_in_msg: The number of WQE opaque handles in this message
 * @number_of_wqes_remaining: The number of WQE opaque handles left to send
 * @wqe_opaque_handles: This is a multiple of 64 because this field will contain
 *		multiple WQE opaque handles. The actual number is indicated by
 *		the Number_of_wqes_in_msg.
 *
 * Set the SRQ Return WQEs UMQ used when deleting a TCP session or tearing
 * down an SRQ to return any WQEs that might still be in play for the SRQ.
 */
void
__hal_srq_return_unused_wqes_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	vxge_hal_opaque_handle_t *srq_opaque_handle,
	u32 number_of_wqes_in_msg,
	u32 number_of_wqes_remaining,
	u64 *wqe_opaque_handles)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;
	u32 i;
	u64 *wqe_handles;

	vxge_assert((umsg != NULL) && ((number_of_wqes_in_msg == 0) ||
	    (wqe_opaque_handles != NULL)));

	if (number_of_wqes_in_msg > 10)
		number_of_wqes_in_msg = 10;

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SRQ_RETURN_UNUSED_WQES_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE +
	    sizeof (__hal_srq_return_unused_wqes_resp_umq_t),
	    0);

	umsg->m.srq_return_wqes.srq_opaque_handle = vxge_os_htonll(
	    (u64)(ptr_t)srq_opaque_handle);

	umsg->m.srq_return_wqes.control_0 = vxge_os_htonll(
	    VXGE_HAL_SRQ_RETURN_WQES_RESP_NUM_IN_MSG(number_of_wqes_in_msg) |
	    VXGE_HAL_SRQ_RETURN_WQES_RESP_NUM_REMAIN(number_of_wqes_remaining));

	wqe_handles = umsg->m.srq_return_wqes.wqe_opaque_handles;

	for (i = 0; i < number_of_wqes_in_msg; i++) {
		*wqe_handles++ = vxge_os_htonll(*wqe_opaque_handles++);
	}
}

/* ========================== STAG MESSAGES ================================= */
/*
 * __hal_stag_allocate_nsmr_req_dmsg_get - Get the values from STAG Allocate
 *					  NSMR Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @pd_id: Buffer to return protection domain to which the new NSMR belongs
 * @num_of_pbles: Buffer to return the number of PBLEs that the new NSMR must be
 *		able to accommodate.
 * @stag_index: Buffer to return the STAG Index of the NSMR to be created.
 * @remote_access_flag: Buffer to return the flag to Specify if this NSMR will
 *		be allowed to be used for any remote access operations (read,
 *		write, invalidate) when it is subsequently registered.
 *
 * Gets the values from STAG Allocate NSMR  Request DMQ message used to
 * implement the Allocate Non-Shared Memory Region memory management verb.
 */
void
__hal_stag_allocate_nsmr_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u64 *pd_id,
	u32 *num_of_pbles,
	u32 *stag_index,
	u32 *remote_access_flag)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (pd_id != NULL) &&
	    (num_of_pbles != NULL) && (stag_index != NULL) &&
	    (remote_access_flag != NULL));

	*pd_id = vxge_os_ntohll(dmsg->m.stag_allocate_nsmr.pd_id);

	val64 = vxge_os_ntohll(dmsg->m.stag_allocate_nsmr.control_0);

	*num_of_pbles = (u32)VXGE_HAL_STAG_ALLOC_NSMR_REQ_GET_PBL_LEN(val64);

	*stag_index = (u32)VXGE_HAL_STAG_ALLOC_NSMR_REQ_STAG_IDX(val64);

	*remote_access_flag =
	    (u32)VXGE_HAL_STAG_ALLOC_NSMR_REQ_GET_RAC_FLAG(val64);

}

/*
 * __hal_stag_allocate_nsmr_resp_umsg_set - Set the STAG Allocate NSMR
 *					   Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @stag_index: The STAG Index of the NSMR to be created.
 * @num_of_pbles: The number of PBLEs that the newly-allocated NSMR is capable
 *		of accommodating.
 * @result: Result code
 *
 * Set the STAG Allocate NSMR Response UMQ used to communicate to the host the
 * success or failure of the corresponding STAG Allocate NSMR Request DMQ
 * message
 */
void
__hal_stag_allocate_nsmr_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 stag_index,
	u32 num_of_pbles,
	u32 result)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_STAG_ALLOCATE_NSMR_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE +
	    sizeof (__hal_stag_allocate_nsmr_resp_umq_t),
	    result);

	umsg->m.stag_allocate_nsmr.control_0 = vxge_os_htonll(
	    VXGE_HAL_STAG_ALLOC_NSMR_RESP_STAG_IDX(stag_index) |
	    VXGE_HAL_STAG_ALLOC_NSMR_RESP_PBL_LEN(num_of_pbles));

}

/*
 * __hal_stag_allocate_mw_req_dmsg_get - Get the values from STAG Allocate
 *					  MW Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @stag_index: Buffer to return the host assigned STag index of the MW being
 *		allocated on the adapter.
 * @pd_id: Buffer to return protection domain to which the new NSMR belongs
 *
 * Gets the values from STAG Allocate MW  Request DMQ message used to
 * implement the Allocate Memory Window memory management verb.
 */
void
__hal_stag_allocate_mw_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *stag_index,
	u64 *pd_id)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (stag_index != NULL) && (pd_id != NULL));

	*stag_index = (u32)VXGE_HAL_STAG_ALLOC_MW_REQ_GET_STAG_IDX(
	    vxge_os_ntohll(dmsg->m.stag_allocate_mw.control_0));

	*pd_id = vxge_os_ntohll(dmsg->m.stag_allocate_mw.pd_id);

}

/*
 * __hal_stag_allocate_mw_resp_umsg_set - Set the STAG Allocate MW Response
 *					 UMQ
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @stag_index: The host assigned STag index of the MW being allocated on
 *		the adapter.
 * @result: Result code
 *
 * Set the STAG Allocate MW Response UMQ used to communicate to the host the
 * success or failure of the corresponding STAG Allocate MW Request DMQ message
 */
void
__hal_stag_allocate_mw_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 stag_index,
	u32 result)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_STAG_ALLOCATE_MW_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE +
	    sizeof (__hal_stag_allocate_mw_resp_umq_t),
	    result);

	umsg->m.stag_allocate_mw.control_0 = vxge_os_htonll(
	    VXGE_HAL_STAG_ALLOC_MW_RESP_STAG_IDX(stag_index));

}

/*
 * __hal_stag_deallocate_req_dmsg_get - Get the values from STAG De-Allocate
 *					Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @stag_index: Buffer to return the index of the STag being de-allocated on the
 *		adapter.
 * @dealloc_pbl: Buffer to return the flag to Specify if the PBL that underlies
 *		this STag is also to be deallocated with the STag.
 *
 * Gets the values from STAG De-Allocate Request DMQ message used to deallocate
 * the specified STag, freeing up any on-adapter resources
 */
void
__hal_stag_deallocate_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *stag_index,
	u32 *dealloc_pbl)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (stag_index != NULL) &&
	    (dealloc_pbl != NULL));

	val64 = vxge_os_ntohll(dmsg->m.stag_deallocate.control_0);

	*stag_index = (u32)VXGE_HAL_STAG_DEALLOC_REQ_GET_STAG_IDX(val64);

	*dealloc_pbl = (u32)VXGE_HAL_STAG_DEALLOC_REQ_GET_PBL_FLAG(val64);

}


/*
 * __hal_stag_deallocate_resp_umsg_set - Set the STAG De-allocate Response
 *					UMQ
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @stag_index: The host assigned STag index of the de-allocated STag.
 * @result: Result code
 *
 * Set the STAG De-allocate Response UMQ used to communicate to the host the
 * success or failure of the corresponding STAG De-allocate Request DMQ message
 */
void
__hal_stag_deallocate_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 stag_index,
	u32 result)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_STAG_DE_ALLOCATE_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE +
	    sizeof (__hal_stag_deallocate_resp_umq_t),
	    result);

	umsg->m.stag_deallocate.control_0 = vxge_os_htonll(
	    VXGE_HAL_STAG_DEALLOC_RESP_STAG_IDX(stag_index));

}

/*
 * __hal_stag_register_nsmr_req_dmsg_get - Get the values from STAG Register
 *					  NSMR Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @stag_index: Buffer to return The host assigned STag index to associate with
 *		 the NSMR being registered on the adapter.
 * @stag_key: Buffer to return Instance identifier for this validation of this
 *		 NSMR
 * @pble_size: Buffer to return the size in bytes of each physical buffer in
 *		 the PBL.
 * @pd_id: Buffer to return protection domain to which the new NSMR belongs
 * @first_pbl_block_ptr: Buffer to return the starting address in host memory
 *		of the 1st PBL block for the NSMRs PBL
 * @first_pbl_block_length: Buffer to return the number of PBLEs in the 1st PBL
 *		block
 * @pbl_length: Buffer to return the total number of PBLEs in the PBL.
 * @va_base_flag: Buffer to return the flag to Specify if the NSMR will use
 *		virtual addressing
 * @remote_access_flag: Buffer to return the flag to Specify if this NSMR will
 *		be allowed to be used for any remote access operations (read,
 *		write, invalidate) on this and subsequent registrations.
 * @local_write_flag: Buffer to return the flag to Specify if the adapter can
 *		write to this NSMR.
 * @local_read_flag: Buffer to return the flag to Specify if the adapter can
 *		read from this NSMR.
 * @remote_write_flag: Buffer to return the flag to Specify if remote peers
 *		are allowed to write into this NSMR
 * @remote_read_flag: Buffer to return the flag to Specify if remote peers are
 *		allowed to read from this NSMR.
 * @enable_mw_binding_flag: Buffer to return the flag to Specify if this NSMR
 *		may have MWs bound to it by subsequent PostSQ.BindMW work
 *		requests.
 * @first_byte_offset: Buffer to return the number of bytes into the first PBLE
 *		where the MR begins.
 * @mr_length: Buffer to return the size in bytes of the NSMR.
 * @base_to: Buffer to return the virtual address of the first byte of the NSMR.
 *
 * Gets the values from STAG Register NSMR  Request DMQ message used to
 * implement the Register Non-Shared Memory Region memory management verb.
 */
void
__hal_stag_register_nsmr_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *stag_index,
	u32 *stag_key,
	u32 *pble_size,
	u64 *pd_id,
	u64 *first_pbl_block_ptr,
	u32 *first_pbl_block_length,
	u32 *pbl_length,
	u32 *va_base_flag,
	u32 *remote_access_flag,
	u32 *local_write_flag,
	u32 *local_read_flag,
	u32 *remote_write_flag,
	u32 *remote_read_flag,
	u32 *enable_mw_binding_flag,
	u32 *first_byte_offset,
	u64 *mr_length,
	u64 *base_to)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (stag_index != NULL) &&
	    (stag_key != NULL) && (pble_size != NULL) &&
	    (pd_id != NULL) && (first_pbl_block_ptr != NULL) &&
	    (first_pbl_block_length != NULL) && (pbl_length != NULL) &&
	    (va_base_flag != NULL) && (remote_access_flag != NULL) &&
	    (local_write_flag != NULL) && (local_read_flag != NULL) &&
	    (remote_write_flag != NULL) && (remote_read_flag != NULL) &&
	    (enable_mw_binding_flag != NULL) &&
	    (first_byte_offset != NULL) && (mr_length != NULL) &&
	    (base_to != NULL));

	val64 = vxge_os_ntohll(dmsg->m.stag_reg_nsmr.control_0);

	*stag_index = (u32)VXGE_HAL_STAG_REG_NSMR_REQ_GET_STAG_IDX(val64);

	*stag_key = (u32)VXGE_HAL_STAG_REG_NSMR_REQ_GET_STAG_KEY(val64);

	*pble_size = (u32)VXGE_HAL_STAG_REG_NSMR_REQ_GET_PBLE_SIZE(val64);

	*pd_id = vxge_os_ntohll(dmsg->m.stag_reg_nsmr.pd_id);

	*first_pbl_block_ptr =
	    vxge_os_ntohll(dmsg->m.stag_reg_nsmr.first_pbl_block_ptr);

	val64 = vxge_os_ntohll(dmsg->m.stag_reg_nsmr.control_1);

	*first_pbl_block_length =
	    (u32)VXGE_HAL_STAG_REG_NSMR_REQ_GET_FPBLB_LEN(val64);

	*pbl_length =
	    (u32)VXGE_HAL_STAG_REG_NSMR_REQ_GET_NUM_PBLS(val64);

	val64 = vxge_os_ntohll(dmsg->m.stag_reg_nsmr.control_2);

	*first_byte_offset =
	    (u32)VXGE_HAL_STAG_REG_NSMR_REQ_GET_FB_OFFSET(val64);

	*va_base_flag =
	    (u32)VXGE_HAL_STAG_REG_NSMR_REQ_GET_VA_BASE(val64);

	*remote_access_flag =
	    (u32)VXGE_HAL_STAG_REG_NSMR_REQ_GET_RAC_FLAG(val64);

	*local_write_flag =
	    (u32)VXGE_HAL_STAG_REG_NSMR_REQ_GET_LW_FLAG(val64);

	*local_read_flag =
	    (u32)VXGE_HAL_STAG_REG_NSMR_REQ_GET_LR_FLAG(val64);

	*remote_write_flag =
	    (u32)VXGE_HAL_STAG_REG_NSMR_REQ_GET_RW_FLAG(val64);

	*remote_read_flag =
	    (u32)VXGE_HAL_STAG_REG_NSMR_REQ_GET_RR_FLAG(val64);

	*enable_mw_binding_flag =
	    (u32)VXGE_HAL_STAG_REG_NSMR_REQ_GET_MW_BIND_EN(val64);

	*mr_length = vxge_os_ntohll(dmsg->m.stag_reg_nsmr.mr_length);

	*base_to = vxge_os_ntohll(dmsg->m.stag_reg_nsmr.base_to);

}

/*
 * __hal_stag_register_nsmr_resp_umsg_set - Set the STAG Register NSMR
 *					   Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @stag_index: The host assigned STag index of the NSMR that was registered
 *		via the original DMQ. This is returned unmodified from the DMQ
 *		for convenience.
 * @stag_key: The host assigned STag key of the NSMR that was registered via
 *		the original DMQ. This is returned unmodified from the DMQ for
 *		convenience.
 * @num_of_pbles: The number of PBLEs that the newly-allocated NSMR is capable
 *		of accommodating.
 * @result: Result code
 *
 * Set the STAG Register NSMR Response UMQ used to communicate to the host the
 * success or failure of the corresponding STAG Register NSMR Request DMQ
 * message
 */
void
__hal_stag_register_nsmr_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 stag_index,
	u32 stag_key,
	u32 num_of_pbles,
	u32 result)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_STAG_REGISTER_NSMR_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE +
	    sizeof (__hal_stag_register_nsmr_resp_umq_t),
	    result);

	umsg->m.stag_reg_nsmr.control_0 = vxge_os_htonll(
	    VXGE_HAL_STAG_REG_NSMR_RESP_STAG_IDX(stag_index) |
	    VXGE_HAL_STAG_REG_NSMR_RESP_STAG_KEY(stag_key) |
	    VXGE_HAL_STAG_REG_NSMR_RESP_PBL_LEN(num_of_pbles));

}

/*
 * __hal_stag_re_register_nsmr_req_dmsg_get - Get the values from STAG
 *					  Re_register NSMR Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @stag_index: Buffer to return STag index of the original NSMR that is being
 *		 re-registered. The adapter will use the same STag index for the
 *		 re-registered NSMR.
 * @stag_key: Buffer to return Instance identifier for this validation of this
 *		 NSMR
 * @pble_size: Buffer to return the size in bytes of each physical buffer in
 *		 the PBL.
 * @pd_id: Buffer to return protection domain to which the new NSMR belongs
 * @first_pbl_block_ptr: Buffer to return the starting address in host memory
 *		of the 1st PBL block for the NSMRs PBL
 * @first_pbl_block_length: Buffer to return the number of PBLEs in the 1st PBL
 *		block
 * @pbl_length: Buffer to return the total number of PBLEs in the PBL.
 * @va_base_flag: Buffer to return the flag to Specify if the NSMR will use
 *		virtual addressing
 * @remote_access_flag: Buffer to return the flag to Specify if this NSMR will
 *		be allowed to be used for any remote access operations (read,
 *		write, invalidate) on this and subsequent registrations.
 * @local_write_flag: Buffer to return the flag to Specify if the adapter can
 *		write to this NSMR.
 * @local_read_flag: Buffer to return the flag to Specify if the adapter can
 *		read from this NSMR.
 * @remote_write_flag: Buffer to return the flag to Specify if remote peers
 *		are allowed to write into this NSMR
 * @remote_read_flag: Buffer to return the flag to Specify if remote peers are
 *		allowed to read from this NSMR.
 * @enable_mw_binding_flag: Buffer to return the flag to Specify if this NSMR
 *		may have MWs bound to it by subsequent PostSQ.BindMW work
 *		requests.
 * @first_byte_offset: Buffer to return the number of bytes into the first PBLE
 *		where the MR begins.
 * @mr_length: Buffer to return the size in bytes of the NSMR.
 * @base_to: Buffer to return the virtual address of the first byte of the NSMR.
 *
 * Gets the values from STAG Re-register NSMR  Request DMQ message used to
 * implement the Re-register Non-Shared Memory Region memory management verb.
 */
void
__hal_stag_re_register_nsmr_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *stag_index,
	u32 *stag_key,
	u32 *pble_size,
	u64 *pd_id,
	u64 *first_pbl_block_ptr,
	u32 *first_pbl_block_length,
	u32 *pbl_length,
	u32 *va_base_flag,
	u32 *remote_access_flag,
	u32 *local_write_flag,
	u32 *local_read_flag,
	u32 *remote_write_flag,
	u32 *remote_read_flag,
	u32 *enable_mw_binding_flag,
	u32 *first_byte_offset,
	u64 *mr_length,
	u64 *base_to)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (stag_index != NULL) &&
	    (stag_key != NULL) && (pble_size != NULL) &&
	    (pd_id != NULL) && (first_pbl_block_ptr != NULL) &&
	    (first_pbl_block_length != NULL) && (pbl_length != NULL) &&
	    (va_base_flag != NULL) && (remote_access_flag != NULL) &&
	    (local_write_flag != NULL) && (local_read_flag != NULL) &&
	    (remote_write_flag != NULL) && (remote_read_flag != NULL) &&
	    (enable_mw_binding_flag != NULL) &&
	    (first_byte_offset != NULL) && (mr_length != NULL) &&
	    (base_to != NULL));

	val64 = vxge_os_ntohll(dmsg->m.stag_rereg_nsmr.control_0);

	*stag_index = (u32)VXGE_HAL_STAG_RREG_NSMR_REQ_GET_STAG_IDX(val64);

	*stag_key = (u32)VXGE_HAL_STAG_RREG_NSMR_REQ_GET_STAG_KEY(val64);

	*pble_size = (u32)VXGE_HAL_STAG_RREG_NSMR_REQ_GET_PBLE_SIZE(val64);

	*pd_id = vxge_os_ntohll(dmsg->m.stag_rereg_nsmr.pd_id);

	*first_pbl_block_ptr =
	    vxge_os_ntohll(dmsg->m.stag_rereg_nsmr.first_pbl_block_ptr);

	val64 = vxge_os_ntohll(dmsg->m.stag_rereg_nsmr.control_1);

	*first_pbl_block_length =
	    (u32)VXGE_HAL_STAG_RREG_NSMR_REQ_GET_FPBLB_LEN(val64);

	*pbl_length = (u32)VXGE_HAL_STAG_RREG_NSMR_REQ_GET_NUM_PBLS(val64);

	val64 = vxge_os_ntohll(dmsg->m.stag_rereg_nsmr.control_2);

	*first_byte_offset =
	    (u32)VXGE_HAL_STAG_RREG_NSMR_REQ_GET_FB_OFFSET(val64);

	*va_base_flag =
	    (u32)VXGE_HAL_STAG_RREG_NSMR_REQ_GET_VA_BASE(val64);

	*remote_access_flag =
	    (u32)VXGE_HAL_STAG_RREG_NSMR_REQ_GET_RAC_FLAG(val64);

	*local_write_flag =
	    (u32)VXGE_HAL_STAG_RREG_NSMR_REQ_GET_LW_FLAG(val64);

	*local_read_flag =
	    (u32)VXGE_HAL_STAG_RREG_NSMR_REQ_GET_LR_FLAG(val64);

	*remote_write_flag =
	    (u32)VXGE_HAL_STAG_RREG_NSMR_REQ_GET_RW_FLAG(val64);

	*remote_read_flag =
	    (u32)VXGE_HAL_STAG_RREG_NSMR_REQ_GET_RR_FLAG(val64);

	*enable_mw_binding_flag =
	    (u32)VXGE_HAL_STAG_RREG_NSMR_REQ_GET_MW_BIND(val64);

	*mr_length = vxge_os_ntohll(dmsg->m.stag_rereg_nsmr.mr_length);

	*base_to = vxge_os_ntohll(dmsg->m.stag_rereg_nsmr.base_to);

}

/*
 * __hal_stag_re_register_nsmr_resp_umsg_set - Set the STAG Re_register NSMR
 *					  Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @stag_index: The STag index of the original NSMR that is being re-registered.
 *		The adapter will use the same STag index for the re-registered
 *		NSMR.
 * @stag_key: Instance identifier for this validation of this NSMR
 * @num_of_pbles: The number of PBLEs that the newly-allocated NSMR is capable
 *		of accommodating.
 * @result: Result code
 *
 * Set the STAG Re_register NSMR Response UMQ used to communicate to the host
 * the success or failure of the corresponding STAG Re_register NSMR Request
 * DMQ message
 */
void
__hal_stag_re_register_nsmr_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 stag_index,
	u32 stag_key,
	u32 num_of_pbles,
	u32 result)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_STAG_RE_REGISTER_NSMR_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE +
	    sizeof (__hal_stag_re_register_nsmr_resp_umq_t),
	    result);

	umsg->m.stag_rereg_nsmr.control_0 = vxge_os_htonll(
	    VXGE_HAL_STAG_RREG_NSMR_RESP_STAG_IDX(stag_index) |
	    VXGE_HAL_STAG_RREG_NSMR_RESP_STAG_KEY(stag_key) |
	    VXGE_HAL_STAG_RREG_NSMR_RESP_PBL_LEN(num_of_pbles));

}

/*
 * __hal_stag_register_smr_req_dmsg_get - Get the values from STAG Register
 *					  SMR Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @base_mr_stag_index: Buffer to return Base Memory region STag Index.
 * @va_base_flag: Buffer to return the flag to Specify if the SMR will use
 *		virtual addressing
 * @remote_access_flag: Buffer to return the flag to Specify if this SMR will
 *		be allowed to be used for any remote access operations (read,
 *		write, invalidate) on this and subsequent registrations.
 * @local_write_flag: Buffer to return the flag to Specify if the adapter can
 *		write to this SMR.
 * @local_read_flag: Buffer to return the flag to Specify if the adapter can
 *		read from this SMR.
 * @remote_write_flag: Buffer to return the flag to Specify if remote peers
 *		are allowed to write into this SMR
 * @remote_read_flag: Buffer to return the flag to Specify if remote peers are
 *		allowed to read from this SMR.
 * @enable_mw_binding_flag: Buffer to return the flag to Specify if this SMR
 *		may have MWs bound to it by subsequent PostSQ.BindMW work
 *		requests.
 * @stag_index: Buffer to return STag Index of the new STag record that is to
 *		 be created.
 * @stag_key: Buffer to return Instance identifier for this validation of this
 *		 SMR
 * @pd_id: Buffer to return protection domain to which the new SMR belongs
 * @base_to: Buffer to return the virtual address of the first byte of the SMR.
 *
 * Gets the values from STAG Register SMR  Request DMQ message used to
 * implement the Register Shared Memory Region memory management verb.
 */
void
__hal_stag_register_smr_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *base_mr_stag_index,
	u32 *va_base_flag,
	u32 *remote_access_flag,
	u32 *local_write_flag,
	u32 *local_read_flag,
	u32 *remote_write_flag,
	u32 *remote_read_flag,
	u32 *enable_mw_binding_flag,
	u32 *stag_index,
	u32 *stag_key,
	u64 *pd_id,
	u64 *base_to)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (base_mr_stag_index != NULL) &&
	    (va_base_flag != NULL) && (remote_access_flag != NULL) &&
	    (local_write_flag != NULL) && (local_read_flag != NULL) &&
	    (remote_write_flag != NULL) && (remote_read_flag != NULL) &&
	    (enable_mw_binding_flag != NULL) && (stag_index != NULL) &&
	    (stag_key != NULL) && (pd_id != NULL) && (base_to != NULL));

	val64 = vxge_os_ntohll(dmsg->m.stag_reg_smr.control_0);

	*base_mr_stag_index =
	    (u32)VXGE_HAL_STAG_REG_SMR_REQ_GET_MR_STAG_IDX(val64);

	*va_base_flag = (u32)VXGE_HAL_STAG_REG_SMR_REQ_GET_VA_BASE(val64);

	*remote_access_flag =
	    (u32)VXGE_HAL_STAG_REG_SMR_REQ_GET_RAC_FLAG(val64);

	*local_write_flag = (u32)VXGE_HAL_STAG_REG_SMR_REQ_GET_LW_FLAG(val64);

	*local_read_flag = (u32)VXGE_HAL_STAG_REG_SMR_REQ_GET_LR_FLAG(val64);

	*remote_write_flag = (u32)VXGE_HAL_STAG_REG_SMR_REQ_GET_RW_FLAG(val64);

	*remote_read_flag = (u32)VXGE_HAL_STAG_REG_SMR_REQ_GET_RR_FLAG(val64);

	*enable_mw_binding_flag =
	    (u32)VXGE_HAL_STAG_REG_SMR_REQ_GET_MW_BIND(val64);

	*stag_index = (u32)VXGE_HAL_STAG_REG_SMR_REQ_GET_STAG_IDX(val64);

	*stag_key = (u32)VXGE_HAL_STAG_REG_SMR_REQ_GET_STAG_KEY(val64);

	*pd_id = vxge_os_ntohll(dmsg->m.stag_reg_smr.pd_id);

	*base_to = vxge_os_ntohll(dmsg->m.stag_reg_smr.base_to);
}

/*
 * __hal_stag_register_smr_resp_umsg_set - Set the STAG Register SMR Response
 *					  UMQ
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @stag_index: The host assigned STag index of the NSMR that was registered via
 *		the original DMQ. This is returned unmodified from the DMQ for
 *		convenience.
 * @stag_key: The host assigned STag key of the NSMR that was registered via the
 *		original DMQ. This is returned unmodified from the DMQ for
 *		convenience.
 * @result: Result code
 *
 * Set the STAG Register SMR Response UMQ used to communicate to the host the
 * success or failure of the corresponding STAG Register SMR Request DMQ message
 */
void
__hal_stag_register_smr_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 stag_index,
	u32 stag_key,
	u32 result)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_STAG_REGISTER_SMR_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE +
	    sizeof (__hal_stag_register_smr_resp_umq_t),
	    result);

	umsg->m.stag_reg_smr.control_0 = vxge_os_htonll(
	    VXGE_HAL_STAG_REG_SMR_RESP_STAG_IDX(stag_index) |
	    VXGE_HAL_STAG_REG_SMR_RESP_STAG_KEY(stag_key));

}

/*
 * __hal_stag_query_req_dmsg_get - Get values from STAG Query Request Down
 *				  Message
 *
 * @dmsgh: Handle to dmsg.
 * @stag_index: Buffer to return the STag index of the existing on-adapter STag
 *		that is being queried
 *
 * Get the values from the STAG Query Request DMQ message used by the host to
 * query an STAG on the adapter.
 */
void
__hal_stag_query_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *stag_index)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (stag_index != NULL));

	*stag_index = (u32)VXGE_HAL_STAG_QUERY_REQ_GET_STAG_IDX(
	    vxge_os_ntohll(dmsg->m.stag_query.control_0));
}

/*
 * __hal_stag_query_resp_umsg_set - Set STAG Query Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result: Result code
 * @stag_index: Field uniquely identifies the STag record.
 * @stag_key: Instance identifier for the STag.
 * @is_allocated: Record in use flag
 * @is_window: STag a MR or an MW flag
 * @is_zero_based: STag zero-based flag
 * @is_valid: STag in the valid state flag
 * @is_shared: MR an SMR flag
 * @allow_bind: MWs be bound to this STag flag
 * @allow_remote_access: STag be used for remote access (read, write,
 *		invalidate) flag
 * @allow_remote_read: STag be used for remote read access flag
 * @allow_remote_write: STag be used for remote write access flag
 * @allow_local_read: STag be used for local read access flag
 * @allow_local_write: STag be used for local write access flag
 * @pd_id: Protection Domain ID
 * @sid: Session ID
 * @sin: Session Instance Number
 * @length: length (in bytes) of the MR or MW described by this STag.
 * @mr_base_to: TO of first byte in the MR
 * @mw_base_to: TO of the first byte in the MW relative to the underlying MR.
 * @fbo: The offset in bytes into the first physical buffer where the MR begins.
 * @pble_size: size in bytes of each underlying physical buffer
 * @root_pbl_record_id: Identify of the Underlying PBL
 * @pbl_tree_depth: Depth of the PBL tree
 * @parent_stag_record_id: Identity of the parent STag
 * @parent_stag_valid: Parent record id flag
 * @left_subtree_valid: Left subtree for this STag record flag
 * @right_subtree_valid: Right subtree for this STag record flag
 * @left_subtree_record_id: Identity of the STag record at the top of the left
 *		subtree
 * @right_subtree_record_id: Identity of the STag record at the top of the right
 *		subtree
 * @num_mws_bound: The number of MWs bound to this MR
 * @subtree_imbalance: Sub tree depths different flag
 * @max_pbl_length: Maximum PBLE capacity of the current PBL tree.
 * @actual_pbl_length: Actual number of PBLEs in the current PBL tree.
 * @underlying_mr_record_id: Identity of the underlying MR
 *
 * Set the STAG Query Response UMQ message used by the adapter to return the
 * configuration of an STAG on the adapter.
 */
void
__hal_stag_query_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result,
	u32 stag_index,
	u32 stag_key,
	u32 is_allocated,
	u32 is_window,
	u32 is_zero_based,
	u32 is_valid,
	u32 is_shared,
	u32 allow_bind,
	u32 allow_remote_access,
	u32 allow_remote_read,
	u32 allow_remote_write,
	u32 allow_local_read,
	u32 allow_local_write,
	u64 pd_id,
	u32 sid,
	u32 sin,
	u64 length,
	u64 mr_base_to,
	u64 mw_base_to,
	u32 fbo,
	u32 pble_size,
	u32 root_pbl_record_id,
	u32 pbl_tree_depth,
	u32 parent_stag_record_id,
	u32 parent_stag_valid,
	u32 left_subtree_valid,
	u32 right_subtree_valid,
	u32 left_subtree_record_id,
	u32 right_subtree_record_id,
	u32 num_mws_bound,
	u32 subtree_imbalance,
	u32 max_pbl_length,
	u32 actual_pbl_length,
	u32 underlying_mr_record_id)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_STAG_QUERY_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE +
	    sizeof (__hal_stag_query_resp_umq_t),
	    result);

	umsg->m.stag_query.control_0 =
	    VXGE_HAL_STAG_QUERY_RESP_STAG_IDX(stag_index) |
	    VXGE_HAL_STAG_QUERY_RESP_STAG_KEY(stag_key);

	if (is_allocated)
		umsg->m.stag_query.control_0 |=
		    VXGE_HAL_STAG_QUERY_RESP_IS_ALLOC;

	if (is_window)
		umsg->m.stag_query.control_0 |=
		    VXGE_HAL_STAG_QUERY_RESP_IS_WINDOW;

	if (is_zero_based)
		umsg->m.stag_query.control_0 |=
		    VXGE_HAL_STAG_QUERY_RESP_IS_ZERO_BASE;

	if (is_valid)
		umsg->m.stag_query.control_0 |=
		    VXGE_HAL_STAG_QUERY_RESP_IS_VALID;

	if (is_shared)
		umsg->m.stag_query.control_0 |=
		    VXGE_HAL_STAG_QUERY_RESP_IS_SHARED;

	if (allow_bind)
		umsg->m.stag_query.control_0 |=
		    VXGE_HAL_STAG_QUERY_RESP_ALLOW_BIND;

	if (allow_remote_access)
		umsg->m.stag_query.control_0 |=
		    VXGE_HAL_STAG_QUERY_RESP_ALLOW_RAC;

	if (allow_remote_read)
		umsg->m.stag_query.control_0 |=
		    VXGE_HAL_STAG_QUERY_RESP_ALLOW_RR;

	if (allow_remote_write)
		umsg->m.stag_query.control_0 |=
		    VXGE_HAL_STAG_QUERY_RESP_ALLOW_RW;

	if (allow_local_read)
		umsg->m.stag_query.control_0 |=
		    VXGE_HAL_STAG_QUERY_RESP_ALLOW_LR;

	if (allow_local_write)
		umsg->m.stag_query.control_0 |=
		    VXGE_HAL_STAG_QUERY_RESP_ALLOW_LW;

	umsg->m.stag_query.control_0 =
	    vxge_os_htonll(umsg->m.stag_query.control_0);

	umsg->m.stag_query.pd_id = vxge_os_htonll(pd_id);

	umsg->m.stag_query.control_1 = vxge_os_htonll(
	    VXGE_HAL_STAG_QUERY_RESP_SID(sid) |
	    VXGE_HAL_STAG_QUERY_RESP_SIN(sin));

	umsg->m.stag_query.length = vxge_os_htonll(length);

	umsg->m.stag_query.mr_base_to = vxge_os_htonll(mr_base_to);

	umsg->m.stag_query.mw_base_to = vxge_os_htonll(mw_base_to);

	umsg->m.stag_query.control_2 = vxge_os_htonll(
	    VXGE_HAL_STAG_QUERY_RESP_FBO(fbo) |
	    VXGE_HAL_STAG_QUERY_RESP_PBLE_SIZE(pble_size));

	umsg->m.stag_query.control_3 =
	    VXGE_HAL_STAG_QUERY_RESP_ROOT_PBL_ID(root_pbl_record_id) |
	    VXGE_HAL_STAG_QUERY_RESP_PBL_DEPTH(pbl_tree_depth) |
	    VXGE_HAL_STAG_QUERY_RESP_PSTAG_ID(parent_stag_record_id);

	if (parent_stag_valid)
		umsg->m.stag_query.control_3 |=
		    VXGE_HAL_STAG_QUERY_RESP_PSTAG_VALID;

	if (left_subtree_valid)
		umsg->m.stag_query.control_3 |=
		    VXGE_HAL_STAG_QUERY_RESP_LEFT_VALID;

	if (right_subtree_valid)
		umsg->m.stag_query.control_3 |=
		    VXGE_HAL_STAG_QUERY_RESP_RIGHT_VALID;

	umsg->m.stag_query.control_3 =
	    vxge_os_htonll(umsg->m.stag_query.control_3);

	umsg->m.stag_query.control_4 = vxge_os_htonll(
	    VXGE_HAL_STAG_QUERY_RESP_LEFT_ID(left_subtree_record_id) |
	    VXGE_HAL_STAG_QUERY_RESP_RIGHT_ID(right_subtree_record_id));

	umsg->m.stag_query.control_5 = vxge_os_htonll(
	    VXGE_HAL_STAG_QUERY_RESP_NUM_MW_BOUND(num_mws_bound) |
	    VXGE_HAL_STAG_QUERY_RESP_ST_IMBALANCE(subtree_imbalance) |
	    VXGE_HAL_STAG_QUERY_RESP_MAX_PBL_LEN(max_pbl_length));

	umsg->m.stag_query.control_6 = vxge_os_htonll(
	    VXGE_HAL_STAG_QUERY_RESP_PBL_LEN(actual_pbl_length) |
	    VXGE_HAL_STAG_QUERY_RESP_MR_ID(underlying_mr_record_id));


}

/*
 * __hal_stag_validate_req_dmsg_get - Get the valuse from STAG Validate
 *			   Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @stag_index: Buffer to return the index of the existing on-adapter STag
 *		that is being validated.
 * @stag_key: Buffer to return the key that to assign to the STag when it
 *		is validated.
 *
 * Gets the values from STAG Validate Request DMQ message.
 */
void
__hal_stag_validate_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *stag_index,
	u32 *stag_key)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (stag_index != NULL) &&
	    (stag_key != NULL));

	val64 = vxge_os_ntohll(dmsg->m.stag_validate.control_0);

	*stag_index = (u32)VXGE_HAL_STAG_VALID_LOCAL_REQ_GET_STAG_IDX(val64);

	*stag_key = (u32)VXGE_HAL_STAG_VALID_LOCAL_REQ_GET_STAG_KEY(val64);

}

/*
 * __hal_stag_validate_resp_umsg_set - Set the STAG Validate Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @stag_index: The host assigned STag index of the STag that was validated
 *		via the original DMQ.
 * @stag_key: The host assigned STag key of the STag that was validated via
 *		the original DMQ.
 * @result: Result code
 *
 * Set the STAG Validate Response UMQ used to communicate to the host the
 * success or failure of the corresponding STAG Validate Request DMQ message
 */
void
__hal_stag_validate_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 stag_index,
	u32 stag_key,
	u32 result)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_STAG_VALID_LOCAL_TAG_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE +
	    sizeof (__hal_stag_validate_resp_umq_t),
	    result);

	umsg->m.stag_validate.control_0 = vxge_os_htonll(
	    VXGE_HAL_STAG_VALID_LOCAL_RESP_STAG_IDX(stag_index) |
	    VXGE_HAL_STAG_VALID_LOCAL_RESP_STAG_KEY(stag_key));
}

/*
 * __hal_stag_invalidate_req_dmsg_get - Get STAG InValidate Request Down
 *					  Message
 *
 * @dmsgh: Handle to dmsg.
 * @stag_index: Buffer to return the index of the existing on-adapter STag
 *		that is being invalidated.
 * @stag_key: Buffer to return the key that is currently assigned to the STag.
 * @pd_id: Buffer to return the PD ID of the session or QP.
 *
 * Gets the values from STAG Validate Request DMQ message.
 */
void
__hal_stag_invalidate_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *stag_index,
	u32 *stag_key,
	u64 *pd_id)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (stag_index != NULL) &&
	    (stag_key != NULL) && (pd_id != NULL));

	val64 = vxge_os_ntohll(dmsg->m.stag_invalidate.control_0);

	*stag_index = (u32)VXGE_HAL_STAG_INVALID_REQ_GET_STAG_IDX(val64);

	*stag_key = (u32)VXGE_HAL_STAG_INVALID_REQ_GET_STAG_KEY(val64);

	*pd_id = vxge_os_ntohll(dmsg->m.stag_invalidate.pd_id);

}

/*
 * __hal_stag_invalidate_resp_umsg_set - Set the STAG InValidate Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @stag_index: The host assigned STag index of the STag that was invalidated.
 * @stag_key: The host assigned STag key of the STag that was invalidated.
 * @result: Result code
 *
 * Set the STAG InValidate Response UMQ used to communicate to the host the
 * success or failure of the corresponding STAG InValidate Request DMQ message
 */
void
__hal_stag_invalidate_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 stag_index,
	u32 stag_key,
	u32 result)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_STAG_INVALID_LOCAL_TAG_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE +
	    sizeof (__hal_stag_invalidate_resp_umq_t),
	    result);

	umsg->m.stag_invalidate.control_0 = vxge_os_htonll(
	    VXGE_HAL_STAG_INVALID_RESP_STAG_IDX(stag_index) |
	    VXGE_HAL_STAG_INVALID_RESP_STAG_KEY(stag_key));

}

/*
 * __hal_stag_bind_mw_req_dmsg_get - Get STAG Bind MW Request Down
 *					  Message
 *
 * @dmsgh: Handle to dmsg.
 * @mr_stag_index: Buffer to return the STag Index of the underlying MR to
 *		bind to.
 * @mr_stag_key: Buffer to return the STag Key of the underlying MR to bind to.
 * @mw_stag_index: Buffer to return the existing Memory Window's STag Index.
 * @mw_stag_key: Buffer to return the STag Key to use for the MW.
 * @mr_stag: Buffer to return the existing Memory Region's STAG index.
 *		Least significant 8 bits is the key.
 * @mw_ctrl: Buffer to return control flags for Memory Window.
 * @mw_length: Buffer to return memory window length in bytes.
 * @tagged_offset: Buffer to return memory window tagged offset in bytes.
 *
 * Sets the STAG Bind MW Request DMQ message.
 */
void
__hal_stag_bind_mw_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *mr_stag_index,
	u32 *mr_stag_key,
	u32 *mw_stag_index,
	u32 *mw_stag_key,
	u32 *mr_stag,
	u32 *mw_ctrl,
	u64 *mw_length,
	u64 *tagged_offset)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (mr_stag_index != NULL) &&
	    (mr_stag_key != NULL) && (mw_stag_index != NULL) &&
	    (mw_stag_key != NULL) && (mr_stag != NULL) &&
	    (mw_ctrl != NULL) && (mw_length != NULL) &&
	    (tagged_offset != NULL));


	val64 = vxge_os_ntohll(dmsg->m.stag_bind_mw.control_0);

	*mr_stag_index = (u32)VXGE_HAL_STAG_BIND_MW_REQ_GET_MR_STAG_IDX(val64);

	*mr_stag_key = (u32)VXGE_HAL_STAG_BIND_MW_REQ_GET_MR_STAG_KEY(val64);

	*mw_stag_index = (u32)VXGE_HAL_STAG_BIND_MW_REQ_GET_MW_STAG_IDX(val64);

	*mw_stag_key = (u32)VXGE_HAL_STAG_BIND_MW_REQ_GET_MW_STAG_KEY(val64);

	val64 = vxge_os_ntohll(dmsg->m.stag_bind_mw.control_1);

	*mr_stag = (u32)VXGE_HAL_STAG_BIND_MW_REQ_GET_MR_STAG(val64);

	*mw_ctrl = (u32)VXGE_HAL_STAG_BIND_MW_REQ_GET_MW_CTRL(val64);

	*mw_length = vxge_os_ntohll(dmsg->m.stag_bind_mw.mw_length);

	*tagged_offset = vxge_os_ntohll(dmsg->m.stag_bind_mw.tagged_offset);

}

/*
 * __hal_stag_bind_mw_resp_umsg_set - Set the STAG Bind MW Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @stag_index: The host assigned STag index MW that was bound. This is
 *		returned unmodified from the request for convenience.
 * @stag_key: The host assigned STag key of the MW that was bound. This is
 *		returned unmodified from the request for convenience.
 * @result: Result code
 *
 * Set the STAG Bind MW Response UMQ used to communicate to the host the
 * success or failure of the corresponding STAG Bind MW Request DMQ message
 */
void
__hal_stag_bind_mw_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 stag_index,
	u32 stag_key,
	u32 result)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_STAG_BIND_MW_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE +
	    sizeof (__hal_stag_bind_mw_resp_umq_t),
	    result);

	umsg->m.stag_bind_mw.control_0 = vxge_os_htonll(
	    VXGE_HAL_STAG_BIND_MW_RESP_STAG_IDX(stag_index) |
	    VXGE_HAL_STAG_BIND_MW_RESP_STAG_KEY(stag_key));
}

/*
 * __hal_stag_fast_register_nsmr_req_dmsg_get - Get values from STAG Fast
 *				Register Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @mr_control_size: Buffer to return length in bytes of the Memory Region
 *		control structure.
 * @mr_control_ptr: buffer to return thr pointer to Memory Region control
 *		structure in host memory.
 *
 * Sets the STAG Fast Register Request DMQ message.
 */
void
__hal_stag_fast_register_nsmr_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *mr_control_size,
	u64 *mr_control_ptr)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (mr_control_size != NULL) &&
	    (mr_control_ptr != NULL));

	*mr_control_size = (u32)VXGE_HAL_STAG_FRNSMR_REQ_GET_MR_CTRL_SIZE(
	    vxge_os_ntohll(dmsg->m.stag_frnsmr.control_0));

	*mr_control_ptr = vxge_os_ntohll(dmsg->m.stag_frnsmr.mr_control_ptr);
}

/*
 * __hal_stag_fast_register_nsmr_resp_umsg_set - Set the STAG Fast Register
 *						Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @stag_index: Host assigned STag index of the NSMR that was fast-registered.
 *		This is returned unmodified from the request for convenience.
 * @stag_key: The host assigned STag key of the NSMR that was fast-registerd.
 *		This is returned unmodified from the request for convenience.
 * @result: Result code
 *
 * Set the STAG Fast Register Response UMQ used to communicate to the host the
 * success or failure of the corresponding STAG Fast Register Request DMQ
 * message
 */
void
__hal_stag_fast_register_nsmr_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 stag_index,
	u32 stag_key,
	u32 result)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_STAG_FAST_REGISTER_NSMR_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE +
	    sizeof (__hal_stag_fast_register_nsmr_resp_umq_t),
	    result);

	umsg->m.stag_frnsmr.control_0 = vxge_os_htonll(
	    VXGE_HAL_STAG_FRNSMR_RESP_STAG_IDX(stag_index) |
	    VXGE_HAL_STAG_FRNSMR_RESP_STAG_KEY(stag_key));

}

/* ======================== SESSION MESSAGES ============================= */
/*
 * __hal_tcp_open_req_dmsg_get - Get the values from TCP Open Request
 *				Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @session_opaque_handle: Buffer to return opaque handle used by the host to
 *		reference data structures in host memory related to this session
 * @active_opener: Buffer to return flags to specify if active opener
 * @tcp_source_port: Buffer to return TCP source port
 * @tcp_dest_port: Buffer to return TCP destination port
 * @use_ipv6: Buffer to return flag to specify if IPv4 or IPv6
 * @use_vlan: Buffer to return Indication whether or not this session
 *		supports VLAN.
 * @vlan_id: Buffer to return the VLAN ID used by this session.
 *		Only valid if use_vlan == 1.
 * @ip_sa_low: Buffer to return lower 64 bits of IPv4 or IPv6 source address
 * @ip_sa_high: Buffer to return higher 64 bits of IPv4 or IPv6 source address
 * @ip_da_low: Buffer to return lower 64 bits of IPv4 or IPv6
 *		destination address
 * @ip_da_high: Buffer to return higher 64 bits of IPv4 or IPv6
 *		destination address
 * @bytestream_cqrq_sin: Buffer to return Instance number of the bytestream
 *		CQRQ for the BS.
 * @bytestream_cqrq_sid: Buffer to return adapter ID of the bytestream CQRQ for
 *		the BS.
 * @bytestream_srq_sin: Buffer to return Instance number of the bytestream SRQ
 *		for the BS.
 * @bytestream_srq_sid: Buffer to return adapter ID of the bytestream SRQ for
 *		the BS.
 * @break_merge_on_psh: Buffer to receive the flag to indicate a merge will be
 *		broken if a frame with PSH == 1 is received.
 * @data_merging_enabled: Buffer to return flag to enable merging of TCP
 *		payloads.
 * @no_wqe_threshold: Buffer to return lower limit for number of WQEs available
 *		for adapter.
 * @cqrq_low_latency: Buffer to return low-latency criteria.
 * @send_vlan_up: Buffer to return flag to force VLAN tag insertion on transmit
 *		frames for this session.
 * @wscale_ok: Buffer to return flag to enable for window scaling.
 * @sack_ok: Buffer to return flag to enable for SACKs.
 * @ecn_ok: Buffer to return flag to enable for Explicit Congestion Notification
 * @tstamp_ok: Buffer to return flag to enable for timestamps.
 * @kpalive_ok: Buffer to return flag to enable for keepalive probes for this
 *		connection.
 * @int_num: Buffer to return interrupt number for the session.
 * @int_bitmap: Buffer to return interrupt bitmap number for the session.
 * @nce_id: Buffer to return NCE # for this session.
 * @vlan_user_priority: Buffer to return value to insert in the user_priority
 *		field when VLAN tags are generated on tx frames.
 * @dup_ack_thres: Buffer to return duplicate ACK threshold for fast retransmit.
 * @ip_tos: Buffer to return value to place in the IPv4 TOS field or the Ipv6
 *		Traffic Class field.
 * @ip_flags: Buffer to return value to place in the IPv4 Flags field.
 * @ip_flowlabel: Buffer to return value to place in the IPv6 Flow Label field.
 * @kpa_idle_time: Buffer to return keepalive idle timer
 * @kpa_retry_time: Buffer to return keepalive retry time interval
 * @retx_time_max: Buffer to return maximum time to spent retransmitting one
 *		segment.
 * @snd_wscale: Buffer to return sender window shift scale value
 * @kpa_probes_max: Buffer to return keepalive maximum probe count
 * @snd_nxt: Buffer to return next sequence number to send.
 * @snd_mss: Buffer to return max segment size for transmit.
 * @snd_mss_limit: Buffer to return value of MSS option received from far end
 *		on the SYN or SYN/ACK segment.
 * @rcv_nxt: Buffer to return next byte expected to be received.
 * @rcv_wscale: Buffer to return receiver window shift scale value
 * @rcv_mss: Buffer to return max segment size for receive.
 * @rcv_buf_size: Buffer to return maximum receive buffer size that the
 *		connection is allowed to consume.
 * @rcv_max_wnd: Buffer to return maximum value for receive window to be
 *		advertised.
 * @rcv_buf: Buffer to return lowest TCP sequence number buffered in the
 *		connection's receive buffer.
 * @rttmin: Buffer to return minimum RTT value for this session.
 * @ts_recent: Buffer to return timestamps Enabled
 * @nxt_towi_ptr: Buffer to return pointer to the next TOWI (or TOWI block
 *		wrapper) to be read from host memory.
 * @nxt_towi_cnt: Buffer to return remaining number of TOWIs in the block
 *		containing *nxt_towi_ptr, including *nxt_towi_ptr.
 * @ack_blk_limit: Buffer to return host supplied limit on the maximum number
 *		of ACK list blocks that can be held by this session at any
 *		one time.
 * @hw_cq_sq: Buffer to return pointer to the "hw-cq-sq" location on
 *		the host.
 *
 * Gets the values from TCP Open Request DMQ message sent by the host to open
 * a TCP connection on the adapter.
 */
void
__hal_tcp_open_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	vxge_hal_opaque_handle_t **session_opaque_handle,
	u32 *active_opener,
	u32 *tcp_source_port,
	u32 *tcp_dest_port,
	u32 *use_ipv6,
	u32 *use_vlan,
	u32 *vlan_id,
	u64 *ip_sa_high,
	u64 *ip_sa_low,
	u64 *ip_da_high,
	u64 *ip_da_low,
	u32 *bytestream_cqrq_sin,
	u32 *bytestream_cqrq_sid,
	u32 *bytestream_srq_sin,
	u32 *bytestream_srq_sid,
	u32 *break_merge_on_psh,
	u32 *data_merging_enabled,
	u32 *no_wqe_threshold,
	u32 *cqrq_low_latency,
	u32 *send_vlan_up,
	u32 *wscale_ok,
	u32 *sack_ok,
	u32 *ecn_ok,
	u32 *tstamp_ok,
	u32 *kpalive_ok,
	u32 *int_num,
	u32 *int_bitmap,
	u32 *nce_id,
	u32 *vlan_user_priority,
	u32 *dup_ack_thres,
	u32 *ip_tos,
	u32 *ip_flags,
	u32 *ip_flowlabel,
	u32 *kpa_idle_time,
	u32 *kpa_retry_time,
	u32 *retx_time_max,
	u32 *snd_wscale,
	u32 *kpa_probes_max,
	u32 *snd_nxt,
	u32 *snd_mss,
	u32 *snd_mss_limit,
	u32 *rcv_nxt,
	u32 *rcv_wscale,
	u32 *rcv_mss,
	u32 *rcv_buf_size,
	u32 *rcv_max_wnd,
	u32 *rcv_buf,
	u32 *rttmin,
	u32 *ts_recent,
	u64 *nxt_towi_ptr,
	u32 *nxt_towi_cnt,
	u32 *ack_blk_limit,
	u64 *hw_cq_sq)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (session_opaque_handle != NULL) &&
	    (active_opener != NULL) && (tcp_source_port != NULL) &&
	    (tcp_dest_port != NULL) && (use_ipv6 != NULL) &&
	    (use_vlan != NULL) && (vlan_id != NULL) &&
	    (ip_sa_high != NULL) && (ip_sa_low != NULL) &&
	    (ip_da_high != NULL) && (ip_da_low != NULL) &&
	    (bytestream_cqrq_sin != NULL) &&
	    (bytestream_cqrq_sid != NULL) &&
	    (bytestream_srq_sin != NULL) &&
	    (bytestream_srq_sid != NULL) &&
	    (break_merge_on_psh != NULL) &&
	    (data_merging_enabled != NULL) &&
	    (no_wqe_threshold != NULL) && (cqrq_low_latency != NULL) &&
	    (send_vlan_up != NULL) && (wscale_ok != NULL) &&
	    (sack_ok != NULL) && (ecn_ok != NULL) &&
	    (tstamp_ok != NULL) && (kpalive_ok != NULL) &&
	    (int_num != NULL) && (int_bitmap != NULL) &&
	    (nce_id != NULL) && (vlan_user_priority != NULL) &&
	    (dup_ack_thres != NULL) && (ip_tos != NULL) &&
	    (ip_flags != NULL) && (ip_flowlabel != NULL) &&
	    (kpa_idle_time != NULL) && (kpa_retry_time != NULL) &&
	    (retx_time_max != NULL) && (snd_wscale != NULL) &&
	    (kpa_probes_max != NULL) && (snd_nxt != NULL) &&
	    (snd_mss != NULL) && (snd_mss_limit != NULL) &&
	    (rcv_nxt != NULL) && (rcv_wscale != NULL) &&
	    (rcv_mss != NULL) && (rcv_buf_size != NULL) &&
	    (rcv_max_wnd != NULL) && (rcv_buf != NULL) &&
	    (rttmin != NULL) && (ts_recent != NULL) &&
	    (nxt_towi_ptr != NULL) && (nxt_towi_cnt != NULL) &&
	    (ack_blk_limit != NULL) && (hw_cq_sq != NULL));

	*session_opaque_handle = (vxge_hal_opaque_handle_t *)(ptr_t)
	    vxge_os_ntohll(dmsg->m.tcp_open.session_opaque_handle);

	val64 = vxge_os_ntohll(dmsg->m.tcp_open.control_0);

	*active_opener = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_ACTIVE_OPENER(val64);

	*tcp_source_port = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_TCP_SRC_PORT(val64);

	*tcp_dest_port = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_TCP_DEST_PORT(val64);

	*use_ipv6 = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_USE_IPV6(val64);

	*use_vlan = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_USE_VLAN(val64);

	*vlan_id = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_VLAN_ID(val64);

	*ip_sa_high = vxge_os_ntohll(dmsg->m.tcp_open.ip_sa.high);

	*ip_sa_low = vxge_os_ntohll(dmsg->m.tcp_open.ip_sa.low);

	*ip_da_high = vxge_os_ntohll(dmsg->m.tcp_open.ip_da.high);

	*ip_da_low = vxge_os_ntohll(dmsg->m.tcp_open.ip_da.low);

	val64 = vxge_os_ntohll(dmsg->m.tcp_open.control_1);

	*bytestream_cqrq_sin =
	    (u32)VXGE_HAL_TCP_OPEN_REQ_GET_BS_CQRQ_INST_NUM(val64);

	*bytestream_cqrq_sid =
	    (u32)VXGE_HAL_TCP_OPEN_REQ_GET_BS_CQRQ_ID(val64);

	*bytestream_srq_sin =
	    (u32)VXGE_HAL_TCP_OPEN_REQ_GET_BS_SRQ_INST_NUM(val64);

	*bytestream_srq_sid =
	    (u32)VXGE_HAL_TCP_OPEN_REQ_GET_BS_SRQ_ID(val64);

	val64 = vxge_os_ntohll(dmsg->m.tcp_open.control_2);

	*break_merge_on_psh =
	    (u32)VXGE_HAL_TCP_OPEN_REQ_GET_BREAK_MERGE_EN(val64);

	*data_merging_enabled =
	    (u32)VXGE_HAL_TCP_OPEN_REQ_GET_DATA_MERGE_EN(val64);

	*no_wqe_threshold = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_NO_WQE_TH(val64);

	*cqrq_low_latency = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_CQRQ_LOW_LAT(val64);

	val64 = vxge_os_ntohll(dmsg->m.tcp_open.control_3);

	*send_vlan_up = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_SEND_VLAN_UP(val64);

	*wscale_ok = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_WSCALE_OK(val64);

	*sack_ok = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_SACK_OK(val64);

	*ecn_ok = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_ECN_OK(val64);

	*tstamp_ok = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_TSTAMP_OK(val64);

	*kpalive_ok = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_KEEPALIVE_OK(val64);

	*int_num = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_INT_NUM(val64);

	*int_bitmap = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_INT_BITMAP(val64);

	*nce_id = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_NCE_ID(val64);

	val64 = vxge_os_ntohll(dmsg->m.tcp_open.control_4);

	*vlan_user_priority =
	    (u32)VXGE_HAL_TCP_OPEN_REQ_GET_VLAN_USER_PRI(val64);

	*dup_ack_thres = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_DUP_ACK_TH(val64);

	*ip_tos = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_IP_TOS(val64);

	*ip_flags = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_IP_FLAGS(val64);

	*ip_flowlabel = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_IP_FLOW_LABEL(val64);

	val64 = vxge_os_ntohll(dmsg->m.tcp_open.control_5);

	*kpa_idle_time = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_KPA_IDLE_TIME(val64);

	*kpa_retry_time = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_KPA_RETRY_TIME(val64);

	val64 = vxge_os_ntohll(dmsg->m.tcp_open.control_6);

	*retx_time_max = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_RETX_TIME_MAX(val64);

	*snd_wscale = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_SND_WSCALE(val64);

	*kpa_probes_max = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_KPA_PROBES_MAX(val64);

	val64 = vxge_os_ntohll(dmsg->m.tcp_open.control_7);

	*snd_nxt = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_SND_NXT(val64);

	*snd_mss = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_SND_MSS(val64);

	*snd_mss_limit = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_SND_MSS_LIMIT(val64);

	val64 = vxge_os_ntohll(dmsg->m.tcp_open.control_8);

	*rcv_nxt = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_RCV_NXT(val64);

	*rcv_wscale = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_RCV_WSCALE(val64);

	*rcv_mss = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_RCV_MSS(val64);

	val64 = vxge_os_ntohll(dmsg->m.tcp_open.control_9);

	*rcv_buf_size = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_RCV_BUF_SIZE(val64);

	*rcv_max_wnd = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_RCV_MAX_WND(val64);

	val64 = vxge_os_ntohll(dmsg->m.tcp_open.control_10);

	*rcv_buf = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_RCV_BUF(val64);

	*rttmin = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_RTT_MIN(val64);

	val64 = vxge_os_ntohll(dmsg->m.tcp_open.control_11);

	*ts_recent = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_TS_RECENT(val64);

	*nxt_towi_ptr = vxge_os_ntohll(dmsg->m.tcp_open.nxt_towi_ptr);

	val64 = vxge_os_ntohll(dmsg->m.tcp_open.control_12);

	*nxt_towi_cnt = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_NXT_TOWI_CNT(val64);

	*ack_blk_limit = (u32)VXGE_HAL_TCP_OPEN_REQ_GET_ACK_BLK_LIMIT(val64);

	*hw_cq_sq = vxge_os_ntohll(dmsg->m.tcp_open.hw_cq_sq);
}

/*
 * __hal_tcp_open_resp_umsg_set - Set the TCP Open Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result: Result code
 *
 * Set the TCP Open Response UMQ used to communicate to the host the
 * success or failure of the corresponding TCP Open Request DMQ message
 */
void
__hal_tcp_open_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_TCP_OPEN_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    result);
}

/*
 * __hal_tcp_promote_to_iwarp_req_dmsg_get - Get values from TCP Promote
 *					to iWARP Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @mpa_crc_valid: Buffer to return flag to specify MPA CRCs are expected
 *		to be valid on received frames.
 * @tx_marker_status: Buffer to return flag to specify markers exist in the
 *		transmit direction.
 * @rx_marker_status: Buffer to return flag to specify markers exist in the
 *		receive direction.
 * @iwarp_local_stag_zero_allowed: Buffer to return flag to specify the session
 *		is allowed to use STag = 0 in Rx WQEs.
 * @iwarp_wire_stag_zero_allowed: Buffer to return flag to specify session to
 *		accept STag = 0 in receive frames.
 * @inbound_write_readresp_enable: Buffer to return flag to specify the session
 *		has remote write permission, incoming RDMA WRITE or RDMA READ
 *		RESPONSE messages can be placed
 * @inbound_readreq_enable: Buffer to return buffer to return flag to specify
 *		when a SEND* WITH INVALIDATE message is received, the STag can
 *		only be invalidated if the session has remote write and/or
 *		remote read access.  Also used to ensure that RIRRs are allowed
 *		on this session
 * @srq_pd_id: Buffer to return protection domain to which the associated iWARP
 *		SRQ belongs.
 * @iwarp_qp_pd_id: Buffer to return protection domain of this session.
 * @cqrq_ll_send_frame_data_offset: Buffer to return offset from the CQE base
 *		address.
 * @iwarp_qp_limit_value: Buffer to return in-progress WQEs limit value.
 * @rq_max_length: Buffer to return maximum number of WQEs that this session
 *		is allowed to have in progress on this SRQ.
 * @iwarp_qp_limit_armed: Buffer to return flag to specify the Queue Pair limit
 *		is armed at the iwarp_qp_limit_value
 * @iwarp_cqrq_sin: Buffer to return Instance number of the CQRQ for the iWARP
 *		session.
 * @iwarp_cqrq_sid: Buffer to return adapter ID of the CQRQ for the iWARP
 *		session.
 * @iwarp_srq_sin: Buffer to return Instance number of the SRQ for the iWARP
 *		session.
 * @iwarp_srq_sid: Buffer to returnAdapter ID of the SRQ for the iWARP session.
 * @rdmap_version: Buffer to return  RDMA version number 01 - IETF RNICs
 *		00 - RDMAC RNICs
 * @ddp_version: Buffer to return Direct Data Placement Protocol Version
 *		01 - for protocol described in
 *		http://ietf.org/internet-drafts/draft-ietf-rddp-ddp-06.txt
 * @mpa_reply_frame_size: Buffer to return the size of TCP payload of MPA
 *		reply frame (in bytes) on the Responder. This value is relevant
 *		only for the MPA Responder. It allows the Tx PE to determine the
 *		TCP SN at which iWarp will begin (ie the TxIMSN).
 * @is_mpa_initiator: Buffer to return whether it is the MPA Initiator or the
 *		MPA Responder.
 * @block_permissions: Buffer to return block permissions used to restrict
 *		session's ability to use special TOWI atoms. It is used to
 *		enforce checks such as preventing the QP from performing a Fast
 *		Register NSMR operation.
 * @ird_limit: Buffer to return host supplied IRD limit for this session.
 * @hw_cq_rr_ir: Buffer to returnthe pointer to "hw-cq-rr" location on the host.
 *
 * Gets the values from TCP Promote to iWARP Request DMQ message sent from the
 * host to the adapter in order to migrate an existing bytestream session to
 * iWARP mode.
 */
void
__hal_tcp_promote_to_iwarp_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *mpa_crc_valid,
	u32 *tx_marker_status,
	u32 *rx_marker_status,
	u32 *iwarp_local_stag_zero_allowed,
	u32 *iwarp_wire_stag_zero_allowed,
	u32 *inbound_write_readresp_enable,
	u32 *inbound_readreq_enable,
	u64 *srq_pd_id,
	u64 *iwarp_qp_pd_id,
	u32 *cqrq_ll_send_frame_data_offset,
	u32 *iwarp_qp_limit_value,
	u32 *rq_max_length,
	u32 *iwarp_qp_limit_armed,
	u32 *iwarp_cqrq_sin,
	u32 *iwarp_cqrq_sid,
	u32 *iwarp_srq_sin,
	u32 *iwarp_srq_sid,
	u32 *rdmap_version,
	u32 *ddp_version,
	u32 *mpa_reply_frame_size,
	u32 *is_mpa_initiator,
	u32 *block_permissions,
	u32 *ird_limit,
	u64 *hw_cq_rr_ir)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (mpa_crc_valid != NULL) &&
	    (tx_marker_status != NULL) && (rx_marker_status != NULL) &&
	    (iwarp_local_stag_zero_allowed != NULL) &&
	    (iwarp_wire_stag_zero_allowed != NULL) &&
	    (inbound_write_readresp_enable != NULL) &&
	    (inbound_readreq_enable != NULL) && (srq_pd_id != NULL) &&
	    (iwarp_qp_pd_id != NULL) &&
	    (cqrq_ll_send_frame_data_offset != NULL) &&
	    (iwarp_qp_limit_value != NULL) && (rq_max_length != NULL) &&
	    (iwarp_qp_limit_armed != NULL) && (iwarp_cqrq_sin != NULL) &&
	    (iwarp_cqrq_sid != NULL) && (iwarp_srq_sin != NULL) &&
	    (iwarp_srq_sid != NULL) && (rdmap_version != NULL) &&
	    (ddp_version != NULL) && (mpa_reply_frame_size != NULL) &&
	    (is_mpa_initiator != NULL) && (block_permissions != NULL) &&
	    (ird_limit != NULL) && (hw_cq_rr_ir != NULL));


	val64 = vxge_os_ntohll(dmsg->m.tcp_promote_iwarp.control_0);

	*mpa_crc_valid = (u32)VXGE_HAL_TCP_PROMO_REQ_GET_MPA_CRC_VALID(val64);

	*tx_marker_status = (u32)VXGE_HAL_TCP_PROMO_REQ_GET_TX_MARK_STAT(val64);

	*rx_marker_status = (u32)VXGE_HAL_TCP_PROMO_REQ_GET_RX_MARK_STAT(val64);

	*iwarp_local_stag_zero_allowed =
	    (u32)VXGE_HAL_TCP_PROMO_REQ_GET_IWLS_ZERO_OK(val64);

	*iwarp_wire_stag_zero_allowed =
	    (u32)VXGE_HAL_TCP_PROMO_REQ_GET_IWWS_ZERO_OK(val64);

	*inbound_write_readresp_enable =
	    (u32)VXGE_HAL_TCP_PROMO_REQ_GET_IBWRR_EN(val64);

	*inbound_readreq_enable =
	    (u32)VXGE_HAL_TCP_PROMO_REQ_GET_IBRR_EN(val64);

	*srq_pd_id = vxge_os_ntohll(dmsg->m.tcp_promote_iwarp.srq_pd_id);

	*iwarp_qp_pd_id =
	    vxge_os_ntohll(dmsg->m.tcp_promote_iwarp.iwarp_qp_pd_id);

	val64 = vxge_os_ntohll(dmsg->m.tcp_promote_iwarp.control_1);

	*cqrq_ll_send_frame_data_offset =
	    (u32)VXGE_HAL_TCP_PROMO_REQ_GET_CQRQ_LLSFD_OFF(val64);

	*iwarp_qp_limit_value =
	    (u32)VXGE_HAL_TCP_PROMO_REQ_GET_QP_LIMIT_VALUE(val64);

	*rq_max_length =
	    (u32)VXGE_HAL_TCP_PROMO_REQ_GET_RQ_MAX_LENGTH(val64);

	*iwarp_qp_limit_armed =
	    (u32)VXGE_HAL_TCP_PROMO_REQ_GET_QP_LIMIT_ARMED(val64);

	val64 = vxge_os_ntohll(dmsg->m.tcp_promote_iwarp.control_2);

	*iwarp_cqrq_sin = (u32)VXGE_HAL_TCP_PROMO_REQ_GET_CQRQ_INST_NUM(val64);

	*iwarp_cqrq_sid = (u32)VXGE_HAL_TCP_PROMO_REQ_GET_CQRQ_ID(val64);

	*iwarp_srq_sin = (u32)VXGE_HAL_TCP_PROMO_REQ_GET_SRQ_INST_NUM(val64);

	*iwarp_srq_sid = (u32)VXGE_HAL_TCP_PROMO_REQ_GET_SRQ_ID(val64);

	val64 = vxge_os_ntohll(dmsg->m.tcp_promote_iwarp.control_3);

	*rdmap_version = (u32)VXGE_HAL_TCP_PROMO_REQ_GET_RDMAP_VERSION(val64);

	*ddp_version = (u32)VXGE_HAL_TCP_PROMO_REQ_GET_DDP_VERSION(val64);

	*mpa_reply_frame_size =
	    (u32)VXGE_HAL_TCP_PROMO_REQ_MPA_FRAME_SIZE(val64);

	*is_mpa_initiator =
	    (u32)VXGE_HAL_TCP_PROMO_REQ_GET_MPA_INITIATOR(val64);

	*block_permissions =
	    (u32)VXGE_HAL_TCP_PROMO_REQ_GET_BLOCK_PERMISSIONS(val64);

	*ird_limit = (u32)VXGE_HAL_TCP_PROMO_REQ_GET_IRD_LIMIT(
	    vxge_os_ntohll(dmsg->m.tcp_promote_iwarp.control_4));

	*hw_cq_rr_ir = vxge_os_ntohll(dmsg->m.tcp_promote_iwarp.hw_cq_rr_ir);
}

/*
 * __hal_tcp_promote_to_iwarp_resp_umsg_set - Set the TCP Promote to iWARP
 *					 Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @rx_imsn: 32b per-session constant that identifies the TCP SN
 * @result: Result code
 *
 * Set the TCP Promote to iWARP Response UMQ used to communicate to the host the
 * success or failure of the corresponding TCP Promote to iWARP Request DMQ
 * message
 */
void
__hal_tcp_promote_to_iwarp_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 rx_imsn,
	u32 result)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_TCP_PROMOTE_TO_IWARP_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE +
	    sizeof (__hal_tcp_promote_to_iwarp_resp_umq_t),
	    result);

	umsg->m.tcp_promote_iwarp.control_0 = vxge_os_htonll(
	    VXGE_HAL_TCP_PROMO_RESP_RX_IMSN(rx_imsn));

}

/*
 * __hal_tcp_modify_req_dmsg_get - Gets the values from TCP Modify Request DMQ
 *
 * @dmsgh: Handle to dmsg.
 * @mpa_crc_valid: Buffer to return the flag, If set then MPA CRCs are expected
 *		to be valid on received frames.The adapter always calculates MPA
 *		CRCs in the transmit direction. This field is ignored for a
 *		bytestream session.
 * @tx_marker_mode: Buffer to return the flag, Whether MPA markers exist in the
 *		transmit direction or not.
 *		-- 1 - enable MPA markers in transmit frames
 *		-- 0 - disable MPA markers in transmit frames
 *		This field is ignored for a bytestream session.
 * @rx_marker_mode: Buffer to return the flag, Whether markers exist in the
 *		receive direction or not.
 *		-- 1 - enable MPA markers in receive frames
 *		-- 0 - disable MPA markers in receive frames
 *		If this is set to zero for an RDMA session, the RxPE drops all
 *		out of order receive segments. This field is ignored for a
 *		bytestream session.
 * @rdmap_version: Buffer to return the RDMA version number
 *		  -- 01 - for IETF RNICs
 *		  -- 00 - for RDMAC RNICs
 * @ddp_version: Buffer to return Direct Data Placement Protocol Version
 *		  -- 01 - for protocol described in
 *		http://ietf.org/internet-drafts/draft-ietf-rddp-ddp-06.txt
 * @iwarp_qp_limit_armed: Buffer to return the flag, If set, the Queue Pair
 *		limit is armed at the iwarp_qp_limit_value. This field is
 *		ignored for a bytestream session.
 * @iwarp_qp_limit_value: Buffer to return iwarp qp limit value.
 * @rq_max_length: Buffer to return the maximum number of WQEs that this session
 *		is allowed to have in progress on this SRQ.
 * @block_permissions: Buffer to return Block Permissions.
 * @ird_limit: Buffer to return Host supplied IRD limit for this session.
 *
 * This routine gets the values from TCP Modify Request DMQ message sent by the
 * host to modify the attributes associated with a bytestream or iWARP session.
 */
void
__hal_tcp_modify_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *mpa_crc_valid,
	u32 *tx_marker_mode,
	u32 *rx_marker_mode,
	u32 *rdmap_version,
	u32 *ddp_version,
	u32 *iwarp_qp_limit_armed,
	u32 *iwarp_qp_limit_value,
	u32 *rq_max_length,
	u32 *block_permissions,
	u32 *ird_limit)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (mpa_crc_valid != NULL) &&
	    (tx_marker_mode != NULL) && (rx_marker_mode != NULL) &&
	    (rdmap_version != NULL) && (ddp_version != NULL) &&
	    (iwarp_qp_limit_armed != NULL) &&
	    (iwarp_qp_limit_value != NULL) &&
	    (rq_max_length != NULL) && (block_permissions != NULL) &&
	    (ird_limit != NULL));

	val64 = vxge_os_ntohll(dmsg->m.tcp_modify.control_0);

	*mpa_crc_valid = (u32)VXGE_HAL_TCP_MODIFY_REQ_GET_MPA_CRC_VALID(val64);

	*tx_marker_mode = (u32)VXGE_HAL_TCP_MODIFY_REQ_GET_TX_MARK_MODE(val64);

	*rx_marker_mode = (u32)VXGE_HAL_TCP_MODIFY_REQ_GET_RX_MARK_MODE(val64);

	*rdmap_version = (u32)VXGE_HAL_TCP_MODIFY_REQ_GET_RDMAP_VERSION(val64);

	*ddp_version = (u32)VXGE_HAL_TCP_MODIFY_REQ_GET_DDP_VERSION(val64);

	*iwarp_qp_limit_armed =
	    (u32)VXGE_HAL_TCP_MODIFY_REQ_GET_QP_LIMIT_ARMED(val64);

	*iwarp_qp_limit_value =
	    (u32)VXGE_HAL_TCP_MODIFY_REQ_GET_QP_LIMIT_VALUE(val64);

	val64 = vxge_os_ntohll(dmsg->m.tcp_modify.control_1);

	*rq_max_length = (u32)VXGE_HAL_TCP_MODIFY_REQ_GET_RQ_MAX_LEN(val64);

	*block_permissions =
	    (u32)VXGE_HAL_TCP_MODIFY_REQ_GET_BLOCK_PERMISSIONS(val64);

	*ird_limit = (u32)VXGE_HAL_TCP_MODIFY_REQ_IRD_LIMIT(val64);
}

/*
 * __hal_tcp_modify_resp_umsg_set - Set TCP Modify Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @session_opaque_handle: Session Opaque handle
 * @result : result
 *
 * Sets the TCP Modify Response Up Message.
 */
void
__hal_tcp_modify_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_TCP_MODIFY_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    result);
}

/*
 * __hal_tcp_delete_resp_umsg_set - Set TCP Delete Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @session_opaque_handle: Session Opaque handle
 * @result : result
 *
 * Sets the TCP Delete Response Up Message.
 */
void
__hal_tcp_delete_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	vxge_hal_opaque_handle_t *session_opaque_handle,
	u32 result)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_TCP_DELETE_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE+sizeof (__hal_tcp_delete_resp_umq_t),
	    result);

	umsg->m.tcp_delete.session_opaque_handle = vxge_os_htonll(
	    (u64)(ptr_t)session_opaque_handle);
}

/*
 * __hal_tcp_abort_resp_umsg_set - Set TCP Abort Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result : result
 *
 * Sets the TCP Abort Response Up Message.
 */
void
__hal_tcp_abort_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_TCP_ABORT_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    result);
}

/*
 * __hal_tcp_established_umsg_set - Set TCP Established Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 *
 * Sets the TCP Established Up Message.
 */
void
__hal_tcp_established_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_TCP_ESTABLISHED,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    0);
}

/*
 * __hal_tcp_fin_received_umsg_set - Set TCP FIN Received Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 *
 * Sets the TCP FIN Received Up Message.
 */
void
__hal_tcp_fin_received_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_TCP_FIN_RECEIVED,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    0);
}

/*
 * __hal_tcp_time_wait_done_umsg_set - Set TCP Time Wait Done Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 *
 * Sets the TCP Time Wait Done Up Message.
 */
void
__hal_tcp_time_wait_done_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_TCP_TIME_WAIT_DONE,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    0);
}

/*
 * __hal_tcp_update_rxwin_req_dmsg_get - Get the values from TCP Update
 *					Rxwin Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @rcv_buf_valid: Buffer to return the recv buf valid flag
 * @rcv_buf_size_valid: Buffer to return the recv buf size valid flag
 * @rcv_max_wnd_valid: Buffer to return the recv max window valid flag
 * @rcv_buf: Buffer to return recv Buf
 * @rcv_buf_size: Buffer to return maximum receive buffer size.
 * @rcv_max_wnd: Buffer to return maximum value for receive window to be
 *		advertised.
 *
 * Get the values from TCP Update Rxwin Request Down Message
 */
void
__hal_tcp_update_rxwin_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *rcv_buf_valid,
	u32 *rcv_buf_size_valid,
	u32 *rcv_max_wnd_valid,
	u32 *rcv_buf,
	u32 *rcv_buf_size,
	u32 *rcv_max_wnd)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (rcv_buf_valid != NULL) &&
	    (rcv_buf_size_valid != NULL) &&
	    (rcv_max_wnd_valid != NULL) && (rcv_buf != NULL) &&
	    (rcv_buf_size != NULL) && (rcv_max_wnd != NULL));

	val64 = vxge_os_ntohll(dmsg->m.tcp_rxwin.control_0);

	*rcv_buf_valid = (u32)VXGE_HAL_TCP_UPDATE_RXWIN_GET_RB_VALID(val64);

	*rcv_buf_size_valid =
	    (u32)VXGE_HAL_TCP_UPDATE_RXWIN_GET_RBSZ_VALID(val64);

	*rcv_max_wnd_valid = (u32)VXGE_HAL_TCP_UPDATE_RXWIN_GET_RW_VALID(val64);

	*rcv_buf = (u32)VXGE_HAL_TCP_UPDATE_RXWIN_GET_RBUF(val64);

	val64 = vxge_os_ntohll(dmsg->m.tcp_rxwin.control_1);

	*rcv_buf_size = (u32)VXGE_HAL_TCP_UPDATE_RXWIN_GET_RBUF_SIZE(val64);

	*rcv_max_wnd = (u32)VXGE_HAL_TCP_UPDATE_RXWIN_GET_RWND_MAX(val64);

}

/*
 * __hal_tcp_update_mss_req_dmsg_get - Get values from TCP Update Mss
 *					Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @snd_mss_valid: Buffer to return the snd_mss valid flag
 * @snd_mss_limit_valid: Buffer to return snd_mss_limit valid flag
 * @snd_mss: Buffer to return MSS
 * @snd_mss_limit: Buffer to return MSS Limit
 *
 * Get values from TCP Update Mss Request Down Message
 */
void
__hal_tcp_update_mss_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *snd_mss_valid,
	u32 *snd_mss_limit_valid,
	u32 *snd_mss,
	u32 *snd_msss_limit)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (snd_mss_valid != NULL) &&
	    (snd_mss_limit_valid != NULL) && (snd_mss != NULL) &&
	    (snd_msss_limit != NULL));

	val64 = vxge_os_ntohll(dmsg->m.tcp_mss.control_0);

	*snd_mss_valid = (u32)VXGE_HAL_TCP_UPDATE_MSS_GET_MSS_VALID(val64);

	*snd_mss_limit_valid =
	    (u32)VXGE_HAL_TCP_UPDATE_MSS_GET_MSS_LIM_VALID(val64);

	*snd_mss = (u32)VXGE_HAL_TCP_UPDATE_MSS_GET_MSS(val64);

	*snd_msss_limit = (u32)VXGE_HAL_TCP_UPDATE_MSS_MSS_LIMIT(val64);

}

/*
 * __hal_tcp_update_ip_header_req_dmsg_get - Get values from TCP Update
 *				  IP Header Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @ip_tos_valid: Buffer to return ip_tos valid flag
 * @ip_flowlabel_valid: Buffer to return ip_flowlable valid flag
 * @ip_tos: Buffer to return IP TOS.
 * @ip_flowlabel: Buffer to return IP Flow Label.
 *
 * Get values from TCP Update IP Header Request Down Message
 */
void
__hal_tcp_update_ip_header_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *ip_tos_valid,
	u32 *ip_flowlabel_valid,
	u32 *ip_tos,
	u32 *ip_flowlabel)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (ip_tos_valid != NULL) &&
	    (ip_flowlabel_valid != NULL) &&
	    (ip_tos != NULL) && (ip_flowlabel != NULL));

	val64 = vxge_os_ntohll(dmsg->m.tcp_ipheader.control_0);

	*ip_tos_valid = (u32)VXGE_HAL_TCP_UPDATE_IPHEAD_GET_TOS_VALID(val64);

	*ip_flowlabel_valid =
	    (u32)VXGE_HAL_TCP_UPDATE_IPHEAD_GET_FL_VALID(val64);

	*ip_tos = (u32)VXGE_HAL_TCP_UPDATE_IPHEAD_TOS(val64);

	*ip_flowlabel = (u32)VXGE_HAL_TCP_UPDATE_IPHEAD_FLOW_LABEL(val64);
}

/*
 * __hal_tcp_update_keepalive_req_dmsg_get - Get values from TCP Update
 *				Keepalive Request Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @kpa_ok: Buffer to return Enable for keepalive probes for this connection.
 * @probes_max: Buffer to return Keepalive maximum probe count: maximum number
 *		of unanswered keepalive probes before giving up on session.
 * @idle_time: Buffer to return Keepalive idle timer: the idle time before
 *		sending the first keepalive probe (default is 2 hours).
 * @retry_time: Buffer to returnKeepalive retry time interval: idle time before
 *		resending an unanswered keepalive probe.
 *
 * Get values from TCP Update Keepalive Request Down Message
 */
void
__hal_tcp_update_keepalive_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *kpa_ok,
	u32 *probes_max,
	u32 *idle_time,
	u32 *retry_time)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) &&
	    (kpa_ok != NULL) &&
	    (probes_max != NULL) &&
	    (idle_time != NULL) &&
	    (retry_time != NULL));

	val64 = vxge_os_ntohll(dmsg->m.tcp_keepalive.control_0);

	*kpa_ok = (u32)VXGE_HAL_TCP_UPDATE_KPA_GET_KPA_OK(val64);

	*probes_max = (u32)VXGE_HAL_TCP_UPDATE_KPA_GET_PROBES_MAX(val64);

	val64 = vxge_os_ntohll(dmsg->m.tcp_keepalive.control_1);

	*idle_time = (u32)VXGE_HAL_TCP_UPDATE_KPA_GET_IDLE_TIME(val64);

	*retry_time = (u32)VXGE_HAL_TCP_UPDATE_KPA_RETRY_TIME(val64);
}

/*
 * __hal_tcp_update_failure_umsg_set - Set TCP Update Failure Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result : result
 *
 * Sets the TCP Update Failure Up Message.
 */
void
__hal_tcp_update_failure_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_TCP_UPDATE_FAILURE,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    result);
}

/*
 * __hal_tcp_fin_ack_received_umsg_set - Set TCP FIN ACK Received UMQ
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 *
 * Sets the TCP FIN ACK Received Up Message.
 */
void
__hal_tcp_fin_ack_received_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_TCP_FIN_ACK_RECEIVED,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    0);

}

/*
 * __hal_tcp_relink_to_nce_req_dmsg_get - Set TCP Relink to NCE Request DMQ
 *
 * @dmsgh: Handle to dmsg.
 * @nce_id : Buffer to return NCE Id
 *
 * Gets the values from TCP Relink to NCE Request down Message.
 */
void
__hal_tcp_relink_to_nce_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	u32 *nce_id)
{
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (nce_id != NULL));

	*nce_id = (u32)VXGE_HAL_TCP_RLINK_NCE_REQ_GET_NCE_ID(
	    vxge_os_ntohll(dmsg->m.tcp_relink.control_0));
}

/*
 * __hal_tcp_relink_to_nce_resp_umsg_set - Set TCP Relink to NCE Response
 *					  Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result : result
 *
 * Sets the TCP Relink to NCE Response Up Message.
 */
void
__hal_tcp_relink_to_nce_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_TCP_RELINK_TO_NCE_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    result);
}

/*
 * __hal_tcp_qp_limit_exceeded_umsg_set - Set TCP QP Limit Exceeded
 *					  Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result : result
 *
 * Sets the TCP Relink to NCE Response Up Message.
 */
void
__hal_tcp_qp_limit_exceeded_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_TCP_QP_LIMIT_EXCEEDED,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    result);
}

/*
 * __hal_tcp_rdma_terminate_received_umsg_set - Set TCP RDMA Terminate Received
 *					  Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result : result
 *
 * Sets the TCP Relink to NCE Response Up Message.
 */
void
__hal_tcp_rdma_terminate_received_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_TCP_RDMA_TERMINATE_RECEIVED,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    result);
}

/*
 * __hal_lro_open_req_dmsg_get - Get the values from LRO Open Request
 *				Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @session_opaque_handle: Buffer to return opaque handle used by the host to
 *		reference data structures in host memory related to this session
 * @lro_mode: Buffer to return LRO Mode. Preconfigured or Adaptive.
 * @max_sessions: Buffer to return maximum LRO sessions
 * @tcp_src_port: Buffer to return TCP source port
 * @tcp_dest_port: Buffer to return TCP destination port
 * @ip_sa_low: Buffer to return lower 64 bits of IPv4 or IPv6 source address
 * @ip_sa_high: Buffer to return higher 64 bits of IPv4 or IPv6 source address
 * @ip_da_low: Buffer to return lower 64 bits of IPv4 or IPv6
 *		destination address
 * @ip_da_high: Buffer to return higher 64 bits of IPv4 or IPv6
 *		destination address
 * @use_ipv6: Buffer to return flag that specifies if IPv6 is used
 * @use_vlan: Buffer to return flag that specifies if vlan id is used
 * @vlan_id: Buffer to return vlan id
 * @no_wqe_threshold: Lower limit for number of WQEs available for adapter.
 * @lro_cqrq_sin: Buffer to return instance number of the lro CQRQ
 * @lro_cqrq_sid: Buffer to return adapter ID of the lro CQRQ
 * @lro_srq_sin: Buffer to return instance number of the lro SRQ
 * @lro_srq_sid: Buffer to return adapter ID of the lro SRQ
 * @dupack_detection_enabled: Buffer to return flags to specify RPE to detects
 *		duplicate frames.
 * @data_merging_enabled2: Buffer to return flag to enable merging of TCP
 *		payloads.
 * @ack_merging_enabled: Buffer to return flag to enable merging of frames
 *		without TCP payloads.
 * @llc_hdr_mode: Buffer to return llc_header_mode
 * @snap_hdr_mode: Buffer to return snap_header_mode
 * @session_ecn_enabled: Buffer to return flag to enable explicit Congestion
 *		Notification check for this session.
 * @session_ecn_enabled_nonce: Buffer to return flag to enable for ECN with
 *		nonces.
 * @scatter_mode: Buffer to return scatter mode A
 * @buffer_mode: Buffer to return buffer mode
 * @data_lm_mask: Buffer to return left mergeability mask for frames with TCP
 *		payload
 * @dupack_mask: Buffer to return mask used when detecting duplicate or window
 *		update Acks
 * @lro_nm_mask: Buffer to return Mask used when detecting maskable
 *		non-mergeability conditions
 * @ack_lm_mask: Buffer to return left mergeability mask for frames without TCP
 *		payload
 * @max_ip_datagram_size: Buffer to return max size of the merged IP datagram.
 * @lro_rm_mask: Buffer to return Right mergeability mask for all LRO frames
 * @frame_threshold: Buffer to return number of frames allowed in the merge.
 * @psh_threshold: Buffer to return push flag threshold in merge.
 * @mtu_threshold: Buffer to return MTU threshold in merge.
 * @mss_threshold: Buffer to return MSS threshold in merge.
 * @max_tcp_tsval_delta: Buffer to return "monotonically non-decreasing"
 *		timestamp comparisons.
 * @max_ack_nbr_delta: Buffer to return ack number advancein the merge.
 * @buffer0_length: Buffer to return length of buffer 0 for this S-RQ.
 * @buffer1_length: Buffer to return length of buffer 1 for this S-RQ
 * @buffer2_length: Buffer to return length of buffer 2 for this S-RQ
 * @buffer3_length: Buffer to return length of buffer 3 for this S-RQ
 * @buffer4_length: Buffer to return length of buffer 4 for this S-RQ
 * @min_spare_wqe_capacity: Buffer to return he minimum number of bytes that
 *		must be free in an in-progress WQE if it is to be considered
 *		right-mergeable
 *
 * Gets the values from LRO Open Request DMQ message sent by the host to open
 * a LRO connection on the adapter.
 */
void
__hal_lro_open_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	vxge_hal_opaque_handle_t **session_opaque_handle,
	u32 *lro_mode,
	u32 *max_sessions,
	u32 *tcp_src_port,
	u32 *tcp_dest_port,
	u64 *ip_sa_high,
	u64 *ip_sa_low,
	u64 *ip_da_high,
	u64 *ip_da_low,
	u32 *use_ipv6,
	u32 *use_vlan,
	u32 *vlan_id,
	u32 *no_wqe_threshold,
	u32 *lro_cqrq_sin,
	u32 *lro_cqrq_sid,
	u32 *lro_srq_sin,
	u32 *lro_srq_sid,
	u32 *dupack_detection_enabled,
	u32 *data_merging_enabled,
	u32 *ack_merging_enabled,
	u32 *llc_hdr_mode,
	u32 *snap_hdr_mode,
	u32 *session_ecn_enabled,
	u32 *session_ecn_enabled_nonce,
	u32 *scatter_mode,
	u32 *buffer_mode,
	u32 *data_lm_mask,
	u32 *dupack_mask,
	u32 *lro_nm_mask,
	u32 *ack_lm_mask,
	u32 *max_ip_datagram_size,
	u32 *lro_rm_mask,
	u32 *frame_threshold,
	u32 *psh_threshold,
	u32 *mtu_threshold,
	u32 *mss_threshold,
	u32 *max_tcp_tsval_delta,
	u32 *max_ack_nbr_delta,
	u32 *buffer0_length,
	u32 *buffer1_length,
	u32 *buffer2_length,
	u32 *buffer3_length,
	u32 *buffer4_length,
	u32 *min_spare_wqe_capacity)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (session_opaque_handle != NULL) &&
	    (lro_mode != NULL) && (max_sessions != NULL) &&
	    (tcp_src_port != NULL) && (tcp_dest_port != NULL) &&
	    (ip_sa_high != NULL) && (ip_sa_low != NULL) &&
	    (ip_da_high != NULL) && (ip_da_low != NULL) &&
	    (use_ipv6 != NULL) && (use_vlan != NULL) &&
	    (vlan_id != NULL) && (no_wqe_threshold != NULL) &&
	    (lro_cqrq_sin != NULL) && (lro_cqrq_sid != NULL) &&
	    (lro_srq_sin != NULL) && (lro_srq_sid != NULL) &&
	    (dupack_detection_enabled != NULL) &&
	    (data_merging_enabled != NULL) &&
	    (ack_merging_enabled != NULL) &&
	    (llc_hdr_mode != NULL) && (snap_hdr_mode != NULL) &&
	    (session_ecn_enabled != NULL) &&
	    (session_ecn_enabled_nonce != NULL) &&
	    (scatter_mode != NULL) && (buffer_mode != NULL) &&
	    (data_lm_mask != NULL) && (dupack_mask != NULL) &&
	    (lro_nm_mask != NULL) && (ack_lm_mask != NULL) &&
	    (max_ip_datagram_size != NULL) && (lro_rm_mask != NULL) &&
	    (frame_threshold != NULL) &&
	    (psh_threshold != NULL) && (mtu_threshold != NULL) &&
	    (mss_threshold != NULL) && (max_tcp_tsval_delta != NULL) &&
	    (max_ack_nbr_delta != NULL) && (buffer0_length != NULL) &&
	    (buffer1_length != NULL) && (buffer2_length != NULL) &&
	    (buffer3_length != NULL) && (buffer4_length != NULL) &&
	    (min_spare_wqe_capacity != NULL));

	*session_opaque_handle = (vxge_hal_opaque_handle_t *)(ptr_t)
	    vxge_os_ntohll(dmsg->m.lro_open.session_opaque_handle);

	val64 = vxge_os_ntohll(dmsg->m.lro_open.control_0);

	*lro_mode =  (u32)VXGE_HAL_LRO_OPEN_REQ_GET_LRO_MODE(val64);

	*max_sessions = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_NUM_LRO_SESSIONS(val64);

	*tcp_src_port = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_TCP_SRC_PORT(val64);

	*tcp_dest_port = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_TCP_DEST_PORT(val64);

	*ip_sa_high = vxge_os_ntohll(dmsg->m.lro_open.ip_sa.high);

	*ip_sa_low = vxge_os_ntohll(dmsg->m.lro_open.ip_sa.low);

	*ip_da_high = vxge_os_ntohll(dmsg->m.lro_open.ip_da.high);

	*ip_da_low = vxge_os_ntohll(dmsg->m.lro_open.ip_da.low);

	val64 = vxge_os_ntohll(dmsg->m.lro_open.control_1);

	*use_ipv6 = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_USE_IPV6(val64);

	*use_vlan = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_USE_VLAN(val64);

	*vlan_id = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_VLAN_ID(val64);

	*no_wqe_threshold =
	    (u32)VXGE_HAL_LRO_OPEN_REQ_GET_NO_WQE_THRESHOLD(val64);

	val64 = vxge_os_ntohll(dmsg->m.lro_open.control_2);

	*lro_cqrq_sin = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_LRO_CQRQ_INST_NUM(val64);

	*lro_cqrq_sid = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_LRO_CQRQ_ID(val64);

	*lro_srq_sin = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_LRO_SRQ_INST_NUM(val64);

	*lro_srq_sid = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_LRO_SRQ_ID(val64);

	val64 = vxge_os_ntohll(dmsg->m.lro_open.control_3);

	*dupack_detection_enabled =
	    (u32)VXGE_HAL_LRO_OPEN_REQ_GET_DUP_ACK_DET_EN(val64);

	*data_merging_enabled =
	    (u32)VXGE_HAL_LRO_OPEN_REQ_GET_DATA_MERGE_EN(val64);

	*ack_merging_enabled =
	    (u32)VXGE_HAL_LRO_OPEN_REQ_GET_ACK_MERGE_EN(val64);

	*llc_hdr_mode = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_LLC_HDR_MODE(val64);

	*snap_hdr_mode = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_SNAP_HDR_MODE(val64);

	*session_ecn_enabled =
	    (u32)VXGE_HAL_LRO_OPEN_REQ_GET_SESS_ECN_EN(val64);

	*session_ecn_enabled_nonce =
	    (u32)VXGE_HAL_LRO_OPEN_REQ_GET_S_ECN_EN_NONCE(val64);

	*scatter_mode = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_SCATTER_MODE(val64);

	*buffer_mode = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_BUFFER_MODE(val64);

	val64 = vxge_os_ntohll(dmsg->m.lro_open.control_4);

	*data_lm_mask = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_DATA_LM_MASK(val64);

	*dupack_mask = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_DUPACK_MASK(val64);

	*lro_rm_mask = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_LRO_NM_MASK(val64);

	*ack_lm_mask = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_ACK_LM_MASK(val64);

	val64 = vxge_os_ntohll(dmsg->m.lro_open.control_5);

	*max_ip_datagram_size =
	    (u32)VXGE_HAL_LRO_OPEN_REQ_GET_MAX_IP_DG_SIZE(val64);

	*lro_rm_mask = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_LRO_RM_MASK(val64);

	*frame_threshold = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_FRAME_TH(val64);

	*psh_threshold = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_PSH_TH(val64);

	val64 = vxge_os_ntohll(dmsg->m.lro_open.control_6);

	*mtu_threshold = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_MTU_TH(val64);

	*mss_threshold = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_MSS_TH(val64);

	*max_tcp_tsval_delta =
	    (u32)VXGE_HAL_LRO_OPEN_REQ_GET_MAX_TCP_TSV_DEL(val64);

	val64 = vxge_os_ntohll(dmsg->m.lro_open.control_7);

	*max_ack_nbr_delta =
	    (u32)VXGE_HAL_LRO_OPEN_REQ_GET_MAX_ACK_NBR_DEL(val64);

	*buffer0_length = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_BUF0_LENGTH(val64);

	*buffer1_length = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_BUF1_LENGTH(val64);

	val64 = vxge_os_ntohll(dmsg->m.lro_open.control_8);

	*buffer2_length = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_BUF2_LENGTH(val64);

	*buffer3_length = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_BUF3_LENGTH(val64);

	*buffer4_length = (u32)VXGE_HAL_LRO_OPEN_REQ_GET_BUF4_LENGTH(val64);

	*min_spare_wqe_capacity =
	    (u32)VXGE_HAL_LRO_OPEN_REQ_GET_MIN_SPARE_WQE_CAPACITY(val64);
}

/*
 * __hal_lro_open_resp_umsg_set - Set the LRO Open Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result: Result code
 *
 * Set the LRO Open Response UMQ used to communicate to the host the
 * success or failure of the corresponding LRO Open Request DMQ message
 */
void
__hal_lro_open_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_LRO_OPEN_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    result);
}

/*
 * __hal_lro_end_classification_resp_umsg_set - Set LRO End Classification
 *						Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result : result
 *
 * Set the LRO End Classification Response UMQ message is sent in response to a
 * LRO End Classification Request DMQ message to indicate that classification
 * has been stopped for the LRO session and the host can proceed with deleting
 * the LRO session.
 */
void
__hal_lro_end_classification_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_LRO_END_CLASSIF_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    result);
}

/*
 * __hal_lro_delete_resp_umsg_set - Set LRO Delete Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @session_opaque_handle: Session Opaque handle
 * @result : result
 *
 * Sets the LRO Delete Response Up Message.
 */
void
__hal_lro_delete_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	vxge_hal_opaque_handle_t *session_opaque_handle,
	u32 result)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_LRO_DELETE_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE+sizeof (__hal_lro_delete_resp_umq_t),
	    result);

	umsg->m.lro_delete.session_opaque_handle = vxge_os_htonll(
	    (u64)(ptr_t)session_opaque_handle);
}

/*
 * __hal_lro_session_candidate_umsg_set - Set LRO Session Candidate
 *						Notification Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @l4sp: Layer 4 source port
 * @l4dp: Layer 4 destination port
 * @ipv4_not_ipv6: Flag to specify IP Address type
 * @ip_sa_low: Lower 64 bits of IPv4 or IPv6 source address
 * @ip_sa_high: Higher 64 bits of IPv4 or IPv6 source address
 * @ip_da_low: Lower 64 bits of IPv4 or IPv6 destination address
 * @ip_da_high: Higher 64 bits of IPv4 or IPv6 destination address
 *
 * Sets the msg that indicates to the host that the adapter's autoLRO feature
 * has identified a candidate LRO session.
 */
void
__hal_lro_session_candidate_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 l4sp,
	u32 l4dp,
	u32 ipv4_not_ipv6,
	u64 ip_sa_low,
	u64 ip_sa_high,
	u64 ip_da_low,
	u64 ip_da_high)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_LRO_DELETE_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    0);

	umsg->m.lro_sess_cand.control_0 =
	    VXGE_HAL_LRO_SESS_CAND_L4_SRC_PORT(l4sp) |
	    VXGE_HAL_LRO_SESS_CAND_L4_DEST_PORT(l4dp);

	if (ipv4_not_ipv6)
		umsg->m.lro_sess_cand.control_0 |=
		    VXGE_HAL_LRO_SESS_CAND_IPV4_NOT_IPV6;

	umsg->m.lro_sess_cand.control_0 =
	    vxge_os_htonll(umsg->m.lro_sess_cand.control_0);

	umsg->m.lro_sess_cand.ip_sa.low = vxge_os_htonll(ip_sa_low);

	umsg->m.lro_sess_cand.ip_sa.high = vxge_os_htonll(ip_sa_high);

	umsg->m.lro_sess_cand.ip_da.low = vxge_os_htonll(ip_da_low);

	umsg->m.lro_sess_cand.ip_da.high = vxge_os_htonll(ip_da_high);
}

/*
 * __hal_spdm_open_req_dmsg_get - Get the values from SPDM Open Request
 *				Down Message
 *
 * @dmsgh: Handle to dmsg.
 * @session_opaque_handle: Buffer to return opaque handle used by the host to
 *		reference data structures in host memory related to this session
 * @l4sp: Buffer to return layer 4 source port
 * @l4dp: Buffer to return layer 4 destination port
 * @use_ipv6: Buffer to return the flag to specify if IPv6 is used
 * @use_vlan: Buffer to return the flag to specify if vlan is used
 * @vlan_id: Buffer to return the vlan id
 * @ip_sa_low: Buffer to return lower 64 bits of IPv4 or IPv6 source address
 * @ip_sa_high: Buffer to return higher 64 bits of IPv4 or IPv6 source address
 * @ip_da_low: Buffer to return lower 64 bits of IPv4 or IPv6
 *		destination address
 * @ip_da_high: Buffer to return higher 64 bits of IPv4 or IPv6
 *		destination address
 *
 * Gets the values from SPDM Open Request DMQ message sent by the host to open
 * a SPDM connection on the adapter.
 */
void
__hal_spdm_open_req_dmsg_get(
	vxge_hal_down_msg_h dmsgh,
	vxge_hal_opaque_handle_t **session_opaque_handle,
	u32 *l4sp,
	u32 *l4dp,
	u32 *use_ipv6,
	u32 *use_vlan,
	u32 *vlan_id,
	u64 *ip_sa_high,
	u64 *ip_sa_low,
	u64 *ip_da_high,
	u64 *ip_da_low)
{
	u64 val64;
	__hal_down_msg_t *dmsg = (__hal_down_msg_t *)dmsgh;

	vxge_assert((dmsg != NULL) && (session_opaque_handle != NULL) &&
	    (l4sp != NULL) && (l4dp != NULL) && (use_ipv6 != NULL) &&\
	    (use_vlan != NULL) && (vlan_id != NULL) &&
	    (ip_sa_high != NULL) && (ip_sa_low != NULL) &&
	    (ip_da_high != NULL) && (ip_da_low != NULL));

	*session_opaque_handle = (vxge_hal_opaque_handle_t *)(ptr_t)
	    vxge_os_ntohll(dmsg->m.spdm_open.session_opaque_handle);

	val64 = vxge_os_ntohll(dmsg->m.spdm_open.control_0);

	*l4sp = (u32)VXGE_HAL_SPDM_OPEN_REQ_GET_L4_SRC_PORT(val64);

	*l4dp = (u32)VXGE_HAL_SPDM_OPEN_REQ_GET_L4_DEST_PORT(val64);

	*use_ipv6 = (u32)VXGE_HAL_SPDM_OPEN_REQ_GET_USE_IPV6(val64);

	*use_vlan = (u32)VXGE_HAL_SPDM_OPEN_REQ_GET_USE_VLAN(val64);

	*vlan_id = (u32)VXGE_HAL_SPDM_OPEN_REQ_GET_VLAN_ID(val64);

	*ip_sa_high = vxge_os_ntohll(dmsg->m.spdm_open.ip_sa.high);

	*ip_sa_low = vxge_os_ntohll(dmsg->m.spdm_open.ip_sa.low);

	*ip_da_high = vxge_os_ntohll(dmsg->m.spdm_open.ip_da.high);

	*ip_da_low = vxge_os_ntohll(dmsg->m.spdm_open.ip_da.low);

}

/*
 * __hal_spdm_open_resp_umsg_set - Set the SPDM Open Response UMQ
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @session_hash: Classification hash used to classify the session context.
 * @result: Result code
 *
 * Set the SPDM Open Response UMQ used to communicate to the host the
 * success or failure of the corresponding SPDM Open Request DMQ message
 */
void
__hal_spdm_open_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 session_hash,
	u32 result)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SPDM_OPEN_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE +
	    sizeof (__hal_spdm_open_resp_umq_t),
	    result);

	umsg->m.spdm_open.control_0 = vxge_os_htonll(
	    VXGE_HAL_SPDM_OPEN_RESP_SESS_HASH(session_hash));

}

/*
 * __hal_spdm_delete_resp_umsg_set - Set SPDM Delete Response Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @session_opaque_handle: Session Opaque handle
 * @result : result
 *
 * Sets the SPDM Delete Response Up Message.
 */
void
__hal_spdm_delete_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	vxge_hal_opaque_handle_t *session_opaque_handle,
	u32 result)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SPDM_DELETE_RESP,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE+sizeof (__hal_spdm_delete_resp_umq_t),
	    result);

	umsg->m.spdm_delete.session_opaque_handle = vxge_os_htonll(
	    (u64)(ptr_t)session_opaque_handle);
}

/*
 * __hal_session_event_notification_umsg_set - Set session event notification
 *					  Up Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result : result
 *
 * Sets the TCP Relink to NCE Response Up Message.
 */
void
__hal_session_event_notification_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result)
{
	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SESSION_EVENT_NOTIF,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE,
	    result);
}


/*
 * vxge_hal_session_query_resp_umsg_set - Set the values of Session query
 *				 response UMQ message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @result : result
 * @sess_type: Session type
 * @vp_id: VP Id
 * @sess_inst_num: Session instance number
 * @is_spdm: The flag to indicate if it is spdm session
 * @use_ipv6: The flag indicating if IPv6 address is used
 * @use_vlan: The flag indicating if vlan is used
 * @vlan_id: vlan Id
 * @rx_imsn: Receive direction Initial Marker Sequence Number
 * @fw_ctrl: The session hash
 * @src_port: TCP source port
 * @dest_port: TCP destination port
 * @ip_sa_high: High part of source address
 * @ip_sa_low: Low part of source address
 * @ip_da_high: High part of destination address
 * @ip_da_low: Low part of destination address
 * @use_mpa_crc232: The flag Indicating if this session uses MPA CRC32C
 * @use_mpa_markers: The flag Indicating if this session uses MPA markers
 *		in the Rx direction.
 *
 * Set the values of the Session Query Response UMQ message is sent in
 * response to a Session Query Request DMQ message to return the attributes
 * associated with the specified session
 */
void
vxge_hal_session_query_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result,
	u32 sess_type,
	u32 vp_id,
	u32 sess_inst_num,
	u32 is_spdm,
	u32 use_ipv6,
	u32 use_vlan,
	u32 vlan_id,
	u32 rx_imsn,
	u32 fw_ctrl,
	u32 src_port,
	u32 dest_port,
	u64 ip_sa_high,
	u64 ip_sa_low,
	u64 ip_da_high,
	u64 ip_da_low,
	u32 use_mpa_crc232,
	u32 use_mpa_markers)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	umsg->m.session_query.control_0 =
	    VXGE_HAL_SESSION_QUERY_RESP_SESS_TYPE(sess_type) |
	    VXGE_HAL_SESSION_QUERY_RESP_VP_ID(vp_id) |
	    VXGE_HAL_SESSION_QUERY_RESP_INST_NUM(sess_inst_num) |
	    VXGE_HAL_SESSION_QUERY_RESP_VLAN_ID(vlan_id) |
	    VXGE_HAL_SESSION_QUERY_RESP_RX_IMSN(rx_imsn);

	if (is_spdm)
		umsg->m.session_query.control_0 |=
		    VXGE_HAL_SESSION_QUERY_RESP_IS_SPDM;

	if (use_ipv6)
		umsg->m.session_query.control_0 |=
		    VXGE_HAL_SESSION_QUERY_RESP_USE_IPV6;

	if (use_vlan)
		umsg->m.session_query.control_0 |=
		    VXGE_HAL_SESSION_QUERY_RESP_USE_VLAN;

	umsg->m.session_query.control_0 =
	    vxge_os_htonll(umsg->m.session_query.control_0);

	umsg->m.session_query.control_1 = vxge_os_htonll(
	    VXGE_HAL_SESSION_QUERY_RESP_FW_CTRL(fw_ctrl) |
	    VXGE_HAL_SESSION_QUERY_RESP_SRC_PORT(src_port) |
	    VXGE_HAL_SESSION_QUERY_RESP_DEST_PORT(dest_port));

	umsg->m.session_query.ip_sa.high = vxge_os_htonll(ip_sa_high);

	umsg->m.session_query.ip_sa.low = vxge_os_htonll(ip_sa_low);

	umsg->m.session_query.ip_da.high = vxge_os_htonll(ip_da_high);

	umsg->m.session_query.ip_da.low = vxge_os_htonll(ip_da_low);

	umsg->m.session_query.control_2 = 0;

	if (use_mpa_crc232)
		umsg->m.session_query.control_2 |=
		    VXGE_HAL_SESSION_QUERY_RESP_USE_MPA_CRC32C;

	if (use_mpa_markers)
		umsg->m.session_query.control_2 |=
		    VXGE_HAL_SESSION_QUERY_RESP_USE_MPA_MARKERS;

	umsg->m.session_query.control_2 =
	    vxge_os_htonll(umsg->m.session_query.control_2);
}

/*
 * __hal_session_return_in_progress_wqes_umsg_set - Set the Session Return
 *				in progress WQEs UMQ message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @number_of_wqes_in_msg: The number of WQE opaque handles in this message
 * @number_of_wqes_remaining: The number of WQE opaque handles left to send
 * @wqe_opaque_handles: This is a multiple of 64 because this field will contain
 *		multiple WQE opaque handles. The actual number is indicated by
 *		the Number_of_wqes_in_msg.
 *
 * Set the Session Return in progress WQEs UMQ message generated by the adapter
 * during deletion of a session to return any WQEs that may be in the in-progres
 * list for the session. If a WQE is in the in-progress list it is owned by the
 * session and cannot be returned to the head of the WQE list for an SRQ because
 * of ordering issues. Therefore, it must be returned to the host at which point
 * the host may choose to destroy the resource or simply re-post the WQE for
 * re-use.
 */
void
__hal_session_return_in_progress_wqes_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 number_of_wqes_in_msg,
	u32 number_of_wqes_remaining,
	u64 *wqe_opaque_handles)
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;
	u32 i;
	u64 *wqe_handles;

	vxge_assert((umsg != NULL) && ((number_of_wqes_in_msg == 0) ||
	    (wqe_opaque_handles != NULL)));

	if (number_of_wqes_in_msg > 10)
		number_of_wqes_in_msg = 10;

	__hal_up_msg_header_set(umsgh, msg_id,
	    VXGE_HAL_MSG_TYPE_SESSION_RETURN_IN_PROG_WQES,
	    compl_dmq, opaque_handle,
	    VXGE_HAL_UP_MSG_HEADER_SIZE +
	    sizeof (__hal_session_return_in_progress_wqes_umq_t),
	    0);

	umsg->m.session_return_wqes.control_0 = vxge_os_htonll(
	    VXGE_HAL_SESSION_RETURN_WQES_NUM_IN_MSG(number_of_wqes_in_msg) |
	    VXGE_HAL_SESSION_RETURN_WQES_NUM_REMAIN(number_of_wqes_remaining));

	wqe_handles = umsg->m.session_return_wqes.wqe_opaque_handles;

	for (i = 0; i < number_of_wqes_in_msg; i++) {
		*wqe_handles++ = vxge_os_htonll(*wqe_opaque_handles++);
	}
}

/*
 * __hal_session_frame_write_umsg_set - Set the values of Session Frame
 *					  Write UMQ Message
 *
 * @umsgh: Handle to umsg.
 * @msg_id: Message Id
 * @compl_dmq: Indicates that a DMQ message from the host is required upon
 *		completion of message processing.
 * @opaque_handle: Opaque handle
 * @umq_opaque_data: A code provided to the RPE by the RxPE firmware.
 * @num_markers: The number of markers contained in the frame, or portion of
 *		the frame, that will be passed up the UMQ.
 * @first_marker_index: The first byte of frame data passed to the host is
 *		considered to be Byte 0.
 * @fdata: The variable length data of the frame pulled from the frame buffer
 *		by the ROCRC.
 *
 * Get the values from Session Frame Write UMQ used by adapter to send
 * certain frames to the host via the UMQ path instead of the normal path..
 */
void __hal_session_frame_write_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 umq_opaque_data,
	u32 num_markers,
	u32 first_marker_index,
	u64 fdata[2])
{
	__hal_up_msg_t *umsg = (__hal_up_msg_t *)umsgh;

	vxge_assert(umsg != NULL);

	umsg->m.session_frame_write.control_0 = vxge_os_htonll(
	    VXGE_HAL_SESSION_FRAME_WRITE_UMQ_OPAQUE_DATA(umq_opaque_data) |
	    VXGE_HAL_SESSION_FRAME_WRITE_NUM_MARKERS(num_markers) |
	    VXGE_HAL_SESSION_FRAME_WRITE_FIRST_MARKER_IDX(first_marker_index));

	umsg->m.session_frame_write.fdata[0] = vxge_os_htonll(fdata[0]);

	umsg->m.session_frame_write.fdata[1] = vxge_os_htonll(fdata[1]);
}


void
__hal_session_query_resp_umsg_set(
	vxge_hal_up_msg_h umsgh,
	u32 msg_id,
	u32 compl_dmq,
	vxge_hal_opaque_handle_t *opaque_handle,
	u32 result,
	u32 sess_type,
	u32 vp_id,
	u32 sess_inst_num,
	u32 is_spdm,
	u32 use_ipv6,
	u32 use_vlan,
	u32 vlan_id,
	u32 rx_imsn,
	u32 fw_ctrl,
	u32 src_port,
	u32 dest_port,
	u64 ip_sa_high,
	u64 ip_sa_low,
	u64 ip_da_high,
	u64 ip_da_low,
	u32 use_mpa_crc232,
	u32 use_mpa_markers)
{

}

#endif
